# cython: language_level=3

import math
import itertools


PI_2 = math.pi / 2


cpdef inline int nC2(int N):
    """Return the number of distinct pairs from a set of N."""

    return N * (N - 1) // 2


cpdef inline float dist_xy(xy A, xy B):
    """Return the distance between A and B."""

    cdef float dx, dy
    cdef float ret_2

    dx = A.x - B.x
    dy = A.y - B.y

    ret_2 = (dx * dx) + (dy * dy)
    return math.sqrt(ret_2)


cpdef inline float dist_2D(float x1, float y1, float x2, float y2):
    """Return the distance between xy1 and xy2."""

    cdef float dx, dy
    cdef float ret_2

    dx = x1 - x2
    dy = y1 - y2

    ret_2 = (dx * dx) + (dy * dy)
    return math.sqrt(ret_2)


cpdef inline float dist_L2P_2D(
    float Ax, float Ay,
    float Bx, float By,
    float Px, float Py
):
    """Return the distance from the point (Px, Py) to the line that
    passes through (Ax, Ay) and (Bx, By)."""

    # Adapted from:
    # http://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line

    cdef float normal

    normal = dist_2D(Ax, Ay, Bx, By)
    return abs((Px - Ax) * (By - Ay) - (Py - Ay) * (Bx - Ax)) / normal


cpdef inline xy absrel_xy(
    float x, float y, float angle,
    float dx, float dy
):
    """Return the absolute coordinates of the point (dx, dy)
    rotated angle radians around the point (x, y).

    """

    cdef xy ret
    cdef float c = math.cos(angle)
    cdef float s = math.sin(angle)

    ret.x = (dx * c) - (dy * s) + x
    ret.y = (dx * s) + (dy * c) + y

    return ret


cpdef inline float absrel_dist(
    float x1, float y1, float angle1, float dx1, float dy1,
    float x2, float y2, float angle2, float dx2, float dy2
):
    """Return the distance between two points:
        (x1, y1) + rotated(dx1, dy1)
    and
        (x2, y2) + rotated(dx2, dy2)

    """

    cdef xy a1 = absrel_xy(x1, y1, angle1, dx1, dy1)
    cdef xy a2 = absrel_xy(x2, y2, angle2, dx2, dy2)
    cdef float diff_x, diff_y
    cdef float ret_2

    diff_x = a1.x - a2.x
    diff_y = a1.y - a2.y

    ret_2 = (diff_x * diff_x) + (diff_y * diff_y)
    return math.sqrt(ret_2)


cpdef object bbox_size(float w, float h, float angle):
    """Return the bounding box size for the given rectangle.

    Arguments:
    w -- width
    h -- height
    angle -- rotation angle in radians

    Return (bbox_width, bbox_height)

    """

    # Coords for the +x-most and +y-most points
    cdef float xm_x, xm_y
    cdef float ym_x, ym_y

    # Precomputed cos(angle) and sin(angle)
    cdef float c, s

    # Computed bounding box width, height
    cdef float bb_w, bb_h

    angle %= 2 * math.pi
    c = math.cos(angle)
    s = math.sin(angle)

    # Do not divide by 2: want full bounding box size
    if angle <= PI_2:
        xm_x = + w
        xm_y = - h

        ym_x = + w
        ym_y = + h

    elif angle <= math.pi:
        xm_x = - w
        xm_y = - h

        ym_x = + w
        ym_y = - h

    elif angle <= 3 * PI_2:
        xm_x = - w
        xm_y = + h

        ym_x = - w
        ym_y = - h

    else:
        xm_x = + w
        xm_y = + h

        ym_x = - w
        ym_y = + h

    bb_w = (xm_x * c) - (xm_y * s)
    bb_h = (ym_x * s) + (ym_y * c)

    assert(bb_w >= 0)
    assert(bb_h >= 0)

    return (bb_w, bb_h)


cpdef float overlap_penalty(DIN_A, DIN_B, placement, DIN2dim):
    """Return the penalty for overlap between two DevInsts.

    Arguments:
    DIN_A -- one humanname(dev_instance)
    DIN_B -- the other humanname(dev_instance)
    placement -- {DIN: (x, y, angle)}
    DIN2dim -- {DIN: (width, height)}

    References:
    DIN2dim -- for computing overlap

    Return a cost based on overlap (0 if no overlap).

    """

    cdef float x1, y1, x2, y2
    cdef float angle1, angle2
    cdef float width1, height1
    cdef float width2, height2
    cdef float dx, dy
    cdef float s1, e1, n2, w2

    # Compute area of intersection
    # Support rotated DevInsts by computing overlap of bounding boxes

    x1 = placement[DIN_A][0]
    y1 = placement[DIN_A][1]
    angle1 = placement[DIN_A][2]
    x2 = placement[DIN_B][0]
    y2 = placement[DIN_B][1]
    angle2 = placement[DIN_B][2]

    width1, height1 = DIN2dim[DIN_A]
    width1, height1 = bbox_size(width1, height1, angle1)
    width2, height2 = DIN2dim[DIN_B]
    width2, height2 = bbox_size(width2, height2, angle2)

    # Project relative coordinates into quadrant I
    dx = abs(x2 - x1)
    dy = abs(y2 - y1)

    # Want SE corner of A, NW corner of B (since quadrant I)
    s1 = height1 / 2
    e1 = width1 / 2

    n2 = dy - (height2 / 2)
    w2 = dx - (width2 / 2)

    if (n2 <= s1) and (w2 <= e1):
        # Overlap!
        return (s1 - n2) * (e1 - w2)    # This is >= 0

    return 0


cpdef float total_cost(
    placement, show_overlap,
    DIN2dim, int max_connectivity,
    CoM, B2D2numreqpins,
    int weight_dist, int weight_overlap
):

    """Return the total cost based on the DevInsts' locations.

    Arguments:
    placement -- {DIN: (x, y, angle)}
    show_overlap -- if true, print overlaps
    DIN2dim -- {DIN: (width, height)}
    max_connectivity -- for weighting costs
    CoM -- {bundlename: {DIN: {bank: (center_dx, center_dy)}}}
    B2D2numreqpins -- {bundlename: {DIN: conn_numreqpins}}
    weight_dist -- weighting factor for total distance
    weight_overlap -- weighting factor for total overlap

    """

    cdef float ret = 0.0
    cdef float cost_dist = 0.0, cost_overlap = 0.0
    cdef float dist, bundle_dist
    cdef float overlap
    cdef float weight

    assert(max_connectivity != 0)

    for bundlename, DIN2CoMs in CoM.items():
        weight = 1. / (nC2(len(DIN2CoMs)) * max_connectivity)
        bundle_dist = 0.0

        DIN2numreqpins = B2D2numreqpins[bundlename]

        for pair in itertools.combinations(DIN2CoMs, 2):
            DIN_L, DIN_R = pair

            x1, y1, angle1 = placement[DIN_L]
            x2, y2, angle2 = placement[DIN_R]

            bank2CoM_L = DIN2CoMs[DIN_L]
            bank2CoM_R = DIN2CoMs[DIN_R]

            # Find the shortest distance between the two DevInsts
            # (any bank's CoM to any bank's CoM)
            best_dist = None
            for bank_L, bank_R in itertools.product(bank2CoM_L, bank2CoM_R):
                dx1, dy1 = bank2CoM_L[bank_L]
                dx2, dy2 = bank2CoM_R[bank_R]

                dist = absrel_dist(
                    x1, y1, angle1, dx1, dy1,
                    x2, y2, angle2, dx2, dy2
                )

                if (best_dist is None) or (dist < best_dist):
                    best_dist = dist

            cnp1 = DIN2numreqpins[DIN_L]
            cnp2 = DIN2numreqpins[DIN_R]

            cnp = (cnp1 + cnp2) / 2 # cnp is the same for all banks' CoMs

            bundle_dist += cnp * best_dist

        cost_dist += weight * bundle_dist


    for pair in itertools.combinations(placement, 2):
        DIN_L, DIN_R = pair

        overlap = overlap_penalty(DIN_L, DIN_R, placement, DIN2dim)

        if show_overlap:
            print('Overlap', pair, overlap)

        cost_overlap += overlap

    ret = (cost_dist * weight_dist) + (cost_overlap * weight_overlap)
    return round(ret, 4)
