#!/usr/bin/env python3

"""Circuit devices, packages, pins, etc."""

import sys
import math
import collections

import wallpy.cleanobject
import wallpy.misc
import wallpy.prettyprint
import wallpy.units

import devmanager




# Signal types
ANALOG = 'ana'
DIGITAL = 'dig'
POWER = 'pow'

# Signal directions
INPUT = 'in'
OUTPUT = 'out'
BIDIR = 'bidir'

# Single-ended vs. differential
SINGLE = 'single'
DIFF_P = 'diff_p'
DIFF_N = 'diff_n'

DIFF_SIGCONFIGS = frozenset({DIFF_P, DIFF_N})



moduleprefix = __name__ + '.'







class DevCollection(wallpy.cleanobject.CleanObject):
    posarg_names = []
    kwarg_names = [ 'devices' ]

    moduleprefix = moduleprefix


    def __init__(self, devices = None):

        if devices:
            devices = wallpy.cleanobject.CleanDict(devices)
        else:
            devices = wallpy.cleanobject.CleanDict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def filter_devices(self, func, args = None, kwargs = None):
        if not args:
            args = tuple()
        if not kwargs:
            kwargs = dict()

        for k, v in self.devices.items():
            ret = func(k, v, *args, **kwargs)

            if ret:
                yield ret




class Device(wallpy.cleanobject.CleanObject):
    posarg_names = []
    kwarg_names = [ 'packages', 'bank2iostd' ]

    moduleprefix = moduleprefix


    def __init__(self, packages=None, bank2iostd=None):

        if packages:
            packages = wallpy.cleanobject.CleanDict(packages)
        else:
            packages = wallpy.cleanobject.CleanDict()


        if bank2iostd:
            bank2iostd = wallpy.cleanobject.CleanDict(bank2iostd)
        else:
            bank2iostd = wallpy.cleanobject.CleanDict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if not self.packages:
            errcallback('Device has no packages')
            return False


        return True



    def filter_packages(self, func, args = None, kwargs = None):
        if not args:
            args = tuple()
        if not kwargs:
            kwargs = dict()

        for k, v in self.packages.items():
            ret = func(k, v, *args, **kwargs)

            if ret:
                yield ret




class LogicalPackage(wallpy.cleanobject.CleanObject):
    posarg_names = []
    kwarg_names = [ 'pins' ]

    moduleprefix = moduleprefix


    def __init__(self, pins = None):

        if pins:
            pins = wallpy.cleanobject.CleanDict(pins)
        else:
            pins = wallpy.cleanobject.CleanDict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


        # {pinfuncname: {bank: {lpk}}}
        def dictset():
            return collections.defaultdict(set)
        self.pf2pins = collections.defaultdict(dictset)


        # {pinfuncname: {bank: diff_pair_lpk}}
        self.diff_pairs = collections.defaultdict(dict)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if not self.pins:
            errcallback('Package has no pins')
            return False


        return True


    def postprocess_pins(self):
        """Populate self.pf2pins and self.diff_pairs."""

        self.pf2pins.clear()

        for lpk, pin in self.pins.items():
            for pfname, pf in pin.pinfuncs.items():
                bank = pf.params.get('bank', None)

                self.pf2pins[pfname][bank].add(lpk)


        self.diff_pairs.clear()

        for lpk, pin in self.pins.items():
            for pfname, pf in pin.pinfuncs.items():
                bank = pf.params.get('bank', None)

                diff_pair = pf.params.get('diff_pair', None)
                if diff_pair is not None:
                    other_lpks = self.pf2pins[diff_pair][bank]
                    if len(other_lpks) != 1:
                        raise AssertionError(
                            'Multiple diff. pin pairs',
                            lpk,
                            pfname,
                            other_lpks
                        )

                    self.diff_pairs[pfname][bank] = next(iter(other_lpks))


    def filter_pins(self, func, args = None, kwargs = None):
        if not args:
            args = tuple()
        if not kwargs:
            kwargs = dict()

        for k, v in self.pins.items():
            ret = func(k, v, *args, **kwargs)

            if ret:
                yield ret



class LogicalPin(wallpy.cleanobject.CleanObject):
    posarg_names = []
    kwarg_names = [ 'pinfuncs' ]

    moduleprefix = moduleprefix


    def __init__(self, pinfuncs = None):

        if pinfuncs:
            pinfuncs = wallpy.cleanobject.CleanDict(pinfuncs)
        else:
            pinfuncs = wallpy.cleanobject.CleanDict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False

        return True


    def required_p(self):
        """Return True if any PinFunc is required."""

        for v in self.pinfuncs.values():
            if v.required:
                return True

        return False


    def filter_pinfuncs(self, func, args = None, kwargs = None):
        if not args:
            args = tuple()
        if not kwargs:
            kwargs = dict()

        for k, v in self.pinfuncs.items():
            ret = func(k, v, *args, **kwargs)

            if ret:
                yield ret


    @property
    def bank(self):
        """Return the bank containing this pin, or None otherwise.

        Raise ValueError if this LogicalPin is in more than one bank.

        """

        banks = set()
        for pf in self.pinfuncs.values():
            try:
                banks.add(pf.params['bank'])
            except KeyError:
                pass

        if len(banks) > 1:
            raise ValueError('Multiple banks for single pin', banks)

        if banks:
            return next(iter(banks))

        return None



class PinFunc(wallpy.cleanobject.CleanObject):
    posarg_names = [ 'sigtype', 'sigdir', 'sigconfig' ]
    kwarg_names = [ 'required', 'params' ]

    valid_sigtypes = { ANALOG, DIGITAL, POWER }
    valid_sigdirs = { INPUT, OUTPUT, BIDIR }
    valid_sigconfigs = { SINGLE, DIFF_P, DIFF_N }

    moduleprefix = moduleprefix


    def __init__(self, sigtype, sigdir, sigconfig,
                 required = False, params = None):

        required = bool(required)

        if not params:
            params = dict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if self.sigtype not in self.valid_sigtypes:
            errcallback('Invalid sigtype', self.sigtype)
            return False

        if self.sigdir not in self.valid_sigdirs:
            errcallback('Invalid sigdir', self.sigdir)
            return False

        if self.sigconfig not in self.valid_sigconfigs:
            errcallback('Invalid sigconfig', self.sigconfig)
            return False


        return True



class DevInst(wallpy.cleanobject.CleanObject):
    posarg_names = [
        'coords', 'dev_fn',
        'logpackname', 'phypack_fn', 'log2phy_fn'
    ]
    kwarg_names = [
        'bank2pin_reserve_frac',
        'bank2diffpair_reserve_frac',
    ]

    extra_names = [ 'logpack', 'phypack', 'log2phy', 'bank2iostd' ]

    moduleprefix = moduleprefix


    def __init__(
        self,
        coords, dev_fn, logpackname, phypack_fn, log2phy_fn,
        bank2pin_reserve_frac=None,
        bank2diffpair_reserve_frac=None,
    ):

        coords = tuple(coords)

        if bank2pin_reserve_frac is None:
            bank2pin_reserve_frac = collections.defaultdict(float)

        if bank2diffpair_reserve_frac is None:
            bank2diffpair_reserve_frac = collections.defaultdict(float)

        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


        dev = devmanager.load_device(self.dev_fn)
        self.logpack = dev.packages[self.logpackname]
        self.bank2iostd = dev.bank2iostd

        self.phypack = devmanager.load_phypack(self.phypack_fn)

        self.log2phy = devmanager.load_log2phy(self.log2phy_fn)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if len(self.coords) != 4:
            errcallback('Invalid number of dimensions in coords',
                        self.coords)
            return False


        ex_Length = wallpy.units.Length(0, 'mm')
        ex_Angle = wallpy.units.Angle(0, 'deg')

        proper_units = (ex_Length, ex_Length, ex_Length, ex_Angle)


        for i, val in enumerate(self.coords):
            if not proper_units[i].same_measure_type(val):
                errcallback('Invalid coord value', val)
                return False



        # Make sure every logical pin has at least one physical pin
        for logpinkey in self.logpack.pins:
            phypinkeys = self.log2phy[logpinkey]

            if not phypinkeys:
                errcallback('No physical pin key given for logical pin',
                            logpinkey)
                return False

            for k in phypinkeys:
                if k not in self.phypack.pins:
                    errcallback(
                        'Logical pin has no corresponding physical pin(s)',
                        logpinkey, phypinkeys,
                        tuple(self.phypack.pins.keys())
                    )
                    return False


        return True



    def abs_coords(self, relcoords, precision=0, float_units=None):
        """Compute the absolute coordinates from the relative coordinates.

        Arguments:
        relcoords -- tuple(dx, dy, dz), relative to self's coords.
        precision -- round to this many digits after the decimal point
        float_units -- if not None, return tuple of floats in given unit

        Return the absolute position (X, Y, Z), where each coordinate is
            a wallpy.units.Length"""


        relcoords_floatable = []

        for c in relcoords:
            if hasattr(c, 'native_val'):
                if c.same_measure_type(wallpy.units.Length(0, 'mm')):
                    f = float(c.native_val)
                else:
                    raise TypeError('Must be float or wallpy.units.Length!')

            else:
                f = float(c)

                if float_units:
                    f *= wallpy.units.Length(1, float_units).native_val

            relcoords_floatable.append(f)


        dx, dy, dz = (float(c) for c in relcoords_floatable)

        angle_rad = math.pi * (
            self.coords[3] / wallpy.units.Angle(1, 'pi_rad')
        )
        c = math.cos(angle_rad)
        s = math.sin(angle_rad)

        rotcoords = (
            dx * c - dy * s,
            dx * s + dy * c,
            dz,
        )


        if float_units:
            native_mag = wallpy.units.Length(1, float_units).native_val
        else:
            native_mag = 1

        ret_float = tuple(
            (
                round((self.coords[i].native_val + rotcoords[i]), precision)
                / native_mag
            )
            for i in range(3)
        )

        if float_units:
            return ret_float

        return tuple(
            wallpy.units.Length(v, wallpy.units.Length.native_unit)
            for v in ret_float
        )



    def lpk2coords(self, lpk, precision=0, float_units=None):
        """Compute the absolute coordinates of the physical pins
        corresponding to the given logical pin key.

        Arguments:
        lpk -- logical pin key
        precision -- round to this many digits after the decimal point
        float_units -- if not None, return tuple of floats in given unit

        Return a dict: {physical pin key: absolute coordinates}

        """

        phypinkeys = self.log2phy(lpk)

        return {
            ppk: self.abs_coords(self.phypack.pins[ppk].coords,
                                 precision=precision,
                                 float_units=float_units)
            for ppk in phypinkeys
        }





class PhysicalPackage(wallpy.cleanobject.CleanObject):
    posarg_names = [ 'dimensions' ]
    kwarg_names = [ 'pins' ]

    moduleprefix = moduleprefix


    def __init__(self, dimensions, pins = None):

        dimensions = tuple(dimensions)

        if pins:
            pins = wallpy.cleanobject.CleanDict(pins)
        else:
            pins = wallpy.cleanobject.CleanDict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if len(self.dimensions) != 3:
            errcallback('Invalid number of dimensions in dimensions',
                        self.dimensions)
            return False


        ex_Length = wallpy.units.Length(0, 'mm')

        proper_units = (ex_Length, ex_Length, ex_Length)


        for i, val in enumerate(self.dimensions):
            if not proper_units[i].same_measure_type(val):
                errcallback('Invalid dimension value', val)
                return False

            if val <= wallpy.units.Length(0, 'mm'):
                errcallback('Negative dimension value', val)
                return False


        return True



class PhysicalPin(wallpy.cleanobject.CleanObject):
    posarg_names = [ 'coords' ]
    kwarg_names = [ 'dimensions', 'attr' ]

    moduleprefix = moduleprefix


    def __init__(self, coords, dimensions=None, attr=None):

        coords = tuple(coords)

        if dimensions:
            dimensions = tuple(dimensions)
        else:
            dimensions = (wallpy.units.Length(0, 'mm') for i in range(3))

        if attr:
            attr = dict(attr)
        else:
            attr = dict()

        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if len(self.coords) != 3:
            errcallback('Invalid number of dimensions in coords',
                        self.coords)
            return False


        ex_Length = wallpy.units.Length(0, 'mm')

        proper_units = (ex_Length, ex_Length, ex_Length)

        for i, val in enumerate(self.coords):
            if not proper_units[i].same_measure_type(val):
                errcallback('Invalid coord value', val)
                return False



        proper_units = (ex_Length, ex_Length, ex_Length)
        num_pos_dim = 0 # Number of dimensions with positive value

        for i, val in enumerate(self.dimensions):
            if not proper_units[i].same_measure_type(val):
                errcallback('Invalid dimension value', val)
                return False

            if val > wallpy.units.Length(0, 'mm'):
                num_pos_dim += 1
            elif val < wallpy.units.Length(0, 'mm'):
                errcallback('Negative dimension value', val)
                return False


        if num_pos_dim < 2:
            errcallback('Pin must have at least 2 nonzero dimensions',
                        self.dimensions)
            return False


        return True



class IOStandard(wallpy.cleanobject.CleanObject):
    posarg_names = []
    kwarg_names = [ 'vcco', 'vref' ]

    moduleprefix = moduleprefix

    supplynames = ('vcco', 'vref')


    def __init__(self, vcco=None, vref=None):

        if vcco is not None:
            vcco = float(vcco)

        if vref is not None:
            vref = float(vref)

        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


    def supplies_compatible_p(self, supplies):
        """Return True if the given supplies support this I/O standard.

        Arguments:
        supplies: {supplyname: value}

        Valid supply names:
        vcco -- I/O output supply voltage
        vref -- I/O voltage reference

        Return True if this I/O standard is compatible with the given
        supply voltages, or False otherwise.

        """

        if self.vcco is not None and self.vcco != supplies.get('vcco', None):
            return False

        if self.vref is not None and self.vref != supplies.get('vref', None):
            return False

        return True
