#!/usr/bin/env python3

"""Automatically generate components from text."""

import sys

import device



moduleprefix = __name__ + '.'



NOCONN = '<NC>'
REQUIRED = '<REQUIRED>'
DIFF_P = '<DIFF_P>'
DIFF_N = '<DIFF_N>'
DIFF_PAIR = '<DIFF_PAIR>'
BANK = '<BANK>'




def __parse_error(errmsg, linenum, specific = None):
    if specific:
        print(
            'ERR @ ',
            linenum,
            ': ',
            errmsg,
            ' : ',
            repr(specific),
            sep='')

    else:
        print(
            'ERR @ ',
            linenum,
            ': ',
            errmsg,
            sep='')

    sys.exit(1)



def lines2package(lines):

    sigtypes = {
        'd' : device.DIGITAL,
        'a' : device.ANALOG,
        'p' : device.POWER,
    }
    sigdirs = {
        'i' : device.INPUT,
        'o' : device.OUTPUT,
        'b' : device.BIDIR,
    }

    signalspecs = dict()

    for t in sigtypes:
        for d in sigdirs:
            signalspecs[t + d] = (sigtypes[t], sigdirs[d])


    pack = device.LogicalPackage()

    for linenum, L in enumerate(lines, start=1):
        L = L.strip()
        words = list(w.strip('"\'') for w in L.split())

        if not words:
            continue

        pinnum = words[0]
        if pinnum.isdigit():
            pinnum = int(pinnum)

        if pinnum in pack.pins:
            __parse_error('pin already seen', linenum, pinnum)

        pinfuncs = words[1:]


        if NOCONN in pinfuncs:
            if len(pinfuncs) != 1:
                __parse_error('noconn specified, but other functions given', linenum)

            pinfuncs.remove(NOCONN)


        pin = device.LogicalPin()


        for pf in pinfuncs:
            if not pf:
                continue

            try:
                name, info = pf.split('=', 1)
            except ValueError:
                __parse_error('no pin function info given', linenum, name)


            sigconfig = device.SINGLE
            params = dict()
            required_p = False
            signalspec = None

            infobits = info.split(',')
            for i in infobits:
                if i == REQUIRED:
                    required_p = True

                elif i == DIFF_P:
                    sigconfig = device.DIFF_P

                elif i == DIFF_N:
                    sigconfig = device.DIFF_N

                elif i.startswith(DIFF_PAIR):
                    k, v = i.split('=')
                    assert k == DIFF_PAIR

                    params['diff_pair'] = v

                elif i.startswith(BANK):
                    k, v = i.split('=')
                    assert k == BANK

                    try:
                        v = int(v)
                    except ValueError:
                        pass

                    params['bank'] = v

                elif i.startswith('<'):
                    __parse_error('invalid parameter', linenum, i)

                elif i in signalspecs:
                    signalspec = signalspecs[i]

                else:
                    __parse_error('invalid signal spec', linenum, i)

            if not signalspec:
                __parse_error('no signal spec given', linenum, name)

            sigtype, sigdir = signalspec

            # Add the pin function
            pin.pinfuncs[name] = device.PinFunc(sigtype, sigdir, sigconfig,
                                                required=required_p,
                                                params=params)


        # Add the pin
        pack.pins[pinnum] = pin


    if not pack.valid_p(errcallback = print):
        __parse_error('resulting package not valid', 0)

    return pack
