#!/usr/bin/env python3

import os.path



devpaths = [
    'devlib',
]


device_cache = dict()
phypack_cache = dict()
log2phy_cache = dict()
iostandard_cache = dict()


# WARNING: Do not modify loaded objects (or fields, subfields, etc.)!
#   Instead, modify deep copies (copy.deepcopy(...))


def find_file(filename):

    for path in devpaths:
        filename = os.path.join(path, filename)
        if os.path.exists(filename):
            return filename

    raise KeyError('File not found: ' + filename)



def __load_modulevar(cachedict, module_varname,
                     filename, force_reload = False, verbose = False):

    """Load a module by the given name and cache it in cachedict.  Return module_varname from the loaded module.

    Arguments:
    cachedict - cache dictionary {filename : module[module_varname]}
    module_varname - return the module's variable given by this name
    filename - filename (including relative path from devlib) to given module, WITHOUT ending .py
    force_reload - given by user: if asserted, reload the module from file rather than used cached copy
    verbose - given by user: print debugging messages
    """

    if (not force_reload) and (filename in cachedict):
        if verbose:
            print('Using cached', filename)

        return cachedict[filename]

    filename_full = find_file(filename + '.py')

    if filename_full:
        if verbose:
            print('Found', filename, ':', filename_full)

        L = dict(globals())

        exec(compile(open(filename_full).read(), filename_full, 'exec'), L)

        cachedict[filename] = L[module_varname]
        return cachedict[filename]

    return None


def load_device(filename, force_reload = False, verbose = False):
    """Return the Device given by filename (including relative path from devlib; WITHOUT ending .py)"""

    global device_cache

    return __load_modulevar(device_cache, 'dev',
                            filename,
                            force_reload = force_reload,
                            verbose = verbose)


def load_phypack(filename, force_reload = False, verbose = False):
    """Return the PhysicalPackage given by filename (including relative path from devlib/phypack; WITHOUT ending .py)"""

    global phypack_cache

    return __load_modulevar(phypack_cache, 'phypack',
                            os.path.join('phypack', filename),
                            force_reload = force_reload,
                            verbose = verbose)


def load_log2phy(filename, force_reload = False, verbose = False):
    """Return the log2phy map given by filename (including relative path from devlib/log2phy; WITHOUT ending .py)"""

    global log2phy_cache

    return __load_modulevar(log2phy_cache, 'log2phy',
                            os.path.join('log2phy', filename),
                            force_reload = force_reload,
                            verbose = verbose)


def load_iostandard(filename, force_reload=False, verbose=False):
    """Return the I/O standard given by filename (including relative path from devlib/iostandard; WITHOUT ending .py)"""

    global iostandard_cache

    return __load_modulevar(iostandard_cache, 'iostandard',
                            os.path.join('iostandard', filename),
                            force_reload=force_reload,
                            verbose=verbose)


def clear_caches():
    global device_cache
    global phypack_cache
    global log2phy_cache
    global iostandard_cache

    device_cache = dict()
    phypack_cache = dict()
    log2phy_cache = dict()
    iostandard_cache = dict()
