#!/usr/bin/env python3

import math
import cairo


def draw_simple_placement(
    filename,
    D2coords,
    D2dim,
    spacesize,
    imagesize,
    extratext='',
    colors=None
):
    """Draw DevInst placement to PNG image with given size.

    Arguments:
    filename -- write to this filename
    D2coords -- {DIN: (x, y, angle_radians)}
    D2dim -- {DIN: (width, height)}
    spacesize -- (space_width, space_height) in arbitary units
    imagesize -- (image_width, image_height) in pixels
    extratext -- place this text at the bottom-left of the image
    colors -- override default colors from this dict

    Assume DevInsts are scattered on the range:
    [-space_width/2, space_width/2] x [-space_height/2, space_height/2]

    """

    space_width, space_height = spacesize
    image_width, image_height = imagesize

    colors_default = {
        'bg': (1, 1, 1),
        'DI_bg': (0.75, 0.75, 0.75),
        'DI_corner': (0.25, 0.25, 1),
        'DI_text': (0, 0, 0),
        'extratext': (0, 0.5, 0),
    }

    if colors:
        colors_default.update(colors)

    colors = colors_default


    surface = cairo.ImageSurface(
        cairo.FORMAT_RGB24,
        image_width,
        image_height
    )
    ctx = cairo.Context(surface)

    ctx.set_source_rgb(*colors['bg'])
    ctx.paint()

    ctx.select_font_face('Sans')
    ctx.set_font_size(4)

    ctx.save()

    ctx.translate(image_width/2, image_height/2)
    ctx.scale(image_width / space_width, image_height / space_height)

    for DIN, coords in D2coords.items():
        x, y, angle = coords
        w, h = D2dim[DIN][:2]

        ctx.save()
        ctx.translate(x, y)

        ctx.save()
        ctx.rotate(angle)

        ctx.set_source_rgb(*colors['DI_bg'])
        ctx.rectangle(-w/2, -h/2, w, h)
        ctx.fill()

        ctx.set_source_rgb(*colors['DI_corner'])
        radius = min(min(w, h)/5, 1)
        ctx.arc(-w/2 + 1.1*radius, -h/2 + 1.1*radius, radius, 0, 2*math.pi)
        ctx.fill()

        ctx.restore()

        ctx.set_source_rgb(*colors['DI_text'])
        xb, yb, w, h = ctx.text_extents(DIN)[:4]
        ctx.new_path()
        ctx.move_to(-w/2-xb, -h/2-yb)
        ctx.text_path(DIN)
        ctx.fill()

        ctx.restore()

    ctx.restore()

    if extratext:
        ctx.set_source_rgb(*colors['extratext'])
        ctx.set_font_size(16)
        xb, yb, w, h = ctx.text_extents(extratext)[:4]
        ctx.new_path()
        ctx.move_to(1, image_height - h/2)
        ctx.text_path(extratext)
        ctx.fill()

    with open(filename, 'wb') as F:
        surface.write_to_png(F)
