#!/usr/bin/env python3

"""PhysicalPackage generator functions"""

import fractions
import os.path
import itertools

import wallpy.units

import device


ROW_LETTERS = tuple("A B C D E F G H J K L M N P R T U V W Y".split())

def iter_rowname():
    def tup2str(tup):
        return ''.join(tup)

    return map(tup2str, itertools.chain.from_iterable(
        itertools.product(ROW_LETTERS, repeat=i)
        for i in itertools.count(1)
    ))



def gen_qfp(Nx, Ny, P, Bx, By, C, D, Hx, Hy, Hz):
    """Generate a QFP PhysicalPackage given the specified parameters.

    Arguments:
    Nx -- number of pins in the X dimension
    Ny -- number of pins in the Y dimension
    P -- pin pitch
    Bx -- inner distance between pins on opposite sides in the X dimension
    By -- inner distance between pins on opposite sides in the Y dimension
    C -- length of pin pad
    D -- width of pin pad
    Hx -- distance between outermost sides of package in the X dimension
    Hy -- distance between outermost sides of package in the Y dimension
    Hz -- package height (thickness)

    All dimensions (P, B*, C, D, H*) must be represented as
        wallpy.units.Length.

    Return the generated PhysicalPackage."""

    ### Parameters from http://ics.nxp.com/packaging/handbook/ chapter 5



    # Pin dimensions for each side
    pindim_WE = (C, D, wallpy.units.Length(0, 'mm'))
    pindim_NS = (D, C, wallpy.units.Length(0, 'mm'))

    side_pindim = (pindim_WE, pindim_NS, pindim_WE, pindim_NS)



    # Convert X and Y named parameters to tuples over dimensions

    N = (Nx, Ny)
    B = (Bx, By)

    major_coord = tuple( ((B[i] + C)/2) for i in range(2))

    side_major_coord = (
        -major_coord[0],    # W
        major_coord[1],     # S
        major_coord[0],     # E
        -major_coord[1],    # N
    )



    ### Package plan:
    #                               -Y
    #           SIDE N              |
    #         +--------+            |
    #        >|        |            |
    # SIDE W  |        |  SIDE E    |
    #         |        |            |
    #         +--------+            |
    #           SIDE S              |
    #                               +Y
    #   -X <--------------> +X
    #
    # Where > indicates pin 1


    W_range = range(0,          Ny)
    S_range = range(Ny,         Ny+Nx)
    E_range = range(Ny+Nx,      Ny+Nx+Ny)
    N_range = range(Ny+Nx+Ny,   Ny+Nx+Ny+Nx)

    side_range = (W_range, S_range, E_range, N_range)



    # Compute index of "middle pin" on each side
    #   (will be actual pin for odd number of pins on the side,
    #   or between two pins for even number of pins on the side)
    # The math for pin coordinates works very well for this setup.

    side_mid = tuple(fractions.Fraction((2 * sum(SR)) // len(SR), 2) for SR in side_range)


    majmin2xy_W = ((1, 0), (0, 1))
    majmin2xy_S = ((0, 1), (1, 0))
    majmin2xy_E = ((1, 0), (0, -1))
    majmin2xy_N = ((0, 1), (-1, 0))

    side_majmin2xy = (majmin2xy_W, majmin2xy_S, majmin2xy_E, majmin2xy_N)




    ### Generate the pins!
    pins = {}   # key is index + 1 (want one-based counting)

    for side in range(4):
        R = side_range[side]
        middlepin = side_mid[side]
        majmin2xy = side_majmin2xy[side]
        major = side_major_coord[side]
        pindim = side_pindim[side]

        for pinindex in R:
            minor = (pinindex - side_mid[side]) * P

            coords_majmin = (major, minor)
            coords_x = sum(
                (coords_majmin[i] * majmin2xy[i][0] for i in range(2)),
                wallpy.units.Length(0, 'mm')   # Start value for sum
            )
            coords_y = sum(
                (coords_majmin[i] * majmin2xy[i][1] for i in range(2)),
                wallpy.units.Length(0, 'mm')   # Start value for sum
            )

            coords_xyz = (coords_x, coords_y, wallpy.units.Length(0, 'mm'))

            pin = device.PhysicalPin(coords_xyz, dimensions = pindim)

            pins[pinindex + 1] = pin


    return device.PhysicalPackage((Hx, Hy, Hz), pins = pins)






def gen_dualrow(N, P, B, C, D, Hx, Hy, Hz):
    """Generate a dual-row PhysicalPackage given the specified parameters.
    Pins are placed on the N and S sides, with pin 1 at NE corner.

    Arguments:
    N -- total number of pins (must be even)
    P -- pin pitch
    B -- inner distance between pins on opposite sides
    C -- length of pin pad
    D -- width of pin pad
    Hx -- distance between outermost sides of package in the X dimension
    Hy -- distance between outermost sides of package in the Y dimension
    Hz -- package height (thickness)

    All dimensions (P, B, C, D, H*) must be represented as
        wallpy.units.Length.

    Return the generated PhysicalPackage."""

    ### Parameters from http://ics.nxp.com/packaging/handbook/ chapter 5


    if (N % 2) != 0:
        raise ValueError('N must be even')


    # Pin dimensions for each side
    pindim_NS = (D, C, wallpy.units.Length(0, 'mm'))

    side_pindim = (pindim_NS, pindim_NS)



    major_coord = (B + C)/2

    side_major_coord = (
        -major_coord,    # N
        major_coord,     # S
    )



    ### Package plan:
    #                               -Y
    #           SIDE N              |
    #         +-------^+            |
    #         |        |            |
    #         |        |            |
    #         |        |            |
    #         +--------+            |
    #           SIDE S              |
    #                               +Y
    #   -X <--------------> +X
    #
    # Where ^ indicates pin 1


    N_range = range(0,      N // 2)
    S_range = range(N // 2, N)

    side_range = (N_range, S_range)



    # Compute index of "middle pin" on each side
    #   (will be actual pin for odd number of pins on the side,
    #   or between two pins for even number of pins on the side)
    # The math for pin coordinates works very well for this setup.

    side_mid = tuple(fractions.Fraction((2 * sum(SR)) // len(SR), 2) for SR in side_range)


    majmin2xy_N = ((0, 1), (-1, 0))
    majmin2xy_S = ((0, 1), (1, 0))

    side_majmin2xy = (majmin2xy_N, majmin2xy_S)




    ### Generate the pins!
    pins = {}   # key is index + 1 (want one-based counting)

    for side in range(2):
        R = side_range[side]
        middlepin = side_mid[side]
        majmin2xy = side_majmin2xy[side]
        major = side_major_coord[side]
        pindim = side_pindim[side]

        for pinindex in R:
            minor = (pinindex - side_mid[side]) * P

            coords_majmin = (major, minor)
            coords_x = sum(
                (coords_majmin[i] * majmin2xy[i][0] for i in range(2)),
                wallpy.units.Length(0, 'mm')   # Start value for sum
            )
            coords_y = sum(
                (coords_majmin[i] * majmin2xy[i][1] for i in range(2)),
                wallpy.units.Length(0, 'mm')   # Start value for sum
            )

            coords_xyz = (coords_x, coords_y, wallpy.units.Length(0, 'mm'))

            pin = device.PhysicalPin(coords_xyz, dimensions = pindim)

            pins[pinindex + 1] = pin


    return device.PhysicalPackage((Hx, Hy, Hz), pins = pins)



def gen_bga(M, N, P, R, Hx, Hy, Hz):
    """Generate a BGA PhysicalPackage given the specified parameters.

    Arguments:
    M -- number of rows
    N -- number of cols
    P -- ball pitch
    R -- radius of ball pad
    Hx -- distance between outermost sides of package in the X dimension
    Hy -- distance between outermost sides of package in the Y dimension
    Hz -- package height (thickness)

    All dimensions (P, R, H*) must be represented as
        wallpy.units.Length.

    Return the generated PhysicalPackage.

    """

    it_row = iter_rowname()
    rownames = tuple(next(it_row) for i in range(M))


    # Compute index of "middle pin" on each side
    #   (will be actual pin for odd number of pins on the side,
    #   or between two pins for even number of pins on the side)
    # The math for pin coordinates works very well for this setup.

    midpoint_row = fractions.Fraction(M - 1, 2)
    midpoint_col = fractions.Fraction(N - 1, 2)


    pindim = (R, R, wallpy.units.Length(0, 'mm'))
    pin_attr = {'shape': 'circ'}

    ### Generate the pins!
    pins = {}

    for row in range(M):
        coords_y = (row - midpoint_row) * P

        for col in range(N):
            coords_x = (col - midpoint_col) * P

            coords_xyz = (coords_x, coords_y, wallpy.units.Length(0, 'mm'))

            pin = device.PhysicalPin(
                coords_xyz, dimensions=pindim, attr=pin_attr
            )

            # Want 1-based counting for columns
            pinkey = rownames[row] + str(col+1)

            pins[pinkey] = pin

    return device.PhysicalPackage((Hx, Hy, Hz), pins = pins)



