# cython: language_level=3
# cython: profile=True

cimport cython

import math
import itertools
import collections
import signal
import sys
import textwrap

cimport cost
import device

cdef extern from "Python.h":
    int PyErr_CheckSignals()


class ReachedCancelTemp(Exception):
    """Reached cancel_temp and therefore exit."""


cdef float INF = float('inf')


cdef int max_inner_iter = 1000

cdef warn_inner_iter(funcname, specific):
    print('WARNING: max_inner_iter:', funcname, specific)


cdef warn_outer_iter(funcname, specific):
    print('WARNING: max_outer_iter:', funcname, specific)


class FakeDict(object):
    def __init__(self, default=None):
        self.default = default

    def __getitem__(self, key):
        return self.default


cdef struct sa_params:
    float weight_dist
    float weight_boardsize
    float weight_overlap_start
    float weight_iostandard_start
    float weight_pin_start
    float weight_diffpair_start
    float weight_boardoversize_start
    float weight_overlap_increase_factor
    float weight_iostandard_increase_factor
    float weight_pin_increase_factor
    float weight_diffpair_increase_factor
    float weight_boardoversize_increase_factor
    float init_temp
    float cold_temp
    float cancel_temp
    int moves_per_devinst
    int quant_rotate_steps
    float translate_reduce_factor
    float min_move_limit_x
    float min_move_limit_y
    bint use_best_each_step
    int num_move_conn

    bint debug_move_conn
    bint debug_accept_p
    bint debug_majorchoice

    bint show_weight_increase
    bint show_movetypes
    int moves_attempted_progress

    int max_outer_iter
    int max_inner_iter


cdef _param_keys():
    cdef sa_params tmp = FakeDict(0)
    return frozenset(tmp)
param_keys = _param_keys()


cdef struct sa_move_ratios:
    int move_DI
    int move_conn
    int move_MAX

    int DI_translate
    int DI_rotate
    int DI_exchange
    int DI_MAX

    int conn_translate
    int conn_exchange
    int conn_MAX


cdef _move_ratio_keys():
    cdef sa_move_ratios tmp = FakeDict(0)
    ret = set()
    for n in tmp:
        if not n.endswith('_MAX'):
            ret.add(n)
    return frozenset(ret)
move_ratio_keys = _move_ratio_keys()


cdef class SAState(object):
    cdef public float weight_dist
    cdef public float weight_boardsize
    cdef public float weight_overlap
    cdef public float weight_iostandard
    cdef public float weight_pin
    cdef public float weight_diffpair
    cdef public float weight_boardoversize
    cdef public float T
    cdef public float translate_limit_x
    cdef public float translate_limit_y
    cdef public int move_conn_left
    cdef public list moveable_DINs  # [DIN]
    cdef public dict D2movesleft    # {DIN: # of moves left at this temp}
    cdef public list conn_bundlechoices # [bundleindex]
    cdef public dict D2b2validsupplies  # {DIN: {bank: {supplyname: {value}}}}
    cdef public sa_move_ratios move_ratios_cumulative
    cdef public object rand
    cdef public object user_cost_func
    cdef public object user_cost_func_arg


    def __init__(
        self,
        weight_dist=None,
        weight_boardsize=None,
        weight_overlap=None,
        weight_iostandard=None,
        weight_pin=None,
        weight_diffpair=None,
        weight_boardoversize=None,
        T=None,
        translate_limit_x=None,
        translate_limit_y=None,
        move_conn_left=None,
        moveable_DINs=None,
        D2movesleft=None,
        conn_bundlechoices=None,
        D2b2validsupplies=None,
        move_ratios_cumulative=FakeDict(0),
        rand=None,
        user_cost_func=None,
        user_cost_func_arg=None,
    ):
        if weight_dist is None:
            self.weight_dist = 0.0
        else:
            self.weight_dist = weight_dist

        if weight_boardsize is None:
            self.weight_boardsize = 0.0
        else:
            self.weight_boardsize = weight_boardsize

        if weight_overlap is None:
            self.weight_overlap = 0.0
        else:
            self.weight_overlap = weight_overlap

        if weight_iostandard is None:
            self.weight_iostandard = 0.0
        else:
            self.weight_iostandard = weight_iostandard

        if weight_pin is None:
            self.weight_pin = 0.0
        else:
            self.weight_pin = weight_pin

        if weight_diffpair is None:
            self.weight_diffpair = 0.0
        else:
            self.weight_diffpair = weight_diffpair

        if weight_boardoversize is None:
            self.weight_boardoversize = 0.0
        else:
            self.weight_boardoversize = weight_boardoversize

        if T is None:
            self.T = 0.0
        else:
            self.T = T

        if translate_limit_x is None:
            self.translate_limit_x = 0.0
        else:
            self.translate_limit_x = translate_limit_x

        if translate_limit_y is None:
            self.translate_limit_y = 0.0
        else:
            self.translate_limit_y = translate_limit_y

        if move_conn_left is None:
            self.move_conn_left = 0
        else:
            self.move_conn_left = move_conn_left

        if moveable_DINs is None:
            self.moveable_DINs = list()
        else:
            self.moveable_DINs = moveable_DINs

        if D2movesleft is None:
            self.D2movesleft = dict()
        else:
            self.D2movesleft = D2movesleft

        if conn_bundlechoices is None:
            self.conn_bundlechoices = list()
        else:
            self.conn_bundlechoices = conn_bundlechoices

        if D2b2validsupplies is None:
            self.D2b2validsupplies = dict()
        else:
            self.D2b2validsupplies = D2b2validsupplies

        self.move_ratios_cumulative = move_ratios_cumulative

        self.rand = rand

        self.user_cost_func = user_cost_func
        self.user_cost_func_arg = user_cost_func_arg


    cpdef update_penalties(self, PlacementCosts costs, sa_params params):
        """Increase various penalties as needed."""

        if costs.overlap > 0.0:
            self.weight_overlap *= params.weight_overlap_increase_factor

            if params.show_weight_increase:
                print(
                    '\tIncreasing overlap weight to',
                    self.weight_overlap
                )

        if costs.iostandard > 0.0:
            self.weight_iostandard *= \
                    params.weight_iostandard_increase_factor

            if params.show_weight_increase:
                print(
                    '\tIncreasing iostandard weight to',
                    self.weight_iostandard
                )

        if costs.pin > 0.0:
            self.weight_pin *= params.weight_pin_increase_factor

            if params.show_weight_increase:
                print('\tIncreasing pin weight to', self.weight_pin)

        if costs.diffpair > 0.0:
            self.weight_diffpair *= \
                    params.weight_diffpair_increase_factor

            if params.show_weight_increase:
                print(
                    '\tIncreasing diffpair weight to',
                    self.weight_diffpair
                )

        if costs.boardoversize > 0.0:
            self.weight_boardoversize *= \
                    params.weight_boardoversize_increase_factor

            if params.show_weight_increase:
                print(
                    '\tIncreasing boardoversize weight to',
                    self.weight_boardoversize
                )


    cpdef update_limits(self, sa_params params):
        """Reduce maximum translate distances."""

        self.translate_limit_x = max(
            self.translate_limit_x * params.translate_reduce_factor,
            params.min_move_limit_x
        )
        self.translate_limit_y = max(
            self.translate_limit_y * params.translate_reduce_factor,
            params.min_move_limit_y
        )


    cpdef mark_DINs_moved(self, DINs):
        for DIN in DINs:
            try:
                self.D2movesleft[DIN] -= 1
            except KeyError:
                continue

            if self.D2movesleft[DIN] < 1:
                del self.D2movesleft[DIN]
                self.moveable_DINs.remove(DIN)


    cpdef mark_conn_moved(self):
        self.move_conn_left -= 1
        if self.move_conn_left < 0:
            self.move_conn_left = 0


cdef class PlacementCosts(object):
    cdef readonly float dist
    cdef readonly float boardsize
    cdef readonly float overlap
    cdef readonly float iostandard
    cdef readonly float pin
    cdef readonly float diffpair
    cdef readonly float boardoversize

    def __init__(
        self,
        default=0.0,
        dist=None,
        boardsize=None,
        overlap=None,
        iostandard=None,
        pin=None,
        diffpair=None,
        boardoversize=None
    ):
        if dist is None:
            self.dist = default
        else:
            self.dist = dist

        if boardsize is None:
            self.boardsize = default
        else:
            self.boardsize = boardsize

        if overlap is None:
            self.overlap = default
        else:
            self.overlap = overlap

        if iostandard is None:
            self.iostandard = default
        else:
            self.iostandard = iostandard

        if pin is None:
            self.pin = default
        else:
            self.pin = pin

        if diffpair is None:
            self.diffpair = default
        else:
            self.diffpair = diffpair

        if boardoversize is None:
            self.boardoversize = default
        else:
            self.boardoversize = boardoversize


    def __repr__(self):
        return textwrap.dedent(
            '''
            <PlacementCosts(
                dist={dist},
                boardsize={boardsize},
                overlap={overlap},
                iostandard={iostandard},
                pin={pin},
                diffpair={diffpair},
                boardoversize={boardoversize}
            )>
            '''.format(
                dist=self.dist,
                boardsize=self.boardsize,
                overlap=self.overlap,
                iostandard=self.iostandard,
                pin=self.pin,
                diffpair=self.diffpair,
                boardoversize=self.boardoversize
            )).strip()


    def clone(self):
        cdef PlacementCosts ret = PlacementCosts(
            default=None,
            dist=(self.dist),
            boardsize=(self.boardsize),
            overlap=(self.overlap),
            iostandard=(self.iostandard),
            pin=(self.pin),
            diffpair=(self.diffpair),
            boardoversize=(self.boardoversize),
        )

        return ret


    def __add__(x, y):
        cdef PlacementCosts ret = PlacementCosts(
            default=None,
            dist=(x.dist + y.dist),
            boardsize=(x.boardsize + y.boardsize),
            overlap=(x.overlap + y.overlap),
            iostandard=(x.iostandard + y.iostandard),
            pin=(x.pin + y.pin),
            diffpair=(x.diffpair + y.diffpair),
            boardoversize=(x.boardoversize + y.boardoversize)
        )

        return ret


    def __sub__(x, y):
        cdef PlacementCosts ret = PlacementCosts(
            default=None,
            dist=(x.dist - y.dist),
            boardsize=(x.boardsize - y.boardsize),
            overlap=(x.overlap - y.overlap),
            iostandard=(x.iostandard - y.iostandard),
            pin=(x.pin - y.pin),
            diffpair=(x.diffpair - y.diffpair),
            boardoversize=(x.boardoversize - y.boardoversize)
        )

        return ret


    def isfinite(self):
        if not math.isfinite(self.dist): return False
        if not math.isfinite(self.boardsize): return False
        if not math.isfinite(self.overlap): return False
        if not math.isfinite(self.iostandard): return False
        if not math.isfinite(self.pin): return False
        if not math.isfinite(self.diffpair): return False
        if not math.isfinite(self.boardoversize): return False

        return True


    def valid_p(self):
        if not self.isfinite(): return False

        if self.dist < 0.0: return False
        if self.boardsize < 0.0: return False
        if self.overlap < 0.0: return False
        if self.iostandard < 0.0: return False
        if self.pin < 0.0: return False
        if self.diffpair < 0.0: return False
        if self.boardoversize < 0.0: return False

        return True


cpdef sa_move_ratios accumulate_move_ratios(sa_move_ratios ratios):
    """Create cumulative distrivution of move ratios."""

    cdef int acc

    ### Create cumulative distribution of move ratios.
    acc = 0

    ratios.move_DI += acc
    acc = ratios.move_DI

    ratios.move_conn += acc
    acc = ratios.move_conn

    ratios.move_MAX = acc


    acc = 0

    ratios.DI_translate += acc
    acc = ratios.DI_translate

    ratios.DI_rotate += acc
    acc = ratios.DI_rotate

    ratios.DI_exchange += acc
    acc = ratios.DI_exchange

    ratios.DI_MAX = acc


    acc = 0

    ratios.conn_translate += acc
    acc = ratios.conn_translate

    ratios.conn_exchange += acc
    acc = ratios.conn_exchange

    ratios.conn_MAX = acc


    return ratios


def empty_k2int():
    return collections.defaultdict(int)


def empty_k2k2int():
    return collections.defaultdict(empty_k2int)


def empty_k2float():
    return collections.defaultdict(float)


def empty_k2k2float():
    return collections.defaultdict(empty_k2float)


cdef class PlacementInfo(object):
    cdef public dict D2bundles  # {DIN: {bundle indices}}
    cdef public dict D2b2maxpins    # {DIN: {bank: max # of pins}}
    cdef public dict D2b2maxdiffpairs   # {DIN: {bank: max # of diff. pairs}}
    cdef public cost.xy boardmaxsize    # Board size limit
    cdef public dict D2dim  # {DIN: (width, height)}

    def __init__(
        self,
        D2bundles=None,
        D2b2maxpins=None,
        D2b2maxdiffpairs=None,
        boardmaxsize=FakeDict(INF),
        D2dim=None,
    ):
        if D2bundles is None:
            D2bundles = dict()

        if D2b2maxpins is None:
            D2b2maxpins = dict()

        if D2b2maxdiffpairs is None:
            D2b2maxdiffpairs = dict()

        if D2dim is None:
            D2dim = dict()

        self.D2bundles = D2bundles
        self.D2b2maxpins = D2b2maxpins
        self.D2b2maxdiffpairs = D2b2maxdiffpairs
        self.boardmaxsize = boardmaxsize
        self.D2dim = D2dim


cdef class PlacedBundleInfo(object):
    cdef public bint differential
    cdef public object iostd
    cdef public float weight
    cdef public dict D2b2CoM    # {DIN: {bank: (dx, dy)}}
    cdef public dict D2b2numpins    # {DIN: {bank: # of pins}}
    cdef public dict D2b2numdiffpairs   # {DIN: {bank: # of diff. pairs}}
    cdef public list conn_netchoices    # [netindex], automatically filled
    cdef public str bundlename  # For user-friendliness only.

    def __init__(
        self,
        differential=False,
        iostd=None,
        weight=1.0,
        D2b2CoM=None,
        D2b2numpins=None,
        D2b2numdiffpairs=None,
        bundlename='',
    ):
        if D2b2CoM is None:
            D2b2CoM = dict()

        if D2b2numpins is None:
            D2b2numpins = dict()

        if D2b2numdiffpairs is None:
            D2b2numdiffpairs = dict()

        self.differential = differential
        self.iostd = iostd
        self.weight = weight
        self.D2b2CoM = D2b2CoM
        self.D2b2numpins = D2b2numpins
        self.D2b2numdiffpairs = D2b2numdiffpairs
        self.bundlename = bundlename


cdef class Placement(object):
    cdef public list bundles    # [PlacedBundle]
    cdef public dict D2coords   # {DIN: (x, y, abs)}
    cdef public object D2b2supplies   # {DIN: {bank: {supplyname: value}}}
    cdef readonly PlacementCosts costs
    cdef readonly float total_cost
    cdef public PlacementInfo info
    cdef public object Db2supply2val2numnets
    cdef public list terminals_supply_moved # NOT COPIED ON CLONE
    cdef public object movetype # NOT COPIED ON CLONE
    cdef public object debug_move   # NOT COPIED ON CLONE

    def __init__(
        self,
        bundles=None,
        D2coords=None,
        D2b2supplies=None,
        costs=None,
        total_cost=INF,
        info=None,
        Db2supply2val2numnets=None,
    ):
        if bundles is None:
            bundles = list()

        if D2coords is None:
            D2coords = dict()

        if D2b2supplies is None:
            D2b2supplies = collections.defaultdict(empty_k2k2float)

        if costs is None:
            costs = PlacementCosts(default=INF)

        if Db2supply2val2numnets is None:
            Db2supply2val2numnets = collections.defaultdict(empty_k2k2int)

        self.bundles = bundles
        self.D2coords = D2coords
        self.D2b2supplies = D2b2supplies
        self.costs = costs
        self.total_cost = total_cost
        self.info = info
        self.Db2supply2val2numnets = Db2supply2val2numnets

        self.terminals_supply_moved = list()

        self.movetype = None
        self.debug_move = None


    def clone(self):
        cdef Placement newplace
        cdef PlacedBundle B

        newplace = Placement(
            D2coords=dict(self.D2coords),
            D2b2supplies={
                DIN: {
                    bank: {
                        supplyname: supplyval
                        for supplyname, supplyval in supplies.items()
                    }
                    for bank, supplies in b2supplies.items()
                }
                for DIN, b2supplies in self.D2b2supplies.items()
            },
            costs=self.costs.clone(),
            total_cost=self.total_cost,
            info=self.info,
            Db2supply2val2numnets=collections.defaultdict(empty_k2k2int, {
                Db: collections.defaultdict(empty_k2int, {
                    supply: collections.defaultdict(int, {
                        val: numnets
                        for val, numnets in val2numnets.items()
                    })
                    for supply, val2numnets in supply2val2numnets.items()
                })
                for Db, supply2val2numnets in self.Db2supply2val2numnets.items()
            }),
        )

        newplace.bundles = list(
            B.clone(new_place=newplace)
            for B in self.bundles
        )

        return newplace


    def debug_nulls(self):
        cdef int bundleindex
        cdef PlacedBundle bundle
        cdef int netindex
        cdef PlacedNet net

        sep = '*' * 75


        print(sep)
        print(sep)
        print('DEBUG NULLS BEGIN')
        print()

        for bundleindex, bundle in enumerate(self.bundles):
            if bundle.info is None:
                bundlename = '???? <NULL INFO> ????'
            else:
                bundlename = bundle.info.bundlename

            print('Bundle[{}] "{}" BEGIN'.format(bundleindex, bundlename))

            print('\tinfo', 'NULL' if bundle.info is None else 'OK')
            print('\tplace', 'NULL' if bundle.place is None else 'OK')

            for netindex, net in enumerate(bundle.nets):
                print('\tNet[{}] BEGIN'.format(netindex))

                print('\t\tplace', 'NULL' if net.place is None else 'OK')
                print('\t\tbundle', 'NULL' if net.bundle is None else 'OK')

                print('\tNet[{}] END'.format(netindex))
                print()

            print('Bundle[{}] "{}" END'.format(bundleindex, bundlename))
            print()
            print(sep)
            print()

        print('DEBUG NULLS END')
        print()
        print(sep)
        print(sep)


    def export(self, file, sep=('&'*75)):

        print(file=file)
        print(file=file)
        print(sep, file=file)
        print(file=file)
        print(file=file)

        print('D2coords:')
        for DIN, coords in sorted(self.D2coords.items()):
            print('\t{}:'.format(DIN), coords, file=file)
        print(file=file)
        print(file=file)

        print('Nets:')
        for bundle in self.bundles:
            bundlename = bundle.info.bundlename
            print('\t{!r}:'.format(bundlename), file=file)

            for net in bundle.nets:
                terms = ['{}.{}'.format(*term) for term in net.terminals]
                print('\t\t{}'.format(', '.join(terms)), file=file)
            print(file=file)
        print(file=file)

        print('Supplies:')
        for DIN, b2supplies in sorted(self.D2b2supplies.items()):
            for bank in sorted(b2supplies, key=banksort_key):
                supplies = b2supplies[bank]

                kv = list()
                for k, v in sorted(supplies.items()):
                    kv.append('{} = {}'.format(k, v))

                if kv:
                    print(
                        '\t{}.{}: {}'.format(DIN, bank, ', '.join(kv)),
                        file=file
                    )

        print(file=file)

        print(file=file)
        print(file=file)
        print(sep)
        print(file=file)
        print(file=file)


    def update(self):
        """Force update of internal data structures.

        This is an expensive operation and is intended for:
            - initialization of data structures as part of setup
            - debugging automatic updates of data structures

        """

        cdef PlacedBundle bundle
        cdef dict req_supplies
        cdef PlacedNet net

        D2bundles = collections.defaultdict(set)

        for index, bundle in enumerate(self.bundles):
            bundle.update()

            if bundle.info.iostd is not None:
                req_supplies = dict()

                for k in bundle.info.iostd.kwarg_names:
                    v = getattr(bundle.info.iostd, k, None)

                    if v is None:
                        continue

                    if v != float(v):
                        raise TypeError(
                            'Expect supply value to be float',
                            (
                                bundle.info.bundlename,
                                k,
                                v,
                            )
                        )

                    req_supplies[k] = float(v)

                if req_supplies:
                    for net in bundle.nets:
                        for term in net.terminals:
                            for k, v in req_supplies.items():
                                self.Db2supply2val2numnets[term][k][v] += 1

            for DIN in bundle.info.D2b2numpins:
                D2bundles[DIN].add(index)

        self.info.D2bundles = dict(D2bundles)

        self.mark_supplies_recalc(self.Db2supply2val2numnets)
        self.recalc_supplies()

        ### Force re-calculate costs.
        self.calc_costs()


    cpdef mark_supplies_recalc(self, term_iter):
        """Indicate that supplies must be recalculated for the terminals
        given in term_iter.

        """

        self.terminals_supply_moved.extend(term_iter)


    cpdef recalc_supplies(self):
        """Recalculate majority vote for supply values.

        Recalculate for terminals listed in terminals_supply_moved.

        """

        cdef set recalc_DINs = set()    # Recompute costs for these DINs.

        for term in self.terminals_supply_moved:
            if term not in self.Db2supply2val2numnets:
                continue

            DIN, bank = term
            assigned_supplies = self.D2b2supplies[DIN][bank]

            supply2val2numnets = self.Db2supply2val2numnets[term]

            for supply, val2numnets in supply2val2numnets.items():
                if val2numnets:
                    val, nn = max(val2numnets.items(), key=valkeysort_key)

                    if nn > 0:
                        chosen = val
                    else:
                        chosen = None

                else:
                    nn = 0
                    chosen = None

                if assigned_supplies[supply] != chosen:
                    print(
                        '\tDDDD Set',
                        term,
                        supply,
                        'from',
                        assigned_supplies[supply],
                        'to',
                        chosen,
                        '(now has {} net{})'.format(
                            nn,
                            '' if nn == 1 else 's',
                        ),
                    )

                    assigned_supplies[supply] = chosen
                    assert(chosen == self.D2b2supplies[DIN][bank][supply])

                    recalc_DINs.add(DIN)

        for DIN in recalc_DINs:
            self.moved_supply(DIN)

        del self.terminals_supply_moved[:]


    cpdef calc_costs(self):
        self.calc_cost_bundle_pin_diffpair()
        self.calc_cost_boardsize_boardoversize()
        self.calc_cost_overlap()

        ### Ensure valid costs.
        assert self.costs.isfinite()


    cpdef calc_cost_bundle_pin_diffpair(self):
        cdef PlacedBundle bundle
        cdef int numconn
        cdef int maxconn

        self.costs.dist = 0.0
        self.costs.iostandard = 0.0
        self.costs.pin = 0.0
        self.costs.diffpair = 0.0


        ### Sum Bundles' contributions to costs.
        for bundle in self.bundles:
            if not bundle.costs.isfinite():
                bundle.calc_costs()

            self.costs += bundle.costs


        ### Update pin, diffpair costs with overall excesses.
        D2b2totalpins = collections.defaultdict(empty_k2int)
        D2b2totaldiffpairs = collections.defaultdict(empty_k2int)

        for bundle in self.bundles:
            for DIN, b2numconn in bundle.D2b2numconn.items():
                b2totalpins = D2b2totalpins[DIN]
                b2totaldiffpairs = D2b2totaldiffpairs[DIN]

                for bank, numconn in b2numconn.items():
                    if bundle.info.differential:
                        b2totaldiffpairs[bank] += numconn
                        numconn *= 2

                    b2totalpins[bank] += numconn

        for DIN, b2totalpins in D2b2totalpins.items():
            b2maxpins = self.info.D2b2maxpins[DIN]

            for bank, numconn in b2totalpins.items():
                maxconn = b2maxpins[bank]

                if numconn > maxconn:
                    self.costs.pin += (numconn - maxconn)

        for DIN, b2totaldiffpairs in D2b2totaldiffpairs.items():
            b2maxdiffpairs = self.info.D2b2maxdiffpairs[DIN]

            for bank, numconn in b2totaldiffpairs.items():
                maxconn = b2maxdiffpairs[bank]

                if numconn > maxconn:
                    self.costs.diffpair += (numconn - maxconn)


    cpdef calc_cost_boardsize_boardoversize(self):
        cdef float x, y, angle
        cdef float width, height
        cdef float min_x = INF, min_y = INF
        cdef float max_x = -INF, max_y = -INF
        cdef cost.xy boardsize = FakeDict(0.0)
        cdef cost.xy boardoversize = FakeDict(0.0)

        self.costs.boardsize = 0.0
        self.costs.boardoversize = 0.0

        for DIN, coords in self.D2coords.items():
            x, y, angle = coords
            width, height = self.info.D2dim[DIN]
            width, height = cost.bbox_size(width, height, angle)

            min_x = min(min_x, x - width/2)
            min_y = min(min_y, y - height/2)
            max_x = max(max_x, x + width/2)
            max_y = max(max_y, y + height/2)

        boardsize.x = max_x - min_x
        boardsize.y = max_y - min_y

        self.costs.boardsize = boardsize.x * boardsize.y

        boardoversize.x = boardsize.x - self.info.boardmaxsize.x
        boardoversize.y = boardsize.y - self.info.boardmaxsize.y

        if boardoversize.x > 0.0:
            self.costs.boardoversize += boardoversize.x
        if boardoversize.y > 0.0:
            self.costs.boardoversize += boardoversize.y


    cpdef calc_cost_overlap(self):
        self.costs.overlap = 0.0

        for L_DIN, R_DIN in itertools.combinations(self.D2coords, 2):
            self.costs.overlap += cost.overlap_penalty(
                L_DIN,
                R_DIN,
                self.D2coords,
                self.info.D2dim
            )


    cpdef moved_supply(self, DIN):
        cdef int index
        cdef PlacedBundle bundle

        for index in self.info.D2bundles.get(DIN, tuple()):
            bundle = self.bundles[index]

            ### Update costs contributed by this bundle.
            self.costs -= bundle.costs
            bundle.moved_supply(DIN)
            self.costs += bundle.costs


    cpdef moved_DINs(self, DINs):
        cdef set indices = set()
        cdef int index
        cdef PlacedBundle bundle

        for DIN in DINs:
            indices.update(self.info.D2bundles.get(DIN, tuple()))

        for index in indices:
            bundle = self.bundles[index]

            ### Update costs contributed by this bundle.
            self.costs -= bundle.costs
            bundle.moved_DINs(DINs)
            self.costs += bundle.costs

        self.calc_cost_boardsize_boardoversize()
        self.calc_cost_overlap()


    cpdef moved_terminal(self, int bundleindex, int netindex, int termindex):
        cdef PlacedBundle bundle = self.bundles[bundleindex]

        bundle.moved_terminal(netindex, termindex)

        self.moved_terminal_postbundle()


    cpdef moved_terminal_postbundle(self):
        """Update costs after bundle's own costs have been updated."""

        self.calc_cost_bundle_pin_diffpair()


    cpdef bint move(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Modify a single aspect of this Placement, one of:
            - DevInst location
            - Net terminal

        Return True if successful, or False if no move was successfully
        completed.

        """

        cdef int outer_iter
        cdef int majorchoice


        for outer_iter from 0 <= outer_iter < params.max_outer_iter:
            if not (
                state.moveable_DINs or
                state.move_conn_left
            ):
                # No more possible moves left at this temperature!
                return False


            majorchoice = state.rand.randrange(
                state.move_ratios_cumulative.move_MAX
            )


            if majorchoice < state.move_ratios_cumulative.move_DI:
                if params.debug_majorchoice:
                    print('\t\tDDDD CHOICE @{} {}'.format(
                        outer_iter,
                        'DI',
                    ))

                if self.move_DI(params, state):
                    return True

            elif majorchoice < state.move_ratios_cumulative.move_conn:
                if params.debug_majorchoice:
                    print('\t\tDDDD CHOICE @{} {}'.format(
                        outer_iter,
                        'conn',
                    ))

                if self.move_conn(params, state):
                    return True

            else:
                raise ValueError('Invalid movement choice', majorchoice)

        if outer_iter >= params.max_outer_iter:
            warn_outer_iter('Placement.move', outer_iter)

        return False


    cpdef bint move_DI(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Move a single DevInst (or two if exchanging positions).

        Return True if successful, or False if no DevInst was moved.

        """

        cdef int minorchoice

        if not state.moveable_DINs:
            return False

        minorchoice = state.rand.randrange(
            state.move_ratios_cumulative.DI_MAX
        )


        if minorchoice < state.move_ratios_cumulative.DI_translate:
            if self.move_DI_translate(params, state):
                return True

        elif minorchoice < state.move_ratios_cumulative.DI_rotate:
            if self.move_DI_rotate(params, state):
                return True

        elif minorchoice < state.move_ratios_cumulative.DI_exchange:
            if self.move_DI_exchange(params, state):
                return True

        else:
            raise ValueError('Invalid movement choice', minorchoice)

        return False


    cpdef bint move_conn(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Move terminal(s) of one net within one Bundle.

        Return True if successful, or False if no terminal was moved.

        """

        cdef int minorchoice

        if state.move_conn_left < 1:
            return False

        if not state.conn_bundlechoices:
            return False

        minorchoice = state.rand.randrange(
            state.move_ratios_cumulative.conn_MAX
        )


        if minorchoice < state.move_ratios_cumulative.conn_translate:
            if self.move_conn_translate(params, state):
                return True

        elif minorchoice < state.move_ratios_cumulative.conn_exchange:
            if self.move_conn_exchange(params, state):
                return True

        else:
            raise ValueError('Invalid movement choice', minorchoice)

        return False


    cpdef bint move_DI_translate(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Translate one DevInst.

        Return True if successful, or False if no DevInst was moved.

        """

        cdef float x, y, angle
        cdef float size_x, size_y
        cdef float min_x, min_y
        cdef float max_x, max_y

        try:
            DIN = state.rand.choice(state.moveable_DINs)
        except (ValueError, IndexError):
            return False

        x, y, angle = self.D2coords[DIN]
        size_x, size_y = self.info.D2dim[DIN]
        size_x, size_y = cost.bbox_size(size_x, size_y, angle)

        min_x = max(
            x - state.translate_limit_x,
            (size_x - self.info.boardmaxsize.x) / 2,
        )
        min_y = max(
            y - state.translate_limit_y,
            (size_y - self.info.boardmaxsize.y) / 2,
        )

        max_x = min(
            x + state.translate_limit_x,
            (self.info.boardmaxsize.x - size_x) / 2,
        )
        max_y = min(
            y + state.translate_limit_y,
            (self.info.boardmaxsize.y - size_y) / 2,
        )

        x = state.rand.uniform(min_x, max_x)
        y = state.rand.uniform(min_y, max_y)

        self.D2coords[DIN] = (x, y, angle)

        self.movetype = 'DI_translate'
        self.debug_move = '{} ({:0.2f},{:0.2f})'.format(
            DIN,
            x,
            y,
        )

        DINs = [DIN]
        self.moved_DINs(DINs)
        state.mark_DINs_moved(DINs)

        return True


    cpdef bint move_DI_rotate(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Rotate one DevInst.

        Return True if successful, or False if no DevInst was moved.

        """

        cdef float x, y, angle
        cdef float angle_quant = 2 * math.pi / params.quant_rotate_steps
        cdef float new_angle
        cdef int inner_iter

        if params.quant_rotate_steps < 2:
            return False

        try:
            DIN = state.rand.choice(state.moveable_DINs)
        except (ValueError, IndexError):
            return False

        x, y, angle = self.D2coords[DIN]


        for inner_iter from 0 <= inner_iter < max_inner_iter:
            new_angle = (
                angle_quant *
                state.rand.randrange(params.quant_rotate_steps)
            )

            if abs(new_angle - angle) > 0.01:
                break

        if inner_iter >= max_inner_iter:
            warn_inner_iter('Placement.move_DI_rotate', inner_iter)
            return False

        self.D2coords[DIN] = (x, y, new_angle)

        self.movetype = 'DI_rotate'
        self.debug_move = '{} ({:0.2f} -> {:0.2f})'.format(
            DIN,
            angle,
            new_angle,
        )

        DINs = [DIN]
        self.moved_DINs(DINs)
        state.mark_DINs_moved(DINs)

        return True


    cpdef bint move_DI_exchange(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Exchange two DevInsts.

        Return True if successful, or False if no DevInst was moved.

        """

        try:
            DINs = state.rand.sample(state.moveable_DINs, 2)
        except ValueError:
            return False

        coordlist = [self.D2coords[DIN] for DIN in DINs]
        coordlist.reverse()

        for DIN, coords in zip(DINs, coordlist):
            self.D2coords[DIN] = coords

        self.movetype = 'DI_exchange'
        self.debug_move = '{} <-> {}'.format(*DINs)

        self.moved_DINs(DINs)
        state.mark_DINs_moved(DINs)

        return True


    cpdef bint move_conn_translate(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Move one terminal of one net within one Bundle.

        Return True if successful, or False if no net was changed.

        """

        cdef int inner_iter
        cdef int bundleindex
        cdef PlacedBundle bundle

        for inner_iter from 0 <= inner_iter < max_inner_iter:
            try:
                bundleindex = state.rand.choice(state.conn_bundlechoices)
            except (ValueError, IndexError):
                return False

            bundle = self.bundles[bundleindex]

            if bundle.move_conn_translate(params, state):
                self.movetype = 'conn_translate'
                self.debug_move = bundle.info.bundlename

                self.moved_terminal_postbundle()
                state.mark_conn_moved()

                self.recalc_supplies()

                return True

        warn_inner_iter('Placement.move_conn_translate', inner_iter)

        return False


    cpdef bint move_conn_exchange(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Exchange one terminal from each of two nets within a Bundle,
        with each terminal being connected to different banks of the
        same DevInst.

        Return True if successful, or False if no net was changed.

        """

        cdef int inner_iter
        cdef int bundleindex
        cdef PlacedBundle bundle

        for inner_iter from 0 <= inner_iter < max_inner_iter:
            try:
                bundleindex = state.rand.choice(state.conn_bundlechoices)
            except (ValueError, IndexError):
                return False

            bundle = self.bundles[bundleindex]

            if bundle.move_conn_exchange(params, state):
                self.movetype = 'conn_exchange'
                self.debug_move = bundle.info.bundlename

                self.moved_terminal_postbundle()
                state.mark_conn_moved()
                return True

        # No warning for exceeding max_inner_iter, as the required
        # conditions may not be present for this movement type.

        return False


    cpdef debug_pin_use(self):
        cdef PlacedBundle bundle
        cdef int numconn
        cdef int maxconn
        cdef int pincost = 0
        cdef int diffcost = 0

        print()

        ### Compute pin, diffpair overall excesses.
        D2b2totalpins = collections.defaultdict(empty_k2int)
        D2b2totaldiffpairs = collections.defaultdict(empty_k2int)

        for bundle in self.bundles:
            for DIN, b2numconn in bundle.D2b2numconn.items():
                b2totalpins = D2b2totalpins[DIN]
                b2totaldiffpairs = D2b2totaldiffpairs[DIN]

                for bank, numconn in b2numconn.items():
                    if bundle.info.differential:
                        b2totaldiffpairs[bank] += numconn
                        numconn *= 2

                    b2totalpins[bank] += numconn

        print()
        print('\tDDDD Bank pin availability and use:')
        for DIN, b2totalpins in sorted(D2b2totalpins.items(), key=str):
            b2maxpins = self.info.D2b2maxpins[DIN]

            for bank in sorted(b2totalpins, key=banksort_key):
                numconn = b2totalpins[bank]
                maxconn = b2maxpins[bank]

                if numconn > maxconn:
                    pincost += (numconn - maxconn)

                print('\t\t({}.{}): {} {} ({})'.format(
                    DIN,
                    bank,
                    numconn,
                    maxconn,
                    max(numconn - maxconn, 0),
                ))

        print()
        print('\tDDDD Bank diffpair availability and use:')
        for DIN, b2totaldiffpairs in sorted(
            D2b2totaldiffpairs.items(),
            key=str
        ):
            b2maxdiffpairs = self.info.D2b2maxdiffpairs[DIN]

            for bank in sorted(b2totaldiffpairs, key=banksort_key):
                numconn = b2totaldiffpairs[bank]
                maxconn = b2maxdiffpairs[bank]

                if numconn > maxconn:
                    diffcost += (numconn - maxconn)

                print('\t\t({}.{}): {} {} ({})'.format(
                    DIN,
                    bank,
                    numconn,
                    maxconn,
                    max(numconn - maxconn, 0),
                ))

        print('\tDDDD Final penalties: {:0.3g}, {:0.3g}'.format(
            pincost,
            diffcost,
        ))
        print()


cdef class PlacedBundle(object):
    cdef public list nets   # [PlacedNet]
    cdef public object D2nets # {DIN: {net indices}}
    cdef public object D2b2numconn    # {DIN: {bank: # of conn. in Bundle}}
    cdef public PlacementCosts costs
    cdef public float total_cost
    cdef public PlacedBundleInfo info
    cdef public Placement place # Refers to containing Placement

    def __init__(
        self,
        nets=None,
        D2nets=None,
        D2b2numconn=None,
        costs=None,
        total_cost=INF,
        info=None,
        place=None
    ):
        if nets is None:
            nets = list()

        if D2nets is None:
            D2nets = collections.defaultdict(set)

        if D2b2numconn is None:
            D2b2numconn = collections.defaultdict(empty_k2int)

        if costs is None:
            costs = PlacementCosts(default=INF)

        self.nets = nets
        self.D2nets = D2nets
        self.D2b2numconn = D2b2numconn
        self.costs = costs
        self.total_cost = total_cost
        self.info = info
        self.place = place


    def clone(self, Placement new_place=None):
        cdef PlacedBundle newbundle
        cdef PlacedNet net

        newbundle = PlacedBundle(
            D2nets=dict(self.D2nets),
            D2b2numconn={
                DIN: {
                    bank: numconn
                    for bank, numconn in b2numconn.items()
                }
                for DIN, b2numconn in self.D2b2numconn.items()
            },
            costs=self.costs.clone(),
            total_cost=self.total_cost,
            info=self.info,
            place=new_place,
        )

        newbundle.nets=list(
            net.clone(new_place=new_place, new_bundle=newbundle)
            for net in self.nets
        )

        return newbundle


    def update(self):
        """Force update of internal data structures.

        This is an expensive operation and is intended for:
            - initialization of data structures as part of setup
            - debugging automatic updates of data structures

        """

        cdef PlacedNet net

        D2nets = collections.defaultdict(set)
        D2b2numconn = {
            DIN: {
                bank: 0
                for bank in b2numpins
            }
            for DIN, b2numpins in self.info.D2b2numpins.items()
        }

        for index, net in enumerate(self.nets):
            net.update()

            for DIN, bank in net.terminals:
                D2nets[DIN].add(index)
                D2b2numconn[DIN][bank] += 1

        self.D2nets = dict(D2nets)
        self.D2b2numconn = D2b2numconn

        ### Force re-calculate costs.
        self.calc_costs()


        ### Populate info.conn_netchoices.
        conn_netchoices = set()

        # Accept any net connecting to a DevInst with multiple banks.
        for DIN, indices in self.D2nets.items():
            if len(self.D2b2numconn[DIN]) > 1:
                conn_netchoices.update(indices)

        self.info.conn_netchoices = sorted(conn_netchoices)


    cpdef calc_costs(self):
        cdef PlacedNet net

        self.costs.dist = 0.0
        self.costs.boardsize = 0.0
        self.costs.overlap = 0.0
        self.costs.boardoversize = 0.0


        for net in self.nets:
            if net.total_dist == INF:
                net.calc_total_dist()

            self.costs.dist += net.total_dist

        self.costs.dist *= self.info.weight

        self.calc_cost_iostandard()

        self.calc_cost_pin_diffpair()


    cpdef calc_cost_pin_diffpair(self):
        cdef int numconn
        cdef int numpins
        cdef int numdiffpair
        cdef int excess

        self.costs.pin = 0.0
        self.costs.diffpair = 0.0

        for DIN, b2numconn in self.D2b2numconn.items():
            b2numpins = self.info.D2b2numpins[DIN]
            b2numdiffpair = self.info.D2b2numdiffpairs[DIN]

            for bank, numconn in b2numconn.items():
                if numconn < 1:
                    continue

                numpins = b2numpins[bank]
                numdiffpair = b2numdiffpair[bank]

                if self.info.differential:
                    excess = numconn - numdiffpair
                    if excess > 0:
                        self.costs.diffpair += excess

                    numconn *= 2

                excess = numconn - numpins
                if excess > 0:
                    self.costs.pin += excess


    cpdef calc_cost_iostandard(self):
        cdef int numconn

        assert(self.place is not None)

        self.costs.iostandard = 0.0

        for DIN, b2numconn in self.D2b2numconn.items():
            b2supplies = self.place.D2b2supplies[DIN]

            for bank, numconn in b2numconn.items():
                if numconn < 1:
                    continue

                try:
                    supplies = b2supplies[bank]
                except KeyError:
                    continue

                if self.info.iostd is not None:
                    if not supplies:
                        # Ignore this DevInst if no supplies.
                        continue

                    if not self.info.iostd.supplies_compatible_p(supplies):
                        self.costs.iostandard += numconn


    cpdef moved_supply(self, DIN):
        self.calc_cost_iostandard()


    cpdef moved_DINs(self, DINs):
        cdef str DIN
        cdef set indices = set()
        cdef int index
        cdef PlacedNet net

        for DIN in DINs:
            indices.update(self.D2nets.get(DIN, set()))

        if not indices:
            return

        for index in indices:
            ### Update distance cost contributed by this net.
            net = self.nets[index]

            self.costs.dist -= net.total_dist * self.info.weight
            net.moved_DINs(DINs, True)
            self.costs.dist += net.total_dist * self.info.weight


    cpdef moved_terminal(
        self,
        int netindex,
        int termindex,
        bint update_costs=True
    ):

        cdef PlacedNet net = self.nets[netindex]

        ### Update distance cost contributed by this net.
        self.costs.dist -= net.total_dist * self.info.weight
        net.moved_terminal(termindex, True)
        self.costs.dist += net.total_dist * self.info.weight

        if update_costs:
            self.moved_terminal_postnet()


    cpdef moved_terminal_postnet(self):
        """Update costs after net's own costs have been updated."""

        self.calc_cost_pin_diffpair()

        # Terminal may have moved to a new bank.
        self.calc_cost_iostandard()


    cpdef bint move_conn_translate(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Move one terminal of one net.

        Return True if successful, or False if no net was changed.

        """

        cdef int netindex
        cdef PlacedNet net
        cdef int inner_iter
        cdef int termindex
        cdef bint recalc_supplies

        try:
            netindex = state.rand.choice(self.info.conn_netchoices)
        except (ValueError, IndexError):
            return False

        net = self.nets[netindex]

        ### Choose a terminal.
        for inner_iter from 0 <= inner_iter < max_inner_iter:
            termindex = state.rand.randrange(len(net.terminals))
            DIN, bank = net.terminals[termindex]
            bank_choices = self.D2b2numconn[DIN]

            if len(bank_choices) > 1:
                break

        if inner_iter >= max_inner_iter:
            warn_inner_iter(
                'PlacedBundle.move_conn_translate <term>',
                inner_iter
            )
            return False

        ### Choose a new bank.
        bank_choices = sorted(bank_choices, key=banksort_key)
        bank_choices.remove(bank)

        new_bank = state.rand.choice(bank_choices)
        assert(new_bank != bank)

        net.terminals[termindex] = (DIN, new_bank)

        self.D2b2numconn[DIN][bank] -= 1
        self.D2b2numconn[DIN][new_bank] += 1

        if self.info.iostd is not None:
            recalc_supplies = False

            for k in self.info.iostd.kwarg_names:
                v = getattr(self.info.iostd, k, None)

                if v is None:
                    continue

                recalc_supplies = True

                self.place.Db2supply2val2numnets[(DIN, bank)][k][v] -= 1
                self.place.Db2supply2val2numnets[(DIN, new_bank)][k][v] += 1

            if recalc_supplies:
                affected_terminals = ((DIN, bank), (DIN, new_bank))
                self.place.mark_supplies_recalc(affected_terminals)

        self.moved_terminal(netindex, termindex)

        return True


    cpdef bint move_conn_exchange(
        self,
        sa_params params,
        SAState state
    ) except? False:

        """Exchange one terminal from each of two nets, with each terminal
        being connected to different banks of the same DevInst.

        Return True if successful, or False if no net was changed.

        """

        cdef list netindices
        cdef int netindex
        cdef PlacedNet net
        cdef list DIN_candidates
        cdef list termindices   # [termindex, termindex]
        cdef list terminals # [(DIN, bank), (DIN, bank)]
        cdef int termindex

        for inner_iter from 0 <= inner_iter < max_inner_iter:
            try:
                netindices = state.rand.sample(self.info.conn_netchoices, 2)
            except (ValueError, IndexError):
                return False

            ### Choose a DevInst common to both nets.
            DIN_candidates = list() # [{DINs}, {DINs}]
            for netindex in netindices:
                net = self.nets[netindex]

                candidates = set()
                for DIN, bank in net.terminals:
                    if len(self.D2b2numconn[DIN]) > 1:
                        candidates.add(DIN)

                DIN_candidates.append(candidates)

            candidates = sorted(
                DIN_candidates[0] & DIN_candidates[1],
                key=banksort_key
            )

            try:
                DIN = state.rand.choice(candidates)
            except (ValueError, IndexError):
                continue

            ### Choose one terminal per net.
            termindices = list()
            terminals = list()
            for netindex in netindices:
                net = self.nets[netindex]

                termindex = state.rand.choice(sorted(net.D2terminals[DIN]))
                termindices.append(termindex)
                terminals.append(net.terminals[termindex])

            ### Swap terminals.
            terminals.reverse()

            for netindex, termindex, terminal in zip(
                netindices,
                termindices,
                terminals,
            ):
                net = self.nets[netindex]
                net.terminals[termindex] = terminal

                self.moved_terminal(netindex, termindex, False)

            self.moved_terminal_postnet()
            return True

        warn_inner_iter(
            'PlacedBundle.move_conn_exchange <nets>',
            inner_iter
        )
        return False


cdef class PlacedNet(object):
    cdef public list terminals  # [(DIN, bank)]
    cdef public list terminalcoords # [xy]
    cdef public object D2terminals  # {DIN: {terminal indices}}
    cdef public cost.xy CoM # Used for nets with >2 terminals
    cdef public float total_dist
    cdef public Placement place # Refers to containing Placement
    cdef public PlacedBundle bundle # Refers to containing PlacedBundle

    def __init__(
        self,
        terminals=None,
        terminalcoords=None,
        D2terminals=None,
        CoM=None,
        total_dist=INF,
        place=None,
        bundle=None
    ):
        if terminals is None:
            terminals = list()

        if terminalcoords is None:
            terminalcoords = list()

        if D2terminals is None:
            D2terminals = collections.defaultdict(set)

        if CoM is None:
            CoM = dict(x=INF, y=INF)

        self.terminals = terminals
        self.terminalcoords = terminalcoords
        self.D2terminals = D2terminals
        self.CoM = CoM
        self.total_dist = total_dist
        self.place = place
        self.bundle = bundle


    cpdef PlacedNet clone(
        self,
        Placement new_place=None,
        PlacedBundle new_bundle=None
    ):
        return PlacedNet(
            terminals=list(self.terminals),
            terminalcoords=list(self.terminalcoords),
            D2terminals=dict(self.D2terminals),
            CoM=self.CoM,
            total_dist=self.total_dist,
            place=new_place,
            bundle=new_bundle,
        )


    def update(self):
        """Force update of internal data structures.

        This is an expensive operation and is intended for:
            - initialization of data structures as part of setup
            - debugging automatic updates of data structures

        """

        D2terminals = collections.defaultdict(set)

        for index, term in enumerate(self.terminals):
            DIN, bank = term
            D2terminals[DIN].add(index)

        self.D2terminals = dict(D2terminals)

        ### Force re-calculate terminal coords and total distance.
        self.terminalcoords = list()
        self.calc_total_dist()


    cpdef calc_total_dist(self):
        cdef cost.xy coords
        cdef cost.xy coords2
        cdef int numterm = len(self.terminals)
        cdef int i

        ### Populate self.terminalcoords if necessary.
        if len(self.terminalcoords) != numterm:
            coords.x = 0.0
            coords.y = 0.0

            self.terminalcoords = list()
            for i from 0 <= i < numterm:
                self.terminalcoords.append(coords)
                self.moved_terminal(i, False)


        if numterm < 2:
            self.total_dist = 0.0

        elif numterm == 2:
            coords, coords2 = self.terminalcoords
            self.total_dist = cost.dist_xy(coords, coords2)

        else:
            self.total_dist = 0.0
            self.CoM.x = 0.0
            self.CoM.y = 0.0

            for coords in self.terminalcoords:
                self.CoM.x += coords.x
                self.CoM.y += coords.y

            self.CoM.x /= numterm
            self.CoM.y /= numterm

            for coords in self.terminalcoords:
                self.total_dist += cost.dist_xy(coords, self.CoM)


    cpdef moved_DINs(self, DINs, bint update_cost=True):
        cdef str DIN
        cdef set indices = set()
        cdef int index

        for DIN in DINs:
            indices.update(self.D2terminals.get(DIN, set()))

        if not indices:
            return

        for index in indices:
            self.moved_terminal(index, False)

        if update_cost:
            self.calc_total_dist()


    cpdef moved_terminal(self, int index, bint update_cost=True):
        cdef str DIN
        cdef float x, y, angle
        cdef float dx, dy
        cdef cost.xy coords

        assert(self.place is not None)
        assert(self.bundle is not None)

        DIN, bank = self.terminals[index]
        x, y, angle = self.place.D2coords[DIN]
        dx, dy = self.bundle.info.D2b2CoM[DIN][bank]

        coords = cost.absrel_xy(x, y, angle, dx, dy)

        self.terminalcoords[index] = coords

        if update_cost:
            self.calc_total_dist()


cpdef float total_placement_cost(
    Placement place,
    SAState cur_state
):
    """Apply weights to the placement cost and add callback.

    Arguments:
    place -- Placement
    cur_state -- contains current weights for placement cost components

    Return sum of weighted placement cost components, callback function.

    """

    cdef PlacementCosts costs = place.costs
    cdef float ret = 0

    if cur_state.user_cost_func is None:
        ret += costs.dist * cur_state.weight_dist
        ret += costs.boardsize * cur_state.weight_boardsize
        ret += costs.overlap * cur_state.weight_overlap
        ret += costs.iostandard * cur_state.weight_iostandard
        ret += costs.pin * cur_state.weight_pin
        ret += costs.diffpair * cur_state.weight_diffpair
        ret += costs.boardoversize * cur_state.weight_boardoversize

    else:
        ret += cur_state.user_cost_func(
            place,
            cur_state,
            cur_state.user_cost_func_arg,
        )

    return ret


def banksort_key(bank):
    """Return sortable values of bank keys.

    If bank is an int, return (0, bank).
    Else return (1, str(bank)).

    """

    try:
        bank = int(bank)
        return (0, bank)
    except (ValueError, TypeError):
        bank = str(bank)
        return (1, bank)


def supplyvalsort_key(val):
    if val is None:
        return (0, None)
    return (1, val)


def valkeysort_key(kv):
    k, v = kv
    return (v, k)


def keys_with_pos_values(D):
    """Yield all keys in D having values > 0."""

    for k, v in D.items():
        if v > 0:
            yield k


def keys_with_long_values(D):
    """Yield all keys in D having values with length > 1."""

    for k, v in D.items():
        if len(v) > 1:
            yield k


def keys_with_values(D):
    """Yield all keys in D having bool(value) == True."""

    for k, v in D.items():
        if v:
            yield k


cdef Placement new_placement(
    Placement cur_placement,
    sa_params params,
    SAState cur_state
):
    """Copy cur_placement and make one move (of a variety of types).

    Arguments:
    cur_placement -- current placement
    params -- parameters for simulated annealing
    cur_state -- current state in simulated annealing

    Return new Placement object, or None if all moves have been exhausted
    at this temperature.

    """

    cdef Placement new_placement = cur_placement.clone()
    cdef bint have_conn_move = False
    cdef int bundleindex
    cdef PlacedBundle old_bundle
    cdef PlacedBundle new_bundle
    cdef int DIN_total_moves
    cdef int old_numconn
    cdef int new_numconn

    if not new_placement.move(params, cur_state):
        return None

    if params.debug_move_conn:
        for bundleindex, new_bundle in enumerate(new_placement.bundles):
            old_bundle = cur_placement.bundles[bundleindex]

            if new_bundle.D2b2numconn != old_bundle.D2b2numconn:
                have_conn_move = True
                break

        if have_conn_move:
            print('\t\tDDDD Movement:')
            for bundleindex, new_bundle in enumerate(new_placement.bundles):
                old_bundle = cur_placement.bundles[bundleindex]

                for DIN, new_b2numconn in new_bundle.D2b2numconn.items():
                    old_b2numconn = old_bundle.D2b2numconn[DIN]

                    DIN_total_moves = 0

                    for bank, new_numconn in new_b2numconn.items():
                        old_numconn = old_b2numconn[bank]

                        if new_numconn != old_numconn:
                            print(
                                '\t\tDDDD\tconn {}:'.format(
                                    (new_bundle.info.bundlename, DIN, bank),
                                ),
                                '{:2} -> {:2} ({:+2})'.format(
                                    old_numconn,
                                    new_numconn,
                                    new_numconn - old_numconn
                                ),
                            )

                            DIN_total_moves += new_numconn - old_numconn

                    if DIN_total_moves != 0:
                        raise ValueError('Mismatched conn_exchange!')

    return new_placement


@cython.profile(False)
cdef inline int accept_p(
    object rand,
    float T,
    float old_cost,
    float new_cost
) except -1:
    """Determine whether or not to keep the latest move.

    Arguments:
    rand -- pseudo-random number generator
    T -- current SA temperature
    old_cost -- cost of previous placement
    new_cost -- cost of newest placement (test whether to keep it)

    Return 1 if newest placement should be kept, or 0 otherwise.

    """

    cdef float neg_diff
    cdef float thresh

    if new_cost < old_cost:
        return 1

    neg_diff = old_cost - new_cost
    thresh = math.exp(neg_diff / T)

    if rand.uniform(0, 1) < thresh:
        return 1

    return 0


cdef object moves_at_temp(
    Placement init_placement,
    list moveable_DINs,
    object post_move_callback,
    object post_move_arg,
    sa_params params,
    SAState cur_state
):
    """Perform simulated annealing steps for a given value of T.

    Arguments:
    init_placement -- placement at beginning of this temperature
    moveable_DINs -- list of moveable DINs (list in predefined order)
    post_move_callback -- if set, then call this function
    post_move_arg -- pass to post_move_callback
    params -- parameters for simulated annealing
    cur_state -- current state in simulated annealing

    Return (end_placement, accept_ratio).

    """

    cdef int moves_attempted = 0
    cdef int moves_accepted = 0
    cdef float accept_ratio
    cdef Placement cur_p, new_p, best_p

    cur_state.D2movesleft = {
        DIN: params.moves_per_devinst
        for DIN in moveable_DINs
    }

    cur_state.moveable_DINs = list(moveable_DINs)

    cur_state.move_conn_left = params.num_move_conn


    ### Compute costs at new temperature.
    assert init_placement.costs.valid_p()

    init_placement.total_cost = total_placement_cost(
        init_placement,
        cur_state
    )


    ### Keep track of movement types.
    movement_types = collections.defaultdict(int)


    cur_p = init_placement.clone()
    cur_p.movetype = init_placement.movetype

    best_p = cur_p

    while True:
        PyErr_CheckSignals()


        new_p = new_placement(cur_p, params, cur_state)

        if new_p is None:
            break

        assert new_p.costs.valid_p()

        new_p.total_cost = total_placement_cost(new_p, cur_state)


        moves_attempted += 1

        accepted = accept_p(
            cur_state.rand,
            cur_state.T,
            cur_p.total_cost,
            new_p.total_cost
        )

        if params.debug_accept_p:
            print('\t\tDDDD accept: @{:02g} ({:04g} - {:04g}) = {:+04g} {} ({})'.format(
                cur_state.T,
                new_p.total_cost,
                cur_p.total_cost,
                new_p.total_cost - cur_p.total_cost,
                'YES' if accepted else 'NO',
                new_p.movetype
            ))
            if new_p.movetype.startswith('conn_'):
                print(
                    '\t\tDDDD\tcosts',
                    'new {:0.4g}, {:0.4g}, {:0.4g}, {:0.4g}, {:0.4g}'.format(
                        new_p.costs['dist'],
                        new_p.costs['overlap'],
                        new_p.costs['iostandard'],
                        new_p.costs['pin'],
                        new_p.costs['diffpair'],
                    )
                )
                print(
                    '\t\tDDDD\tcosts',
                    'old {:0.4g}, {:0.4g}, {:0.4g}, {:0.4g}, {:0.4g}'.format(
                        cur_p.costs['dist'],
                        cur_p.costs['overlap'],
                        cur_p.costs['iostandard'],
                        cur_p.costs['pin'],
                        cur_p.costs['diffpair'],
                    )
                )

        if accepted:
            moves_accepted += 1
            cur_p = new_p

            movement_types[new_p.movetype] += 1

            if new_p.total_cost < best_p.total_cost:
                best_p = new_p


        if post_move_callback is not None:
            post_move_callback(
                new_p,
                accepted,
                cur_state,
                post_move_arg
            )


        if (moves_attempted % params.moves_attempted_progress) == 0:
            print(
                '\t{} / {}.'.format(
                    moves_accepted,
                    moves_attempted
                ),
                '(r: {}, {})'.format(
                    sum(cur_state.D2movesleft.values()),
                    cur_state.move_conn_left,
                )
            )
            sys.stdout.flush()


    if params.show_movetypes:
        print('\tMovement types:')
        for k, v in sorted(movement_types.items()):
            print('\t\t{}: {}'.format(k, v))


    if params.use_best_each_step:
        cur_p = best_p


    if moves_attempted == 0:
        accept_ratio = 0.0
    else:
        accept_ratio = <float>moves_accepted / <float>moves_attempted

    return (cur_p, accept_ratio)


@cython.profile(False)
cpdef inline float next_temp(float T, float accept_ratio):
    """Return a slightly cooler temperature."""

    # Cooling schedule per V. Betz, J. Rose. 1997.
    if accept_ratio > 0.96:
        return T * 0.5
    elif accept_ratio > 0.8:
        return T * 0.9
    elif accept_ratio > 0.15:
        return T * 0.95
    else:
        return T * 0.8


@cython.profile(False)
cdef inline int continue_p(
    PlacementCosts costs,
    sa_params params,
    SAState cur_state
) except -1:
    """Return 1 if simulated annealing should continue, or 0 otherwise.

    Arguments:
    costs -- separate parts of placement cost
    params -- parameters for simulated annealing
    cur_state -- current state in simulated annealing

    Continue if any of the following are true:
        * T > cold_temp
        * costs.overlap > 0
        * costs.iostandard > 0
        * costs.pin > 0
        * costs.diffpair > 0
        * costs.boardoversize > 0

    """

    if cur_state.T > params.cold_temp:
        return 1

    if costs.overlap > 0.0:
        return 1

    if costs.iostandard > 0.0:
        return 1

    if costs.pin > 0.0:
        return 1

    if costs.diffpair > 0.0:
        return 1

    if costs.boardoversize > 0.0:
        return 1

    return 0


cdef int cancel_sa

def handle_hup(signum, frame):
    print('Cancelling the remainder of SA...')
    global cancel_sa
    cancel_sa = True


cpdef Placement solve_place_and_banks(
    Placement init_placement,
    object locked_DINs,
    object D2b2validsupplies,
    object rand,
    object post_move_callback,
    object post_move_arg,
    object params_dict,
    object move_ratios_dict
):
    """Position all unfixed components.

    Iteratively find legal positions for all DevInsts, legal bank assignments
    for all Bundles, and legal bank supplies for all banks.
    Do not move any DevInst listed in locked_DINs.

    Arguments:
    init_placement -- Placement object as starting placement
    locked_DINs -- Do not move any of these DevInsts
    D2b2validsupplies -- {DIN: {bank: {supplyname: {value, ...}}}}
    rand -- Pseudo-random number generator
    post_move_callback -- if set, then call this function of (P, accepted)
    post_move_arg -- pass to post_move_callback
    params_dict -- Parameters for simulated annealing
    move_ratios_dict -- relative probabilities of different types of moves

    Return final placement.

    """

    cdef sa_params params = FakeDict(0)
    cdef sa_move_ratios move_ratios = FakeDict(0)
    cdef SAState cur_state = SAState(
        D2b2validsupplies=D2b2validsupplies,
        rand=rand,
    )
    cdef float accept_ratio
    cdef int bundleindex
    cdef PlacedBundle bundle
    cdef PlacedNet net


    ### Ensure all required parameters, move ratios are provided.
    missing_params = param_keys - set(params_dict)
    if missing_params:
        raise KeyError('Missing param: {}'.format(
            ', '.join(missing_params)
        ))

    missing_ratios = move_ratio_keys - set(move_ratios_dict)
    if missing_ratios:
        raise KeyError('Missing move ratio: {}'.format(
            ','.join(missing_ratios)
        ))


    print()


    ### Populate params and move ratios.
    params = dict(itertools.chain(
        params.items(),
        params_dict.items()
    ))
    move_ratios = dict(itertools.chain(
        move_ratios.items(),
        move_ratios_dict.items()
    ))
    cur_state.move_ratios_cumulative = accumulate_move_ratios(move_ratios)

    print('Params:')
    for k, v in sorted(params.items()):
        print('\t{}: {}'.format(k, v))
    print()

    print('Move ratios:')
    for k, v in sorted(move_ratios.items()):
        print('\t{}: {}'.format(k, v))
    print()

    print('Accumulated move ratios:')
    for k, v in sorted(dict(cur_state.move_ratios_cumulative).items()):
        print('\t{}: {}'.format(k, v))
    print()


    ### Prepare required data structures.
    print('Preparing data structures...')
    all_DINs = set(init_placement.D2coords)
    locked_DINs = set(locked_DINs)
    moveable_DINs = sorted(all_DINs - locked_DINs)
    print('Moveable DINs:', moveable_DINs)


    conn_bundlechoices = list()
    for bundleindex, bundle in enumerate(init_placement.bundles):
        accept = False

        bundlename = bundle.info.bundlename

        for DIN, b2numconn in bundle.D2b2numconn.items():
            if len(b2numconn) > 1:
                print('Accepting bundle', bundlename, 'because of', DIN)
                accept = True
                break

        if accept:
            conn_bundlechoices.append(bundleindex)

    conn_bundlechoices = sorted(conn_bundlechoices)

    print()
    print('Connection bundle names:')
    for bundleindex in conn_bundlechoices:
        print('Bundle[{}]: {}'.format(
            bundleindex,
            init_placement.bundles[bundleindex].info.bundlename,
        ))
    print()


    # Compute initial costs.
    if not init_placement.costs.isfinite():
        init_placement.calc_costs()

    # Update total cost given updated weights.
    init_placement.total_cost = total_placement_cost(
        init_placement,
        cur_state
    )

    init_placement.movetype = ''


    print()
    print(
        'Estimated number of moves at each temperature:',
        params.moves_per_devinst * len(moveable_DINs) +
        params.num_move_conn
    )


    global cancel_sa
    cancel_sa = False
    signal.signal(signal.SIGHUP, handle_hup)


    ### Perform simulated annealing
    print()
    print('Simulated annealing begins!')
    cur_state.weight_dist = params.weight_dist
    cur_state.weight_boardsize = params.weight_boardsize
    cur_state.weight_overlap = params.weight_overlap_start
    cur_state.weight_iostandard = params.weight_iostandard_start
    cur_state.weight_pin = params.weight_pin_start
    cur_state.weight_diffpair = params.weight_diffpair_start
    cur_state.weight_boardoversize = params.weight_boardoversize_start
    cur_state.T = params.init_temp
    cur_state.translate_limit_x = init_placement.info.boardmaxsize.x * 2
    cur_state.translate_limit_y = init_placement.info.boardmaxsize.y * 2
    cur_state.conn_bundlechoices = conn_bundlechoices

    cur_state.user_cost_func = params_dict.get('user_cost_func', None)
    cur_state.user_cost_func_arg = params_dict.get('user_cost_func_arg', None)

    sys.stdout.flush()

    cur_placement = init_placement
    run = True
    while run:
        run = continue_p(cur_placement.costs, params, cur_state)

        if cancel_sa:
            run = False

        print('SA temp: {:0.4g} ({})'.format(
            cur_state.T,
            'continue' if run else 'last'
        ))

        if not run:
            # Finish one last iteration and save the best placement.
            params.use_best_each_step = True

        cur_placement, accept_ratio = moves_at_temp(
            cur_placement,
            moveable_DINs,
            post_move_callback,
            post_move_arg,
            params,
            cur_state
        )

        cur_state.T = next_temp(cur_state.T, accept_ratio)

        cur_state.update_penalties(cur_placement.costs, params)

        print(
            '\tAccept: {:0.5f}'.format(accept_ratio),
            'Cost: {:0.5g}'.format(cur_placement.total_cost),
            'Penalties: {:0.3g} {:0.3g} {:0.3g} {:0.3g} {:0.3g}'.format(
                cur_placement.costs.overlap,
                cur_placement.costs.iostandard,
                cur_placement.costs.pin,
                cur_placement.costs.diffpair,
                cur_placement.costs.boardoversize
            )
        )

        cur_state.update_limits(params)


        ### Ensure continue if nonzero penalties resulted from this last
        ### iteration!
        if not run and not cancel_sa:
            run = continue_p(cur_placement.costs, params, cur_state)
            if run:
                print('\tNeed to continue iterating!')


        ### Cancel if it appears that SA will never finish.
        if cur_state.T < params.cancel_temp:
            print(
                '\tCancelling due to very low temperature: {:0.3g}'.format(
                    cur_state.T
                )
            )
            run = False
            raise ReachedCancelTemp('Reached cancel_temp', cur_state.T)


        # Ensure output for each T.
        sys.stdout.flush()


    cur_placement.debug_pin_use()
    cur_placement.update()
    cur_placement.debug_pin_use()


    return cur_placement
