#!/usr/bin/env python3


"""
Represent schematics and conceptual connections;
Solve conceptual connections to realize complete schematic.

Schematic
    bundles: Bundle
        endpoints: BundleEndpoint
            subgroups: EndpointSubgroup
                search: PinSearch
"""

import sys
import itertools
import string
import math
import collections
import re
import random
import cvxopt.glpk
import csv
import operator
import code
import xml.etree.ElementTree as et

import wallpy.cleanobject
import wallpy.prettyprint
import wallpy.svg
import wallpy.units
import wallpy.version

import device
import devmanager
import cost
import placeandbanks

VER = wallpy.version.get_version()


moduleprefix = __name__ + '.'



AUTO = 'auto'

# Initial placement options
INIT_ASIS = 'asis'
INIT_ZERO = 'zero'
INIT_RAND = 'rand'

DEFAULT_ROUND_DIGITS = 6


good_devinstname_re = re.compile(r'^[a-zA-Z0-9_<>[\]]+$')
good_netname_re = re.compile(r'^[a-zA-Z0-9_<>[\]]+$')
good_bundlename_re = re.compile(r'^[a-zA-Z0-9_<>[\]]+$')



def dev_warning(msg, *extrainfo):
    """Show developer's warning message"""

    print('DEVELOPER WARNING:', msg, *extrainfo)


def dev_info(msg, *extrainfo):
    """Show developer's informational message"""

    print('DEVELOPER INFO:', msg, *extrainfo)



class SchemError(Exception):
    """Base class for schematic errors"""


class DesignError(SchemError):
    """User's design has an insoluble problem"""


class NetConnectError(SchemError):
    """Net could not be connected as requested"""


class AssumptionError(DesignError):
    """Design violates algorithm's assumptions"""



def name2ana(name):
    """Convert an alphanumeric name to (alpha, numeric, alpha).
    If name is fully non-numeric, will return (name, False, '').
    If name is an int, will return ('', int(name), '')."""


    if isinstance(name, int):
        return ('', name, '')


    len_parts = [None, None, None]

    # Find length of first non-numeric portion
    for i, c in enumerate(name):
        if c in string.digits:
            len_parts[0] = i
            break

    if len_parts[0] is None:
        len_parts[0] = len(name)


    # Find length of numeric portion
    for i, c in enumerate(name[len_parts[0]:]):
        if c not in string.digits:
            len_parts[1] = i
            break

    if len_parts[1] is None:
        len_parts[1] = len(name) - len_parts[0]


    # Find length of last non-numeric portion
    len_parts[2] = len(name) - sum(len_parts[0:2])


    parts = [
        name[:len_parts[0]],
        name[len_parts[0]:sum(len_parts[0:2])],
        name[sum(len_parts[0:2]):],
    ]

    if parts[1]:
        parts[1] = int(parts[1])
    else:
        parts[1] = False

    return tuple(parts)


def termsort_key(DIN_bank):
    DIN, bank = DIN_bank
    return (name2ana(DIN), placeandbanks.banksort_key(bank))



def bus2netnames(prefix, suffix, iterable):
    return tuple(''.join([prefix, str(i), suffix]) for i in iterable)



def rows2spmatrix(rowlist, numcols):
    """Create a sparse matrix given list of row elements (colnum, value).

    """

    X = list()
    I = list()
    J = list()

    for rownum, row in enumerate(rowlist):
        # Each element in row is (colnum, value)
        for colnum, value in row:
            X.append(value)
            I.append(rownum)
            J.append(colnum)

    return cvxopt.spmatrix(X, I, J, (len(rowlist), numcols))


def abs_coords_float(abscoords, relcoords):
    """Compute the absolute coordinates: rotated(abscoords) + relcoords.

    Arguments:
    abscoords -- (absolute X, absolute Y, angle in radians)
    relcoords -- (delta X, delta Y)

    Return (result X, result Y).

    """

    x, y, angle = abscoords
    dx, dy = relcoords

    c = math.cos(angle)
    s = math.sin(angle)

    ax = (dx * c) - (dy * s) + x
    ay = (dx * s) + (dy * c) + y

    return (ax, ay)


def coords_angle_div_pi(coords):
    """Divide coords' angle by pi.

    Arguments:
    coords -- (x, y, angle)

    Return (x, y, angle / pi).

    """

    x, y, angle = coords
    return (x, y, angle / math.pi)


def set_placement_asis(dest, DIN2DI, unit_length, unit_angle):
    """Set placement of all DevInsts to their current locations in DIN2DI.

    Arguments:
    dest -- {DIN: (float_x, float_y, float_angle)} to modify
    DIN2DI -- DevInsts with their current positions
    unit_length -- represent x, y as floats in terms of this length
    unit_angle -- represent angle as float in terms of this angle

    Modify dest inplace.

    """

    for DIN, DI in DIN2DI.items():
        x = DI.coords[0]
        y = DI.coords[1]
        angle = DI.coords[3]
        dest[DIN] = (
            float(x / unit_length),
            float(y / unit_length),
            float(angle / unit_angle)
        )


def set_placement_zero(dest, DIN2DI, locked_DINs):
    """Set placement of all DevInsts to (0.0, 0.0, 0.0).

    Arguments:
    dest -- {DIN: (float_x, float_y, float_angle)} to modify
    DIN2DI -- DevInsts with their current positions
    locked_DINs -- DevInsts in this set will not be moved

    Modify dest inplace.

    """

    for DIN, DI in DIN2DI.items():
        if DIN in locked_DINs:
            continue
        dest[DIN] = (0.0, 0.0, 0.0)


def set_placement_rand(dest, DIN2DI, board_dim, DIN2dim, locked_DINs,
                       quant_rotate_steps, RNG):
    """Set placement of all DevInsts to random coordinates within the board.

    Arguments:
    dest -- {DIN: (float_x, float_y, float_angle)} to modify
    DIN2DI -- DevInsts with their current positions
    board_dim -- (width, height) as floats
    DIN2dim -- dimensions of DevInsts as floats
    locked_DINs -- DevInsts in this set will not be moved
    quant_rotate_steps -- number of valid DevInst angles
    RNG -- use this random number generator

    Modify dest inplace.

    """

    boardwidth, boardheight = board_dim

    for DIN, DI in DIN2DI.items():
        if DIN in locked_DINs:
            continue
        # Assume worst rotation
        width, height = DIN2dim[DIN]
        R = math.sqrt(width + height) / 2

        max_x = (boardwidth / 2) - R
        min_x = -max_x
        max_y = (boardheight / 2) - R
        min_y = -max_y

        x = RNG.uniform(min_x, max_x)
        y = RNG.uniform(min_y, max_y)
        angle = RNG.randrange(quant_rotate_steps)
        angle *= 2 * math.pi / quant_rotate_steps

        dest[DIN] = (x, y, angle)


def compute_CoM_noweight(points):
    """Return the center of mass (x, y) of the iterable of points.

    Arguments:
    points -- iterable of (x, y)

    Return (center_x, center_y).

    """

    sum_x = 0
    sum_y = 0
    numpins = 0

    if not points:
        raise ValueError('No points given')

    for x, y in points:
        sum_x += x
        sum_y += y
        numpins += 1

    return (sum_x / numpins, sum_y / numpins)


def compute_CoM(points):
    """Return the center of mass (x, y) of the iterable of points.

    Arguments:
    points -- iterable of (x, y, weight)

    Return (center_x, center_y).

    """

    sum_x = 0
    sum_y = 0
    sum_weight = 0

    if not points:
        raise ValueError('No points given')

    for x, y, weight in points:
        sum_x += x * weight
        sum_y += y * weight
        sum_weight += weight

    if sum_weight <= 0.0:
        raise ValueError('Weights must be positive')

    return (sum_x / sum_weight, sum_y / sum_weight)


def shift_CoM(orig, prev, chosen, weights=(1,1,2)):
    """Shift the center of mass toward the chosen pins.

    Arguments:
    orig -- original center of mass before any placement
    prev -- center of mass from previous iteration
    chosen -- center of mass of currently-chosen pins
    weights -- relative weights for orig, prev, chosen

    Return (x, y) of the new weighted center of mass.

    """

    points = (
        (orig[0], orig[1], weights[0]),
        (prev[0], prev[1], weights[1]),
        (chosen[0], chosen[1], weights[2]),
    )

    return compute_CoM(points)


def relcoords2side(x, y, slope):
    A = 2 if (y > (slope * x)) else 0
    B = 1 if (y > (-slope * x)) else 0
    return A + B


def round_sigdig(x, n):
    """Return x rounded to n significant digits."""

    # From:
    # http://code.activestate.com/lists/python-tutor/70739/
    return round(x, int(n - math.ceil(math.log10(abs(x)))))






zero_length = wallpy.units.Length(0, 'mm')

sign2quadrant = {
    (True, True)    :   0,
    (False, True)   :   1,
    (False, False)  :   2,
    (True, False)   :   3,
}


def coords2qs(coords):
    """Return (quadrant, slope) given coordinates (x, y)"""

    ##########################################
    ####                   |                 
    ####                   |                 
    ####           2       |       3         
    ####                   |                 
    ####    ---------------------------------+
    ####                   |                 
    ####           1       |       0         
    ####                   |                 
    ####                   +                 
    ##########################################

    qk = tuple((coords[i] >= zero_length) for i in range(2))

    if coords[0] == zero_length:
        slope = float('inf') if coords[1] >= zero_length else float('-inf')
    else:
        slope = coords[1] / coords[0]

    return (sign2quadrant[qk], slope)



def lpk2qs(DI, logpinkey):
    """Return (quadrant, slope) given device instance and logpinkey."""

    phypinkeys = DI.log2phy(logpinkey)

    if not phypinkeys:
        raise ValueError('Empty phypinkeys')
    if len(phypinkeys) != 1:
        raise ValueError('Only support single physical pin per logical pin')

    ppk = phypinkeys[0]

    phypin = DI.phypack.pins[ppk]
    coords = phypin.coords[:2]

    return coords2qs(coords)



def quadrants2offset(quadrants):
    """Return an appropriate offset that makes the quadrants properly circularly sortable.

    Raises ValueError if all 4 quadrants are present in the iterable."""

    if len(quadrants) > 3:
        raise ValueError('Cannot correct quadrant sorting with all 4 quadrants')

    if (0 in quadrants) and (3 in quadrants):
        # Offset of 2 (all mod 4) will correct this
        return 2

    return 0



def calc_quadrant_offset(iter_qs):
    """Go through iter_qs and find which quadrants are present in the iterable.
    Return an appropriate offset that makes the quadrants properly circularly sortable.

    Raises ValueError if all 4 quadrants are present in the iterable."""

    quadrants = frozenset(qs[0] for qs in iter_qs)

    return quadrants2offset(quadrants)



def adjust_quadrant(qs, offset):
    """Adjust quadrant for proper circular sorting.

    Arguments:
    qs -- tuple of (quadrant, slope)
    offset -- add this value to the quadrant (mod 4)

    Returns (new_quadrant, slope)."""

    q, s = qs

    q += offset
    q %= 4

    return (q, s)



def relcoords2dist(relcoords):
    rc_f = tuple(float(c.native_val) for c in relcoords)

    return math.sqrt(sum(c ** 2 for c in rc_f))



def netnames2bundlename(nets, groupname):
    if not nets:
        raise ValueError('Need at least one net')

    if groupname:
        return 'G__' + groupname

    if len(nets) == 1:
        return nets[0]

    s = sorted(nets, key = name2ana)

    return ''.join((
        s[0],
        '____',
        s[-1],
    ))



def from_file(filename):


    def design_warning(msg, *args, comment=''):
        if comment:
            comment = ''.join(('(', comment, ')'))

        print('WARN:', msg, list(args), comment)


    S = Schematic()

    propernames = dict()  # Device nickname : devinstname

    def empty_bundle():
        return Bundle(None, None)

    S.bundles = collections.defaultdict(empty_bundle, S.bundles)


    # Groupname : {netnames}
    groups = collections.defaultdict(set)


    # Groupname : {params}
    group_params = collections.defaultdict(dict)


    def AddDevice(
        devnickname, devprefix, str_coords,
        dev_fn, logpackname, phypack_fn, log2phy_fn,
        bank2pin_reserve_frac=None,
        bank2diffpair_reserve_frac=None,
        pin_reserve_frac=None,
        diffpair_reserve_frac=None,
    ):
        print("Adding device", devnickname)

        if devnickname in propernames:
            raise KeyError('Device nickname already exists', devnickname)

        if len(str_coords) != 4:
            raise ValueError('Must have 4 coords (X, Y, Z, rot_angle)')


        propernames[devnickname] = S.next_devinstname(devprefix)


        coords = list()
        for i in str_coords[0:3]:
            coords.append(wallpy.units.Length(i))

        for i in str_coords[3:]:
            coords.append(wallpy.units.Angle(i))


        DI = S.add_devinst(
            propernames[devnickname],
            coords,
            dev_fn,
            logpackname,
            phypack_fn,
            log2phy_fn,
            bank2pin_reserve_frac=bank2pin_reserve_frac,
            bank2diffpair_reserve_frac=bank2diffpair_reserve_frac,
        )


        banks = set()
        for pin in DI.logpack.pins.values():
            banks.add(pin.bank)

        banks = sorted(banks, key=placeandbanks.banksort_key)


        if pin_reserve_frac is not None:
            frac = float(pin_reserve_frac)

            for bank in banks:
                if bank not in DI.bank2pin_reserve_frac:
                    print('\tSet {} pin_reserve_frac = {:0.1%}'.format(
                        bank,
                        frac,
                    ))
                    DI.bank2pin_reserve_frac[bank] = frac


        if diffpair_reserve_frac is not None:
            frac = float(diffpair_reserve_frac)

            for bank in banks:
                if bank not in DI.bank2diffpair_reserve_frac:
                    print('\tSet {} diffpair_reserve_frac = {:0.1%}'.format(
                        bank,
                        frac,
                    ))
                    DI.bank2diffpair_reserve_frac[bank] = frac


        return propernames[devnickname]



    def Search(*args):
        return S.cached_pinsearch(*args)



    def AddSimpleEP(bundlename, devnickname,
                    matchfunc, mf_args, mf_kwargs = None,
                    **kwargs):

        print("Adding EP", bundlename, devnickname)

        S.bundles[bundlename].endpoints.append(
            BundleEndpoint((
                EndpointSubgroup(
                    AUTO,
                    Search(
                        propernames[devnickname],
                        matchfunc,
                        mf_args,
                        mf_kwargs
                    ),
                    params = kwargs
                ),
            ))
        )


    def AddMultigroupEP(bundlename, sgparams):
        """Add Endpoint with multiple Subgruops.

        Arguments:
        bundlename -- Add EP to this Bundle
        sgparams -- Iterable of parameters to Subgroup:
            (numpins, devnickname, matchfunc, mf_args, mf_kwargs, params)

        """

        print('Adding multigroup EP to', bundlename)

        subgroups = (
            EndpointSubgroup(
                numpins,
                Search(
                    propernames[devnickname],
                    matchfunc,
                    mf_args,
                    mf_kwargs
                ),
                params=params
            )
            for numpins, devnickname, matchfunc, mf_args, mf_kwargs, params
            in sgparams
        )

        S.bundles[bundlename].endpoints.append(
            BundleEndpoint(subgroups)
        )



    def BundleNets(bundlename, netnames):
        S.bundles[bundlename].netnames.extend(netnames)



    def Group(groupname, netnames):
        groups[groupname].update(netnames)



    def GroupParams(groupname, params):
        group_params[groupname].update(params)



    def BundleParams(bundlename, params):
        S.bundles[bundlename].params.update(params)



    def BundleSetDifferential(bundlename, p_suffix, n_suffix):
        S.bundles[bundlename].params['differential'] = True
        S.bundles[bundlename].params['diff_p_suffix'] = p_suffix
        S.bundles[bundlename].params['diff_n_suffix'] = n_suffix



    G = dict(globals())

    for n in (
        'AddDevice',
        'AddSimpleEP',
        'AddMultigroupEP',
        'BundleNets',
        'BundleParams',
        'BundleSetDifferential',
        'Search',
        'Group',
        'GroupParams',
        'S',
    ):
        G[n] = locals()[n]

    for n in ('Length', 'Angle'):
        G[n] = getattr(wallpy.units, n)

    exec(compile(open(filename).read(), filename, 'exec'), G)



    for groupname, nets in groups.items():
        for n in nets:
            # TODO: Add to group
            pass

        if groupname not in group_params:
            design_warning(
                'group without defined parameters',
                groupname,
                comment = ''.join((
                    'call GroupParams(',
                    repr(groupname),
                    ', {}) to suppress',
                ))
            )


    for groupname, params in group_params.items():
        if groupname not in groups:
            design_warning(
                'parameters for nonexistent group',
                groupname,
                comment = ''.join((
                    'call Group(',
                    repr(groupname),
                    ', []) to suppress',
                ))
            )




    #### BEGIN DEBUG PRINT
    print('Devices')
    for k, v in propernames.items():
        print(k, ':', v)


    print('Bundles:')
    S.debug_bundles(include_results=True)

    #### END DEBUG PRINT


    print('Checking for sufficient eligible pins...')
    pin_assumptions_failed = False

    for bundlename, B in S.bundles.items():
        try:
            B.compute_subgroup_numpins()
        except:
            print('Could not compute subgroup numpins for', bundlename)
            raise


        for epindex, ep in enumerate(B.endpoints):
            for sgindex, sg in enumerate(ep.subgroups):
                if len(sg.search.resultset) < sg.numpins:
                    pin_assumptions_failed = True

                    dev_warning(
                        'Not enough eligible pins',
                        (
                            bundlename,
                            epindex,
                            sgindex,
                            '{} < {}'.format(
                                len(sg.search.resultset),
                                sg.numpins,
                            ),
                        )
                    )

    if pin_assumptions_failed:
        raise AssumptionError('Bundles violate eligible pin assumptions')


    print('Checking differential assumptions...')
    diff_assumptions_failed = False

    for DIN, DI in S.devinst_n2o.items():
        DI.logpack.postprocess_pins()

        for lpk, pin in DI.logpack.pins.items():
            sigconfig_pins = {
                k: list()
                for k in (device.SINGLE, device.DIFF_P, device.DIFF_N)
            }

            for pfname, pf in pin.pinfuncs.items():
                sigconfig_pins[pf.sigconfig].append(pfname)

            c_s = len(sigconfig_pins[device.SINGLE])
            c_p = len(sigconfig_pins[device.DIFF_P])
            c_n = len(sigconfig_pins[device.DIFF_N])

            sigconfig_pins_filtered = {
                k: v
                for k, v in sigconfig_pins.items()
                if v
            }

            if (c_n > 0) and (c_p > 0):
                diff_assumptions_failed = True

                dev_warning(
                    'Both DIFF_P and DIFF_N pinfuncs',
                    (DIN, lpk, sigconfig_pins_filtered)
                )

            if (c_n > 0) or (c_p > 0):
                ### Ensure exactly one corresponding differential pin
                pf2diffpair = dict()
                for pfname in itertools.chain(
                    sigconfig_pins[device.DIFF_P],
                    sigconfig_pins[device.DIFF_N]
                ):
                    pf = pin.pinfuncs[pfname]
                    bank = pf.params.get('bank', None)

                    # Get lpk of corresponding differential pin
                    pf2diffpair[pfname] = DI.logpack.diff_pairs[pfname][bank]

                other_lpks = set(pf2diffpair.values())
                if len(other_lpks) != 1:
                    diff_assumptions_failed = True

                    dev_warning(
                        'Differential pin corresponds to multiple pins',
                        (DIN, lpk, sigconfig_pins_filtered, other_lpks)
                    )

    if diff_assumptions_failed:
        raise AssumptionError('Devices violate algorothm assumptions')

    print('Done checking differential assumptions.')

    print()


    print('Checking differential pin user requirements...')

    req_failed = False

    for bundlename, B in S.bundles.items():
        diff = B.params.get('differential', False)

        if diff:
            for epindex, ep in enumerate(B.endpoints):
                for sg in ep.subgroups:
                    search = sg.search
                    DIN = search.devinstname
                    DI = S.devinst_n2o[DIN]

                    for lpk in search.resultset:
                        pin = DI.logpack.pins[lpk]

                        other = None
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig in device.DIFF_SIGCONFIGS:
                                bank = pf.params.get('bank', None)
                                other = DI.logpack.diff_pairs[pfname][bank]
                                break

                        if other is not None:
                            if other in search.resultset:
                                req_failed = True

                                print(
                                    'DESIGN ERROR:',
                                    'Includes both pins from a diff pair!',
                                    (
                                        bundlename,
                                        epindex,
                                        DIN,
                                        lpk,
                                        pfname,
                                        other,
                                    )
                                )

    if req_failed:
        raise DesignError('Include only one pin from each differential pair!')

    print('Done checking differential pin user requirements.')


    return S



#### Endpoint filter functions

def any_pin(devinst, logpinkey, logpin):
    return True


def pinkey_in_set(devinst, logpinkey, logpin, acceptset):
    return logpinkey in acceptset


def has_pinfunc(devinst, logpinkey, logpin, pinfuncname):
    return pinfuncname in logpin.pinfuncs


def pinfunc_match(devinst, logpinkey, logpin, pinfunc_re,
                 ignorecase = False):
    flags = 0
    if ignorecase:
        flags |= re.IGNORECASE

    R = re.compile(pinfunc_re, flags = flags)

    for k in logpin.pinfuncs:
        if R.match(k):
            return True

    return False


def pinfunc_match_in_bank(devinst, logpinkey, logpin, pinfunc_re,
                          ignorecase=False, banks=frozenset()):
    flags = 0
    if ignorecase:
        flags |= re.IGNORECASE

    R = re.compile(pinfunc_re, flags=flags)

    for k, pf in logpin.pinfuncs.items():
        if R.match(k):
            if pf.params.get('bank') in banks:
                return True

    return False


def in_bank(devinst, logpinkey, logpin, banks):
    for pf in logpin.pinfuncs.values():
        if pf.params.get('bank') in banks:
            return True

    return False





class Schematic(object):
    moduleprefix = moduleprefix

    def __init__(self, devinstances = None, nets = None):
        """Create a schematic object with the given device instances and net connections.

        Arguments:
        devinstances -- list of (devinstname, coords, dev_fn, logpackname, phypack_fn, log2phy_fn)
        nets -- {netname : set((humanname(dev_instance), logpinkey))}
        """

        self.devinst_id2o = dict()  # id(dev_instance) : dev_instance
        self.devinst_id2n = dict()  # id(dev_instance) : human_friendly_name
        self.devinst_n2o = dict()   # human_friendly_name : dev_instance

        # netname : set((humanname(dev_instance), logpinkey))
        self.nets = collections.defaultdict(set)

        self.logpin2net = dict()    # (humanname(dev_instance), logpinkey) : netname
        self.namegens = dict()      # devprefix : generator
        self.bundles = dict()       # bundlename : Bundle
        self.pinsearches = dict()    # (humanname(dev_instance), matchfunc, args, tuple(kwargs.items())) : PinSearch

        self.lockedpins = set()     # (humanname(dev_instance), logpinkey)
        self.locked_DINs = set()    # DevInsts that have fixed coordinates

        self.potential_pins = set() # Optionally highlight in export_svg
        self.potential_pins2 = set()    # Secondary highlight in export_svg

        # Debugging DevInst-relative points
        # {DIN: [(dx1, dy1), (dx2, dy2), ...]}
        self.di_points = collections.defaultdict(list)

        # If defined, include result from user_cost_func in
        # solve_place_and_banks.
        self.user_cost_func = None
        self.user_cost_func_arg = None



        if devinstances:
            for args in devinstances:
                self.add_devinst(*args)


        if nets:
            for k, v in nets.items():
                self.connect_net(k, v)



    def __eq__(self, other):

        for k in ('devinst_n2o', 'nets', 'bundles'):
            if not hasattr(other, k):
                return False

            if getattr(self, k) != getattr(other, k):
                return False

        return True



    def valid_p(self, errcallback=None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass


        for n, DI in self.devinst_n2o.items():
            if not n:
                errcallback('Empty devinstname', repr(n))
                return False

            if not good_devinstname_re.match(n):
                errcallback('Invalid devinstname', repr(n))
                return False

            if not DI.valid_p(errcallback):
                errcallback('Invalid devinst', repr(n))
                return False


        for netname, connections in self.nets.items():
            if not good_netname_re.match(netname):
                errcallback('Invalid netname', repr(netname))
                return False

            if not connections:
                errcallback('Empty net', repr(netname))
                return False


        for bundlename, B in self.bundles.items():
            if not good_bundlename_re.match(bundlename):
                errcallback('Invalid bundle name', repr(bundlename))
                return False

            if not B.valid_p(errcallback):
                errcallback('Invalid bundle', repr(bundlename))
                return False


        ### Check logpin2net
        expect = dict()

        for netname, connections in self.nets.items():
            for k in connections:
                if k in expect:
                    errcallback(
                        'Already have connection to',
                        repr(k),
                        repr(expect[k]),
                        repr(netname)
                    )
                    return False

                expect[k] = netname

        if expect != self.logpin2net:
            errcallback('logpin2net does not match expected value')
            return False


        ### All Bundles, BundleEndpoints, and EndpointSubgroups must be unique
        def check_id_happy(D, i):
            if i in D:
                return False
            D.add(i)
            return True

        seen_ids = set()
        for bundlename, B in self.bundles.items():

            if not check_id_happy(seen_ids, id(B)):
                errcallback('Non-unique Bundle', bundlename)
                return False

            for i, ep in enumerate(B.endpoints):

                if not check_id_happy(seen_ids, id(ep)):
                    errcallback('Non-unique BundleEndpoint', bundlename, i)
                    return False

                for j, sg in enumerate(ep.subgroups):

                    if not check_id_happy(seen_ids, id(sg)):
                        errcallback(
                            'Non-unique EndpointSubgroup',
                            bundlename,
                            i,
                            j
                        )
                        return False


        return True




    def next_devinstname(self, devprefix):
        if devprefix not in self.namegens:
            self.namegens[devprefix] = (devprefix + str(i) for i in itertools.count(1))

        ret = next(self.namegens[devprefix])

        while ret in self.devinst_n2o:
            ret = next(self.namegens[devprefix])

        return ret



    def add_devinst(
        self,
        devinstname,
        coords,
        dev_fn,
        logpackname,
        phypack_fn,
        log2phy_fn,
        bank2pin_reserve_frac=None,
        bank2diffpair_reserve_frac=None,
    ):

        """Import and add a DevInst given by (coords, dev_fn, ..., log2phy_fn).

        Arguments:
        devinstname -- the human-friendly name for the new DevInst
        coords -- the physical coordinates (X, Y, Z, rot_angle) for the DevInst
        dev_fn -- the filename for the device (as passed to devmanager.load_device)
        logpackname -- which logical package to use
        phypack_fn -- the filename for the physical package (as passed to devmanager.load_phypack)
        log2phy_fn -- the filename for the map from logical to physical pin key (as passed to devmanager.load_log2phy)
        bank2pin_reserve_frac -- {bank: fraction of pins to reserve from SA}
        bank2diffpair_reserve_frac -- {bank: fraction of diff. pairs to
            reserve from SA}

        Returns the DevInst added to the Schematic"""


        DI = device.DevInst(
            coords, dev_fn, logpackname, phypack_fn, log2phy_fn,
            bank2pin_reserve_frac=bank2pin_reserve_frac,
            bank2diffpair_reserve_frac=bank2diffpair_reserve_frac,
        )

        self.devinst_id2o[id(DI)] = DI
        self.devinst_id2n[id(DI)] = devinstname
        self.devinst_n2o[devinstname] = DI

        return DI



    def connect_net(self, netname, iterable):
        """Connect the logical pins in iterable to the given net.

        Arguments:
        netname -- the name for the net connecting the given pins (may already exist)
        iterable -- each element is (humanname(devinst), logpinkey)

        Return True if any changes were made, or
            False if no changes were made (only happens when all the pins in
                iterable already were connected to the net).
        """


        changed = False


        for devinstname, logpinkey in iterable:

            # Make sure LogicalPin exists
            dummy = self.devinst_n2o[devinstname].logpack.pins[logpinkey]

            # Make sure LogicalPin is not already connected to a different net
            k = (devinstname, logpinkey)

            if k in self.logpin2net:
                if netname != self.logpin2net[k]:
                    raise NetConnectError(
                        'Cannot add pin to another net',
                        (
                            devinstname,
                            logpinkey,
                            netname,
                            self.logpin2net[k],
                        )
                    )

            else:
                changed = True

            self.logpin2net[k] = netname
            self.nets[netname].add((devinstname, logpinkey))


        return changed



    def disconnect_net(self, netname, iterable):
        """Disconnect the logical pins in iterable from the given net.

        Arguments:
        netname -- the name for the net from which to disconnect the given pins
                    (use None to disconnect regardless of net name)
        iterable -- each element is (humanname(devinst), logpinkey)

        Return True if any changes were made, or
            False if no changes were made (only happens when none of the pins
                in iterable were connected to the net).
        """

        changed = False


        for devinstname, logpinkey in iterable:

            # Make sure LogicalPin exists
            dummy = self.devinst_n2o[devinstname].logpack.pins[logpinkey]

            # Make sure LogicalPin is not already connected to a different net
            k = (devinstname, logpinkey)

            if k in self.logpin2net:
                if netname != self.logpin2net[k]:
                    raise NetConnectError(
                        'Cannot remove pin from a different net',
                        (devinstname, logpinkey)
                    )

                changed = True

                del self.logpin2net[k]
                self.nets[netname].remove(k)

                if not self.nets[netname]:
                    # Removed last connection to this net!
                    del self.nets[netname]


        return changed



    def disconnect_pins(self, iterable):
        """Disconnect the logical pins in iterable from any nets.

        Arguments:
        iterable -- each element is (humanname(devinst), logpinkey)

        Return True if any changes were made, or
            False if no changes were made (only happens when none of the pins
                in iterable were connected to any net).

        """

        changed = False

        for devinstname, logpinkey in iterable:

            # Make sure LogicalPin exists
            dummy = self.devinst_n2o[devinstname].logpack.pins[logpinkey]

            k = (devinstname, logpinkey)

            netname = self.logpin2net.get(k, None)
            if netname:
                changed = True

                del self.logpin2net[k]
                self.nets[netname].remove(k)


        return changed



    def export(self, file, comments = False):

        import_modules = (
            'sys',
            'fractions',
            'wallpy.cleanobject',
            'wallpy.units',
            'device',
            'schem',
        )

        print('#!/usr/bin/env python3', file=file)
        print(file=file)
        print('"""Schematic', file=file)
        print(file=file)
        print('Automatically generated by schem.py', file=file)
        print(file=file)
        print('Latest tool revision:', VER, file=file)
        print('"""', file=file)
        print(file=file)

        for m in import_modules:
            print('import', m, file=file)

        print(file=file)
        print(file=file)



        PP = wallpy.prettyprint.PrettyPrinter(0, file)

        PP.print('schem = schem.Schematic(', extraspaces = 0)

        PP.print('devinstances = (', extraspaces = 2)

        for devinstname in sorted(self.devinst_n2o, key = name2ana):
            DI = self.devinst_n2o[devinstname]

            args = (devinstname, DI.coords, DI.dev_fn, DI.logpackname, DI.phypack_fn, DI.log2phy_fn)

            PP.print(args, ',', extraspaces = 4)
            PP.print()

        PP.print('),', extraspaces = 2)



        def netkey(netname):
            try:
                prefix, num, suffix = name2ana(netname)
            except ValueError:
                return (netname, )
            return (prefix, suffix, num)


        def logpin_sortkey(pair):
            devinstname, logpinkey = pair

            return (name2ana(devinstname), name2ana(logpinkey))


        PP.print('nets = {', extraspaces = 2)

        for netname in sorted(self.nets, key = netkey):
            if comments:
                comment_str = ' '.join([' # terminal count:', str(len(self.nets[netname]))])
            else:
                comment_str = ''

            PP.print(repr(netname), ': (', comment_str, extraspaces = 4)

            for devinstname, logpinkey in sorted(self.nets[netname], key = logpin_sortkey):
                if comments:
                    comment_str = ' # ' + ', '.join(
                        self.devinst_n2o[devinstname].logpack.pins[logpinkey].pinfuncs
                    )
                else:
                    comment_str = ''

                PP.print((devinstname, logpinkey), ',', comment_str, extraspaces = 6)

            PP.print('),', extraspaces = 4)

        PP.print('}', extraspaces = 2)


        PP.print(')', extraspaces = 0)




    def cached_pinsearch(
        self,
        devinstname,
        matchfunc,
        matchfunc_args,
        matchfunc_kwargs):

        """Create a PinSearch with the given fields, or return the cached PS if present.

        Arguments:
        devinstname -- human-friendly name of the desired device instance
        matchfunc -- match all logical pins for which bool(matchfunc(devinst, logpinkey, logpin, *args, **kwargs)) evaluates to True
        matchfunc_args -- passed as *args to matchfunc
        matchfunc_kwargs -- passed as **kwargs to matchfunc

        Returns the PinSearch matching the given fields"""


        if matchfunc_args:
            matchfunc_args = tuple(matchfunc_args)
        else:
            matchfunc_args = tuple()

        if not matchfunc_kwargs:
            matchfunc_kwargs = dict()


        k = (
            devinstname,
            matchfunc,
            matchfunc_args,
            tuple(matchfunc_kwargs.items()),
        )
        
        
        if k not in self.pinsearches:
            PS = PinSearch(
                devinstname,
                self.devinst_n2o[devinstname],
                matchfunc,
                matchfunc_args,
                matchfunc_kwargs
            )

            self.pinsearches[k] = PS


        return self.pinsearches[k]




    def add_bundle(self, bundlename, netnames, ep_fields, params = None):
        """Add a Bundle to the schematic with endpoints described by ep_fields.

        Arguments:
        bundlename -- the name of this Bundle (used as key in self.bundles)
        netnames -- the names for the nets, once the Bundle has been satisfied
        ep_fields -- [ [(numpins, search, subgroup_params), ...] ]
        params -- parameters for the overall Bundle

        Return the created Bundle"""

        if bundlename in self.bundles:
            raise DesignError('Bundle already exists', bundlename)

        for n in netnames:
            if n in self.nets:
                raise DesignError('Net already exists', n)


        if ep_fields:
            eps = []

            for EP in ep_fields:
                subgroups = []

                for numpins, search, ep_params in EP:
                    S = self.cached_pinsearch(*search)
                    SG = EndpointSubgroup(numpins, S, params = ep_params)
                    subgroups.append(SG)


                eps.append(BundleEndpoint(subgroups))

        else:
            eps = tuple()

        if params:
            params = dict(params)
        else:
            params = dict()


        B = Bundle(netnames, eps, params)
        self.bundles[bundlename] = B


        return B



    def debug_bundles(self, include_results = False):
        if not self.bundles:
            print('<<no Bundles>>')
            return

        for k, v in self.bundles.items():
            print('  ', repr(k), ': ', sep='')
            v.prettyprint(4)

            if include_results:
                print(NotImplemented)

            print()



    def export_svg(self, file, prefs=None, styles=None):
        """Print the Schematic's physical layout to an SVG file.

        Arguments:
        file -- Output SVG to this file (must already be open)
        prefs -- Configure settings (e.g. include pins, ratsnests, etc.)
        styles -- Configure colors, fonts, etc.

        """

        default_prefs = {
            'devinstlabels': True,
            'pins': True,
            'pinlabels': True,
            'rats': True,
            'highlight_net_pins': True,
            'highlight_potential_pins': True,
            'dipoints': True,
            'title': '',
            'main_unit': 'mm',
            'margins': (1.0, 1.0),    # In main units
            'zoom': 1,
            'comments': list(),
        }

        default_styles = {
            'style': {
                'title': {
                    'font-family': 'Sans',
                    'font-size': '0.8px',
                    'text-anchor': 'start',
                },

                'devinst': {
                    'font-family': 'Sans',
                    'font-size': '3px',
                    'text-anchor': 'middle',
                    'dominant-baseline': 'central',
                },

                'pin': {
                    'font-family': 'Sans',
                    'font-size': '0.2px',
                    'text-anchor': 'middle',
                    'dominant-baseline': 'central',
                },
            },

            'attrib': {
                'pin': {
                    'fill': 'aliceblue',
                    'stroke': 'black',
                    'stroke-width': '0.05',
                },

                'pin_highlight1': {
                    'fill': 'gray',
                    'stroke': 'black',
                    'stroke-width': '0.05',
                },

                'pin_highlight2': {
                    'fill': 'lime',
                    'stroke': 'black',
                    'stroke-width': '0.05',
                },

                'pin_highlight3': {
                    'fill': 'yellow',
                    'stroke': 'black',
                    'stroke-width': '0.05',
                },

                'pin_label': {
                    'fill': 'white',
                    'stroke': 'black',
                    'stroke-width': '0.01',
                },

                'package': {
                    'fill': 'lightgray',
                    'stroke': 'black',
                    'stroke-width': '0.5',
                },

                'devinst_label': {
                    'fill': 'white',
                    'stroke': 'black',
                    'stroke-width': '0.02',
                },

                'rats': {
                    'stroke': 'blue',
                    #'stroke-dasharray': '0.2,0.1',
                    'stroke-width': '0.2',
                },

                'layout_area': {
                    'fill': 'none',
                    'stroke': 'black',
                    'stroke-width': '0.1',
                },

                'debug1': {
                    'fill': 'red',
                    'stroke': 'black',
                    'stroke-width': '0.02',
                },

                'debug2': {
                    'fill': 'lime',
                    'stroke': 'black',
                    'stroke-width': '0.02',
                },

                'debug3': {
                    'fill': 'blue',
                    'stroke': 'black',
                    'stroke-width': '0.02',
                },

                'di_point': {
                    'fill': 'blue',
                    'stroke': 'black',
                    'stroke-width': '0.2',
                    #'stroke-dasharray': '0.5,0.4',
                },
            },

            'dim': {
                'devinst_label': {
                    'char_width': 0.4 * 5,
                    'char_height': 0.6 * 5,
                    'offset_x': -0.015,
                    'offset_y': -0.02,
                },

                'pin_label': {
                    'char_width': 0.17,
                    'char_height': 0.2,
                    'offset_x': -0.005,
                    'offset_y': -0.013,
                },

                'di_point': {
                    'radius': 0.5,
                }
            },
        }


        if prefs:
            default_prefs.update(prefs)
        prefs = default_prefs

        if styles:
            for styletype, name2attrs in styles.items():
                for name, attrs in name2attrs.items():
                    default_styles[styletype][name].update(attrs)

        styles = default_styles

        for k in styles['style']:
            attr = styles['style'][k]
            styles['style'][k] = ';'.join(
                (':'.join(kv) for kv in attr.items())
            )


        ### Compute boundaries of layout
        bound_n = None
        bound_s = None
        bound_w = None
        bound_e = None

        unit_length = wallpy.units.Length(1, prefs['main_unit'])
        unit_angle = wallpy.units.Angle(1, 'pi_rad')

        for DIN, DI in self.devinst_n2o.items():
            x, y, z, rot = DI.coords
            Hx, Hy, Hz = DI.phypack.dimensions

            rot_rad = math.pi * float(rot / unit_angle)
            Hx = float(Hx / unit_length)
            Hy = float(Hy / unit_length)
            Hx, Hy = cost.bbox_size(Hx, Hy, rot_rad)
            Hx *= unit_length
            Hy *= unit_length

            w = x - (Hx / 2)
            e = x + (Hx / 2)
            n = y - (Hy / 2)
            s = y + (Hy / 2)

            if (bound_w is None) or (w < bound_w):
                bound_w = w
            if (bound_e is None) or (e > bound_e):
                bound_e = e
            if (bound_n is None) or (n < bound_n):
                bound_n = n
            if (bound_s is None) or (s > bound_s):
                bound_s = s


        # Include margins
        margin_x, margin_y = (
            wallpy.units.Length(i, prefs['main_unit'])
            for i in prefs['margins']
        )

        bound_w -= margin_x
        bound_e += margin_x
        bound_n -= margin_y
        bound_s += margin_y

        width = bound_e - bound_w
        height = bound_s - bound_n

        # _ic indicates inner coordinate system
        width_ic = wallpy.svg.length2float(width, prefs['main_unit'])
        height_ic = wallpy.svg.length2float(height, prefs['main_unit'])

        # _oc indicates outer coordinate system
        width_oc = width_ic * prefs['zoom']
        height_oc = height_ic * prefs['zoom']


        bound_w_ic = wallpy.svg.length2float(bound_w, prefs['main_unit'])
        bound_n_ic = wallpy.svg.length2float(bound_n, prefs['main_unit'])

        offset_x_ic = - bound_w_ic
        offset_y_ic = - bound_n_ic


        ### Start creating layout
        C = wallpy.svg.Canvas(
            (width_oc, height_oc),
            title=prefs['title'],
            main_unit=prefs['main_unit']
        )


        commenttext = '\n'.join([''] + prefs['comments'] + [''])
        C.doc.append(et.Comment(text=commenttext))


        zoom_only = wallpy.svg.SubElement(
            C.doc,
            'g',
            transform='scale({:0.08E})'.format(prefs['zoom'])
        )

        layout_area = wallpy.svg.SubElement(
            zoom_only,
            'g',
            transform='translate({:0.06F},{:0.06F})'.format(
                offset_x_ic,
                offset_y_ic
            )
        )


        # Layout background
        stroke_width = float(styles['attrib']['layout_area']['stroke-width'])

        bound_w_ic += stroke_width / 2
        bound_n_ic += stroke_width / 2
        width_ic -= stroke_width
        height_ic -= stroke_width

        bg_rect = wallpy.svg.rect(
            layout_area, (bound_w_ic, bound_n_ic),
            (width_ic, height_ic),
            attrib=styles['attrib']['layout_area']
        )


        ### Draw DevInst areas
        for DIN, DI in self.devinst_n2o.items():
            x_ic, y_ic, z_ic = (
                wallpy.svg.length2float(v, prefs['main_unit'])
                for v in DI.coords[:3]
            )

            rot_deg = float(DI.coords[3] / wallpy.units.Angle(1, 'deg'))


            transforms = (
                'translate({:0.06F},{:0.06F})'.format(x_ic, y_ic),
                'rotate({:0.08E})'.format(rot_deg),
            )

            pack_area = wallpy.svg.SubElement(
                layout_area,
                'g',
                transform=' '.join(transforms)
            )


            Hx_ic, Hy_ic, Hz_ic = (
                wallpy.svg.length2float(v, prefs['main_unit'])
                for v in DI.phypack.dimensions
            )

            w_ic = - (Hx_ic / 2)
            n_ic = - (Hy_ic / 2)

            pack_rect = wallpy.svg.rect(
                pack_area, (w_ic, n_ic), (Hx_ic, Hy_ic),
                attrib=styles['attrib']['package']
            )


            if prefs['pins']:
                for lpk in DI.logpack.pins:
                    phypinkeys = DI.log2phy[lpk]

                    for ppk in phypinkeys:
                        try:
                            phypin = DI.phypack.pins[ppk]
                        except KeyError:
                            raise AssumptionError(
                                'No matching physical pin',
                                (DIN, lpk, ppk)
                            )

                        x_ic, y_ic, z_ic = (
                            wallpy.svg.length2float(v, prefs['main_unit'])
                            for v in phypin.coords
                        )

                        Hx_ic, Hy_ic, Hz_ic = (
                            wallpy.svg.length2float(v, prefs['main_unit'])
                            for v in phypin.dimensions
                        )


                        # Normal or highlighted pin
                        pinstyle = 'pin'

                        DIN_lpk = (DIN, lpk)

                        if prefs['highlight_potential_pins']:
                            if DIN_lpk in self.potential_pins:
                                pinstyle = 'pin_highlight2'
                            elif DIN_lpk in self.potential_pins2:
                                pinstyle = 'pin_highlight3'

                        if prefs['highlight_net_pins']:
                            if DIN_lpk in self.logpin2net:
                                pinstyle = 'pin_highlight1'

                        pin_attrib = styles['attrib'][pinstyle]

                        shape = phypin.attr.get('shape', 'rect')

                        if shape == 'rect':
                            w_ic = x_ic - (Hx_ic / 2)
                            n_ic = y_ic - (Hy_ic / 2)

                            pin_rect = wallpy.svg.rect(
                                pack_area, (w_ic, n_ic), (Hx_ic, Hy_ic),
                                attrib=pin_attrib
                            )

                        elif shape == 'circ':
                            if Hx_ic != Hy_ic:
                                print(DIN, lpk, 'circle pin, but Hx != Hy')
                                sys.exit(1)

                            radius = Hx_ic

                            pin_circ = wallpy.svg.circle(
                                pack_area, (x_ic, y_ic), radius,
                                attrib=pin_attrib
                            )

                        else:
                            print(DIN, lpk, 'unknown shape', shape)
                            sys.exit(1)



        if prefs['rats']:

            ### Draw ratsnest
            for netname, connections in self.nets.items():
                points = list()

                for DIN, lpk in connections:
                    DIN = self.devinst_n2o[DIN]

                    pincoords = DIN.lpk2coords(
                        lpk,
                        float_units=prefs['main_unit']
                    )

                    for i in pincoords.values():
                        points.append(i[:2])


                points_it = iter(points)


                # First point
                coords_to = next(points_it)

                while True:
                    # Connect to next point
                    coords_from = coords_to

                    try:
                        coords_to = next(points_it)
                    except StopIteration:
                        break

                    line = wallpy.svg.line(
                        layout_area,
                        coords_from,
                        coords_to,
                        attrib=styles['attrib']['rats']
                    )



        ### Draw DevInst labels
        for DIN, DI in self.devinst_n2o.items():

            if prefs['pinlabels']:
                # Draw pin labels
                for lpk in DI.logpack.pins:
                    phypinkeys = DI.log2phy[lpk]

                    for ppk in phypinkeys:
                        phypin = DI.phypack.pins[ppk]

                        x_ic, y_ic, z_ic = DI.abs_coords(
                            phypin.coords, float_units=prefs['main_unit']
                        )

                        pinname = str(ppk)

                        rect_width_ic = (
                            styles['dim']['pin_label']['char_width']
                        )
                        rect_width_ic *= len(pinname)
                        rect_height_ic = (
                            styles['dim']['pin_label']['char_height']
                        )

                        rect_x_ic = x_ic - (rect_width_ic / 2)
                        rect_y_ic = y_ic - (rect_height_ic / 2)

                        rect_x_ic += styles['dim']['pin_label']['offset_x']
                        rect_y_ic += styles['dim']['pin_label']['offset_y']

                        pin_label_rect = wallpy.svg.rect(
                            layout_area,
                            (rect_x_ic, rect_y_ic),
                            (rect_width_ic, rect_height_ic),
                            attrib=styles['attrib']['pin_label']
                        )

                        pin_label = wallpy.svg.text(
                            layout_area,
                            (x_ic, y_ic),
                            text=pinname,
                            style=styles['style']['pin']
                        )


            if prefs['devinstlabels']:
                # Draw DevInst label
                x_ic, y_ic, z_ic = (
                    wallpy.svg.length2float(v, prefs['main_unit'])
                    for v in DI.coords[:3]
                )

                rect_width_ic = styles['dim']['devinst_label']['char_width']
                rect_width_ic *= len(DIN)
                rect_height_ic = styles['dim']['devinst_label']['char_height']

                rect_x_ic = x_ic - (rect_width_ic / 2)
                rect_y_ic = y_ic - (rect_height_ic / 2)

                rect_x_ic += styles['dim']['devinst_label']['offset_x']
                rect_y_ic += styles['dim']['devinst_label']['offset_y']

                pack_label_rect = wallpy.svg.rect(
                    layout_area,
                    (rect_x_ic, rect_y_ic),
                    (rect_width_ic, rect_height_ic),
                    attrib=styles['attrib']['devinst_label']
                )

                pack_label = wallpy.svg.text(
                    layout_area,
                    (x_ic, y_ic),
                    text=DIN,
                    style=styles['style']['devinst']
                )


            if prefs['dipoints']:
                for dx, dy in self.di_points[DIN]:
                    x, y, z = DI.abs_coords(
                        (dx, dy, 0),
                        float_units=prefs['main_unit']
                    )

                    point_circ = wallpy.svg.circle(
                        layout_area,
                        (x, y),
                        styles['dim']['di_point']['radius'],
                        attrib=styles['attrib']['di_point']
                    )


        ### Write to file
        C.export(file)



    def solve_phase1_obvious(self):
        """Connect nets for pins with "obvious" requests.
        "Obvious" includes:
        -- Request endpoints that specify "all" pins in a matchset
        -- Request endpoints that have exactly one available pin in a matchset

        Return True if any connections were made, or
            False if no change was made to the Schematic.
        """


        ret = False
        changed = True


        while changed:
            changed = False


            for bundlename, B in self.bundles.items():

                ### Don't support differential pins in this function
                if B.params.get('differential'):
                    continue

                nn = len(B.netnames)
                changed_this_bundle = False

                if not nn:
                    raise DesignError('Bundle has no netnames', bundlename)


                # TODO: In the future, may want to support partial bundles
                #   (assign pins for some (but not necessarily all) endpoints)


                # Candidate for pin selection at this phase if
                #   Bundle is valid
                #   AND
                #       Exactly two endpoints
                #       OR at least one endpoint
                #           AND exactly one netname
                #           AND all endpoints' subgroups params['all'] == True


                if not B.valid_p():
                    continue


                acceptable = False

                if len(B.endpoints) == 2:
                    acceptable = True

                elif len(B.endpoints) >= 1:
                    if nn == 1:
                        acceptable = True

                        for ep in B.endpoints:
                            for sg in ep.subgroups:
                                if not sg.params.get('all', False):
                                    acceptable = False
                                    break

                else:
                    acceptable = False


                if not acceptable:
                    continue





                for epindex, ep in enumerate(B.endpoints):

                    # Endpoint has (nn) nets.
                    # Candidate for pin selection at this phase if
                    #   nn == 1 AND subgroup has param['all'] == True
                    #   OR
                    #       Each subgroup has a valid number of pins
                    #               (either auto or specified number)
                    #       AND total number of all subgroups' matching pins
                    #               == (nn)



                    have_all = False
                    num_auto = 0
                    num_explicit = 0  # Total number of pins (excluding AUTO)

                    for sg in ep.subgroups:
                        if sg.params.get('all', False):
                            have_all = True

                        if sg.numpins == AUTO:
                            num_auto += 1
                        else:
                            num_explicit += sg.numpins

                    pins_remain = nn - num_explicit


                    if have_all:
                        if nn == 1:
                            acceptable = True
                        else:
                            acceptable = False

                    else:
                        acceptable = True

                        if pins_remain < 0:
                            raise DesignError(
                                (bundlename, epindex),
                                'More nets requested than available',
                                nn,
                                num_explicit
                            )

                        if num_auto:
                            if (pins_remain % num_auto) != 0:
                                raise DesignError(
                                    'Remaining requested pins cannot be '
                                    + 'divided evenly between AUTO '
                                    + 'subgroups',
                                    (bundlename, epindex),
                                    pins_remain,
                                    num_auto
                                )

                        # Now make sure each the total number of subgroups'
                        #   matching pins == (nn)

                        total_matching_pins = 0

                        for sg in ep.subgroups:
                            din = sg.search.devinstname

                            for lpk in sg.search.resultset:

                                # This pin matches if it isn't already connected
                                if (din, lpk) not in self.logpin2net:
                                    total_matching_pins += 1

                        if total_matching_pins != nn:
                            acceptable = False
                            break


                if not acceptable:
                    continue


                print('BEGIN', bundlename)


                for ep in B.endpoints:

                    available_netnames = list(B.netnames)

                    have_all = False
                    num_auto = 0
                    num_explicit = 0  # Total number of pins (excluding AUTO)

                    for sg in ep.subgroups:
                        if sg.numpins == AUTO:
                            num_auto += 1
                        else:
                            num_explicit += sg.numpins

                        if sg.params.get('all', False):
                            have_all = True

                            # Connect all matching pins to the net
                            netname = available_netnames.pop(0)

                            din = sg.search.devinstname

                            changed_this_bundle |= self.connect_net(
                                netname,
                                ((din, lpk) for lpk in sg.search.resultset)
                            )

                            print('CONNECT_ALL', din, sg.search.resultset)


                    if have_all:
                        # Already handled (above loop)
                        continue


                    pins_remain = nn - num_explicit

                    if num_auto:
                        # Compute number of pins for each AUTO subgroup
                        pins_auto = pins_remain // num_auto
                    else:
                        pins_auto = 0


                    for sg in ep.subgroups:
                        din = sg.search.devinstname

                        if sg.numpins == AUTO:
                            numpins = pins_auto
                        else:
                            numpins = sg.numpins


                        possible = set()

                        for lpk in sg.search.resultset:
                            if (din, lpk) not in self.logpin2net:
                                possible.add(lpk)


                        if len(possible) == numpins:
                            # Exactly as many possible pins as requested
                            # Connect pins to first available netnames

                            for lpk in sorted(possible, key = name2ana):

                                netname = available_netnames.pop(0)
                                pins = [ (din, lpk) ]
                                print('CONNECT', netname, din, lpk)
                                changed_this_bundle |= self.connect_net(netname, pins)



                changed |= changed_this_bundle
                ret |= changed_this_bundle

                if changed_this_bundle:
                    self.OrderBundlePins(B)

                print('END', bundlename)


        return ret



    def OrderBundlePins(self, B):

        # TODO: Are there connected pins on both endpoints? (require exactly 2
        # endpoints for now)?

        if len(B.endpoints) != 2:
            return

        netnames = set(B.netnames)
        nn = len(netnames)
        if nn < 2:
            return


        def dist(coords1, coords2):
            d = (coords2[i] - coords1[i] for i in range(len(coords1)))
            return relcoords2dist(d)

        def dinlpk2coords(devinstname, logpinkey):
            return self.devinst_n2o[devinstname].lpk2coords(logpinkey)


        pin_choices = list()    # One list per endpoint

        for ep in B.endpoints:
            ep_pin_choices = list()

            for sg in ep.subgroups:
                devinstname = sg.search.devinstname

                for lpk in sg.search.resultset:
                    k = (devinstname, lpk)

                    if self.logpin2net.get(k, None) in netnames:
                        if k not in self.lockedpins:
                            num_phypins = len(dinlpk2coords(*k))
                            if num_phypins == 1:
                                # Candidate!
                                ep_pin_choices.append(k)

            pin_choices.append(ep_pin_choices)


        assert(len(pin_choices) == 2)
        pins_left = pin_choices[0]
        pins_right = pin_choices[1]


        ### Generate variables and cost matrix
        N = len(pins_right)

        vars = list()
        costs = list()

        for L in pins_left:
            lcoords = next(iter(dinlpk2coords(*L).values()))

            for R in pins_right:
                rcoords = next(iter(dinlpk2coords(*R).values()))
                c = dist(lcoords, rcoords)
                vars.append((L, R))
                costs.append(c)

        c = cvxopt.matrix(costs, (1, N*N), 'd')


        ### Generate A and b matrices
        A_lines = list()
        for i in range(N):
            # Restrict: a given L pin connects to exactly one R pin
            line = [0]*N*(i) + [1]*N + [0]*N*(N-1-i)
            A_lines.append(line)

            # Restrict: a given R pin connects to exactly one L pin
            line = [0] * N
            line[i] = 1
            line *= N
            A_lines.append(line)

        A = cvxopt.matrix(A_lines, tc='d').T
        b = cvxopt.matrix([1]*2*N, (2*N, 1), 'd')


        ### Generate dummy G and h matrices
        G = cvxopt.matrix([0]*N*N, (1, N*N), 'd')
        h = cvxopt.matrix([0], (1, 1), 'd')


        # G*x <= h
        # A*x = b


        # All variables are binary
        B = frozenset(range(N*N))


        sys.stdout.flush()
        (status, x) = cvxopt.glpk.ilp(c, G, h, A, b, B=B)
        sys.stdout.flush()


        # Disconnect pins on right (will reconnect in correct order)
        assert(self.disconnect_pins(pins_right))

        chosen = tuple(bool(round(i)) for i in x)


        print(
            'ILP final cost:', sum(
                cost if var else 0
                for cost, var in zip(c, chosen)
            )
        )


        for i, var in enumerate(vars):
            if chosen[i]:
                L, R = var

                netname = self.logpin2net[L]
                assert(self.connect_net(netname, (R, )))



    def solve_ilp_connections(self,
                              round_digits=DEFAULT_ROUND_DIGITS,
                              debugfilename=None):
        """Determine set of nets to satisfy Bundles

        Solve all Bundle requests simultaneously:
            One ILP var for each possibility:
                (bundlename, (sg0, DIN0, lpk0), (sg1, DIN1, lpk1))
                where sg? is the index of the corresponding EndpointSubgroup
                    in the ?th BundleEndpoint in the Bundle
            Cost for each ILP var is distance
                FROM (coords of DI0's physical pin corresponding to lpk0)
                TO (coords of DI1's physical pin corresponding to lpk1)
            Restrict sum of (*, (sg0, DIN0, lpk0), *) <= 1 (single-ended) OR
                sum of (*, (sg0, DIN0, lpk0), *) +
                    (*, (sg0, DIN0, lpk0diffpair), *) <= 1
            Restrict sum of (*, *, (sg1, DIN1, lpk1)) <= 1 (single-ended) OR
                sum of (*, *, (sg1, DIN1, lpk1)) +
                    (*, *, (sg1, DIN1, lpk1diffpair) <= 1
            Restrict sum of (bundlename,(sg0,DIN0,*),*) == sg0's numpins
            Restrict sum of (bundlename,*,(sg1,DIN1,*)) == sg1's numpins

        Arguments:
        round_digits -- round ILP cost to this many digits
        debugfilename -- if set, then print ILP variables and constraints
                            to this file (automatically opens and closes)

        Return generator of (bundlename, netname, (DIN, lpk)).
        """


        # Postprocess each DevInst's LogicalPackage's pins
        for DIN, DI in self.devinst_n2o.items():
            DI.logpack.postprocess_pins()


        # Require actual number of pins needed per Subgroup
        for B in self.bundles.values():
            B.compute_subgroup_numpins()


        changed = False


        # Each element is (bundlename, (sg0,DIN0,lpk0), (sg1,DIN1,lpk1))
        ilp_vars = list()   # These correspond to result x.
        binvars = set() # indices of binary variables

        c_vals = list() # min c'*x
        G_rows = list() # G*x <= h; elements are ((colnum, value), ...)
        h_vals = list() # "<="
        A_rows = list() # A*x = b;  elements are ((colnum, value), ...)
        b_vals = list() # "="

        # (DIN, lpk): set(var_indices)
        logpin2vars = collections.defaultdict(set)

        # (bundlename, epindex, sgindex): set(var_indices)
        BES2vars = collections.defaultdict(set)

        for bundlename, B in self.bundles.items():
            if len(B.endpoints) != 2:
                print('Skipping Bundle', bundlename, '(# emdpoints)')
                continue

            skip_bundle = False

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    if sg.params.get('all'):
                        skip_bundle = True
                        break

            if skip_bundle:
                print(
                    'Skipping Bundle',
                    bundlename,
                    '(subgroup has ALL set)'
                )
                continue


            ep_it = iter(B.endpoints)

            ### Left-hand endpoint
            ep0_candidates = list() # (sgindex, DIN, lpk)
            ep0 = next(ep_it)
            for sgindex, sg in enumerate(ep0.subgroups):
                DIN = sg.search.devinstname
                DI = self.devinst_n2o[DIN]

                for lpk in sg.search.resultset:
                    if len(DI.log2phy[lpk]) != 1:
                        print('Skipping pin', lpk, '(too many phypins)')
                        continue
                    ep0_candidates.append((sgindex, DIN, lpk))

                    self.potential_pins.add((DIN, lpk))


            ### Right-hand endpoint
            ep1_candidates = list() # (sgindex, DIN, lpk)
            ep1 = next(ep_it)
            for sgindex, sg in enumerate(ep1.subgroups):
                DIN = sg.search.devinstname
                DI = self.devinst_n2o[DIN]

                for lpk in sg.search.resultset:
                    if len(DI.log2phy[lpk]) != 1:
                        print('Skipping pin', lpk, '(too many phypins)')
                        continue
                    ep1_candidates.append((sgindex, DIN, lpk))

                    self.potential_pins.add((DIN, lpk))


            for c0, c1 in itertools.product(
                ep0_candidates,
                ep1_candidates
            ):

                ilp_vars.append((bundlename, c0, c1))
                varindex = len(ilp_vars) - 1

                sgindex0, DIN0, lpk0 = c0
                sgindex1, DIN1, lpk1 = c1

                logpin2vars[(DIN0, lpk0)].add(varindex)
                logpin2vars[(DIN1, lpk1)].add(varindex)

                if B.params.get('differential'):
                    # Choosing this connection prohibits other
                    # use of corresponding differential pins
                    for DIN, lpk in ((DIN0, lpk0), (DIN1, lpk1)):
                        DI = self.devinst_n2o[DIN]
                        pin = DI.logpack.pins[lpk]

                        # Determine corresponding differential pin
                        other_lpks = set()
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig in device.DIFF_SIGCONFIGS:
                                bank = pf.params.get('bank', None)
                                other = DI.logpack.diff_pairs[pfname][bank]
                                other_lpks.add(other)

                        if len(other_lpks) != 1:
                            raise AssumptionError(
                                '{} corresponding diff. pins found!'.format(
                                    len(other_lpks)
                                ),
                                (DIN, lpk, other_lpks)
                            )

                        # Only corresponding differential pin
                        other = next(iter(other_lpks))

                        logpin2vars[(DIN, other)].add(varindex)


                BES2vars[(bundlename, 0, sgindex0)].add(varindex)
                BES2vars[(bundlename, 1, sgindex1)].add(varindex)


        print('# of ILP vars:', len(ilp_vars))
        print('# of distinct (DIN, lpk) pairs:', len(logpin2vars))


        ### Add ILP constraints

        # Restrict vars to at most 1 per (DIN, lpk)
        print('Restrict to at most one net per pin...')

        for pinkey, varindices in logpin2vars.items():
            # elements are (colnum, value)
            G_newrow = tuple((i, 1) for i in varindices)

            if G_newrow:
                G_rows.append(G_newrow)
                h_vals.append(1)    # At most one variable may be chosen
            else:
                print('No matching pins for', pinkey, 'on #', i)


        # For each Bundle, require exactly as many variables as netnames
        print('Require enough pins per bundle...')

        for bundlekey, varindices in BES2vars.items():
            bundlename, epindex, sgindex = bundlekey

            B = self.bundles[bundlename]

            # elements are (colnum, value)
            A_newrow = tuple((i, 1) for i in varindices)

            if A_newrow:
                A_rows.append(A_newrow)
                b_vals.append(
                    B.endpoints[epindex].subgroups[sgindex].numpins
                )
            else:
                print('No matching bundles for', bundlekey)


        # Compute ILP variable costs
        print('Compute costs...')

        base_distance = wallpy.units.Length(1, 'mm')

        coords = dict() # (DIN, lpk): (x, y, z) as float()

        def pin_coord(pinkey):
            if pinkey in coords:
                return coords[pinkey]

            sgindex, DIN, lpk = pinkey

            coordlist = self.devinst_n2o[DIN].lpk2coords(lpk)
            if len(coordlist) == 1:
                c = tuple(
                    float(v / base_distance)
                    for v in next(iter(coordlist.values()))
                )
                coords[pinkey] = c
                return c

            raise ValueError('More than one physical pin', DIN, lpk)

        def ilp_cost(bundlename, pinL, pinR):
            coords_L = pin_coord(pinL)
            coords_R = pin_coord(pinR)
            rel_coords = tuple(
                (coords_R[i] - coords_L[i])
                for i in range(len(coords_L))
            )
            return math.sqrt(sum((v * v) for v in rel_coords))

        c_vals = [round_sigdig(ilp_cost(*v), round_digits) for v in ilp_vars]


        # All ILP vars are binary
        num_vars = len(ilp_vars)
        binvars = frozenset(range(num_vars))


        ### Convert to matrices
        print('Converting ILP constraints to matrices...')
        c = cvxopt.matrix(c_vals, (num_vars, 1), 'd')
        if not debugfilename:
            del c_vals
        h = cvxopt.matrix(h_vals, (len(h_vals), 1), 'd')
        if not debugfilename:
            del h_vals
        b = cvxopt.matrix(b_vals, (len(b_vals), 1), 'd')
        if not debugfilename:
            del b_vals

        G = rows2spmatrix(G_rows, num_vars)
        if not debugfilename:
            del G_rows
        A = rows2spmatrix(A_rows, num_vars)
        if not debugfilename:
            del A_rows


        ### Solve ILP
        print('Solving ILP...')
        sys.stdout.flush()

        status, x = cvxopt.glpk.ilp(c, G, h, A, b, B=binvars)
        sys.stdout.flush()

        print('ILP returned status:', status)
        if not x:
            print('SOLUTION NOT FOUND!')
            return

        chosen = tuple(bool(round(v)) for v in x)


        print(
            'ILP final cost:', sum(
                cost if var else 0
                for cost, var in zip(c, chosen)
            )
        )


        if debugfilename:
            print('Saving ILP variables and constraints to', debugfilename)

            bundlename2id = {
                name: i
                for i, name in enumerate(self.bundles)
            }

            def nicevarname(var):
                bundlename, c0, c1 = var

                sgindex0, DIN0, lpk0 = c0
                sgindex1, DIN1, lpk1 = c1

                bundleid = bundlename2id[bundlename]

                return ' '.join((
                    '{}',
                    '{}:{}:{}',
                    '{}:{}:{}',
                )).format(
                    bundleid,
                    sgindex0, DIN0, lpk0,
                    sgindex1, DIN1, lpk1
                )


            fieldnames = ['head']
            fieldnames.extend(range(len(ilp_vars)))
            fieldnames.append('tail')

            with open(debugfilename, 'w', newline='') as Fcsv:
                writer = csv.DictWriter(
                    Fcsv,
                    fieldnames,
                    dialect=csv.excel_tab
                )

                # Header: row names
                writer.writerow({
                    i: nicevarname(var)
                    for i, var in enumerate(ilp_vars)
                })


                writer.writerow(dict())


                writer.writerow(dict(head='h >= Gx:'))

                for i, row in enumerate(G_rows):
                    h = h_vals[i]

                    D = {'head': h}
                    D.update(row)

                    writer.writerow(D)


                writer.writerow(dict())


                writer.writerow(dict(head='b = Ax:'))

                for i, row in enumerate(A_rows):
                    b = b_vals[i]

                    D = {'head': b}
                    D.update(row)

                    writer.writerow(D)


                writer.writerow(dict())


                writer.writerow(dict(head='chosen:'))

                chosen_row = {
                    i: 'X' if v else ''
                    for i, v in enumerate(chosen)
                }
                writer.writerow(chosen_row)


        ### Connect nets
        netname_iters = dict()  # bundlename: iterator over net names
        for bundlekey in BES2vars:
            bundlename, epindex, sgindex = bundlekey

            if bundlename not in netname_iters:
                B = self.bundles[bundlename]
                netname_iters[bundlename] = iter(B.netnames)

        for i, var in enumerate(ilp_vars):
            if chosen[i]:
                bundlename, c0, c1 = var
                B = self.bundles[bundlename]

                sgindex0, DIN0, lpk0 = c0
                sgindex1, DIN1, lpk1 = c1

                pin0 = (DIN0, lpk0)
                pin1 = (DIN1, lpk1)

                netname = next(netname_iters[bundlename])
                yield (bundlename, netname, pin0)
                yield (bundlename, netname, pin1)


    def solve_ilp(self,
                  round_digits=DEFAULT_ROUND_DIGITS,
                  debugfilename=None):
        """Determine set of nets to satisfy Bundles, then connect the nets.

        Use solve_ilp_connections to determine set of nets.

        Arguments:
        round_digits -- round ILP cost to this many digits
        debugfilename -- if set, then print ILP variables and constraints
                            to this file (automatically opens and closes)

        Return True if any connections were made, or
            False if no change was made to the schematic.
        """

        changed = False

        def dict_set():
            return collections.defaultdict(set)

        # {bundlename: {DIN: {lpks}}}
        BN_DIN2lpks = collections.defaultdict(dict_set)

        for bundlename, netname, DIN_lpk in self.solve_ilp_connections(
            round_digits=round_digits,
            debugfilename=debugfilename
        ):
            changed |= self.connect_net(netname, (DIN_lpk, ))

            DIN, lpk = DIN_lpk
            BN_DIN2lpks[bundlename][DIN].add(lpk)

        print()

        ### Order pins to minimize total distance
        if BN_DIN2lpks:
            changed = True

        for bundlename, DIN2lpks in BN_DIN2lpks.items():
            B = self.bundles[bundlename]

            DIN2search = dict()

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    DIN2search[DIN] = sg.search

            for DIN, lpks in DIN2lpks.items():
                # Overwrite eligible pins with only chosen pins
                DIN2search[DIN].override_results(lpks)

            print('Ordering pins for Bundle {}...'.format(bundlename))
            self.OrderBundlePins(B)
            print('Done ordering pins for Bundle {}.'.format(bundlename))
            print()

            for DIN in DIN2lpks:
                # Restore results for future use of this same Schematic.
                DIN2search[DIN].override_results(DIN2search[DIN].matches())


        ### Reconnect differential pins, including suffixes and negative pins
        for bundlename, B in self.bundles.items():
            if B.params.get('differential'):
                for netname_prefix in B.netnames:
                    se_pins = list(self.nets[netname_prefix])

                    ### Disconnect old (single-ended) nets
                    assert(self.disconnect_net(netname_prefix, se_pins))

                    for DIN_lpk in se_pins:

                        ### Connect positive net (specified pin)
                        netname = netname_prefix + B.params['diff_p_suffix']
                        changed |= self.connect_net(netname, [DIN_lpk])


                        ### Connect negative net (corresponding diff. pin)
                        netname = netname_prefix + B.params['diff_n_suffix']
                        DIN, lpk = DIN_lpk
                        DI = self.devinst_n2o[DIN]
                        pin = DI.logpack.pins[lpk]

                        # Determine corresponding differential pin
                        other = None
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig in device.DIFF_SIGCONFIGS:
                                bank = pf.params.get('bank', None)
                                other = DI.logpack.diff_pairs[pfname][bank]
                                break

                        if other is None:
                            raise DesignError(
                                'Could not find diff. pin',
                                (bundlename, netname, DIN, lpk)
                            )

                        changed |= self.connect_net(netname, [(DIN, other)])

        return changed


    def solve_pins_from_placement(self, placement,
                                  length_in_units, angle_in_units,
                                  round_digits=DEFAULT_ROUND_DIGITS):
        """Update DevInsts' coordinates from placement, then determine
        pin connections that satisfy Bundles.

        Arguments:
        placement -- {DIN: (x, y, angle)} with coords as floats
        length_in_units -- Unit of x, y
        angle_in_units -- Unit of angle
        round_digits -- round ILP cost to this many digits

        Return generator of (bundlename, netname, (DIN, lpk)).

        """

        self.set_placement_2D(placement,
                              length_in_units=length_in_units,
                              angle_in_units=angle_in_units)

        return self.solve_ilp_connections(round_digits=round_digits)


    def estimate_pins_from_placement(self, placement, CoM, method,
                                     length_in_units, angle_in_units):
        """Update DevInsts' coordinates from placement, then guess
        pin connections that satisfy Bundles.

        Arguments:
        placement -- {DIN: (x, y, angle)} with coords as floats
        CoM -- {bundlename: {DIN: {bank: (dx, dy}}}
        method -- CoM estimation method:
                'closestpin_rad': choose pins nearest to closest pin to
                    other CoM
                'closestpin_line': choose pins nearest to line intersecting
                    closest pin to other CoM
                'dist_other': choose pins nearest other CoM
        length_in_units -- Unit of x, y
        angle_in_units -- Unit of angle

        Return generator of (bundlename, (DIN, bank, lpk)).

        """

        self.set_placement_2D(placement,
                              length_in_units=length_in_units,
                              angle_in_units=angle_in_units)


        unit_length = wallpy.units.Length(1, length_in_units)


        # Require actual number of pins needed per Subgroup
        for B in self.bundles.values():
            B.compute_subgroup_numpins()


        for bundlename, B in self.bundles.items():
            if len(B.endpoints) != 2:
                continue

            skip_bundle = False

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    if sg.params.get('all'):
                        skip_bundle = True
                        break

            if skip_bundle:
                continue


            ### Compute CoM of each Endpoint
            ep_CoM = list()

            for ep in B.endpoints:
                points = list()

                for sg in ep.subgroups:
                    DIN = sg.search.devinstname

                    # Use the average position of this DevInst's CoMs
                    one_bank_CoM = next(iter(CoM[bundlename][DIN].values()))
                    numpins = sg.numpins

                    DI_CoM = list(compute_CoM(CoM[bundlename][DIN].values()))
                    DI_CoM.append(numpins)
                    points.append(tuple(DI_CoM))

                ep_CoM.append(compute_CoM(points))

            ### For each DevInst, find the eligible pin closest to
            ### the other Endpoint's CoM.
            for i, ep in enumerate(B.endpoints):
                assert 0 <= i <= 1

                other_index = 1 - i
                other_CoM = ep_CoM[other_index]
                other_x, other_y = other_CoM

                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    DI = self.devinst_n2o[DIN]
                    pins = sg.search.resultset

                    dimensions = DI.phypack.dimensions[:2]
                    slope = dimensions[1] / dimensions[0]

                    lpk2coords = dict()
                    best_lpk = None
                    best_dist = None

                    for lpk in pins:
                        coords = compute_CoM_noweight(
                            v[:2]
                            for v in
                            DI.lpk2coords(
                                lpk, float_units=length_in_units
                            ).values()
                        )

                        lpk2coords[lpk] = coords

                        x, y = coords
                        d = cost.dist_2D(x, y, other_x, other_y)

                        if best_dist is None or d < best_dist:
                            best_lpk = lpk
                            best_dist = d

                    best_x, best_y = lpk2coords[best_lpk]


                    if method == 'closestpin_line':
                        lineA_x, lineA_y = (best_x, best_y)

                        dx = other_x - best_x
                        dy = other_y - best_y

                        # Rotate 90 degrees to form line perpendicular
                        #   to normal vector.
                        lineB_x, lineB_y = (best_x + dy, best_y - dx)


                    ### Then sort this DevInst's eligible pins by cost.

                    def pin_cost_closestpin_rad(lpk):
                        """Return cost (distance to closest pin) and lpk.

                        Include lpk as tiebreaker for consistent results.

                        """

                        x, y = lpk2coords[lpk]
                        return (cost.dist_2D(x, y, best_x, best_y), lpk)


                    def pin_cost_closestpin_line(lpk):
                        """Return cost (distance to parallel line at closest
                        pin) and lpk.

                        Include lpk as tiebreaker for consistent results.

                        """

                        x, y = lpk2coords[lpk]
                        return cost.dist_L2P_2D(
                            lineA_x, lineA_y,
                            lineB_x, lineB_y,
                            x, y
                        )


                    def pin_cost_dist_other(lpk):
                        """Return cost (distance to other endpoint) and lpk.

                        Include lpk as tiebreaker for consistent results.

                        """

                        x, y = lpk2coords[lpk]
                        return (cost.dist_2D(x, y, other_x, other_y), lpk)


                    costfunc = {
                        'closestpin_rad': pin_cost_closestpin_rad,
                        'closestpin_line': pin_cost_closestpin_line,
                        'dist_other': pin_cost_dist_other,
                    }[method]


                    sortedpins = sorted(pins, key=costfunc)

                    ### Yield the closest (conn_numpins) pins.
                    bank_CoM = next(iter(CoM[bundlename][DIN].values()))
                    conn_numpins = sg.numpins

                    for i in range(conn_numpins):
                        lpk = sortedpins[i]

                        phypinkeys = DI.log2phy(lpk)
                        dx = 0.0
                        dy = 0.0
                        num_phypins = 0

                        for ppk in phypinkeys:
                            rc = DI.phypack.pins[ppk].coords
                            dx += float(rc[0] / unit_length)
                            dy += float(rc[1] / unit_length)
                            num_phypins += 1

                        pin_dx = dx / num_phypins
                        pin_dy = dy / num_phypins

                        bank = DI.logpack.pins[lpk].bank

                        yield (bundlename, (DIN, bank, lpk))



    def assign_banks(self, unit_length=wallpy.units.Length(1, 'mm')):
        """Assign connections to banks based on current placement and
        I/O standard requirements.

        Arguments:
        unit_length -- express all distances in terms of this unit

        Yield elements: (solved_value, (typestr, args)) where solved_value > 0

        """

        # Postprocess each DevInst's LogicalPackage's pins
        for DIN, DI in self.devinst_n2o.items():
            DI.logpack.postprocess_pins()


        # Require actual number of pins needed per Subgroup
        for B in self.bundles.values():
            B.compute_subgroup_numpins()


        ### Compile set of required Vcco/Vref pairs
        # {DIN: {bank: {(Vcco, Vref), ...}}}
        DIN2bank2supplies = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )
        DIN2bank2vcco = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )
        DIN2bank2vref = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )
        all_supplies = set()

        for DIN, DI in self.devinst_n2o.items():
            for b, standards in DI.bank2iostd.items():
                for standard in standards:
                    s = devmanager.load_iostandard(standard)
                    supplies = (s.vcco, s.vref)
                    DIN2bank2supplies[DIN][b].add(supplies)

                    if s.vcco is not None:
                        DIN2bank2vcco[DIN][b].add(s.vcco)

                    if s.vref is not None:
                        DIN2bank2vref[DIN][b].add(s.vref)

                    all_supplies.add(supplies)


        supplies_vcco = set()
        supplies_vref = set()
        supplies_both = set()

        for s in all_supplies:
            vcco, vref = s

            if (vcco is not None) and (vref is not None):
                supplies_both.add((vcco, vref))
            elif vcco is not None:
                supplies_vcco.add(vcco)
            elif vref is not None:
                supplies_vref.add(vref)
            else:
                raise AssumptionError('I/O standard: (None, None)')

        supplies_vcco = sorted(supplies_vcco)
        supplies_vref = sorted(supplies_vref)
        supplies_both = sorted(supplies_both)


        ### Compile number of required pins
        B2DIN2reqpins = collections.defaultdict(dict)

        ### Compile number of eligible pins and differential pairs
        ### (number of eligible pins includes differential pairs)
        # {bundlename: {DIN: {bank: numpins}}}
        B2DIN2bank2numpins = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(int)
            )
        )
        # {bundlename: {DIN: {bank: numdiffpairs}}}
        B2DIN2bank2numdiffpairs = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(int)
            )
        )

        for bundlename, B in self.bundles.items():
            for epindex, ep in enumerate(B.endpoints):
                for sgindex, sg in enumerate(ep.subgroups):
                    DIN = sg.search.devinstname
                    DI = sg.search.devinst
                    req_numpins = int(sg.numpins)

                    B2DIN2reqpins[bundlename][DIN] = req_numpins

                    for lpk in sg.search.resultset:
                        pin = DI.logpack.pins[lpk]
                        bank = pin.bank
                        B2DIN2bank2numpins[bundlename][DIN][bank] += 1

                        other = None
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig == device.DIFF_P:
                                other = DI.logpack.diff_pairs[pfname][bank]

                        if other in sg.search.resultset:
                            B2DIN2bank2numdiffpairs[bundlename][DIN][bank] += 1

        DIN_bank2numpins = collections.defaultdict(int)
        for bundlename, DIN2bank2numpins in B2DIN2bank2numpins.items():
            for DIN, bank2numpins in DIN2bank2numpins.items():
                for bank, numpins in bank2numpins.items():
                    k = (DIN, bank)
                    if DIN_bank2numpins[k] < numpins:
                        DIN_bank2numpins[k] = numpins

        DIN_bank2numdiffpairs = collections.defaultdict(int)
        for bundlename, DIN2bank2ndp in B2DIN2bank2numdiffpairs.items():
            for DIN, bank2ndp in DIN2bank2ndp.items():
                for bank, numdiffpairs in bank2ndp.items():
                    k = (DIN, bank)
                    if DIN_bank2numdiffpairs[k] < numdiffpairs:
                        DIN_bank2numdiffpairs[k] = numdiffpairs

        print('='*70)
        print('Supplies:')
        print('='*70)
        print(supplies_vcco)
        print(supplies_vref)
        print(supplies_both)
        print()

        print('='*70)
        print('Bank statistics:')
        print('='*70)
        for bundlename, db2np in B2DIN2bank2numpins.items():
            print('Bundle', 'DIN', 'bank', '(numpins, numdiffpairs)',
                  sep='\t')
            db2ndiff = B2DIN2bank2numdiffpairs[bundlename]

            for DIN, b2np in db2np.items():
                b2ndiff = db2ndiff[DIN]

                for bank, numpins in b2np.items():
                    numdiffpairs = b2ndiff[bank]

                    print(bundlename, DIN, bank, (numpins, numdiffpairs),
                          sep='\t')
        print()


        for bundlename, db2np in B2DIN2bank2numpins.items():
            B = self.bundles[bundlename]

            try:
                iostd = B.params['iostandard']
            except KeyError:
                iostd = None

            if iostd is None:
                vcco = None
                vref = None
            else:
                s = devmanager.load_iostandard(iostd)
                vcco = s.vcco
                vref = s.vref

            k = (vcco, vref)

            for DIN, b2np in db2np.items():
                ### Make sure at least one bank supports required I/O standard
                supporting_banks = dict()

                for bank, numpins in b2np.items():
                    if k in DIN2bank2supplies[DIN][bank]:
                        supporting_banks[bank] = numpins

                if not supporting_banks:
                    raise DesignError(
                        'Bundle requires unsupported I/O standard',
                        (bundlename, DIN, set(b2np), iostd)
                    )


        ### Compile set of ILP variables
        DIN_bank2netvars = collections.defaultdict(list)
        DIN_bank2diffpairvars = collections.defaultdict(list)
        DIN_bank2ioclass2vars = collections.defaultdict(
            lambda: collections.defaultdict(list)
        )
        DIN_bank2vcco2var = collections.defaultdict(dict)
        DIN_bank2vref2var = collections.defaultdict(dict)
        bundle2DIN2vars = collections.defaultdict(
            lambda: collections.defaultdict(list)
        )
        variables = list()
        it_index = itertools.count()

        ### Compile supply voltage ILP variables
        for DIN, bank2vcco in DIN2bank2vcco.items():
            for bank, vccoset in bank2vcco.items():
                for vcco in vccoset:
                    v = ('p', (vcco, DIN, bank))

                    variables.append(v)
                    varindex = next(it_index)

                    assert vcco not in DIN_bank2vcco2var[(DIN, bank)]
                    DIN_bank2vcco2var[(DIN, bank)][vcco] = varindex

        for DIN, bank2vref in DIN2bank2vref.items():
            for bank, vrefset in bank2vref.items():
                for vref in vrefset:
                    v = ('q', (vref, DIN, bank))

                    variables.append(v)
                    varindex = next(it_index)

                    assert vref not in DIN_bank2vref2var[(DIN, bank)]
                    DIN_bank2vref2var[(DIN, bank)][vref] = varindex


        ### Compile bank-to-bank connection variables
        for bundlename, B in self.bundles.items():
            assert len(B.endpoints) == 2

            try:
                iostd = B.params['iostandard']
            except KeyError:
                iostd = None

            if iostd is None:
                vcco = None
                vref = None
            else:
                s = devmanager.load_iostandard(iostd)
                vcco = s.vcco
                vref = s.vref

            supplies = (vcco, vref)

            differential = B.params.get('differential', False)


            ep_it = iter(B.endpoints)

            ### Left-hand endpoint
            ep0_candidates = list() # (DIN, bank)
            ep0 = next(ep_it)
            for sgindex, sg in enumerate(ep0.subgroups):
                DIN = sg.search.devinstname

                for bank in B2DIN2bank2numpins[bundlename][DIN]:
                    ep0_candidates.append((DIN, bank))

            ### Right-hand endpoint
            ep1_candidates = list() # (DIN, bank)
            ep1 = next(ep_it)
            for sgindex, sg in enumerate(ep1.subgroups):
                DIN = sg.search.devinstname

                for bank in B2DIN2bank2numpins[bundlename][DIN]:
                    ep1_candidates.append((DIN, bank))


            for c0, c1 in itertools.product(
                ep0_candidates,
                ep1_candidates
            ):

                variables.append(('io', (supplies, bundlename, c0, c1)))
                varindex = next(it_index)

                for DIN_bank in (c0, c1):
                    if differential:
                        DIN_bank2diffpairvars[DIN_bank].append(varindex)
                        mult = 2
                    else:
                        mult = 1

                    DIN_bank2netvars[DIN_bank].append((varindex, mult))

                    DIN_bank2ioclass2vars[DIN_bank][supplies].append(varindex)

                    DIN, bank = DIN_bank
                    bundle2DIN2vars[bundlename][DIN].append(varindex)


        num_variables = next(it_index)
        assert len(variables) == num_variables


        ### Add ILP constraints
        print('Add ILP constraints')
        c_vals = list() # min c'*x
        A_rows = list() # A*x = b; elements are ((colnum, value), ...)
        b_vals = list() # "="
        G_rows = list() # G*x <= h; elements are ((colnum, value), ...)
        h_vals = list() # "<="

        ### Eqn(1a): Ensure <= maximum number of nets for each bank
        for DIN_bank, netvars in DIN_bank2netvars.items():
            row = list()

            for varindex_mult in netvars:
                row.append(varindex_mult)

            assert(row)
            G_rows.append(row)
            h_vals.append(DIN_bank2numpins[DIN_bank])


        ### Eqn(1b): Ensure <= maximum number of diff. pairs for each bank
        for DIN_bank, netvars in DIN_bank2diffpairvars.items():
            row = list()

            for varindex in netvars:
                row.append((varindex, 1))

            assert(row)
            G_rows.append(row)
            h_vals.append(DIN_bank2numpins[DIN_bank])


        ### Eqn(2-5): Ensure each Bundle's DevInsts have exactly the
        ### required number of connections.
        for bundlename, DIN2reqpins in B2DIN2reqpins.items():
            for DIN, reqpins in DIN2reqpins.items():
                row = list()
                for varindex in bundle2DIN2vars[bundlename][DIN]:
                    row.append((varindex, 1))

                assert(row)
                A_rows.append(row)
                b_vals.append(reqpins)


        ### Eqn(6-9): Ensure each bank's connections adhere to vcco, vref
        for DIN_bank, ioclass2vars in DIN_bank2ioclass2vars.items():
            DIN, bank = DIN_bank

            DI_totalreqpins = 0
            for bundlename, DIN2reqpins in B2DIN2reqpins.items():
                DI_totalreqpins += DIN2reqpins.get(DIN, 0)

            for conn_supplies, conn_vars in ioclass2vars.items():
                conn_vcco, conn_vref = conn_supplies

                if conn_vcco is not None:
                    vcco_var = DIN_bank2vcco2var[DIN_bank][conn_vcco]
                    mult = min(
                        DIN_bank2numpins[DIN_bank],
                        DI_totalreqpins,
                    )

                    row = list()
                    for varindex in conn_vars:
                        row.append((varindex, 1))

                    assert(row)
                    row.append((vcco_var, -mult))

                    G_rows.append(row)
                    h_vals.append(0)

                if conn_vref is not None:
                    vref_var = DIN_bank2vref2var[DIN_bank][conn_vref]
                    mult = min(
                        DIN_bank2numpins[DIN_bank],
                        DI_totalreqpins,
                    )

                    row = list()
                    for varindex in conn_vars:
                        row.append((varindex, 1))

                    assert(row)
                    row.append((vref_var, -mult))

                    G_rows.append(row)
                    h_vals.append(0)


        ### Eqn(10): Ensure exactly one Vcco per bank
        for DIN_bank, vcco2var in DIN_bank2vcco2var.items():
            row = list()
            for vcco, var in vcco2var.items():
                row.append((var, 1))

            assert(row)
            A_rows.append(row)
            b_vals.append(1)


        ### Eqn(11): Ensure at most one Vref per bank
        for DIN_bank, vref2var in DIN_bank2vref2var.items():
            row = list()
            for vref, var in vref2var.items():
                row.append((var, 1))

            assert(row)
            A_rows.append(row)
            b_vals.append(1)


        ### Eqn(12): Each connection assignment variable is integer >= 0
        int_vars = list()
        for i, varparts in enumerate(variables):
            typestr, args = varparts
            if typestr == 'io':
                int_vars.append(i)
                G_rows.append([(i, -1)])
                h_vals.append(0)


        ### Eqn(13): Each supply voltage assignment variable is binary
        bin_vars = list()
        for i, varparts in enumerate(variables):
            typestr, args = varparts
            if typestr in {'p', 'q'}:
                bin_vars.append(i)


        print('Compute costs...')
        CoM = self.compute_subgroups_CoM(unit_length)
        c_vals = list()

        for varparts in variables:
            typestr, args = varparts

            if typestr in {'p', 'q'}:
                c_vals.append(0)

            elif typestr == 'io':
                supplies, bundlename, c0, c1 = args

                A, B = (
                    CoM[bundlename][DIN][bank][:2]
                    for DIN, bank in (c0, c1)
                )
                Ax, Ay = A
                Bx, By = B

                c_vals.append(cost.dist_2D(Ax, Ay, Bx, By))

            else:
                raise AssumptionError(
                    'Unknown variable type',
                    varparts
                )


        print('Converting ILP constraints to matrices...')
        assert(len(c_vals) == len(variables))
        assert(len(A_rows) == len(b_vals))
        assert(len(G_rows) == len(h_vals))

        c = cvxopt.matrix(c_vals, (len(c_vals), 1), 'd')
        del c_vals
        b = cvxopt.matrix(b_vals, (len(b_vals), 1), 'd')
        del b_vals
        h = cvxopt.matrix(h_vals, (len(h_vals), 1), 'd')
        del h_vals

        A = rows2spmatrix(A_rows, len(variables))
        del A_rows
        G = rows2spmatrix(G_rows, len(variables))
        del G_rows


        print('Solving ILP...')
        int_vars = frozenset(int_vars)
        bin_vars = frozenset(bin_vars)
        sys.stdout.flush()

        status, x = cvxopt.glpk.ilp(c, G, h, A, b, I=int_vars, B=bin_vars)
        sys.stdout.flush()

        assert(len(x) == len(variables))

        for val, var in zip(x, variables):
            val = int(round(val))

            if val > 0:
                yield (val, var)


    @property
    def DIN2coords(self):
        """Return {DIN: coords} for all DevInsts in this Schematic."""

        return {DIN: DI.coords for DIN, DI in self.devinst_n2o.items()}


    def compute_subgroups_CoM(self, unit_length):
        """Compute centers of mass for each EndpointSubgroup.

        Arguments:
        unit_length -- express all centers of mass in terms of this unit

        Require no Bundle to have the same DIN more than once.

        Return {bundlename: {DIN:
            {bank: (center_dx, center_dy)}
        }}

        """

        CoM = collections.defaultdict(dict)

        for bundlename, B in self.bundles.items():
            B_CoM = CoM[bundlename]

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    DI = self.devinst_n2o[DIN]

                    if DIN in B_CoM:
                        raise NotImplementedError('Repeated DIN in Bundle')

                    dimensions = DI.phypack.dimensions[:2]
                    for v in dimensions:
                        if v <= zero_length:
                            raise ValueError(
                                'Dimensions must be positive!',
                                (bundlename, DIN)
                            )

                    slope = dimensions[1] / dimensions[0]


                    bank2pincoords = collections.defaultdict(list)

                    if not sg.search.resultset:
                        raise AssumptionError(
                            'No pins in resultset',
                            (bundlename, DIN, sg.search.resultset)
                        )

                    for lpk in sg.search.resultset:
                        phypinkeys = DI.log2phy(lpk)
                        dx = 0.0
                        dy = 0.0
                        num_phypins = 0

                        for ppk in phypinkeys:
                            rc = DI.phypack.pins[ppk].coords
                            dx += float(rc[0] / unit_length)
                            dy += float(rc[1] / unit_length)
                            num_phypins += 1

                        pin_dx = dx / num_phypins
                        pin_dy = dy / num_phypins

                        bank = DI.logpack.pins[lpk].bank
                        bank2pincoords[bank].append((pin_dx, pin_dy))


                    bank2CoMs = dict()
                    for bank, pincoords in bank2pincoords.items():
                        center_dx, center_dy = compute_CoM_noweight(pincoords)
                        bank2CoMs[bank] = (center_dx, center_dy)

                    B_CoM[DIN] = bank2CoMs

        return CoM


    def shift_all_CoMs(self, new_placement, orig_CoM, CoM,
                       method, weights,
                       length_in_units, angle_in_units):
        """Update CoM inplace with computed new centers of mass.

        Arguments:
        new_placement -- {DIN: (x, y, angle)} with coords as floats
        orig_CoM -- {bundlename: {DIN: {bank: (dx, dy}}} (orig)
        CoM -- {bundlename: {DIN: {bank: (dx, dy}}}
            (modify inplace)
        method -- CoM estimation method:
                'closestpin_rad': choose pins nearest to closest pin to
                    other CoM
                'closestpin_line': choose pins nearest to line intersecting
                    closest pin to other CoM
                'dist_other': choose pins nearest other CoM
        weights -- (orig, prev, chosen)
        length_in_units -- Unit of x, y
        angle_in_units -- Unit of angle

        Modify CoM inplace with new centers of mass.

        """

        ### Choose pins based on the new placement
        bp_gen = self.estimate_pins_from_placement(
            {
                DIN: coords_angle_div_pi(coords)
                for DIN, coords in new_placement.items()
            },
            CoM,
            method,
            length_in_units,
            angle_in_units
        )


        ### Compute CoM of chosen pins
        # {bundlename: {DIN: {bank: (dx, dy)}}}
        bundle_DIN_chosenpins = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(list)
            )
        )

        unit_length = wallpy.units.Length(1, length_in_units)

        for bundlename, pin in bp_gen:
            DIN, bank, lpk = pin
            DI = self.devinst_n2o[DIN]

            phypinkeys = DI.log2phy(lpk)

            phypincoords = (
                tuple(
                    float(v / unit_length)
                    for v in DI.phypack.pins[ppk].coords[:2]
                )
                for ppk in phypinkeys
            )

            bundle_DIN_chosenpins[bundlename][DIN][bank].append(
                compute_CoM_noweight(phypincoords)
            )


        ### Update CoM based on original, previous, and chosen-pin CoM
        # Do not update CoMs for banks with no pins estimated to be chosen
        for bundlename, DIN2chosenpins in bundle_DIN_chosenpins.items():
            for DIN, bank2pincoordset in DIN2chosenpins.items():
                for bank, pincoordset in bank2pincoordset.items():
                    chosen_dx, chosen_dy = compute_CoM_noweight(pincoordset)

                    prev_dx, prev_dy = CoM[bundlename][DIN][bank][:2]

                    orig_dx, orig_dy = orig_CoM[bundlename][DIN][bank][:2]

                    dx, dy = shift_CoM(
                        (orig_dx, orig_dy),
                        (prev_dx, prev_dy),
                        (chosen_dx, chosen_dy),
                        weights=weights
                    )

                    CoM[bundlename][DIN][bank] = (dx, dy)


    def find_best_placements(self, seeds, *args, **kwargs):
        """Call solve_placement once per seed and sort results by cost.

        Arguments:
        seeds -- iterable of seed values to pass to solve_placement
        *args -- positional arguments to pass to solve_placement
        **kwargs -- keyword arguments to pass to solve_placement

        Return sorted list of (total_cost, SimplePlacement, seed).

        """

        results = list()

        for seed in seeds:
            P = self.solve_placement(*args, seed=seed, **kwargs)
            results.append((P.total_cost, P, seed))

        results.sort(key=operator.itemgetter(0))

        return results


    def solve_placement(self, board_dim, ref_unit='mm', allow_rotate=True,
                        seed=None,
                        params=None):
        """Position all unfixed components.

        Iteratively find legal (non-overlapping) positions for all DevInsts.
        Do not move any DevInst listed in self.locked_DINs.

        Arguments:
        board_dim -- Maximum board dimensions as (float, float)
        ref_unit -- All length floats measured in this unit
        allow_rotate -- If true, then rotations are valid moves for SA
        seed -- Random seed, if desired
        params -- Override placement parameters

        Return the final best placement.

        """

        param_defaults = {
            # Increase space around DevInst by this factor
            # per connected or requested net
            'extra_margin_net_factor': 1.0005,

            'weight_dist': 1,

            'weight_overlap_start': 0,
            'weight_overlap_end': 1000000,

            'buffer_margin': 0.5,

            'init_temp': 1E6,

            'cold_temp': 1E-6,

            # Disable slow lowest T region
            'temp_thresh_c': -1,

            'max_iter_no_improvement': 50,

            # Per iteration decreasing T
            'moves_per_devinst': 300,

            # Percentage of exchanges (vs. translations)
            # Start at _start and increase to _end at end
            'exchange_percent_start': 0,
            'exchange_percent_end': 75,

            # Percentage of rotations (vs. exchanges or translations)
            'rotate_percent': 15,

            # Number of quantized angles to try at end
            'quant_rotate_steps': 4,

            # Spread out DevInsts by this factor at end
            'spread_factor': 4,

            # Condense by this factor each iteration
            'condense_factor': 0.02,

            # Method for setting initial DevInst locations
            'init_placement': INIT_RAND,

            # Weights for (orig, prev, chosen) CoM for shifting CoM
            'shift_CoM_weights': (1, 0, 2),

            # Percentage of SA iterations at which to shift CoMs
            'shift_CoM_percent': 0,

            # Method by which to estimate chosen pins
            'shift_CoM_method': 'dist_other',

            # Keep move limits at least this value (in terms of ref_unit)
            'min_move_limits': (1.0, 1.0),

            # If True, use the best placement at each temperature
            'use_best_each_step': False,
        }

        if params:
            params = dict(params)
        else:
            params = dict()

        # Fill in unspecified params with defaults
        for k, v in param_defaults.items():
            if k not in params:
                params[k] = v

        # Parameters that default to functions of other parameters
        param_variable_defaults = {
            'temp_thresh_h': params['init_temp'] / 4,
            'temp_thresh_m': math.sqrt(params['init_temp']),
        }

        for k, v in param_variable_defaults.items():
            if k not in params:
                params[k] = v


        known_params = set(param_defaults) | set(param_variable_defaults)
        unknown_params = set(params) - known_params

        if unknown_params:
            raise KeyError('Unknown param: {}'.format(
                ', '.join(unknown_params)
            ))


        own_rand = random.Random(seed)


        # TODO: Heed allow_rotate

        unit_length = wallpy.units.Length(1, ref_unit)
        unit_angle = wallpy.units.Angle(1, 'pi_rad')

        # frozenset({DIN1, DIN2}): number of nets connecting the DevInsts
        connectivity = collections.defaultdict(float)

        # DIN: number of nets to be connected to that DevInst
        DIN2numnets = collections.defaultdict(int)


        ### Update DIN2numnets from currently-connected nets
        for DIN, lpk in self.logpin2net:
            DIN2numnets[DIN] += 1


        ### First compute all EndpointSubgroups' numpins
        for bundlename, B in self.bundles.items():
            try:
                B.compute_subgroup_numpins()
            except:
                print('Could not compute subgroup numpins for', bundlename)
                raise


        ### Compute connectivity graph
        for bundlename, B in self.bundles.items():
            # ep2DINs holds all the distinct DINs by endpoint index
            ep2DINs = collections.defaultdict(set)  # epnum: {DINs}

            for epnum, ep in enumerate(B.endpoints):
                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    ep2DINs[epnum].add(DIN)
                    DIN2numnets[DIN] += sg.numpins


            ### Split connectivity strength across all pairs of DevInsts
            #
            # We have N BundleEndpoints, each of which has >= 1 DIN.
            #
            # Suppose we have ep2DINs = {0: {'U1', 'U2'}, 1: {'U3', 'U4'}}.
            # Pairs are U1,U3; U1,U4; U2,U3; U2,U4.
            #
            # Suppose we have ep2DINs = {0: {'U1'}, 1: {'U2'}, 3: {'U3'}}.
            # Pairs are U1,U2; U1,U3; U2,U3
            #
            # First, pair endpoints. Then pair DINs between each endpoint.

            ep_pairs = itertools.combinations(ep2DINs, 2)

            # frozenset({DIN1, DIN2}): frequency of occurrence
            DIN_pairs = collections.defaultdict(int)

            for epnum1, epnum2 in ep_pairs:
                DINs1 = ep2DINs[epnum1]
                DINs2 = ep2DINs[epnum2]

                for pair in itertools.product(DINs1, DINs2):
                    DIN_pairs[frozenset(pair)] += 1


            ### Update connectivity graph
            conn_strength = len(B.netnames) / sum(DIN_pairs.values())

            for pair, freq in DIN_pairs.items():
                connectivity[pair] += conn_strength * freq

        max_connectivity = max(connectivity.values())

        print()
        print('Connectivity graph:')
        for pair, strength in connectivity.items():
            print('\t', pair, ': ', strength, sep='')
        print()


        # Pre-convert dimensions
        DIN2dim = dict()

        for DIN, DI in self.devinst_n2o.items():
            DIN2dim[DIN] = [
                float(v / unit_length)
                for v in DI.phypack.dimensions[:2]
            ]


        # Keep track of DevInsts' present locations (2D only)
        DIN2coords = dict()     # DIN: (float, float, float)

        ### Set all placements to current location.
        # This accomodates locked DevInsts.
        set_placement_asis(DIN2coords, self.devinst_n2o,
                           unit_length, unit_angle)

        ### Override placements for non-locked DevInsts if desired.
        if params['init_placement'] == INIT_ZERO:
            set_placement_zero(DIN2coords, self.devinst_n2o,
                               self.locked_DINs)

        elif params['init_placement'] == INIT_RAND:
            set_placement_rand(DIN2coords, self.devinst_n2o,
                               board_dim, DIN2dim,
                               self.locked_DINs,
                               params['quant_rotate_steps'],
                               own_rand)

        else:
            raise ValueError('Invalid value for init_placement')

        for DIN, coords in DIN2coords.items():
            x, y, angle = coords

            if DIN in self.locked_DINs:
                # Convert to radians, since later will be converted back.
                angle *= math.pi

            DIN2coords[DIN] = (x, y, angle)


        original_coords = {
            DIN: tuple(round(c, 2) for c in coords)
            for DIN, coords in DIN2coords.items()
        }


        ### Compute the center of mass for each DevInst in each Bundle
        CoM = self.compute_subgroups_CoM(unit_length)

        ### Highlight centers of mass
        for bundlename, DIN2CoM in CoM.items():
            for DIN, bank2coords in DIN2CoM.items():
                for bank, coords in bank2coords.items():
                    center_dx, center_dy = coords[:2]
                    self.di_points[DIN].append((center_dx, center_dy))


        ### Compute the number of required pins per DevInst per Bundle
        B2D2numreqpins = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )

        for bundlename, B in self.bundles.items():
            for ep in B.endpoints:
                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    B2D2numreqpins[bundlename][DIN] += sg.numpins


        ################################################
        ### Placement Algorithm: Simulated Annealing ###
        ################################################
        #
        # Reference:
        #   Sarrafzadeh, Majid.
        #   An Introduction to VLSI Physical Design
        #   The McGraw-Hill Companies, Inc.
        #   1996
        #   pp. 57-58, 73-74
        #
        # 1. Start at a high temperature T.
        # 2. Reset move count to 0.
        # 3. Select a placement change to make, either:
        #   A. Move a random DevInst to a random position, or
        #   B. Exchange the positions of two random DevInsts
        # 4. If new placement gives a total cost decrease, use it.
        #    Otherwise, use it if f(cost increase, T) > random(0, 1).
        # 5. If move count < limit, loop back to 3.
        # 6. Once enough moves have been made, decrease T.
        # 7. Loop back to 2.
        #
        # Note: Limit how far DevInsts can move based on g(T)



        def slide_T(T):
            """Return a sliding value from 0 to 1 as T decreases from
            init_temp to cold_temp.

            Increases piecewise linearly as T decreases exponentially.

            Arguments:
            T -- current SA temperature

            Return a float on [0, 1].

            """

            if T >= params['init_temp']:
                return 0.0
            if T <= params['cold_temp']:
                return 1.0

            return math.log(
                params['init_temp'] / T,
                params['init_temp'] / params['cold_temp']
            )


        def slide_interval(T, start, end):
            """Return a sliding value from start to end as T decreases
            from init_temp to cold_temp.

            Arguments:
            T -- current SA temperature
            start -- value for T=init_temp
            end -- value for T=cold_temp

            Return a float between start and end, inclusive.

            """

            f = slide_T(T)
            return (((1 - f) * start) + (f * end))


        def max_rotate(T):
            """Return maximum rotation (in radians).

            Arguments:
            T -- SA temperature

            Maximum rotation decreases logarithmically.

            """

            return 2 * math.pi
            return 8 * math.pi * slide_T(T)


        def next_move(T, max_move, cur_placement, DIN2movesleft,
                      exchange_percent, DIN2dim,
                      quant_rotate_steps,
                      num_tries=20):
            """Return a new placement based on current DIN2coords, T.

            Arguments:
            T -- current SA temperature
            cur_placement -- current placement
            DIN2movesleft -- how many more moves allowed for this T
            exchange_percent -- probability of exchanging two DevInsts
            max_move -- maximum move as (float, float)
            DIN2dim -- {DIN: (width, height)}
            quant_rotate_steps -- number of valid DevInst angles

            Either exchange two randomly-chosen DevInsts' positions, or
            move a randomly-chosen DevInst somewhere on the board.

            Return None if no new placement could be generated.

            """

            if not DIN2movesleft:
                # No movable DevInsts!
                return None

            # Start with current placement
            ret = cur_placement.copy()

            if own_rand.randint(0, 100) < params['rotate_percent']:
                move_op = 'rotate'
            elif own_rand.randint(0, 100) < exchange_percent:
                move_op = 'exchange'
            else:
                move_op = 'translate'


            if move_op == 'exchange':
                ### Exchange two DevInsts

                for i in range(num_tries):
                    DINs = list(DIN2movesleft)

                    DIN_L = own_rand.choice(DINs)
                    DINs.remove(DIN_L)

                    if not DINs:
                        # No more DevInsts from which to choose
                        return None

                    DIN_R = own_rand.choice(DINs)

                    # Perform the exchange
                    ret[DIN_L] = cur_placement[DIN_R]
                    ret[DIN_R] = cur_placement[DIN_L]

                    DIN2movesleft[DIN_L] -= 1
                    if DIN2movesleft[DIN_L] < 1:
                        del DIN2movesleft[DIN_L]

                    DIN2movesleft[DIN_R] -= 1
                    if DIN2movesleft[DIN_R] < 1:
                        del DIN2movesleft[DIN_R]

                    return ret

            elif (move_op == 'rotate') or (move_op == 'translate'):
                ### Move a DevInst
                DINs = list(DIN2movesleft)
                DIN = own_rand.choice(DINs)

                cur_xy = ret[DIN]

                angle = cur_placement[DIN][-1]
                w, h = DIN2dim[DIN][:2]

                ### Rotate or translate?
                rotate = False
                translate = False



                if move_op == 'rotate':
                    # Try to get a different angle
                    old_angle = angle
                    for i in range(num_tries):
                        angle = own_rand.randrange(quant_rotate_steps)
                        angle *= 2 * math.pi / quant_rotate_steps

                        if angle != old_angle:
                            break


                dim = cost.bbox_size(w, h, angle)

                min_xy = tuple(
                    ((dim[i] - board_dim[i]) / 2)
                    for i in range(2)
                )

                max_xy = tuple(
                    ((board_dim[i] - dim[i]) / 2)
                    for i in range(2)
                )


                if move_op == 'translate':
                    new_coords = [
                        round(
                            own_rand.uniform(
                                cur_xy[i] - max_move[i],
                                cur_xy[i] + max_move[i]
                            ),
                            2
                        )
                        for i in range(2)
                    ]

                else:
                    new_coords = list(cur_xy[:2])


                # Keep within board boundaries
                for i in range(2):
                    if new_coords[i] < min_xy[i]:
                        new_coords[i] = min_xy[i]
                    elif new_coords[i] > max_xy[i]:
                        new_coords[i] = max_xy[i]

                new_coords.append(angle)

                ret[DIN] = tuple(new_coords)

                DIN2movesleft[DIN] -= 1
                if DIN2movesleft[DIN] < 1:
                    del DIN2movesleft[DIN]

                return ret

            # Couldn't generate a new placement
            return None


        def next_temp(T, accept_ratio):
            """Return a slightly cooler temperature."""

            # Cooling schedule per V. Betz, J. Rose. 1997.
            if accept_ratio > 0.96:
                return T * 0.5
            elif accept_ratio > 0.8:
                return T * 0.9
            elif accept_ratio > 0.15:
                return T * 0.95
            else:
                return T * 0.8


        def next_max_move(max_move, accept_ratio, board_dim, min_limits):
            """Return smaller maximum move distances.

            Arguments:
            max_move -- current value of (max_move_x, max_move_y)
            accept_ratio -- ratio of accepted moves to attempted moves
            board_dim -- maximum board dimensions
            min_limit -- keep max_move[i] >= min_limit[i]

            Return (new_max_move_x, new_max_move_y).
            """

            # Cooling schedule per V. Betz, J. Rose. 1997.
            factor = 1 - 0.44 + accept_ratio

            ret = []
            for i in range(2):
                v = max_move[i] * factor

                # Clamp limit within minimum limit, board dimension
                if v < min_limits[i]:
                    v = min_limits[i]
                elif v > board_dim[i]:
                    v = board_dim[i]

                ret.append(v)

            return tuple(ret)


        def keep_change_p(T, old_cost, new_cost):
            if new_cost < old_cost:
                return True

            neg_diff = old_cost - new_cost
            thresh = math.exp(neg_diff / T)

            return (own_rand.uniform(0, 1) < thresh)


        def terminate(T, iter_since_improvement):
            """Return True if SA should terminate."""

            if T < params['cold_temp']:
                print('Terminating SA due to cold temp', T)
                return True
            if iter_since_improvement > params['max_iter_no_improvement']:
                print('Terminating SA due to no recent improvement')
                return True
            return False


        def calc_exchange_percent(T, init_temp, start, end):
            """Calculate the probability of DevInst exchange.

            Arguments:
            T -- current SA temperature
            init_temp -- initial SA temperature XXX unused
            start -- starting probability (in percent)
            end -- ending probability (in percent)

            Return probability in percent rounded to the nearest integer.

            """

            # f starts at 0 and increases to near 1
            f = slide_T(T)
            ret = (start * (1 - f)) + (end * f)
            return int(round(ret))


        def calc_shift_CoM_weights(T, init_temp):
            """Calculate weights for shifting centers of mass.

            Start with full weight on original CoMs, end with full weight
            on chosen CoMs.

            Arguments:
            T -- current SA temperature
            init_temp -- initial SA temperature XXX unused

            Return weights: (orig, prev, chosen).

            """

            # f starts at 0 and increases to near 1
            f = slide_T(T)
            return ((1 - f), 0, f)


        def perform_moves(
            T, max_move,
            cur_placement, cur_cost,
            best_placement, best_cost,
            params,
            DIN2dim,
            max_connectivity,
            CoM,
            B2D2numreqpins
        ):
            """Perform simulated annealing steps for a given value of T.

            Arguments:
            T -- current SA temperature
            max_move -- maximum move as (float, float)
            cur_placement -- current placement
            cur_cost -- cost of current placement
            best_placement -- placement with lowest total cost seen thus far
            best_cost -- cost of best placement seen thus far
            params -- SA parameters
            DIN2dim -- {DIN: (width, height)}
            max_connectivity -- maximum connectivity value
            CoM -- {bundlename: {DIN: (center_dx, center_dy)}}
            B2D2numreqpins -- {bundlename: {DIN: conn_numreqpins}}

            Return (
                best_placement, best_cost,
                cur_placement, cur_cost,
                accept_ratio
            ).

            """

            # Allow all DevInsts to move again
            DIN2movesleft = dict()
            for DIN in self.devinst_n2o:
                if DIN not in self.locked_DINs:
                    DIN2movesleft[DIN] = params['moves_per_devinst']

            print('SA:', T)

            exchange_percent = calc_exchange_percent(
                T, params['init_temp'],
                params['exchange_percent_start'],
                params['exchange_percent_end']
            )

            weight_overlap = slide_interval(
                T,
                params['weight_overlap_start'],
                params['weight_overlap_end']
            )

            moves_total = 0
            moves_accepted = 0

            new_placement = next_move(T, max_move, cur_placement,
                                      DIN2movesleft, exchange_percent,
                                      DIN2dim,
                                      params['quant_rotate_steps'])
            while new_placement:
                new_cost = cost.total_cost(
                    new_placement, False, DIN2dim, max_connectivity, CoM,
                    B2D2numreqpins,
                    params['weight_dist'], weight_overlap
                )

                moves_total += 1

                if keep_change_p(T, cur_cost, new_cost):
                    moves_accepted += 1

                    cur_placement = new_placement
                    cur_cost = new_cost

                    if new_cost < best_cost:
                        best_cost = new_cost
                        best_placement = new_placement.copy()
                        print('\t', 'New best: ', best_cost, sep='')

                new_placement = next_move(T, max_move, cur_placement,
                                          DIN2movesleft, exchange_percent,
                                          DIN2dim,
                                          params['quant_rotate_steps'])

            assert (moves_total > 0)
            accept_ratio = moves_accepted / moves_total

            return (
                best_placement, best_cost,
                cur_placement, cur_cost,
                accept_ratio
            )


        ### Simulated Annealing variables
        T = params['init_temp']
        max_move = board_dim[:2]
        best_placement = DIN2coords.copy()
        best_cost = cost.total_cost(
            best_placement, False, DIN2dim, max_connectivity, CoM,
            B2D2numreqpins,
            params['weight_dist'], params['weight_overlap_start']
        )
        iter_since_improvement = 0
        new_cost = best_cost

        # Update DIN2dim to include params['buffer_margin']
        #   and extra buffer for high-connectivity devices
        for DIN in self.devinst_n2o:
            for i in range(2):
                net_factor = (
                    params['extra_margin_net_factor'] ** DIN2numnets[DIN]
                )
                DIN2dim[DIN][i] *= net_factor
                DIN2dim[DIN][i] += params['buffer_margin']


        print('Initial cost:', cost.total_cost(
            DIN2coords, False, DIN2dim, max_connectivity, CoM,
            B2D2numreqpins,
            params['weight_dist'], params['weight_overlap_start']
        ))
        print('Max move:', max_move)


        ### Perform SA
        cur_placement = DIN2coords
        cur_cost = new_cost

        orig_CoM = {
            bundlename: DIN2CoM.copy()
            for bundlename, DIN2CoM in CoM.items()
        }

        # Current CoM is just CoM

        while not terminate(T, iter_since_improvement):
            iter_since_improvement += 1

            weight_overlap = slide_interval(
                T,
                params['weight_overlap_start'],
                params['weight_overlap_end']
            )

            # Recompute best cost at this new temperature
            best_placement = cur_placement
            best_cost = cost.total_cost(
                best_placement, False, DIN2dim, max_connectivity, CoM,
                B2D2numreqpins,
                params['weight_dist'], weight_overlap
            )

            old_best_cost = best_cost


            ret = perform_moves(
                T, max_move,
                cur_placement, cur_cost, best_placement, best_cost,
                params, DIN2dim, max_connectivity, CoM,
                B2D2numreqpins
            )

            (
                best_placement, best_cost,
                new_placement, new_cost,
                accept_ratio,
            ) = ret

            print('ACCEPT RATIO', accept_ratio)

            if best_cost < old_best_cost:
                iter_since_improvement = 0


            if own_rand.randint(0, 100) < params['shift_CoM_percent']:
                self.shift_all_CoMs(new_placement, orig_CoM, CoM,
                                    params['shift_CoM_method'],
                                    calc_shift_CoM_weights(
                                        T, params['init_temp']
                                    ),
                                    ref_unit, 'pi_rad')

            if params['use_best_each_step']:
                cur_placement = best_placement
            else:
                cur_placement = new_placement

            T = next_temp(T, accept_ratio)
            max_move = next_max_move(max_move, accept_ratio,
                                     board_dim, params['min_move_limits'])


        best_placement = cur_placement
        DIN2coords = best_placement


        print('SA final best cost', best_cost,
              cost.total_cost(
                  best_placement, True, DIN2dim, max_connectivity, CoM,
                  B2D2numreqpins,
                  params['weight_dist'], params['weight_overlap_end']
              ))


        ### Final stage: greedy translate, rotate
        if params.get('use_greedy_postprocessing'):
            raise NotImplementedError(
                'Greedy postprocessing does not support new CoM',
                'Refs #47'
            )

            def greedy_rotate(initial_placement, initial_cost, DINs,
                              locked_DINs, angles):
                """Greedily rotate DevInsts to find local minimum cost.

                Arguments:
                initial_placement -- Initial {DIN: (x, y, angle)}
                initial_cost -- Cost of initial_placement
                DINs -- list of DINs to use as rotation possibilities
                locked_DINs -- DevInsts in this set will not be moved
                angles -- list of angle values to evaluate for each DevInst

                Return (best_placement, best_cost, improved_p).

                """

                best_placement = initial_placement
                best_cost = initial_cost

                improved = False

                for DIN in DINs:
                    if DIN in locked_DINs:
                        continue

                    best_rot = best_placement
                    best_rot_cost = None
                    best_angle = None

                    for angle in angles:
                        new_placement = best_placement.copy()

                        new_coords = list(best_placement[DIN])
                        new_coords[-1] = angle
                        new_placement[DIN] = new_coords

                        new_cost = cost.total_cost(
                            new_placement, False, DIN2dim, max_connectivity,
                            CoM,
                            params['weight_dist'], params['weight_overlap_end']
                        )

                        if (best_rot_cost is None) or (new_cost < best_rot_cost):
                            best_rot_cost = new_cost
                            best_rot = new_placement

                    if best_rot_cost < best_cost:
                        # Improvement found!
                        improved = True

                        print(
                            'Rotate Cost Difference',
                            best_rot_cost - best_cost
                        )

                        best_cost = best_rot_cost
                        best_placement = best_rot

                return (best_placement, best_cost, improved)


            ### Spread out DevInsts to enable rotational improvements
            spread_placement(best_placement, self.locked_DINs,
                             params['spread_factor'])


            sorted_devinsts = sorted(self.devinst_n2o)

            angle_choices = tuple(
                (i * 2 * math.pi / params['quant_rotate_steps'])
                for i in range(params['quant_rotate_steps'])
            )


            ### Force all angles to be acceptable choices
            for DIN in sorted_devinsts:
                coords = list(best_placement[DIN])

                if DIN not in self.locked_DINs:
                    coords[-1] = angle_choices[0]

                best_placement[DIN] = coords


            best_cost = cost.total_cost(
                best_placement, False, DIN2dim, max_connectivity, CoM,
                params['weight_dist'], params['weight_overlap_end']
            )


            outer_improved = True
            while outer_improved:
                outer_best_cost = best_cost

                ### Greedily rotate each DevInst until no improvement
                improved = True
                while improved:
                    best_placement, best_cost, improved = greedy_rotate(
                        best_placement, best_cost, sorted_devinsts,
                        self.locked_DINs, angle_choices
                    )


                ### Condense DevInsts via forces
                force_condense_placement(
                    best_placement, self.locked_DINs,
                    CoM, DIN2dim, params['condense_factor']
                )

                best_cost = cost.total_cost(
                    best_placement, False, DIN2dim, max_connectivity, CoM,
                    params['weight_dist'], params['weight_overlap_end']
                )


                # Continue while still making improvements
                outer_improved = (best_cost < outer_best_cost)



        print('Final best cost', best_cost,
              cost.total_cost(
                  best_placement, True, DIN2dim, max_connectivity, CoM,
                  B2D2numreqpins,
                  params['weight_dist'], params['weight_overlap_end']
              ))


        print()

        print('DevInst movements:')
        for DIN, coords in best_placement.items():
            v = tuple(round(c, 2) for c in coords)
            orig = original_coords[DIN]

            if v == orig:
                print('{} did not move from {}'.format(DIN, orig))
            else:
                print('{} moved from {} to {}'.format(DIN, orig, v))

        print()


        ### Update DevInsts' coordinates

        placement_pi_rad = {
            DIN: coords_angle_div_pi(coords)
            for DIN, coords in best_placement.items()
        }


        self.set_placement_2D(placement_pi_rad,
                              length_in_units=ref_unit,
                              angle_in_units='pi_rad')


        ret = SimplePlacement(D2coords={
            DIN: DI.coords
            for DIN, DI in self.devinst_n2o.items()
        })

        ret.total_cost = best_cost


        return ret


    def set_placement_2D(self, new_placement,
                         length_in_units='mm',
                         angle_in_units='pi_rad'):
        """Set DevInsts' coordinates to those in new_placement.

        Arguments:
        new_placement -- {DIN: (x, y, angle)} in units or floats
        length_in_units -- Unit of x, y if given as floats
        angle_in_units -- Unit of angle if given as float

        """

        for DIN, newcoords in new_placement.items():
            DI = self.devinst_n2o[DIN]

            z = DI.coords[2]    # Leave Z as-is

            x = wallpy.units.Length(newcoords[0], length_in_units)
            y = wallpy.units.Length(newcoords[1], length_in_units)
            ang = wallpy.units.Angle(newcoords[-1], angle_in_units)

            DI.coords = (x, y, z, ang)


    def find_best_place_and_banks(self, seeds, *args, **kwargs):
        """Call solve_place_and_banks once per seed and sort results by cost.

        Arguments:
        seeds -- iterable of seed values to pass to solve_place_and_banks
        *args -- positional arguments to pass to solve_place_and_banks
        **kwargs -- keyword arguments to pass to solve_place_and_banks

        Return sorted list of (total_cost, Placement, seed).

        """

        results = list()

        for seed in seeds:
            try:
                P = self.solve_place_and_banks(*args, seed=seed, **kwargs)
                results.append((P.total_cost, P, seed))

            except placeandbanks.ReachedCancelTemp:
                print('SA did not complete! Moving past seed', seed)
                print()

            sys.stdout.flush()

        results.sort(key=operator.itemgetter(0))

        return results


    def solve_place_and_banks(
        self,
        board_dim,
        ref_unit='mm',
        seed=None,
        params=None
    ):
        """Position all unfixed components.

        Iteratively find legal (non-overlapping) positions for all DevInsts,
        bank assignments for all Bundles, and bank supplies for all banks.
        Do not move any DevInst listed in self.locked_DINs.

        Arguments:
        board_dim -- Maximum board dimensions as (float, float)
        ref_unit -- All length floats measured in this unit
        seed -- Random seed, if desired
        params -- Override placement parameters

        Return the final best placement.

        """

        param_defaults = {
            # Increase space around DevInst by this factor
            # per connected or requested net
            'extra_margin_net_factor': 1.0005,

            'weight_dist': 1,

            'weight_boardsize': 1,

            'weight_overlap_start': 1E-2,

            'weight_iostandard_start': 1E-1,

            'weight_pin_start': 1E-1,
            'weight_diffpair_start': 1E-1,

            'weight_boardoversize_start': 1E-1,

            # If there is any overlap at the end of the current temperature,
            # multiply weight_overlap by this factor.
            'weight_overlap_increase_factor': 1.1,

            # If there is any I/O standard incompatibility at the end of the
            # current temperature, multiply weight_iostandard by this factor.
            'weight_iostandard_increase_factor': 1.1,

            # If there is any pin / diff. pair overuse at the end of the
            # current temperature, multiply weight_? by this factor.
            'weight_pin_increase_factor': 1.1,
            'weight_diffpair_increase_factor': 1.1,

            # If the board is larger than its maximum size at the end of the
            # current temperature, multiply weight_boardoversize by this
            # factor.
            'weight_boardoversize_increase_factor': 1.1,

            'buffer_margin': 0.5,

            'init_temp': 1E9,

            'cold_temp': 1E-6,

            # If the temperature drops below this value, cancel (failure)!
            'cancel_temp': 1E-10,

            # Per iteration decreasing T
            'moves_per_devinst': 300,

            # Number of quantized angles
            'quant_rotate_steps': 4,

            # Method for setting initial DevInst locations
            'init_placement': INIT_RAND,

            # At each temperature change, decrease translate limits by
            # multiplying by this factor.
            'translate_reduce_factor': 0.9,

            # Keep move limits at least this value (in terms of ref_unit)
            'min_move_limits': (10.0, 10.0),

            # If True, use the best placement at each temperature
            'use_best_each_step': False,

            # Number of bank connection moves at each temperature.
            # If None, default to total number of requested nets * 2.
            'num_move_conn': None,

            # Major movement ratios:
            'move_DI': 10,
            'move_conn': 10,

            # DevInst movement ratios:
            'DI_translate': 10,
            'DI_rotate': 2,
            'DI_exchange': 3,

            # Bank connection movement ratios:
            'conn_translate': 10,
            'conn_exchange': 3,

            # If True, drop to interactive shell after finished.
            'postplace_interact': False,

            # If True, show all connection movements as the occur.
            'debug_move_conn': False,

            # If True, show all inputs to and outputs from accept_p.
            'debug_accept_p': False,

            # If True, show all major choices.
            'debug_majorchoice': False,

            # If True, call Placement.debug_nulls.
            'debug_nulls': False,

            # If True, show all weight increases.
            'show_weight_increase': False,

            # If True, show total of each movement type for each T.
            'show_movetypes': True,

            # Show progress every (#) attempted moves for each T.
            'moves_attempted_progress': 32768,

            # If set, call this function after each attempted move.
            'post_move_callback': None,

            # Pass this object to post_move_callback.
            'post_move_arg': None,

            # Maximum iteration counts.
            'max_outer_iter': 1000,
            'max_inner_iter': 20,

            # If defined, include result from user_cost_func in SA cost.
            'user_cost_func': self.user_cost_func,
            'user_cost_func_arg': self.user_cost_func_arg,
        }

        if params:
            params = dict(params)
        else:
            params = dict()

        # Fill in unspecified params with defaults
        for k, v in param_defaults.items():
            if k not in params:
                params[k] = v


        known_params = set(param_defaults)
        unknown_params = set(params) - known_params

        if unknown_params:
            raise KeyError('Unknown param: {}'.format(
                ', '.join(unknown_params)
            ))


        own_rand = random.Random(seed)


        unit_length = wallpy.units.Length(1, ref_unit)
        unit_angle = wallpy.units.Angle(1, 'pi_rad')


        # DIN: number of nets to be connected to that DevInst
        DIN2numnets = collections.defaultdict(int)


        ### Update DIN2numnets from currently-connected nets
        for DIN, lpk in self.logpin2net:
            DIN2numnets[DIN] += 1


        ### First compute all EndpointSubgroups' numpins
        for bundlename, B in self.bundles.items():
            try:
                B.compute_subgroup_numpins()
            except:
                print('Could not compute subgroup numpins for', bundlename)
                raise


        ### Update DIN2numnets from Bundles' required connections
        for bundlename, B in self.bundles.items():
            for ep in B.endpoints:
                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    DIN2numnets[DIN] += sg.numpins


        # Pre-convert dimensions
        DIN2dim = dict()

        for DIN, DI in self.devinst_n2o.items():
            DIN2dim[DIN] = [
                float(v / unit_length)
                for v in DI.phypack.dimensions[:2]
            ]


        init_placement = placeandbanks.Placement()


        ### Set all placements to current location.
        # This accomodates locked DevInsts.
        set_placement_asis(init_placement.D2coords, self.devinst_n2o,
                           unit_length, unit_angle)

        ### Override placements for non-locked DevInsts if desired.
        if params['init_placement'] == INIT_ZERO:
            set_placement_zero(init_placement.D2coords, self.devinst_n2o,
                               self.locked_DINs)

        elif params['init_placement'] == INIT_RAND:
            set_placement_rand(init_placement.D2coords, self.devinst_n2o,
                               board_dim, DIN2dim,
                               self.locked_DINs,
                               params['quant_rotate_steps'],
                               own_rand)

        else:
            raise ValueError('Invalid value for init_placement')

        for DIN, coords in init_placement.D2coords.items():
            x, y, angle = coords

            if DIN in self.locked_DINs:
                # Convert to radians, since later will be converted back.
                angle *= math.pi

            init_placement.D2coords[DIN] = (x, y, angle)


        # Update DIN2dim to include params['buffer_margin']
        #   and extra buffer for high-connectivity devices
        for DIN in self.devinst_n2o:
            net_factor = (
                params['extra_margin_net_factor'] ** DIN2numnets[DIN]
            )
            for i in range(2):
                DIN2dim[DIN][i] *= net_factor
                DIN2dim[DIN][i] += params['buffer_margin']


        ### Compile number of eligible pins and differential pairs
        ### (number of eligible pins includes differential pairs)
        # {bundlename: {DIN: {bank: numpins}}}
        B2D2b2numpins = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(int)
            )
        )
        # {bundlename: {DIN: {bank: numdiffpairs}}}
        B2D2b2numdiffpairs = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(int)
            )
        )

        # {DIN: {bank: {lpk, ...}}}
        D2b2pinkeys = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )
        D2b2pairkeys = collections.defaultdict(
            lambda: collections.defaultdict(set)
        )

        for bundlename, B in self.bundles.items():
            diff_p = bool(B.params.get('differential'))

            for epindex, ep in enumerate(B.endpoints):
                for sgindex, sg in enumerate(ep.subgroups):
                    DIN = sg.search.devinstname
                    DI = sg.search.devinst
                    req_numpins = int(sg.numpins)

                    for lpk in sg.search.resultset:
                        pin = DI.logpack.pins[lpk]
                        bank = pin.bank
                        B2D2b2numpins[bundlename][DIN][bank] += 1

                        D2b2pinkeys[DIN][bank].add(lpk)

                        other = None
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig in device.DIFF_SIGCONFIGS:
                                other = DI.logpack.diff_pairs[pfname][bank]
                                break

                        if (other is not None) and diff_p:
                            B2D2b2numpins[bundlename][DIN][bank] += 1
                            B2D2b2numdiffpairs[bundlename][DIN][bank] += 1
                            D2b2pinkeys[DIN][bank].add(other)
                            D2b2pairkeys[DIN][bank].add(lpk)

        D2b2maxpins = {
            DIN: {bank: len(v) for bank, v in b2pinkeys.items()}
            for DIN, b2pinkeys in D2b2pinkeys.items()
        }

        D2b2maxdiffpairs = {
            DIN: {
                bank: len(D2b2pairkeys[DIN][bank])
                for bank in b2pinkeys
            }
            for DIN, b2pinkeys in D2b2pinkeys.items()
        }


        ### Reserve pins and diffpairs as required.
        for bundlename, D2b2numpins in B2D2b2numpins.items():
            for DIN, b2numpins in D2b2numpins.items():
                DI = self.devinst_n2o[DIN]

                for bank, numpins in b2numpins.items():
                    b2numpins[bank] -= int(float(
                        numpins * DI.bank2pin_reserve_frac[bank]
                    ))

        for bundlename, D2b2numdiffpairs in B2D2b2numdiffpairs.items():
            for DIN, b2numdiffpairs in D2b2numdiffpairs.items():
                DI = self.devinst_n2o[DIN]

                for bank, numdiffpairs in b2numdiffpairs.items():
                    b2numdiffpairs[bank] -= int(float(
                        numdiffpairs * DI.bank2diffpair_reserve_frac[bank]
                    ))

        for DIN, b2maxpins in D2b2maxpins.items():
            DI = self.devinst_n2o[DIN]

            for bank, maxpins in b2maxpins.items():
                b2maxpins[bank] -= int(float(
                    maxpins * DI.bank2pin_reserve_frac[bank]
                ))

        for DIN, b2maxdiffpairs in D2b2maxdiffpairs.items():
            DI = self.devinst_n2o[DIN]

            for bank, maxdiffpairs in b2maxdiffpairs.items():
                b2maxdiffpairs[bank] -= int(float(
                    maxdiffpairs * DI.bank2diffpair_reserve_frac[bank]
                ))


        ### Compute D2b2validsupplies:
        ### {DIN: {bank: {supplyname: {value, ...}}}}
        D2b2validsupplies = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(set)
            )
        )

        total_supply_choices = 0
        for DIN, DI in self.devinst_n2o.items():
            for b, standards in DI.bank2iostd.items():
                for standard in standards:
                    s = devmanager.load_iostandard(standard)
                    for k in s.supplynames:
                        D2b2validsupplies[DIN][b][k].add(getattr(s, k))
                        total_supply_choices += 1


        ### Create and partially populate init_placement.info.
        bx, by = board_dim

        init_placement.info = placeandbanks.PlacementInfo(
            D2b2maxpins = D2b2maxpins,
            D2b2maxdiffpairs = D2b2maxdiffpairs,
            boardmaxsize = dict(x=bx, y=by),
            D2dim = DIN2dim,
        )


        ### Prepare init_placement.D2b2supplies.
        for DIN, b2validsupplies in D2b2validsupplies.items():
            for b, validsupplies in b2validsupplies.items():
                for k in validsupplies:
                    init_placement.D2b2supplies[DIN][b][k] = None


        CoM = self.compute_subgroups_CoM(unit_length)


        ### Prepare bundlename2index and init_placement.bundles.
        bundlename2index = dict()
        for bundleindex, bundlename in enumerate(
            sorted(self.bundles, key=name2ana)
        ):

            bundlename2index[bundlename] = bundleindex

            B = self.bundles[bundlename]

            try:
                iostd = devmanager.load_iostandard(B.params['iostandard'])
            except KeyError:
                iostd = None

            info = placeandbanks.PlacedBundleInfo(
                differential=B.params.get('differential', False),
                iostd=iostd,
                weight=float(B.params.get('weight', 1.0)),
                D2b2CoM={
                    DIN: {
                        bank: coords
                        for bank, coords in b2CoM.items()
                    }
                    for DIN, b2CoM in CoM[bundlename].items()
                },
                D2b2numpins={
                    DIN: {
                        bank: numpins
                        for bank, numpins in b2numpins.items()
                    }
                    for DIN, b2numpins in B2D2b2numpins[bundlename].items()
                },
                D2b2numdiffpairs={
                    DIN: {
                        bank: B2D2b2numdiffpairs[bundlename][DIN][bank]
                        for bank in b2numpins
                    }
                    for DIN, b2numpins in
                    B2D2b2numpins[bundlename].items()
                },
                bundlename=bundlename,
            )

            PB = placeandbanks.PlacedBundle(
                info=info,
                place=init_placement,
            )


            ### Add nets.
            ep_terminals = list()   # [[term,...], ...]
            D2bankiter = dict()
            for DIN, b2numpins in info.D2b2numpins.items():
                banks = sorted(b2numpins, key=placeandbanks.banksort_key)
                own_rand.shuffle(banks)
                D2bankiter[DIN] = itertools.cycle(banks)

            for ep in B.endpoints:
                terminals = list()

                for sg in ep.subgroups:
                    DIN = sg.search.devinstname
                    for i in range(sg.numpins):
                        terminals.append((DIN, next(D2bankiter[DIN])))

                ep_terminals.append(terminals)

            ep_numterms = tuple(map(len, ep_terminals))
            for numterms in ep_numterms[1:]:
                if numterms != ep_numterms[0]:
                    raise AssumptionError(
                        'Mismatch in number of terminals per endpoint',
                        ep_numterms
                    )

            for terminals in zip(*ep_terminals):
                PB.nets.append(
                    placeandbanks.PlacedNet(
                        terminals=list(terminals),
                        place=init_placement,
                        bundle=PB,
                    )
                )


            assert(len(init_placement.bundles) == bundleindex)
            init_placement.bundles.append(PB)


        ### Update and use consistent data structures.
        init_placement.update()

        print()
        print('Initial placement:')
        init_placement.export(sys.stdout, sep='i'*75)
        print()
        print()

        if params['debug_nulls']:
            init_placement.debug_nulls()

        init_placement = init_placement.clone()

        if params['debug_nulls']:
            init_placement.debug_nulls()


        ### Prepare params and move_ratios.
        mx, my = params['min_move_limits']
        params['min_move_limit_x'] = mx
        params['min_move_limit_y'] = my

        if params['num_move_conn'] is None:
            params['num_move_conn'] = 2 * sum(DIN2numnets.values())
            print('Set num_move_conn to', params['num_move_conn'])



        print('Performing simulated annealing...')

        ### Perform simulated annealing!
        best_placement = placeandbanks.solve_place_and_banks(
            init_placement,
            self.locked_DINs,
            {
                DIN: {
                    bank: {
                        name: frozenset(validvalues)
                        for name, validvalues in validsupplies.items()
                    }
                    for bank, validsupplies in b2validsupplies.items()
                }
                for DIN, b2supplies in D2b2validsupplies.items()
            },
            own_rand,
            params['post_move_callback'],
            params['post_move_arg'],
            params, # Specify params twice, once for params_dict,
            params  # and once for move_ratios_dict.
        )

        print('Simulated annealing complete!')


        print()
        for DIN in sorted(init_placement.D2coords):
            orig = init_placement.D2coords[DIN]
            now = best_placement.D2coords[DIN]

            diff = (round(operator.sub(*d), 3) for d in zip(orig, now))
            maxdiff = max(map(abs, diff))

            if maxdiff > 0.0:
                print(DIN, 'moved with max diff of', maxdiff)
            else:
                print(DIN, 'did not move')


        ### Print pin overuse stats.
        print()
        D2b2totalsingleconn = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )
        D2b2totaldiffconn = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )
        D2b2totalconn = collections.defaultdict(
            lambda: collections.defaultdict(int)
        )

        for bundle in best_placement.bundles:
            bundlename = bundle.info.bundlename
            B = self.bundles[bundlename]
            diff_p = bool(B.params.get('differential'))
            assert(diff_p == bundle.info.differential)

            D2b2numpins = bundle.info.D2b2numpins

            for DIN, b2nc in bundle.D2b2numconn.items():
                b2numpins = D2b2numpins[DIN]

                for bank, nc in b2nc.items():
                    numpins = b2numpins[bank]

                    usedpins = nc
                    if diff_p:
                        usedpins *= 2
                    D2b2totalconn[DIN][bank] += usedpins

                    if usedpins > numpins:
                        print(
                            'Bundle pin overuse:',
                            (bundlename, DIN, bank),
                            '{} > {} by {}'.format(
                                usedpins,
                                numpins,
                                usedpins - numpins
                            )
                        )

                    if diff_p:
                        D2b2totaldiffconn[DIN][bank] += nc
                    else:
                        D2b2totalsingleconn[DIN][bank] += nc


        for DIN, b2totalconn in D2b2totalconn.items():
            for bank, totalconn in b2totalconn.items():
                avail = D2b2maxpins[DIN][bank]

                if totalconn > avail:
                    print(
                        'Overall pin overuse:',
                        (DIN, bank),
                        '{} > {} by {}'.format(
                            totalconn,
                            avail,
                            totalconn - avail
                        )
                    )


        best_placement.export(sys.stdout)


        ### Update DevInsts' coordinates

        best_placement.D2coords = {
            DIN: coords_angle_div_pi(coords)
            for DIN, coords in best_placement.D2coords.items()
        }

        self.set_placement_2D(best_placement.D2coords,
                              length_in_units=ref_unit,
                              angle_in_units='pi_rad')

        best_placement.D2coords = {
            DIN: DI.coords
            for DIN, DI in self.devinst_n2o.items()
        }


        if params['postplace_interact']:
            print()
            banner = 'Interacting with results from seed {} ' + \
                    '(press CTRL-D to exit):'.format(seed)
            code.interact(
                banner=banner,
                local=locals()
            )


        return best_placement



    def solve_b2b_ilp_connections(
        self,
        placement,
        round_digits=DEFAULT_ROUND_DIGITS,
        debugfilename=None
    ):

        """Determine set of nets to satisfy Bundles

        Solve all Bundle requests simultaneously:
            One ILP var for each possibility:
                (bundlename, (DIN0, lpk0), (DIN1, lpk1))
            Cost for each ILP var is distance
                FROM (coords of DI0's physical pin corresponding to lpk0)
                TO (coords of DI1's physical pin corresponding to lpk1)
            Restrict sum of (*, (DIN0, lpk0), *) <= 1 (single-ended) OR
                sum of (*, (DIN0, lpk0), *) +
                    (*, (DIN0, lpk0diffpair), *) <= 1
            Restrict sum of (*, *, (DIN1, lpk1)) <= 1 (single-ended) OR
                sum of (*, *, (DIN1, lpk1)) +
                    (*, *, (DIN1, lpk1diffpair) <= 1
            Restrict sum of (bundlename,(DIN0,lpk0),(DIN1,lpk1)) == number
                of pins connecting bank0 of DIN0 to bank1 of DIN1
                for bank0 in DIN0, bank1 in DIN1

        Arguments:
        placement -- bank-to-bank connections
        round_digits -- round ILP cost to this many digits
        debugfilename -- if set, then print ILP variables and constraints
                            to this file (automatically opens and closes)

        Return generator of (bundlename, netname, (DIN, lpk)).
        """

        print()
        print('Preparing to solve pin assignment...')

        # Postprocess each DevInst's LogicalPackage's pins
        for DIN, DI in self.devinst_n2o.items():
            DI.logpack.postprocess_pins()


        # Prepare B2T2conn: {bundlename: {((DIN0,bank0),...): # of nets}}.
        B2T2conn = dict()

        for bundleindex, placedbundle in enumerate(placement.bundles):
            bundlename = placedbundle.info.bundlename
            B = self.bundles[bundlename]

            if len(B.endpoints) != 2:
                print('Skipping Bundle {!r} ({})'.format(
                    bundlename,
                    '# endpoints',
                ))
                continue

            skip_bundle = False

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    if sg.params.get('all'):
                        skip_bundle = True
                        break

            if skip_bundle:
                print('Skipping Bundle {!r} ({})'.format(
                    bundlename,
                    'subgroup has ALL set',
                ))
                continue


            T2conn = collections.defaultdict(int) # {(term0, term1): numconn}

            for netindex, net in enumerate(placedbundle.nets):
                if len(net.terminals) != 2:
                    raise AssumptionError(
                        'Unexpected number of terminals',
                        (bundlename, netindex, net.terminals)
                    )

                terms = tuple(net.terminals)
                T2conn[terms] += 1

            if bundlename in B2T2conn:
                raise AssumptionError(
                    'Repeated bundlename',
                    (bundlename, bundleindex)
                )

            B2T2conn[bundlename] = dict(T2conn)


            print()
            print('B2T2conn[{}]:'.format(bundlename))
            for T in sorted(T2conn):
                conn = T2conn[T]
                print('\t{}: {}'.format(T, conn))
            print()


        changed = False


        # Each element is (bundlename, (DIN0,lpk0), (DIN1,lpk1)).
        ilp_vars = list()   # These correspond to result x.
        binvars = set() # indices of binary variables

        c_vals = list() # min c'*x
        G_rows = list() # G*x <= h; elements are ((colnum, value), ...)
        h_vals = list() # "<="
        A_rows = list() # A*x = b;  elements are ((colnum, value), ...)
        b_vals = list() # "="

        # (DIN, lpk): set(var_indices)
        logpin2vars = collections.defaultdict(set)

        # (bundlename, ((DIN0,bank0), (DIN1,bank1))): set(var_indices)
        BT2vars = collections.defaultdict(set)

        it_varindex = itertools.count()


        def add_var(bundlename, B, terminals, pins):
            terminals = tuple(terminals)
            pins = tuple(pins)

            assert(len(terminals) == len(pins))

            for term, pin in zip(terminals, pins):
                t_DIN, t_bank = term
                p_DIN, p_lpk = pin
                if t_DIN != p_DIN:
                    raise AssumptionError(
                        'DINs do not match',
                        (bundlename, terminals, pins)
                    )

            pin0, pin1 = pins


            ilp_vars.append((bundlename, pin0, pin1))

            varindex = next(it_varindex)
            assert(varindex == len(ilp_vars) - 1)

            logpin2vars[pin0].add(varindex)
            logpin2vars[pin1].add(varindex)

            if B.params.get('differential'):
                # Choosing this connection prohibits other
                # use of corresponding differential pins
                for DIN, lpk in pins:
                    DI = self.devinst_n2o[DIN]
                    pin = DI.logpack.pins[lpk]

                    # Determine corresponding differential pin
                    other_lpks = set()
                    for pfname, pf in pin.pinfuncs.items():
                        if pf.sigconfig in device.DIFF_SIGCONFIGS:
                            bank = pf.params.get('bank', None)
                            other = DI.logpack.diff_pairs[pfname][bank]
                            other_lpks.add(other)

                    if len(other_lpks) != 1:
                        raise AssumptionError(
                            '{} corresponding diff. pins found!'.format(
                                len(other_lpks)
                            ),
                            (DIN, lpk, other_lpks)
                        )

                    # Only corresponding differential pin
                    other = next(iter(other_lpks))

                    logpin2vars[(DIN, other)].add(varindex)

            BT2vars[(bundlename, terminals)].add(varindex)


        print()

        for bundlename, T2conn in sorted(B2T2conn.items()):
            B = self.bundles[bundlename]

            if len(B.endpoints) != 2:
                raise AssumptionError(
                    'Should already have excluded Bundle (# endpoints)',
                    (bundlename, len(B.endpoints))
                )

            skip_bundle = False

            for ep in B.endpoints:
                for sg in ep.subgroups:
                    if sg.params.get('all'):
                        skip_bundle = True
                        break

            if skip_bundle:
                raise AssumptionError(
                    'Should already have excluded Bundle (ALL set)',
                    bundlename
                )


            for epindex, ep in enumerate(B.endpoints):
                ep.D2sgindex = dict()

                for sgindex, sg in enumerate(ep.subgroups):
                    DIN = sg.search.devinstname
                    DI = self.devinst_n2o[DIN]

                    if DIN in ep.D2sgindex:
                        raise AssumptionError(
                            'DIN in more than one subgroup in endpoint',
                            (bundlename, epindex, sgindex, DIN)
                        )

                    ep.D2sgindex[DIN] = sgindex
                    sg.D2b2lpks = dict()

                    b2lpks = collections.defaultdict(set)
                    for lpk in sg.search.resultset:
                        if len(DI.log2phy[lpk]) != 1:
                            print('Skipping pin', lpk, '(too many phypins)')
                            continue

                        pin = DI.logpack.pins[lpk]
                        b2lpks[pin.bank].add(lpk)

                        self.potential_pins2.add((DIN, lpk))

                    sg.D2b2lpks[DIN] = dict(b2lpks)

                    for bank, lpks in sg.D2b2lpks[DIN].items():
                        print('{}: [{}][{}]: {}.{}: {}'.format(
                            bundlename,
                            epindex,
                            sgindex,
                            DIN,
                            bank,
                            len(lpks),
                        ))


            for T in T2conn:
                candidates = list()

                for termindex, term in enumerate(T):
                    DIN, bank = term

                    ep = B.endpoints[termindex]
                    sg = ep.subgroups[ep.D2sgindex[DIN]]

                    assert(sg.search.devinstname == DIN)

                    ep_candidates = list()

                    try:
                        lpks = sg.D2b2lpks[DIN][bank]
                    except KeyError:
                        raise AssumptionError(
                            'No pins in bank',
                            (
                                DIN,
                                bank,
                                sorted(
                                    sg.D2b2lpks[DIN],
                                    key=placeandbanks.banksort_key
                                ),
                            )
                        )

                    for lpk in lpks:
                        ep_candidates.append((DIN, lpk))

                        self.potential_pins.add((DIN, lpk))

                    candidates.append(ep_candidates)


                for pins in itertools.product(*candidates):
                    add_var(bundlename, B, T, pins)

        print()


        print('# of ILP vars:', len(ilp_vars))
        print('# of distinct (DIN, lpk) pairs:', len(logpin2vars))


        ### Add ILP constraints

        # Restrict vars to at most 1 per (DIN, lpk)
        print('Restrict to at most one net per pin...')

        for pinkey, varindices in logpin2vars.items():
            # elements are (colnum, value)
            G_newrow = tuple((i, 1) for i in varindices)

            if G_newrow:
                G_rows.append(G_newrow)
                h_vals.append(1)    # At most one variable may be chosen
            else:
                print('No matching pins for', pinkey, 'on #', i)


        # For each bank-to-bank set of connections, require exactly as many
        # variables as specified.
        print('Require enough pins per bank-to-bank connections...')

        num_conn_rows = 0

        for bundlename, T2conn in B2T2conn.items():
            for terminals, numconn in T2conn.items():
                if numconn < 1:
                    raise AssumptionError(
                        'numconn < 1',
                        (bundlename, terminals)
                    )

                BT = (bundlename, terminals)

                if BT not in BT2vars:
                    raise AssertionError(
                        'BT not in BT2vars',
                        (BT, sorted(BT2vars))
                    )

                varindices = BT2vars[BT]

                if not varindices:
                    raise AssumptionError(
                        'No varindices',
                        (BT, varindices)
                    )

                num_conn_rows += 1

                # elements are (colnum, value)
                A_newrow = tuple((i, 1) for i in varindices)

                assert(A_newrow)

                A_rows.append(A_newrow)
                b_vals.append(numconn)


        if num_conn_rows != len(BT2vars):
            raise AssumptionError(
                'Mismatch in number of rows vs. BT2vars',
                (num_conn_rows, len(BT2vars))
            )


        # Compute ILP variable costs
        print('Compute costs...')

        base_distance = wallpy.units.Length(1, 'mm')

        coords = dict() # (DIN, lpk): (x, y, z) as float()

        def pin_coord(pinkey):
            if pinkey in coords:
                return coords[pinkey]

            DIN, lpk = pinkey

            coordlist = self.devinst_n2o[DIN].lpk2coords(lpk)
            if len(coordlist) == 1:
                c = tuple(
                    float(v / base_distance)
                    for v in next(iter(coordlist.values()))
                )
                coords[pinkey] = c
                return c

            raise ValueError('More than one physical pin', DIN, lpk)

        def ilp_cost(bundlename, pinL, pinR):
            coords_L = pin_coord(pinL)
            coords_R = pin_coord(pinR)
            rel_coords = tuple(
                (coords_R[i] - coords_L[i])
                for i in range(len(coords_L))
            )
            return math.sqrt(sum((v * v) for v in rel_coords))

        c_vals = [round_sigdig(ilp_cost(*v), round_digits) for v in ilp_vars]


        # All ILP vars are binary
        num_vars = len(ilp_vars)
        binvars = frozenset(range(num_vars))


        ### Convert to matrices
        print('Converting ILP constraints to matrices...')
        c = cvxopt.matrix(c_vals, (num_vars, 1), 'd')
        if not debugfilename:
            del c_vals
        h = cvxopt.matrix(h_vals, (len(h_vals), 1), 'd')
        if not debugfilename:
            del h_vals
        b = cvxopt.matrix(b_vals, (len(b_vals), 1), 'd')
        if not debugfilename:
            del b_vals

        G = rows2spmatrix(G_rows, num_vars)
        if not debugfilename:
            del G_rows
        A = rows2spmatrix(A_rows, num_vars)
        if not debugfilename:
            del A_rows


        ### Solve ILP
        print('Solving ILP...')
        sys.stdout.flush()

        status, x = cvxopt.glpk.ilp(c, G, h, A, b, B=binvars)
        sys.stdout.flush()

        print('ILP returned status:', status)
        if not x:
            print('SOLUTION NOT FOUND!')
            return

        chosen = tuple(bool(round(v)) for v in x)


        print(
            'ILP final cost:', sum(
                cost if var else 0
                for cost, var in zip(c, chosen)
            )
        )


        if debugfilename:
            print('Saving ILP variables and constraints to', debugfilename)

            bundlename2id = {
                name: i
                for i, name in enumerate(self.bundles)
            }

            def nicevarname(var):
                bundlename, c0, c1 = var

                DIN0, lpk0 = c0
                DIN1, lpk1 = c1

                bundleid = bundlename2id[bundlename]

                return ' '.join((
                    '{}',
                    '{}:{}',
                    '{}:{}',
                )).format(
                    bundleid,
                    DIN0, lpk0,
                    DIN1, lpk1
                )


            fieldnames = ['head']
            fieldnames.extend(range(len(ilp_vars)))
            fieldnames.append('tail')

            with open(debugfilename, 'w', newline='') as Fcsv:
                writer = csv.DictWriter(
                    Fcsv,
                    fieldnames,
                    dialect=csv.excel_tab
                )

                # Header: row names
                writer.writerow({
                    i: nicevarname(var)
                    for i, var in enumerate(ilp_vars)
                })


                writer.writerow(dict())


                writer.writerow(dict(head='h >= Gx:'))

                for i, row in enumerate(G_rows):
                    h = h_vals[i]

                    D = {'head': h}
                    D.update(row)

                    writer.writerow(D)


                writer.writerow(dict())


                writer.writerow(dict(head='b = Ax:'))

                for i, row in enumerate(A_rows):
                    b = b_vals[i]

                    D = {'head': b}
                    D.update(row)

                    writer.writerow(D)


                writer.writerow(dict())


                writer.writerow(dict(head='chosen:'))

                chosen_row = {
                    i: 'X' if v else ''
                    for i, v in enumerate(chosen)
                }
                writer.writerow(chosen_row)


        ### Connect nets
        netname_iters = dict()  # bundlename: iterator over net names
        for bundlename in B2T2conn:

            if bundlename not in netname_iters:
                B = self.bundles[bundlename]
                netname_iters[bundlename] = iter(B.netnames)

        for i, var in enumerate(ilp_vars):
            if chosen[i]:
                bundlename, pin0, pin1 = var
                B = self.bundles[bundlename]

                netname = next(netname_iters[bundlename])
                yield (bundlename, netname, pin0)
                yield (bundlename, netname, pin1)


    def solve_b2b_ilp(
        self,
        placement,
        round_digits=DEFAULT_ROUND_DIGITS,
        debugfilename=None
    ):

        """Determine set of nets to satisfy Bundles, then connect the nets.

        Use solve_b2b_ilp_connections to determine set of nets.

        Arguments:
        placement -- bank-to-bank connections
        round_digits -- round ILP cost to this many digits
        debugfilename -- if set, then print ILP variables and constraints
                            to this file (automatically opens and closes)

        Return True if any connections were made, or
            False if no change was made to the schematic.
        """

        changed = False

        def dict_set():
            return collections.defaultdict(set)

        # {bundlename: {DIN: {lpks}}}
        BN_DIN2lpks = collections.defaultdict(dict_set)

        for bundlename, netname, DIN_lpk in self.solve_b2b_ilp_connections(
            placement,
            round_digits=round_digits,
            debugfilename=debugfilename
        ):
            changed |= self.connect_net(netname, (DIN_lpk, ))

            DIN, lpk = DIN_lpk
            BN_DIN2lpks[bundlename][DIN].add(lpk)

        print()


        ### Reconnect differential pins, including suffixes and negative pins
        for bundlename, B in self.bundles.items():
            if B.params.get('differential'):
                for netname_prefix in B.netnames:
                    se_pins = list(self.nets[netname_prefix])

                    ### Disconnect old (single-ended) nets
                    assert(self.disconnect_net(netname_prefix, se_pins))

                    for DIN_lpk in se_pins:

                        ### Connect positive net (specified pin)
                        netname = netname_prefix + B.params['diff_p_suffix']
                        changed |= self.connect_net(netname, [DIN_lpk])


                        ### Connect negative net (corresponding diff. pin)
                        netname = netname_prefix + B.params['diff_n_suffix']
                        DIN, lpk = DIN_lpk
                        DI = self.devinst_n2o[DIN]
                        pin = DI.logpack.pins[lpk]

                        # Determine corresponding differential pin
                        other = None
                        for pfname, pf in pin.pinfuncs.items():
                            if pf.sigconfig in device.DIFF_SIGCONFIGS:
                                bank = pf.params.get('bank', None)
                                other = DI.logpack.diff_pairs[pfname][bank]
                                break

                        if other is None:
                            raise DesignError(
                                'Could not find diff. pin',
                                (bundlename, netname, DIN, lpk)
                            )

                        changed |= self.connect_net(netname, [(DIN, other)])

        return changed



def empty_b2numconn(): return collections.defaultdict(int)
def empty_D2b2numconn(): return collections.defaultdict(empty_b2numconn)
def empty_B2D2b2numconn(): return collections.defaultdict(empty_D2b2numconn)

def empty_None(): return None
def empty_supplies(): return collections.defaultdict(empty_None)
def empty_b2supplies(): return collections.defaultdict(empty_supplies)
def empty_D2b2supplies(): return collections.defaultdict(empty_b2supplies)


class SimplePlacement(wallpy.cleanobject.CleanObject):
    """Positions and assignments of DevInsts and bank connections"""

    posarg_names = []
    kwarg_names = [ 'D2coords', 'B2D2b2numconn', 'D2b2supplies' ]

    moduleprefix = moduleprefix


    def __init__(
        self,
        D2coords=None,
        B2D2b2numconn=None,
        D2b2supplies=None
    ):
        """Define the positions and assignments of DevInsts and connections.

        Arguments:
        D2coords -- {DIN: (x, y, angle)}
        B2D2b2numconn - {bundlename: {DIN: {bank: num_connections}}}
        D2b2supplies - {DIN: {bank: {supplyname: value}}}

        """

        if D2coords:
            D2coords = dict(D2coords)
        else:
            D2coords = dict()

        if B2D2b2numconn:
            B2D2b2numconn = {
                bundlename: {
                    DIN: {
                        bank: numconn
                        for bank, numconn in b2numconn.items()
                    } for DIN, b2numconn in D2b2numconn.items()
                } for bundlename, D2b2numconn in B2D2b2numconn.items()
            }
        else:
            B2D2b2numconn = empty_B2D2b2numconn()

        if D2b2supplies:
            D2b2supplies = {
                DIN: {
                    bank: {
                        supplyname: vals
                        for supplyname, vals in supplies.items()
                    } for bank, supplies in b2supplies.items()
                } for DIN, b2supplies in D2b2supplies.items()
            }
        else:
            D2b2supplies = empty_D2b2supplies()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


    def clone(self):
        return type(self)(
            D2coords=self.D2coords,
            B2D2b2numconn=self.B2D2b2numconn,
            D2b2supplies=self.D2b2supplies,
        )


class Bundle(wallpy.cleanobject.CleanObject):
    """Group of nets connecting groups of pins from each endpoint"""

    posarg_names = [ 'netnames', 'endpoints' ]
    kwarg_names = [ 'params' ]

    moduleprefix = moduleprefix


    def __init__(self, netnames, endpoints, params = None):
        """Request a connection between one pin in each of the endpoints.

        Arguments:
        netnames -- the names for the nets, once the Bundle has been satisfied
        endpoints -- [BundleEndpoints]
        params -- named parameters
        """


        if netnames:
            netnames = list(netnames)
        else:
            netnames = list()

        if endpoints:
            endpoints = wallpy.cleanobject.CleanList(endpoints)
        else:
            endpoints = wallpy.cleanobject.CleanList()

        if params:
            params = dict(params)
        else:
            params = dict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)


        self.finished = False



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if not self.netnames:
            errcallback('Must have at least one netname')
            return False

        if not self.endpoints:
            errcallback('Must have at least one endpoint')
            return False


        if len(self.netnames) != len(set(self.netnames)):
            errcallback('netnames must not contain duplicate names')
            return False


        space_re = re.compile(r'\s')


        for n in self.netnames:
            if not n:
                errcallback('netnames must be non-empty')
                return False

            if str(n) != n:
                errcallback('netnames must be strings')
                return False

            if space_re.search(n):
                errcallback('netnames cannot include space characters')
                return False



        ### Check subgroups for 'all' parameter

        subgroup_all = False    # Does any subgroup have all = True?

        for ep in self.endpoints:
            for sg in ep.subgroups:
                if sg.params.get('all', False):
                    subgroup_all = True

        if subgroup_all:
            # Require exactly one netname

            if len(self.netnames) != 1:
                errcallback(
                    'For all=True, require exactly one netname',
                    self.netnames
                )

                return False



        ### Make sure each BundleEndpoint's subgroups have valid numpins
        for epnum, ep in enumerate(self.endpoints):
            numnets_left = len(self.netnames)
            num_autogroups = 0

            subgroup_all = False    # Does any subgroup have all = True?
            for sg in ep.subgroups:
                if sg.params.get('all', False):
                    subgroup_all = True


            if subgroup_all:
                if len(ep.subgroups) != 1:
                    errcallback(
                        'For all=True, require exactly one subgroup',
                        epnum,
                        len(ep.subgroups)
                    )

                    return False

                # No need to test this endpoint for pin counts
                continue



            for sg in ep.subgroups:
                np = sg.numpins

                if np == AUTO:
                    num_autogroups += 1
                else:
                    numnets_left -= np

            if numnets_left < 0:
                errcallback(
                    'Requested more pins than nets',
                    epnum,
                    len(self.netnames) - numnets_left,
                    len(self.netnames)
                )

                return False


            if num_autogroups:
                extranets = numnets_left % num_autogroups

                if extranets:
                    errcallback(
                        'Could not split up pins among AUTO subgroups',
                        epnum,
                        numnets_left,
                        num_autogroups,
                        extranets
                    )

                    return False

        return True



    def compute_subgroup_numpins(self):
        """Replace subgroups' numpins=ANY with actual number of pins
        the subgroups will use.

        """

        numnets = len(self.netnames)

        for epnum, ep in enumerate(self.endpoints):

            num_auto = 0
            num_explicit = 0    # Total nets EXCLUDING auto

            for sgnum, sg in enumerate(ep.subgroups):
                if sg.numpins == AUTO:
                    num_auto += 1
                else:
                    num_explicit += sg.numpins

            num_available = numnets - num_explicit

            if num_available < 0:
                raise DesignError(
                    'More connections requested than nets in Bundle',
                    epnum,
                    numnets,
                    num_explicit,
                    num_available
                )

            if num_auto:
                if num_available < num_auto:
                    raise DesignError(
                        'Not enough nets for AUTO subgroups',
                        epnum,
                        numnets,
                        num_explicit,
                        num_available,
                        num_auto
                    )

                nets_per_auto, extranets = divmod(num_available, num_auto)

                if extranets:
                    raise DesignError(
                        'Could not split up pins among AUTO subgroups',
                        epnum,
                        numnets,
                        num_explicit,
                        num_available,
                        num_auto,
                        extranets
                    )


                ### Update all subgroups having numpins == AUTO
                for sgnum, sg in enumerate(ep.subgroups):
                    if sg.numpins == AUTO:
                        sg.numpins = nets_per_auto



    def __str__(self):
        parts = [
            type(self).__name__ + '(',
            repr(self.netnames) + ',',
            '[',
        ]

        parts.extend((str(ep) + ',') for ep in self.endpoints)

        parts.extend([
            '],',
            repr(self.params),
            ')',
        ])

        return ' '.join(parts)



    def DIN2banks(self, exclude_banks=frozenset({'NA'})):
        """Return {DIN: {banks in DevInst}}.

        Arguments:
        exclude_banks -- Do not return bank names in this set

        """

        ret = collections.defaultdict(set)

        for epnum, ep in enumerate(self.endpoints):
            for sgnum, sg in enumerate(ep.subgroups):
                DIN = sg.search.devinstname
                DI = sg.search.devinst
                pack = DI.logpack

                for lpk in sg.search.resultset:
                    pin = pack.pins[lpk]

                    pinbanks = set()
                    for pf in pin.pinfuncs.values():
                        try:
                            pinbanks.add(pf.params['bank'])
                        except KeyError:
                            pass

                    if len(pinbanks) > 1:
                        raise AssumptionError(
                            'More than one bank for pin',
                            (DIN, lpk)
                        )

                    ret[DIN] |= pinbanks

                ret[DIN] -= exclude_banks

        return ret





class BundleEndpoint(wallpy.cleanobject.CleanObject):
    """One end of a Bundle of nets."""

    posarg_names = [ 'subgroups' ]
    kwarg_names = []

    def __init__(self, subgroups):
        if subgroups:
            subgroups = wallpy.cleanobject.CleanList(subgroups)
        else:
            subgroups = wallpy.cleanobject.CleanList()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if not self.subgroups:
            errcallback('Must have at least one subgroup')
            return False


        # If any subgroup has params['all'] = True
        #   then this endpoint must have exactly one subgroup.

        have_all = False
        for sg in self.subgroups:
            if sg.params.get('all', False):
                have_all = True
                break

        if have_all and len(self.subgroups) != 1:
            errcallback(
                'Endpoint must have exactly one subgroup if all = True'
            )
            return False

        return True





class EndpointSubgroup(wallpy.cleanobject.CleanObject):
    """One group of pins; part of a BundleEndpoint."""

    posarg_names = [ 'numpins', 'search' ]
    kwarg_names = [ 'params' ]

    moduleprefix = moduleprefix


    def __init__(self, numpins, search, params = None):
        """Create a group of pins given by search.

        Arguments:
        numpins -- number of pins to be chosen from search: may be AUTO
                    (but see note)
        search -- PinSearch to provide matching pins
        params -- named parameters

        Note:
            numpins may be AUTO if any of the following criteria are met:
                1.  params['all'] is True (meaning all matching pins are
                        requested to be connected)
                2.  Bundle's nets are to be distributed evenly between
                        EndpointSubgroups
                    AND every EndpointSubgroup with numpins=AUTO will
                        receive the same number of net connections
        """


        if numpins != AUTO:
            numpins = int(numpins)


        if params:
            params = dict(params)
        else:
            params = dict()


        posargs, kwargs = wallpy.misc.args2pos_kw(type(self), locals())

        super().__init__(*posargs, **kwargs)



    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args, **kwargs):
                pass

        if not super().valid_p(errcallback = errcallback):
            return False


        if self.numpins != AUTO:
            try:
                self.numpins = int(self.numpins)
            except ValueError:
                errcallback('numpins must be AUTO or an int >= 1',
                                 self.numpins)
                return False

            if self.numpins < 1:
                errcallback('numpins must be AUTO or an int >= 1',
                                 self.numpins)
                return False


        if not hasattr(self.search, 'resultset'):
            errcallback('search must include a resultset')
            return False

        return True



class PinSearch(object):
    """Compile a set of pins matching the desired filter function.
    
    Arguments:
    devinstname -- human-friendly name of the desired device instance
    devinst -- Device instance
    matchfunc -- match all logical pins for which bool(matchfunc(devinst, logpinkey, logpin, *args, **kwargs)) evaluates to True
    matchfunc_args -- passed as *args to matchfunc
    matchfunc_kwargs -- passed as **kwargs to matchfunc
    """


    moduleprefix = moduleprefix

    def __init__(self, devinstname, devinst,
                 matchfunc, match_args = None, match_kwargs = None):

        self.__devinstname = devinstname
        self.__devinst = devinst
        self.__matchfunc = matchfunc

        if match_args:
            self.__match_args = tuple(match_args)
        else:
            self.__match_args = tuple()

        if match_kwargs:
            self.__match_kwargs = dict(match_kwargs)
        else:
            self.__match_kwargs = dict()


        self.__resultset = None



    def __str__(self):

        args = ', '.join((repr(getattr(self, k)) for k in (
            'devinstname', 'matchfunc',
        )))

        kwargs = ', '.join(((k + ' = ' + repr(getattr(self, k))) for k in (
            'match_args', 'match_kwargs'
        )))


        return ''.join((
            type(self).__name__,
            '(',
            args,
            ', ',
            kwargs,
            ')',
        ))



    def __repr__(self):
        return ''.join((
            '<',
            str(self),
            '>',
        ))




    def __eq__(self, other):
        for k in ('devinstname', 'devinst',
                  'matchfunc', 'match_args', 'match_kwargs',
                 ):
            if not hasattr(other, k):
                return False
            if getattr(self, k) != getattr(other, k):
                return False
        return True


    def __ne__(self, other):
        return not (self == other)



    @property
    def devinstname(self):
        return self.__devinstname


    @property
    def devinst(self):
        return self.__devinst


    @property
    def matchfunc(self):
        return self.__matchfunc


    @property
    def match_args(self):
        return self.__match_args


    @property
    def match_kwargs(self):
        return dict(self.__match_kwargs)



    @property
    def resultset(self):
        """Return a frozenset of logpinkeys that match matchfunc."""

        if self.__resultset is None:
            self.__resultset = frozenset(self.matches())

        return self.__resultset


    def matches(self):
        """Iterate through device instance's logical pins and generate logpinkeys that match matchfunc."""

        for logpinkey, logpin in self.__devinst.logpack.pins.items():
            if self.__matchfunc(self.__devinst, logpinkey, logpin,
                                *self.__match_args, **self.__match_kwargs):
                yield logpinkey


    def override_results(self, iterable):
        """Override resultset with given iterable."""

        self.__resultset = frozenset(iterable)

