#!/usr/bin/env python3

import unittest
import math
import wallpy.svg

import cost




class Test_bbox(unittest.TestCase):

    def setUp(self):
        pass


    def test_bbox_size(self):
        GRIDX = 15
        GRIDY = 15
        NUM_ANGLES = 49
        NUM_PER_ROW = math.ceil(math.sqrt(NUM_ANGLES))
        ANGLE_INC = 2 * math.pi / NUM_ANGLES
        CANVAS_WIDTH = GRIDX * (NUM_PER_ROW + 0)
        CANVAS_HEIGHT = GRIDY * (math.ceil(NUM_ANGLES / NUM_PER_ROW) + 0)
        SVG_ZOOM = 2
        PINONE_RADIUS = 0.25

        attrib = {
            'bg': {
                'fill': 'lightgrey',
            },

            'devinst': {
                'fill': 'white',
                'stroke': 'blue',
                'stroke-width': '0.2',
            },

            'devinst_center': {
                'fill': 'black',
            },

            'devinst_pinone': {
                'fill': 'blue',
            },

            'bbox': {
                'fill': 'none',
                'stroke': 'red',
                'stroke-width': '0.2',
            },
        }

        w = 10.0
        h = 5.0

        def index2coords(i):
            return (
                GRIDX * (0.5 + (i % NUM_PER_ROW)),
                GRIDY * (0.5 + (i // NUM_PER_ROW)),
            )

        C = wallpy.svg.Canvas(
            (SVG_ZOOM * CANVAS_WIDTH, SVG_ZOOM * CANVAS_HEIGHT),
            title='bbox_size'
        )

        zoom_area = wallpy.svg.SubElement(
            C.doc,
            'g',
            transform='scale({:0.08E})'.format(SVG_ZOOM)
        )

        bg_rect = wallpy.svg.rect(
            zoom_area,
            (0, 0),
            (CANVAS_WIDTH, CANVAS_HEIGHT),
            attrib=attrib['bg']
        )


        for i in range(NUM_ANGLES):
            angle = ANGLE_INC * i

            x, y = index2coords(i)
            bb_w, bb_h = cost.bbox_size(w, h, angle)

            transforms = (
                'translate({:0.06F},{:0.06F})'.format(x, y),
                'rotate({:0.08E})'.format(180 * angle / math.pi),
            )

            rotated = wallpy.svg.SubElement(
                zoom_area,
                'g',
                transform=' '.join(transforms)
            )

            bb_rect = wallpy.svg.rect(
                zoom_area,
                (x - (bb_w/2), y - (bb_h/2)),
                (bb_w, bb_h),
                attrib=attrib['bbox']
            )

            devinst_rect = wallpy.svg.rect(
                rotated,
                (-w/2, -h/2),
                (w, h),
                attrib=attrib['devinst']
            )

            devinst_pinone = wallpy.svg.circle(
                rotated,
                (PINONE_RADIUS*2 - (w/2), PINONE_RADIUS*2 - (h/2)),
                PINONE_RADIUS,
                attrib=attrib['devinst_pinone']
            )

            devinst_center = wallpy.svg.circle(
                zoom_area,
                (x, y),
                0.5,
                attrib=attrib['devinst_center']
            )

        with open('/tmp/bbox_size.svg', 'w') as F:
            C.export(F)


class Test_dist(unittest.TestCase):

    def setUp(self):
        pass


    def test_dist_2D(self):
        def check_dist(p1, p2, expect):
            x1, y1 = p1
            x2, y2 = p2
            actual = cost.dist_2D(x1, y1, x2, y2)
            self.assertAlmostEqual(expect, actual, places=6)

        check_dist((-2, -2), (2, 2), math.sqrt(32))
        check_dist((-1, 1), (1, 1), 2)
        check_dist((-1, 1), (1, 3), math.sqrt(8))


    def test_dist_L2P_2D(self):
        def check_points(A, B, P, expect):
            Ax, Ay = A
            Bx, By = B
            Px, Py = P
            actual = cost.dist_L2P_2D(Ax, Ay, Bx, By, Px, Py)
            self.assertAlmostEqual(expect, actual)

        A = (-2, 0)
        B = (-1, -2)
        P_expect = (
            ((0, -1.5), math.sqrt(1.25)),
            ((1, -1), 2 * math.sqrt(1.25)),
            ((0, -4), 0),
            ((-3.5, -2), math.sqrt(5)),
            ((-1.5, -1), 0),
        )

        for P, expect in P_expect:
            check_points(A, B, P, expect)


__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_bbox),
    __TL.loadTestsFromTestCase(Test_dist),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)

