#!/usr/bin/env python3

import unittest
import sys
import io
import copy
import math

import wallpy.units
import wallpy.random

import device
import devmanager




class Test_DevCollection(unittest.TestCase):

    def setUp(self):
        pass

    def test_filter_devices(self):
        DC = device.DevCollection(devices = {
            'Dev1' : device.Device(),
            'Dev2' : device.Device(),
            'Dev3' : device.Device(),
            'Dev4' : device.Device(),
        })


        def filter_endeven(devname, dev):
            lastletter = devname[-1]
            if lastletter.isdigit():
                i = int(lastletter)
                if (i % 2) == 0:
                    return (devname, )
            return False

        expect = frozenset((i,) for i in ['Dev2', 'Dev4'])
        actual = frozenset(DC.filter_devices(filter_endeven))

        self.assertSetEqual(expect, actual, 'filter (endeven) mismatch')



class Test_Device(unittest.TestCase):

    def setUp(self):
        pin_gnd = device.LogicalPin(pinfuncs = {
            'GND' :
            device.PinFunc(
                device.POWER, device.INPUT, device.SINGLE,
                required=True,
            ),
        })
        pin_vcc = device.LogicalPin(pinfuncs = {
            'VCC' :
            device.PinFunc(
                device.POWER, device.INPUT, device.SINGLE,
                required=True,
            ),
        })
        pin_din = device.LogicalPin(pinfuncs = {
            'DIN' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
        })
        pin_adin = device.LogicalPin(pinfuncs = {
            'DIN' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
            'AIN' : device.PinFunc(
                device.ANALOG, device.INPUT, device.SINGLE,
            ),
        })
        pin_dout = device.LogicalPin(pinfuncs = {
            'DOUT' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
        })
        pin_adout = device.LogicalPin(pinfuncs = {
            'DOUT' : device.PinFunc(
                device.DIGITAL, device.OUTPUT, device.SINGLE,
            ),
            'AOUT' : device.PinFunc(
                device.ANALOG, device.OUTPUT, device.SINGLE,
            ),
        })

        pack = device.LogicalPackage(pins = {
            1 : pin_din,
            2 : pin_vcc,
            3 : pin_adin,
            4 : pin_adout,
            5 : pin_gnd,
            6 : pin_dout,
        })

        self.D = device.Device(packages = {
            'DIP-6' : pack,
            'SOIC-6' : pack,
        })


    def test_prettyprint(self):
        with io.StringIO() as buf:
            self.D.prettyprint(file = buf)

            resultstring = buf.getvalue()

        exec_globals = dict(globals())
        exec_locals = dict()

        exec('result = ' + resultstring, exec_globals, exec_locals)

        result = exec_locals['result']

        self.assertEqual(self.D, result, 'prettyprint result does not match')


    def test_package_eq(self):

        dip = self.D.packages['DIP-6']
        soic = self.D.packages['SOIC-6']

        self.assertEqual(dip, soic, 'packages do not match')



    def test_filter_packages(self):

        def filter_passkeys(packname, pack):
            return (packname,)

        expect = frozenset((i,) for i in ['DIP-6', 'SOIC-6'])
        actual = frozenset(self.D.filter_packages(filter_passkeys))

        self.assertSetEqual(expect, actual, 'filter (pass) mismatch')


        tD = device.Device(packages = {
            'DIP-5' : device.LogicalPackage(pins = {
                1 : device.LogicalPin(pinfuncs = {
                    'A': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                2 : device.LogicalPin(pinfuncs = {
                    'B': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                3 : device.LogicalPin(pinfuncs = {
                    'C': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                4 : device.LogicalPin(pinfuncs = {
                    'D': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                5 : device.LogicalPin(pinfuncs = {
                    'E': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
            }),

            'DIP-6' : device.LogicalPackage(pins = {
                1 : device.LogicalPin(pinfuncs = {
                    'A': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                2 : device.LogicalPin(pinfuncs = {
                    'B': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                3 : device.LogicalPin(pinfuncs = {
                    'C': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                4 : device.LogicalPin(pinfuncs = {
                    'D': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                5 : device.LogicalPin(pinfuncs = {
                    'E': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
                6 : device.LogicalPin(pinfuncs = {
                    'F': device.PinFunc(
                        device.DIGITAL, device.BIDIR, device.SINGLE,
                    ),
                }),
            }),

        })


        def filter_even(packname, pack):
            if (len(pack.pins) % 2) == 0:
                return (packname,)
            return False

        expect = frozenset((i,) for i in ['DIP-6'])
        actual = frozenset(tD.filter_packages(filter_even))

        self.assertSetEqual(expect, actual, 'filter (even) mismatch')




    def test_keep_only(self):
        D = devmanager.load_device('fpga/xilinx/xc3s200a')

        DI = copy.deepcopy(D)

        self.assertIsNot(D, DI, 'deepcopy returned same object')
        self.assertEqual(D, DI, 'deepcopy returned inequal object')

        expect_orig = frozenset(['fg320', 'ft256', 'vq100'])
        actual = frozenset(DI.packages)

        self.assertSetEqual(expect_orig, actual, 'packages differ')


        actual = frozenset(D.packages)

        self.assertSetEqual(expect_orig, actual, 'packages differ')



        DI.packages.keep_only_matches(lambda packname, pack : packname == 'vq100')

        expect_onlyone = frozenset(['vq100'])
        actual = frozenset(DI.packages)

        self.assertSetEqual(expect_onlyone, actual, 'packages differ')


        actual = frozenset(D.packages)

        self.assertSetEqual(expect_orig, actual, 'packages differ')





class Test_LogicalPackage(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_filter_pins(self):
        tP = device.LogicalPackage(pins = {
            1 : device.LogicalPin(pinfuncs = {
                'VCC' : device.PinFunc(
                    device.POWER, device.INPUT, device.SINGLE,
                    required=True,
                ),
            }),

            2 : device.LogicalPin(pinfuncs = {
                'DigIn' : device.PinFunc(
                    device.DIGITAL, device.INPUT, device.SINGLE,
                ),
            }),

            3 : device.LogicalPin(pinfuncs = {
                'DigOut' : device.PinFunc(
                    device.DIGITAL, device.OUTPUT, device.SINGLE,
                ),
            }),

            4 : device.LogicalPin(pinfuncs = {
                'GND' : device.PinFunc(
                    device.POWER, device.INPUT, device.SINGLE,
                    required=True,
                ),
            }),
        })


        def filter_req(pinkey, pin):
            def filter_pinfunc(pfkey, pf):
                if pf.required:
                    return (pfkey, )
                return False

            # Return just the first PinFunc that is required
            ret_pf = next(pin.filter_pinfuncs(filter_pinfunc), False)

            if ret_pf:
                # Return (pinkey, pfkey)
                return (pinkey, ret_pf[0])
            return False

        expect = frozenset((
            (1, 'VCC'),
            (4, 'GND')
        ))
        actual = frozenset(tP.filter_pins(filter_req))

        self.assertSetEqual(expect, actual, 'filter (req) mismatch')





class Test_LogicalPin(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_valid(self):

        i = device.LogicalPin(pinfuncs ={
            'Only_req' : device.PinFunc(
                device.POWER, device.INPUT, device.SINGLE,
                required=True,
            ),
        })

        self.assertEqual(i.valid_p(), True,
                         'valid viewed as invalid')


        i = device.LogicalPin(pinfuncs = {
            'pf1' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
            'pf2' : device.PinFunc(
                device.ANALOG, device.INPUT, device.SINGLE,
            ),
            'pf3' : device.PinFunc(
                device.POWER, device.OUTPUT, device.SINGLE,
            ),

        })

        self.assertEqual(i.valid_p(), True,
                         'valid viewed as invalid')



        i = device.LogicalPin(pinfuncs = {
            'pf1' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
            'pf2_req' : device.PinFunc(
                device.ANALOG, device.INPUT, device.SINGLE,
                required=True,
            ),
            'pf3' : device.PinFunc(
                'x' + device.POWER, device.OUTPUT, device.SINGLE,
            ),

        })

        self.assertEqual(i.valid_p(), False,
                         'invalid viewed as valid')


        i = device.LogicalPin(pinfuncs = {
            'pf1' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
            'pf2' : device.PinFunc(
                device.ANALOG, device.INPUT, device.SINGLE,
            ),
            'pfx' : device.PinFunc(
                'abc', 'def', 'ghi',
            ),
            'pf3' : device.PinFunc(
                device.POWER, device.OUTPUT, device.SINGLE,
            ),

        })

        self.assertEqual(i.valid_p(), False,
                         'invalid viewed as valid')


    def test_filter_pinfuncs(self):

        tP = device.LogicalPin(pinfuncs = {
            'StartMode' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
                required=True,
            ),
            'DigIn' : device.PinFunc(
                device.DIGITAL, device.INPUT, device.SINGLE,
            ),
            'DigOut' : device.PinFunc(
                device.DIGITAL, device.OUTPUT, device.SINGLE,
            ),
        })


        def filter_req(pfkey, pf):
            if pf.required:
                return (pfkey, )
            return False

        expect = frozenset((i,) for i in ['StartMode'])
        actual = frozenset(tP.filter_pinfuncs(filter_req))

        self.assertSetEqual(expect, actual, 'filter (req) mismatch')


        def filter_digin(pfkey, pf):
            if ((pf.sigtype == device.DIGITAL)
                and (pf.sigdir == device.INPUT)):

                return (pfkey, )
            return False

        expect = frozenset((i,) for i in ['StartMode', 'DigIn'])
        actual = frozenset(tP.filter_pinfuncs(filter_digin))

        self.assertSetEqual(expect, actual, 'filter (digin) mismatch')



class Test_PinFunc(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_valid(self):
        PF_valid = []
        for t in device.PinFunc.valid_sigtypes:
            for d in device.PinFunc.valid_sigdirs:
                for c in device.PinFunc.valid_sigconfigs:
                    PF_valid.append(device.PinFunc(t, d, c))


        PF_invalid = []
        for t in device.PinFunc.valid_sigtypes:
            for d in device.PinFunc.valid_sigdirs:
                for c in device.PinFunc.valid_sigconfigs:
                    PF_invalid.append(device.PinFunc(t, d, c + 'X'))
                    PF_invalid.append(device.PinFunc(t, d + 'X', c))
                    PF_invalid.append(device.PinFunc(t + 'X', d, c))
                    PF_invalid.append(device.PinFunc(t + 'X', d + 'X', c))
                    PF_invalid.append(device.PinFunc(t, d + 'X', c + 'X'))
                    PF_invalid.append(device.PinFunc(t + 'X', d, c + 'X'))
                    PF_invalid.append(device.PinFunc(t + 'X', d + 'X', c + 'X'))


        for i in PF_valid:
            self.assertEqual(i.valid_p(), True,
                             'valid viewed as invalid')

        for i in PF_invalid:
            self.assertEqual(i.valid_p(), False,
                             'invalid viewed as valid')



class Test_DevInst(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_fields(self):
        X, Y, Z, rot = (next(self.RD) for i in range(4))
        rot = rot % 360

        dev_fn = 'led/led'
        logpackname = 'AK'
        phypack_fn = '0402'
        log2phy_fn = 'ak'

        UUT = device.DevInst((X, Y, Z, rot), dev_fn, logpackname,
                             phypack_fn, log2phy_fn)

        self.assertTupleEqual(UUT.coords, (X, Y, Z, rot))

        self.assertIs(UUT.dev_fn, dev_fn)
        self.assertIs(UUT.logpackname, logpackname)
        self.assertIs(UUT.phypack_fn, phypack_fn)
        self.assertIs(UUT.log2phy_fn, log2phy_fn)


        dev = devmanager.load_device(dev_fn)
        logpack = dev.packages[logpackname]
        phypack = devmanager.load_phypack(phypack_fn)
        log2phy = devmanager.load_log2phy(log2phy_fn)

        self.assertEqual(UUT.logpack, logpack)
        self.assertEqual(UUT.phypack, phypack)
        self.assertEqual(UUT.log2phy, log2phy)


    def test_valid(self):

        dev_fn = 'led/led'
        logpackname = 'AK'
        phypack_fn = '0402'
        log2phy_fn = 'ak'

        UUT_valid = [
            device.DevInst(
                tuple(wallpy.units.Length(next(self.RD), 'micron') for i in range(3))
                + (wallpy.units.Angle(next(self.RD) % 360, 'deg'), ),
                dev_fn,
                logpackname,
                phypack_fn,
                log2phy_fn
            )
            for i in range(10)
        ]




        ### Invalid: start with valid and then make one bad setting per DevInst
        UUT_invalid = [
            device.DevInst(
                tuple(wallpy.units.Length(next(self.RD), 'micron') for i in range(3))
                + (wallpy.units.Angle(next(self.RD) % 360, 'deg'), ),
                dev_fn,
                logpackname,
                phypack_fn,
                log2phy_fn
            )
            for i in range(4)
        ]


        UUT_invalid[0].phypack = copy.deepcopy(UUT_invalid[0].phypack)
        UUT_invalid[0].phypack.pins.clear()

        UUT_invalid[1].coords = tuple(
            wallpy.units.Length(next(self.RD), 'micron')
            for i in range(4)
        )


        class BadMeasure(wallpy.units.Measure):
            native_unit = 'bogo'
            unit2native_mult = {
                'bogo'  :   1,
            }


        UUT_invalid[2].coords = (
            wallpy.units.Length(next(self.RD), 'micron'),
            wallpy.units.Length(next(self.RD), 'micron'),
            BadMeasure(next(self.RD), 'bogo'),
            wallpy.units.Angle(next(self.RD) % 360, 'deg'),
        )

        UUT_invalid[3].coords = (
            wallpy.units.Length(next(self.RD), 'micron'),
            wallpy.units.Length(next(self.RD), 'micron'),
            wallpy.units.Length(next(self.RD), 'micron'),
            BadMeasure(next(self.RD), 'bogo'),
        )




        for i in UUT_valid:
            self.assertEqual(i.valid_p(), True,
                             'valid viewed as invalid')

        for i in UUT_invalid:
            self.assertEqual(i.valid_p(), False,
                             'invalid viewed as valid')




    def test_mod(self):

        dev_fn = 'led/led'
        logpackname = 'AK'
        phypack_fn = '0402'
        log2phy_fn = 'ak'


        coords = (
            tuple(wallpy.units.Length(next(self.RD), 'micron') for i in range(3))
            + (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )
        )

        UUTs = [
            device.DevInst(
                coords,
                dev_fn,
                logpackname,
                phypack_fn,
                log2phy_fn
            )
            for i in range(2)
        ]


        self.assertEqual(UUTs[0], UUTs[1])

        for U in UUTs:
            U.logpack.flag = next(self.RD)

        # Logical packages are used repeatedly (to save on memory)
        self.assertEqual(UUTs[0], UUTs[1])




    def test_abscoords(self):

        dev_fn = 'led/led'
        logpackname = 'AK'
        phypack_fn = '0402'
        log2phy_fn = 'ak'

        UUTs = [
            device.DevInst(
                tuple(wallpy.units.Length(next(self.RD), 'micron') for i in range(3))
                + (wallpy.units.Angle(next(self.RD) % 360, 'deg'), ),
                dev_fn,
                logpackname,
                phypack_fn,
                log2phy_fn
            )
            for i in range(10)
        ]



        for U in UUTs:
            for i in range(10):
                X, Y, Z = tuple(
                    wallpy.units.Length(next(self.RD), 'micron')
                    for i in range(3)
                )

                X_nat, Y_nat, Z_nat = (float(c.native_val) for c in (X, Y, Z))

                rad = math.pi * (U.coords[3] / wallpy.units.Angle(1, 'pi_rad'))
                c = math.cos(rad)
                s = math.sin(rad)

                relcoords = (
                    c * X_nat - s * Y_nat,
                    s * X_nat + c * Y_nat,
                    Z_nat,
                )

                f_e = tuple(relcoords[i] + U.coords[i].native_val for i in range(3))
                expect = tuple(
                    wallpy.units.Length(round(c, 0), wallpy.units.Length.native_unit)
                    for c in f_e
                )

                actual = U.abs_coords((X, Y, Z))

                self.assertEqual(expect, actual)



    def test_lpk2coords(self):

        dev_fn = 'led/led'
        logpackname = 'AK'
        phypack_fn = '0402'
        log2phy_fn = 'ak'

        UUTs = [
            device.DevInst(
                tuple(wallpy.units.Length(next(self.RD), 'micron')
                      for i in range(3))
                + (wallpy.units.Angle(next(self.RD) % 360, 'deg'), ),
                dev_fn,
                logpackname,
                phypack_fn,
                log2phy_fn
            )
            for i in range(10)
        ]


        for index, U in enumerate(UUTs):
            for lpk in U.logpack.pins:
                phypinkeys = U.log2phy(lpk)

                phypins = {k: U.phypack.pins[k] for k in phypinkeys}
                relcoords = {k: P.coords for k, P in phypins.items()}

                expect = {
                    k: U.abs_coords(C) for k, C in relcoords.items()
                }

                actual = U.lpk2coords(lpk)

                self.assertEqual(expect, actual)




class Test_PhysicalPackage(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_dimensions(self):
        X, Y, Z = (next(self.RD) for i in range(3))

        tP = device.PhysicalPackage((X, Y, Z))

        self.assertTupleEqual(tP.dimensions, (X, Y, Z))


    def test_valid(self):

        pin_dimensions = tuple(wallpy.units.Length(next(self.RD), 'micron') for i in range(3))

        PP_valid = [
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (1, 4, 1)),
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (1, 1, 1)),
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (1, 2, 5)),
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (1, 123, 456)),
            device.PhysicalPackage(wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
            device.PhysicalPackage(wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
            device.PhysicalPackage(wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
            device.PhysicalPackage(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                pins = {
                    1 : device.PhysicalPin(
                        (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                        dimensions = pin_dimensions),
                    2 : device.PhysicalPin(
                        (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                        dimensions = pin_dimensions),
                    3 : device.PhysicalPin(
                        (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                        dimensions = pin_dimensions),
                }
            ),
        ]

        PP_invalid = [
            device.PhysicalPackage(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                pins = {
                    1 : device.PhysicalPin(
                        (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                        dimensions = pin_dimensions),
                    2 : device.PhysicalPin(
                        (None, wallpy.units.Length(0, 'mm'), wallpy.units.Length(0, 'mm')),
                        dimensions = pin_dimensions),
                    3 : device.PhysicalPin(
                        (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                        dimensions = pin_dimensions),
                }
            ),
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (-1, 2, 5)),
            device.PhysicalPackage(wallpy.units.Length(i, 'micron') for i in (0, 123, 456)),
        ]
        for badval in (None, 'ABC', 'x', object()):
            for baddim in range(3):
                dimensions = [wallpy.units.Length(next(self.RD), 'micron') for i in range(3)]
                dimensions[baddim] = badval

                PP_invalid.append(device.PhysicalPackage(dimensions))



        for i in PP_valid:
            self.assertEqual(i.valid_p(), True,
                             'valid viewed as invalid')

        for i in PP_invalid:
            self.assertEqual(i.valid_p(), False,
                             'invalid viewed as valid')




class Test_PhysicalPin(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_fields(self):
        X, Y, Z = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))
        L, W, H = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))

        tP = device.PhysicalPin((X, Y, Z), dimensions = (L, W, H))

        self.assertTupleEqual(tP.coords, (X, Y, Z))
        self.assertTupleEqual(tP.dimensions, (L, W, H))


    def test_valid(self):

        PP_valid = [
            device.PhysicalPin(
                (wallpy.units.Length(i, 'micron') for i in (0, 0, 0)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(i, 'micron') for i in (1, 0, 0)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(i, 'micron') for i in (-1, 2, 55)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(i, 'micron') for i in (0, 123, 456)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3))),
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(i, 'micron') for i in (0, 1, 1))),
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(i, 'micron') for i in (1, 0, 1))),
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(i, 'micron') for i in (1, 1, 0))),
        ]

        PP_invalid = [
            device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)))
            for i in range(2)
        ]

        i = 1
        tmp_dim = list(PP_invalid[i].coords)
        tmp_dim[0] = wallpy.units.Measure(1, '')
        PP_invalid[i].coords = tuple(tmp_dim)

        i = 0
        tmp_dim = list(PP_invalid[i].dimensions)
        tmp_dim[0] = wallpy.units.Measure(1, '')
        PP_invalid[i].dimensions = tuple(tmp_dim)


        for badval in (None, 'ABC', 'x', object()):
            for baddim in range(3):
                coords = [wallpy.units.Length(next(self.RD), 'micron') for i in range(3)]
                coords[baddim] = badval
                dimensions = [wallpy.units.Length(next(self.RD), 'micron') for i in range(3)]

                PP_invalid.append(device.PhysicalPin(coords, dimensions = dimensions))


        for badval in (None, 'ABC', 'x', object(), wallpy.units.Length(-1, 'micron')):
            for baddim in range(3):
                coords = [wallpy.units.Length(next(self.RD), 'micron') for i in range(3)]
                dimensions = [wallpy.units.Length(next(self.RD), 'micron') for i in range(3)]
                dimensions[baddim] = badval

                PP_invalid.append(device.PhysicalPin(coords, dimensions = dimensions))




        for gooddim in range(3):
            dimensions = [wallpy.units.Length(0, 'mm') for i in range(3)]
            dimensions[gooddim] = wallpy.units.Length(next(self.RD), 'micron')

            PP_invalid.append(device.PhysicalPin(
                (wallpy.units.Length(next(self.RD), 'micron') for i in range(3)),
                dimensions = dimensions))


        for i in PP_valid:
            self.assertEqual(i.valid_p(), True,
                             'valid viewed as invalid')

        for i in PP_invalid:
            self.assertEqual(i.valid_p(), False,
                             'invalid viewed as valid')



__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_PinFunc),
    __TL.loadTestsFromTestCase(Test_LogicalPin),
    __TL.loadTestsFromTestCase(Test_LogicalPackage),
    __TL.loadTestsFromTestCase(Test_Device),
    __TL.loadTestsFromTestCase(Test_DevCollection),
    __TL.loadTestsFromTestCase(Test_PhysicalPin),
    __TL.loadTestsFromTestCase(Test_PhysicalPackage),
    __TL.loadTestsFromTestCase(Test_DevInst),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
