#!/usr/bin/env python3

import unittest
import re

import wallpy.units

import devmanager
import schem



class Test_OrderBundlePins(unittest.TestCase):
    def setUp(self):
        self.S = schem.Schematic()


    def test_attiny85_io3(self):
        devs = (
            (
                'U1',
                (0, 0, 0, 10),
                ('microcontroller/attiny85', 'SO-8', 'so8_n', 'direct'),
            ),

            (
                'U2',
                (0, 10, 0, 170),
                ('microcontroller/attiny85', 'SO-8', 'so8_n', 'direct'),
            ),
        )


        for d in devs:
            devinstname, coord_nums, devargs = d
            coords = [wallpy.units.Length(i, 'mm') for i in coord_nums[:3]]
            coords.append(wallpy.units.Angle(coord_nums[3], 'deg'))

            self.S.add_devinst(devinstname, coords, *devargs)


        def match_io_not_reset(devinst, pinkey, pin):
            re_io = re.compile(r'^pb[0-2]*$', flags=re.IGNORECASE)
            re_reset = re.compile(r'^reset', flags=re.IGNORECASE)

            for pfname in pin.pinfuncs:
                if re_reset.match(pfname):
                    return False

            for pfname in pin.pinfuncs:
                if re_io.match(pfname):
                    return True

            return False


        ep_fields = list()

        for d in devs:
            devinstname = d[0]

            pinsearch = (devinstname, match_io_not_reset, None, None)

            ep_fields.append([(schem.AUTO, pinsearch, dict())])

        B = self.S.add_bundle(
            'B_io',
            schem.bus2netnames('d<', '>', range(3)),
            ep_fields
        )


        print('Solver...')
        self.assertTrue(self.S.solve_phase1_obvious())


        expect = {
            'd<0>': {('U1', 5), ('U2', 7)},
            'd<1>': {('U1', 6), ('U2', 6)},
            'd<2>': {('U1', 7), ('U2', 5)},
        }

        actual = self.S.nets

        self.assertDictEqual(expect, actual)



        for netname in sorted(self.S.nets, key=schem.name2ana):
            print(netname, '{')

            for DIN, lpk in sorted(
                self.S.nets[netname],
                key=lambda s: (schem.name2ana(s[0]), s[1:])
            ):
                print('\t', DIN, '[', lpk, ']', sep='')

            print('} //', netname)


        svg_prefs = {
            'title': 'attiny85_io3',
            'zoom': 12,
            'include_rats': True,
            'include_pins': True,
            'include_pinlabels': True,
        }

        with open('/tmp/schem.svg', 'w') as Fout:
            self.S.export_svg(Fout, prefs=svg_prefs)



    def test_attiny85_io4(self):
        devs = (
            (
                'U1',
                (0, 0, 0, 0),
                ('microcontroller/attiny85', 'SO-8', 'so8_n', 'direct'),
            ),

            (
                'U2',
                (0, 10, 0, 180),
                ('microcontroller/attiny85', 'SO-8', 'so8_n', 'direct'),
            ),
        )


        for d in devs:
            devinstname, coord_nums, devargs = d
            coords = [wallpy.units.Length(i, 'mm') for i in coord_nums[:3]]
            coords.append(wallpy.units.Angle(coord_nums[3], 'deg'))

            self.S.add_devinst(devinstname, coords, *devargs)


        def match_io_not_reset(devinst, pinkey, pin):
            re_io = re.compile(r'^pb[0-3]*$', flags=re.IGNORECASE)
            re_reset = re.compile(r'^reset', flags=re.IGNORECASE)

            for pfname in pin.pinfuncs:
                if re_io.match(pfname):
                    if not re_reset.match(pfname):
                        return True

            return False


        ep_fields = list()

        for d in devs:
            devinstname = d[0]

            pinsearch = (devinstname, match_io_not_reset, None, None)

            ep_fields.append([(schem.AUTO, pinsearch, dict())])

        B = self.S.add_bundle(
            'B_io',
            schem.bus2netnames('d<', '>', range(4)),
            ep_fields
        )


        print('Solver...')
        self.assertTrue(self.S.solve_phase1_obvious())


        expect = {
            'd<0>': {('U1', 2), ('U2', 5)},
            'd<1>': {('U1', 5), ('U2', 2)},
            'd<2>': {('U1', 6), ('U2', 7)},
            'd<3>': {('U1', 7), ('U2', 6)},
        }

        actual = self.S.nets

        self.assertDictEqual(expect, actual)


        for netname in sorted(self.S.nets, key=schem.name2ana):
            print(netname, '{')

            for DIN, lpk in sorted(
                self.S.nets[netname],
                key=lambda s: (schem.name2ana(s[0]), s[1:])
            ):
                print('\t', DIN, '[', lpk, ']', sep='')

            print('} //', netname)


        svg_prefs = {
            'title': 'attiny85_io4',
            'zoom': 12,
            'include_rats': True,
            'include_pins': True,
            'include_pinlabels': True,
        }

        with open('/tmp/schem2.svg', 'w') as Fout:
            self.S.export_svg(Fout, prefs=svg_prefs)


        self.S.solve_ilp()



__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_OrderBundlePins),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
