#!/usr/bin/env python3

import sys

import wallpy.units

import device
import schem



S = schem.Schematic()

FPGAs = tuple(S.next_devinstname('U') for i in range(4))


args = ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')
for i, devinstname in enumerate(FPGAs):

    xi = i
    yi = i >> 1

    xi += 1
    yi += 1

    xmult = (xi >> 1)
    if (xi % 2):
        xmult = -xmult

    ymult = (yi >> 1)
    if (yi % 2):
        ymult = -ymult

    coords = (
        xmult * wallpy.units.Length(30, 'mm'),
        ymult * wallpy.units.Length(30, 'mm'),
        wallpy.units.Length(0, 'mm'),
        wallpy.units.Angle(0, 'deg')
    )

    S.add_devinst(devinstname, coords, *args)



def filter_pfname(logpinkey, logpin, devinstname, name):
    if name in logpin.pinfuncs:
        return (devinstname, logpinkey)
    return False



stdpins = {
    'GND'       :   'GND',
    'VCCINT'    :   '1V2',
    'VCCAUX'    :   '3V3',
    'VCCO'      :   '3V3',
    'TMS'       :   'JTAG_TMS',
    'TCK'       :   'JTAG_TCK',
    'CCLK'      :   'FPGA_CCLK',
    'INIT_B'    :   'FPGA_INIT_B',
    'PROG_B'    :   'FPGA_PROG_B',

    'PUDC_B'    :   'GND',  # Activate pullups

    'M2'        :   '3V3',  # M[2:0] = 2_101 for JTAG mode
    'M1'        :   'GND',
    'M0'        :   '3V3',
}


for pfname, netname in stdpins.items():
    for devinstname in FPGAs:
        DI = S.devinst_n2o[devinstname]

        S.connect_net(netname, DI.logpack.filter_pins(
            filter_pfname,
            args = [devinstname, pfname]
        ))


nets_by_name = [
    ('JTAG_TDI', {(FPGAs[0], 'TDI')}),
    ('JTAG_TDO', {(FPGAs[-1], 'TDO')}),
]

iter_FPGA = iter(FPGAs)
prev_FPGA = next(iter_FPGA)
for bitnum, next_FPGA in enumerate(iter_FPGA):
    netname = 'TD_between[{:}]'.format(bitnum)

    nets_by_name.append(
        (netname, {
            (prev_FPGA, 'TDO'),
            (next_FPGA, 'TDI'),
        })
    )

    prev_FPGA = next_FPGA



for netname, pinset in nets_by_name:
    for devinstname, pfname in pinset:
        DI = S.devinst_n2o[devinstname]

        S.connect_net(netname, DI.logpack.filter_pins(
            filter_pfname,
            args = [devinstname, pfname]
        ))






S.export(sys.stdout, comments = True)



print('Unconnected required pins:')
for devinstname in FPGAs:
    for pinkey, pin in sorted(
        S.devinst_n2o[devinstname].logpack.pins.items(),
        key = lambda tup : tup[0]
    ):

        if pin.required_p() and ((devinstname, pinkey) not in S.logpin2net):
            print((devinstname, pinkey), ', '.join(pin.pinfuncs), sep='\t')
print('End of unconnected required pins')
print()




def match_available_pfprefix(devinst, logpinkey, logpin, prefix, case_sensitive):
    devinstname = S.devinst_id2n[id(devinst)]

    if case_sensitive:
        prefix = prefix.lower()


    if (devinstname, logpinkey) in S.logpin2net:
        return False

    for k in logpin.pinfuncs:
        if case_sensitive:
            k = k.lower()

        if k.startswith(prefix):
            return True

    return False



print('#### NEEDS TO BE UPDATED FOR BUNDLES, etc. ####')
print('OK', file=sys.stderr)
sys.exit(0)


# Bus between each FPGA
for DI1 in FPGAs:
    for DI2 in FPGAs:
        if DI1 >= DI2:
            continue

        NRname = 'NR_databus_{:}_{:}'.format(DI1, DI2)

        NR = S.add_netreq(
            NRname,
            schem.bus2netnames('DB_{:}_{:}['.format(DI1, DI2), ']', range(8)),
            ((n, match_available_pfprefix, ['IO', False], None) for n in (DI1, DI2))
        )

        print('Added', NRname)



print()

print('Mockup:', NRname)
for devinstname, ep in sorted(NR.endpoints.items(),
                              key = lambda nv : schem.name2ana(nv[0])):

    print('EP', devinstname)
    DI = ep.devinst
    for logpinkey in ep.resultset:
        comment_str = ' # ' + ', '.join(
            DI.logpack.pins[logpinkey].pinfuncs
        )

        print('\t', logpinkey, comment_str, sep='')
    print()






# Endpoint coordinates start out at center of devinst
for devinstname, ep in NR.endpoints.items():
    ep.coords = S.devinst_n2o[devinstname].coords[:3]



def cost_simpledist(DI, devinstname, logpinkey, endpoints):
    phypinkeys = DI.log2phy(logpinkey)

    for ppk in phypinkeys:
        pin_coords = tuple(
            endpoints[devinstname].coords[i] + DI.phypack.pins[ppk].coords[i]
            for i in range(3)
        )

        ret = wallpy.units.Length(0, 'mm')
        for n, ep in endpoints.items():
            if n != devinstname:
                for i in range(3):
                    ret += abs(ep.coords[i] - pin_coords[i])

    return (ret / len(phypinkeys)) // wallpy.units.Length(1, 'nm')












# Choose pins for each FPGA
reverse = False
for devinstname, ep in sorted(NR.endpoints.items(),
                              key = lambda nv : schem.name2ana(nv[0])):

    print(devinstname)

    DI = ep.devinst
    DIN = ep.devinstname
    costs = dict()

    for logpinkey in ep.resultset:
        costs[logpinkey] = cost_simpledist(DI, DIN, logpinkey, NR.endpoints)


    best = tuple((k, v) for k, v in sorted(costs.items(),
                                           key = lambda kv: (kv[1], schem.name2ana(kv[0]))))


    best_N = best[:len(NR.netnames)]

    k2qs = {lpk[0] : schem.lpk2qs(DI, lpk[0]) for lpk in best_N}
    q_offset = schem.calc_quadrant_offset(k2qs.values())
    adjusted_qs = {k : schem.adjust_quadrant(k2qs[k], q_offset) for k in k2qs}


    new_relcoords = [wallpy.units.Length(0, 'mm') for i in range(3)]

    for k, v in best_N:

        group_relcoords = tuple(schem.zero_length for i in range(3))

        phypinkeys = DI.log2phy(k)

        for ppk in phypinkeys:
            for i in range(3):
                new_relcoords[i] += DI.phypack.pins[ppk].coords[i]

        for i in range(3):
            new_relcoords[i] += group_relcoords[i] / len(phypinkeys)


    for i in range(3):
        new_relcoords[i] /= len(best_N)


    new_abscoords = tuple(
        S.devinst_n2o[devinstname].coords[i] + new_relcoords[i]
        for i in range(3)
    )

    ep.coords = tuple(new_abscoords)
    print(tuple(float(C / wallpy.units.Length(1, 'mm')) for C in ep.coords))


    for k, v in sorted(best_N,
                       key = lambda kv: adjusted_qs[kv[0]],
                       reverse = reverse):
        print(k, v)


    print()


    # Alternate CW / CCW sorting
    reverse = not reverse



print('Calling solve...')
S.solve()




print('OK', file=sys.stderr)
