#!/usr/bin/env python3

import unittest
import sys
import random
import io

import wallpy.random
import wallpy.units
import wallpy.cleanobject

import schem
import device




class ArgedObject(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return ''.join((
            '<',
            type(self).__name__,
            ' ',
            repr(self.__dict__),
            '>',
        ))






class Test_Schematic(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)
        self.S = schem.Schematic()



    def test_next_devinstname_simple(self):

        prefix2curnum = dict()

        def DIN_next(prefix):
            num = prefix2curnum.get(prefix, 0) + 1
            prefix2curnum[prefix] = num
            return prefix + str(num)

        prefixes = ('U', 'R', 'C', 'J', 'P', 'L', )


        for i in range(1000):
            prefix = random.choice(prefixes)

            expect = DIN_next(prefix)
            actual = self.S.next_devinstname(prefix)
            self.assertEqual(expect, actual)





    def test_add_devinst(self):

        possible_devices = (
            ('U', ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')),
            ('D', ('led/led', 'AK', '0402', 'ak')),
        )


        expect = dict()


        for i in range(64):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            prefix, args = random.choice(possible_devices)

            devinstname = prefix + str(i)

            added = self.S.add_devinst(devinstname, coords, *args)


            expect[devinstname] = added

            self.assertDictEqual(expect, self.S.devinst_n2o)






    def test_next_devinstname_complex(self):

        # Make sure don't overlap with manually-added devinstnames


        possible_devices = (
            ('U', ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')),
            ('D', ('led/led', 'AK', '0402', 'ak')),
        )


        for i in range(64):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            prefix, args = random.choice(possible_devices)

            devinstname = prefix + str(i)

            added = self.S.add_devinst(devinstname, coords, *args)



        for prefix, dummy in possible_devices:
            for i in range(64):
                newname = self.S.next_devinstname(prefix)

                self.assertNotIn(newname, self.S.devinst_n2o,
                                 'got existing devinst name: ' + newname)



    def test_connect_net(self):

        devices = (
            ('U1', ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')),
            ('D1', ('led/led', 'AK', '0402', 'ak')),
            ('D2', ('led/led', 'AK', '0402', 'ak')),
            ('D3', ('led/led', 'AK', '0402', 'ak')),
            ('D4', ('led/led', 'AK', '0402', 'ak')),
        )


        for devinstname, args in devices:
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            added = self.S.add_devinst(devinstname, coords, *args)


        U1 = self.S.devinst_n2o['U1']
        D1 = self.S.devinst_n2o['D1']
        D2 = self.S.devinst_n2o['D2']
        D3 = self.S.devinst_n2o['D3']
        D4 = self.S.devinst_n2o['D4']


        expect = dict()



        # Connect FPGA GND pins
        def match_gnd(pinkey, pin):
            if 'GND' in pin.pinfuncs:
                return pinkey

        gnd_pins = tuple(('U1', logpinkey) for logpinkey in U1.logpack.filter_pins(match_gnd))

        expect['GND'] = set(gnd_pins)

        self.S.connect_net('GND', gnd_pins)

        self.assertDictEqual(expect, self.S.nets)





        # Connect LED anodes
        LED_nets = (
            ('LED1', [('U1', 3), ('D1', 'A')]),
            ('LED2', [('U1', 4), ('D2', 'A')]),
            ('LED3', [('U1', 5), ('D3', 'A')]),
            ('LED4', [('U1', 6), ('D4', 'A')]),
        )

        for netname, pinset in LED_nets:
            expect[netname] = set(pinset)

            self.S.connect_net(netname, pinset)

            self.assertDictEqual(expect, self.S.nets)



        # Connect LED cathodes
        LEDs = tuple('D' + str(i+1) for i in range(4))

        for LED_name in LEDs:
            expect['GND'].add((LED_name, 'K'))

            self.S.connect_net('GND', [(LED_name, 'K'), ])

            self.assertDictEqual(expect, self.S.nets)


        # Disconnect LED cathodes
        for LED_name in LEDs:
            expect['GND'].remove((LED_name, 'K'))

            ret = self.S.disconnect_net('GND', [(LED_name, 'K'), ])
            self.assertEqual(True, ret)

            self.assertDictEqual(expect, self.S.nets)

            # Removing the same net a second time should make no change
            ret = self.S.disconnect_net('GND', [(LED_name, 'K'), ])
            self.assertEqual(False, ret)

            self.assertDictEqual(expect, self.S.nets)



        # Reconnect LED cathodes
        LEDs = tuple('D' + str(i+1) for i in range(4))

        for LED_name in LEDs:
            expect['GND'].add((LED_name, 'K'))

            self.S.connect_net('GND', [(LED_name, 'K'), ])

            self.assertDictEqual(expect, self.S.nets)


        # Disconnect LED cathodes (using disconnect_pins)
        for LED_name in LEDs:
            expect['GND'].remove((LED_name, 'K'))

        ret = self.S.disconnect_pins([(n, 'K') for n in LEDs])
        self.assertEqual(True, ret)

        self.assertDictEqual(expect, self.S.nets)

        # Disconnecting the same pins a second time should make no change
        ret = self.S.disconnect_pins([(n, 'K') for n in LEDs])
        self.assertEqual(False, ret)

        self.assertDictEqual(expect, self.S.nets)



        # Make sure we can't add net connections to nonexistent pins
        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('D1', 1), ])

        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('D1', 'd'), ])

        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('D1', 99), ])

        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('D1', 0), ])

        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('D1', 101), ])

        with self.assertRaises(KeyError):
            self.S.connect_net('GND', [('NonExistent', 1), ])




    @unittest.skip('Old version')
    def test_export(self):


        # Add devices
        args = ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')


        ICs = []

        for i in range(12):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            devinstname = self.S.next_devinstname('U')

            added = self.S.add_devinst(devinstname, coords, *args)
            ICs.append(devinstname)


        args = ('led/led', 'AK', '0402', 'ak')
        for i in range(16):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            added = self.S.add_devinst(self.S.next_devinstname('D'), coords, *args)



        # Add net requests
        def match_pinfuncname(devinst, logpinkey, logpin, desired_funcname):
            if desired_funcname in logpin.pinfuncs:
                return True
            return False

        req_GNDs = tuple(schem.NetReqEndpoint(n, self.S.devinst_n2o[n], match_pinfuncname,
                                              match_args = ['GND'])
                         for n in ICs)


        for req in req_GNDs:
            DIN = req.devinstname
            self.S.connect_net('GND', ((DIN, lpk) for lpk in req.matches()))


        # Add nets
        self.S.connect_net('GND', [('U1', 8)])

        for i in range(128):
            netname = 'N_{:03}'.format(i)

            connections = set()

            for i in range(8):

                for itercount in range(1000):

                    # Focus on ICs before other devices
                    for trycount in range(4):
                        devname = random.choice(tuple(self.S.devinst_n2o))
                        if devname.startswith('U'):
                            break

                    logpinkey = random.choice(tuple(self.S.devinst_n2o[devname].logpack.pins))

                    # Make sure the pin is not already connected to a net
                    if (
                        ((devname, logpinkey) in self.S.logpin2net)
                        or ((devname, logpinkey) in connections)
                    ):
                        devname = None
                        logpinkey = None
                    else:
                        break

                if not (devname and logpinkey):
                    self.fail('Ran out of pins to connect to nets')

                connections.add((devname, logpinkey))


            self.S.connect_net(netname, connections)



        # Try adding net to pin with existing, different net
        with self.assertRaises(ValueError):
            self.S.connect_net('VCC', [('U1', 8)])



        # Make sure exported-imported schematic matches current schematic
        with io.StringIO() as buf:
            self.S.export(buf, comments = True)

            bufval = buf.getvalue()


        exec_globals = dict()
        exec_locals = dict()

        exec(bufval, exec_globals, exec_locals)


        actual = exec_locals['schem']


        self.assertEqual(self.S.devinst_n2o, actual.devinst_n2o)
        self.assertEqual(self.S.nets, actual.nets)
        self.assertEqual(self.S.net_reqs, actual.net_reqs)
        self.assertEqual(self.S.netreq_eps, actual.netreq_eps)
        self.assertEqual(self.S, actual)




    def test_set_placement(self):
        self.maxDiff = None

        devinstnames = tuple('D{}'.format(i) for i in range(1, 32))
        devargs = ('led/led', 'AK', '0402', 'ak')

        expect = dict() # DIN: coords

        ### Add DevInsts at random initial positions
        for DIN in devinstnames:
            coords = tuple(
                wallpy.units.Length(next(self.RD), 'mm')
                for i in range(3)
            )

            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            self.S.add_devinst(DIN, coords, *devargs)

            expect[DIN] = coords

        for i in range(64):
            ### Now update some DevInsts' coordinates (as units)
            new_placement = dict()  # DIN: coords

            for DIN in random.sample(devinstnames, 8):
                coords = tuple(
                    wallpy.units.Length(next(self.RD), 'mm')
                    for i in range(2)
                )

                coords += (expect[DIN][2], )    # Z does not change

                coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

                new_placement[DIN] = coords

                expect[DIN] = coords

            self.S.set_placement_2D(new_placement)


            actual = {
                DIN: DI.coords
                for DIN, DI in self.S.devinst_n2o.items()
            }

            self.assertDictEqual(expect, actual)



            ### Now update some DevInsts' coordinates (as floats)
            new_placement = dict()  # DIN: coords

            for DIN in random.sample(devinstnames, 8):
                coords_float = tuple(
                    next(self.RD)
                    for i in range(2)
                )

                coords_float += (next(self.RD) % 360, )

                new_placement[DIN] = coords_float

                coords = tuple(
                    wallpy.units.Length(v, 'mm')
                    for v in coords_float[:2]
                )

                coords += (expect[DIN][2], )    # Z does not change

                coords += (wallpy.units.Angle(coords_float[-1], 'deg'), )

                expect[DIN] = coords

            self.S.set_placement_2D(new_placement,
                                    length_in_units='mm',
                                    angle_in_units='deg')


            actual = {
                DIN: DI.coords
                for DIN, DI in self.S.devinst_n2o.items()
            }

            self.assertDictEqual(expect, actual)




class Test_helperfunctions(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_bus2netnames(self):
        expect = tuple('D<{:}>'.format(i) for i in range(2, 20))
        actual = schem.bus2netnames('D<', '>', range(2, 20))
        self.assertTupleEqual(expect, actual)


    def test_name2ana(self):
        expect = {
            'U1'    :   ('U', 1, ''),
            'abc'   :   ('abc', False, ''),
            '123'   :   ('', 123, ''),
            'D[56]' :   ('D[', 56, ']'),
            'D<56>' :   ('D<', 56, '>'),
        }

        actual = {k : schem.name2ana(k) for k in expect}

        self.assertDictEqual(expect, actual)


    def test_compute_CoM_noweight_fixed(self):
        points = [
            (1, -1),
            (3, 3),
            (-5, 5),
            (2, -4),
        ]

        expect = (0.25, 0.75)

        actual = schem.compute_CoM_noweight(points)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])


    def test_compute_CoM_noweight_rand(self):
        points = [
            (next(self.RD), next(self.RD))
            for i in range(16)
        ]

        sum_x = sum(p[0] for p in points)
        sum_y = sum(p[1] for p in points)
        n = len(points)

        expect = (sum_x / n, sum_y / n)

        actual = schem.compute_CoM_noweight(points)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])


    def test_compute_CoM_fixed(self):
        points = [
            (1, -1, 5),
            (2, -2, 4),
            (3, -3, 5),
        ]

        expect = (2.0, -2.0)

        actual = schem.compute_CoM(points)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])


    def test_compute_CoM_rand(self):
        points = [
            (next(self.RD), next(self.RD), next(self.RD))
            for i in range(16)
        ]

        sum_x = sum(p[0]*p[2] for p in points)
        sum_y = sum(p[1]*p[2] for p in points)
        sum_weights = sum(p[2] for p in points)

        expect = (sum_x / sum_weights, sum_y / sum_weights)

        actual = schem.compute_CoM(points)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])


    def test_shift_CoM_fixed(self):
        orig = (1, -2)
        prev = (3, 1)
        chosen = (-2, 1)
        weights = (1, 1, 2)

        expect = (0, 0.25)

        actual = schem.shift_CoM(orig, prev, chosen, weights=weights)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])


    def test_shift_CoM_rand(self):
        orig = tuple(next(self.RD) for i in range(2))
        prev = tuple(next(self.RD) for i in range(2))
        chosen = tuple(next(self.RD) for i in range(2))
        weights = tuple(next(self.RD) for i in range(3))

        points = (orig, prev, chosen)
        sum_x = 0
        sum_y = 0
        sum_weights = sum(weights)

        for i, p in enumerate(points):
            sum_x += p[0] * weights[i]
            sum_y += p[1] * weights[i]

        expect = (sum_x / sum_weights, sum_y / sum_weights)

        actual = schem.shift_CoM(orig, prev, chosen, weights=weights)

        for i in range(len(expect)):
            self.assertAlmostEqual(expect[i], actual[i])




class Test_PinSearch(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)



    def test_match_evens(self):
        coords = tuple(wallpy.units.Length(0, 'mm') for i in range(3))
        coords += (wallpy.units.Angle(0, 'deg'), )
        DI = device.DevInst(coords, 'fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')
        devname = 'U' + str(next(self.RD))

        def match_evens(devinst, logpinkey, logpin):
            if (logpinkey % 2) == 0:
                return True
            return False


        UUT_evens = schem.PinSearch(devname, DI, match_evens)

        expect = frozenset(range(2, 101, 2))
        actual = UUT_evens.resultset

        self.assertSetEqual(expect, actual, 'UUT evens mismatch')



        def match_reqs(devinst, logpinkey, logpin):
            return logpin.required_p()


        UUT_reqs = schem.PinSearch(devname, DI, match_reqs)

        required_pins = [
            1, 2, 8, 11, 14, 17, 18, 22, 23, 24, 25, 26, 38, 42, 45, 47, 48,
            53, 55, 58, 63, 66, 67, 69, 74, 75, 76, 79, 80, 81, 87, 91, 92,
            95, 96, 99, 100,
        ]

        expect = frozenset(required_pins)
        actual = UUT_reqs.resultset

        self.assertSetEqual(expect, actual, 'UUT reqs mismatch')



    def test_eq(self):
        coords = tuple(wallpy.units.Length(0, 'mm') for i in range(3))
        coords += (wallpy.units.Angle(0, 'deg'), )
        DI = device.DevInst(coords, 'led/led', 'AK', '0402', 'ak')

        def match_anode(devinst, logpinkey, logpin):
            return logpinkey == 'A'

        UUT1 = schem.PinSearch('D1', DI, match_anode)
        UUT2 = schem.PinSearch('D1', DI, match_anode)

        self.assertEqual(UUT1, UUT2)

        def match_cathode(devinst, logpinkey, logpin):
            return logpinkey == 'K'

        UUT3 = schem.PinSearch('D1', DI, match_cathode)

        self.assertNotEqual(UUT1, UUT3)





class Test_EndpointSubgroup(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)



    def test_init(self):

        happy_numpins = tuple(next(self.RD) for i in range(10))
        bad_numpins = (
            'a',
            '1b',
            '',
            None,
            (1, 2),
        )


        for v in happy_numpins:
            UUT = schem.EndpointSubgroup(v, None)

        for v in bad_numpins:
            with self.assertRaises((ValueError, TypeError)):
                UUT = schem.EndpointSubgroup(v, None)




    def test_valid(self):


        vals = {
            'numpins' : {
                True :  (
                    1,
                    2,
                    99,
                    schem.AUTO,
                ),
                False : (
                    0,
                    -1,
                    -99,
                ),
            },

            'search' :  {
                True :  (
                    ArgedObject(a=1, b=2, resultset=3),
                    ArgedObject(resultset=None),
                ),
                False : (
                    None,
                    '1',
                    set(),
                    ArgedObject(a=1, b=2, search=3),
                    object(),
                ),
            },
        }



        # Valid, Valid
        numpins = vals['numpins'][True]
        search = vals['search'][True]

        for n in numpins:
            for s in search:
                UUT = schem.EndpointSubgroup(n, s)
                self.assertEqual(True, UUT.valid_p())


        # Valid, Invalid
        numpins = vals['numpins'][True]
        search = vals['search'][False]

        for n in numpins:
            for s in search:
                UUT = schem.EndpointSubgroup(n, s)
                self.assertEqual(False, UUT.valid_p())


        # Invalid, Valid
        numpins = vals['numpins'][False]
        search = vals['search'][True]

        for n in numpins:
            for s in search:
                UUT = schem.EndpointSubgroup(n, s)
                self.assertEqual(False, UUT.valid_p())


        # Invalid, Invalid
        numpins = vals['numpins'][False]
        search = vals['search'][False]

        for n in numpins:
            for s in search:
                UUT = schem.EndpointSubgroup(n, s)
                self.assertEqual(False, UUT.valid_p())





class Test_BundleEndpoint(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_valid(self):
        def yes_valid(*args):
            return True

        def not_valid(*args):
            return False

        valid_elements = (
            ArgedObject(),
            ArgedObject(valid_p = yes_valid),
        )

        invalid_elements = (
            ArgedObject(valid_p = not_valid),
        )


        for i in valid_elements:
            i.params = dict()

        for i in invalid_elements:
            i.params = dict()


        valid = [valid_elements]
        invalid = [tuple()]


        for good in valid_elements:
            valid.append([good])
            valid.append([good, good])

        for good in valid_elements:
            for bad in invalid_elements:
                invalid.append([good, bad])
                invalid.append([bad, bad])
                invalid.append([bad, good])

        for bad in invalid_elements:
            invalid.append([bad])


        for i, good in enumerate(valid):
            UUT = schem.BundleEndpoint(good)
            self.assertEqual(True, UUT.valid_p(), (i, UUT.subgroups))

        for i, bad in enumerate(invalid):
            UUT = schem.BundleEndpoint(bad)
            self.assertEqual(False, UUT.valid_p(), (i, UUT.subgroups))





class Test_Bundle(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_valid(self):
        def yes_valid(*args):
            return True

        def not_valid(*args):
            return False

        valid_elements = (
            ArgedObject(),
            ArgedObject(valid_p = yes_valid),
        )

        invalid_elements = (
            ArgedObject(valid_p = not_valid),
        )


        good_netnames = (
            'GND',
            'LED<0>',
            'D[0]',
            'RESET_n',
        )

        bad_netnames = (
            '',
            None,
            123,
            object(),
            'space here',
            ' ',
        )


        for i in valid_elements:
            i.subgroups = list()

        for i in invalid_elements:
            i.subgroups = list()



        valid = [valid_elements]
        invalid = [tuple()]


        for good in valid_elements:
            valid.append([good])
            valid.append([good, good])

        for good in valid_elements:
            for bad in invalid_elements:
                invalid.append([good, bad])
                invalid.append([bad, bad])
                invalid.append([bad, good])

        for bad in invalid_elements:
            invalid.append([bad])


        for i, good in enumerate(valid):
            UUT = schem.Bundle(['net1'], good)
            self.assertEqual(True, UUT.valid_p(), i)

        for i, bad in enumerate(invalid):
            UUT = schem.Bundle(['net1'], bad)
            self.assertEqual(False, UUT.valid_p(), i)

        for i, good in enumerate(good_netnames):
            ep = ArgedObject(subgroups=list())
            UUT = schem.Bundle([good], [ep])
            self.assertEqual(True, UUT.valid_p(), repr(good))

        for i, bad in enumerate(bad_netnames):
            UUT = schem.Bundle([bad], [ArgedObject()])
            self.assertEqual(False, UUT.valid_p(), repr(bad))

        UUT = schem.Bundle([], [ArgedObject()])
        self.assertEqual(False, UUT.valid_p())


    def test_compute_subgroup_numpins(self):

        class FakeSearch(object):
            """Return a fixed, fake set of results"""

            def __init__(self, num_results):
                self.num_results = num_results
                self.resultset = frozenset(range(num_results))

        NN = 12     # Nets in this Bundle

        subgroups1 = [
            schem.EndpointSubgroup(schem.AUTO, FakeSearch(8))
            for i in range(NN // 3)
        ]

        subgroups2 = [
            schem.EndpointSubgroup(schem.AUTO, FakeSearch(8))
            for i in range(NN // 4)
        ]

        subgroups3 = [
            schem.EndpointSubgroup(schem.AUTO, FakeSearch(8))
            for i in range(2)
        ]
        subgroups3.append(schem.EndpointSubgroup(6, FakeSearch(8)))

        B = schem.Bundle(
            ('N<{:02}>'.format(i) for i in range(NN)),
            (
                schem.BundleEndpoint(sg)
                for sg in (subgroups1, subgroups2, subgroups3)
            )
        )

        B.compute_subgroup_numpins()

        expect = [
            [3, 3, 3, 3],
            [4, 4, 4],
            [3, 3, 6],
        ]

        actual = list()
        for ep in B.endpoints:
            sg_numpins = list()

            for sg in ep.subgroups:
                sg_numpins.append(sg.numpins)

            actual.append(sg_numpins)

        self.assertEqual(expect, actual)

        self.assertEqual(True, B.valid_p(errcallback=print))




class Test_Schematic_Reqs(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)
        self.S = schem.Schematic()


        # Add devices
        FPGAs = []
        args = ('fpga/xilinx/xc3s200a', 'vq100', 'vqfp100', 'direct')
        for i in range(12):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            devinstname = self.S.next_devinstname('U')

            added = self.S.add_devinst(devinstname, coords, *args)
            FPGAs.append(devinstname)


        LEDs = []
        args = ('led/led', 'AK', '0402', 'ak')
        for i in range(16):
            coords = tuple(wallpy.units.Length(next(self.RD), 'mm') for i in range(3))
            coords += (wallpy.units.Angle(next(self.RD) % 360, 'deg'), )

            devinstname = self.S.next_devinstname('D')

            added = self.S.add_devinst(devinstname, coords, *args)
            LEDs.append(devinstname)



        # Connect GNDs
        def match_pinfuncname(devinst, logpinkey, logpin, desired_funcname):
            if desired_funcname in logpin.pinfuncs:
                return True
            return False

        req_GNDs = tuple(schem.PinSearch(n, self.S.devinst_n2o[n], match_pinfuncname,
                                         match_args = ['GND'])
                         for n in FPGAs)


        for req in req_GNDs:
            DIN = req.devinstname
            self.S.connect_net('GND', ((DIN, lpk) for lpk in req.matches()))

        self.S.connect_net('GND', ((i, 'K') for i in LEDs))


        # Add nets
        for i in range(128):
            netname = 'N_{:03}'.format(i)

            connections = set()

            for i in range(8):

                for itercount in range(1000):

                    # Focus on ICs before other devices
                    for trycount in range(4):
                        devname = random.choice(tuple(self.S.devinst_n2o))
                        if devname.startswith('U'):
                            break

                    logpinkey = random.choice(tuple(self.S.devinst_n2o[devname].logpack.pins))

                    # Make sure the pin is not already connected to a net
                    if (
                        ((devname, logpinkey) in self.S.logpin2net)
                        or ((devname, logpinkey) in connections)
                    ):
                        devname = None
                        logpinkey = None
                    else:
                        break

                if not (devname and logpinkey):
                    self.fail('Ran out of pins to connect to nets')

                connections.add((devname, logpinkey))


            self.S.connect_net(netname, connections)




    def test_add_bundle(self):

        def match_req(devinst, logpinkey, logpin):
            return logpin.required_p()


        reqname = 'B_' + str(next(self.RD))
        netnames = tuple('N_' + str(next(self.RD)) for i in range(3))

        PS1 = schem.PinSearch('U1', self.S.devinst_n2o['U1'], match_req)
        SG1 = schem.EndpointSubgroup(len(netnames), PS1)
        EP1 = schem.BundleEndpoint([SG1])

        expect = schem.Bundle(netnames, wallpy.cleanobject.CleanList([EP1]))
        B1 = self.S.add_bundle(reqname, netnames, [
            [
                (len(netnames), ('U1', match_req, None, None), {}),
            ],
        ])
        self.assertEqual(expect, B1)


        expect = {reqname : B1}
        self.assertDictEqual(expect, self.S.bundles)


        expect = {('U1', match_req, tuple(), tuple()) : PS1}
        self.assertDictEqual(expect, self.S.pinsearches)




    def test_cached_pinsearch(self):

        def match_req(devinst, logpinkey, logpin):
            return logpin.required_p()


        PS1 = self.S.cached_pinsearch('U1', match_req, None, None)
        PS2 = self.S.cached_pinsearch('U1', match_req, None, None)

        self.assertEqual(PS1, PS1)
        self.assertIs(PS1, PS1)

        expect = {
            ('U1', match_req, tuple(), tuple()) : PS1,
        }

        self.assertDictEqual(expect, self.S.pinsearches)


        PS3 = self.S.cached_pinsearch('U3', match_req, None, None)
        self.assertNotEqual(PS1, PS3)

        expect = {
            ('U1', match_req, tuple(), tuple()) : PS1,
            ('U3', match_req, tuple(), tuple()) : PS3,
        }

        self.assertDictEqual(expect, self.S.pinsearches)




__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_Schematic),
    __TL.loadTestsFromTestCase(Test_helperfunctions),
    __TL.loadTestsFromTestCase(Test_PinSearch),
    __TL.loadTestsFromTestCase(Test_EndpointSubgroup),
    __TL.loadTestsFromTestCase(Test_BundleEndpoint),
    __TL.loadTestsFromTestCase(Test_Bundle),
    __TL.loadTestsFromTestCase(Test_Schematic_Reqs),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
