#!/usr/bin/env python3


"""
Import textual results into proper data structures:
-- placeandbanks.Place*

"""

import sys
import re
import ast

import wallpy.units

import placeandbanks


def import_Placements(F):
    """Import placeandbanks.Place from text output of 
    Schematic.solve_place_and_banks.

    Arguments:
    F -- file-like object, to be read line by line

    Yield placeandbanks.Placement trees.

    """

    IDLE = '1: idle'
    W_COORDS = '2: w_coords'
    COORDS = '3: coords'
    NETS = '4: nets'
    SUPPLIES = '5: supplies'
    sepchar = '&'

    re_coord = re.compile(r'^([^.:]+):\s*(.+)$')
    re_terminals = re.compile(r'^([^.:]+\.\w+)(?:(?:,\s*)([^.:]+\.\w+))')
    re_supplies = re.compile(r'^([^.:]+\.\w+):')

    state_next = IDLE
    state_now = IDLE

    bundlename = None
    bundle = None

    place = placeandbanks.Placement()

    for linenum, rawline in enumerate(F, start=1):
        state_now = state_next

        rawline = rawline.rstrip()
        line = rawline.strip()
        if not line:
            continue

        if line.startswith('#'):
            continue

        issep = (line == (sepchar * len(line)))


        if state_now == IDLE:
            if issep:
                state_next = W_COORDS

        else:
            if issep:
                yield place
                place = placeandbanks.Placement()
                bundlename = None
                bundle = None

                state_next = IDLE

            elif state_now == W_COORDS:
                if rawline.lower() == 'd2coords:':
                    state_next = COORDS

            elif state_now == COORDS:
                if rawline.lower() == 'nets:':
                    state_next = NETS

                else:
                    match = re_coord.match(line)
                    if match:
                        DIN, coords = match.groups()
                        place.D2coords[DIN] = ast.literal_eval(coords)
                    else:
                        raise ValueError('BAD LINE', linenum, state_now)

            elif state_now == NETS:
                if rawline.lower() == 'supplies:':
                    state_next = SUPPLIES

                elif line.endswith(':'):
                    bundlename = ast.literal_eval(line[:-1])

                    info = placeandbanks.PlacedBundleInfo(
                        bundlename=bundlename
                    )

                    bundle = placeandbanks.PlacedBundle(
                        info=info,
                        place=place,
                    )

                    place.bundles.append(bundle)

                else:
                    match = re_terminals.match(line)
                    if match:
                        net = placeandbanks.PlacedNet(
                            place=place,
                            bundle=bundle,
                        )

                        for terminaltext in match.groups():
                            DIN, banktext = terminaltext.split('.', 1)

                            if DIN not in place.D2coords:
                                raise ValueError(
                                    'BAD DIN',
                                    linenum,
                                    state_now,
                                    DIN,
                                )

                            if banktext.lower() == 'none':
                                bank = None
                            elif banktext.isdigit():
                                bank = int(banktext)
                            else:
                                raise ValueError(
                                    'BAD BANK',
                                    linenum,
                                    state_now,
                                    banktext,
                                )

                            net.terminals.append((DIN, bank))

                        bundle.nets.append(net)

            elif state_now == SUPPLIES:
                match = re_supplies.match(line)
                if not match:
                    raise ValueError('BAD SUPPLIES', linenum, state_now)

                DIN_bank, supplies = line.split(':')
                DIN, banktext = DIN_bank.split('.', 1)

                if DIN not in place.D2coords:
                    raise ValueError(
                        'BAD DIN',
                        linenum,
                        state_now,
                        DIN,
                    )

                if banktext.lower() == 'none':
                    bank = None
                elif banktext.isdigit():
                    bank = int(banktext)
                else:
                    raise ValueError(
                        'BAD BANK',
                        linenum,
                        state_now,
                        banktext,
                    )

                for supply in supplies.split(','):
                    name, val = (s.strip() for s in supply.split('='))
                    val = ast.literal_eval(val)

                    place.D2b2supplies[DIN][bank][name] = val
