#!/usr/bin/env python3

import sys
import string
import os
import re

import wallpy.version

sys.path.append('.')

import device
import devmanager

VER = wallpy.version.get_version()



args = sys.argv[1:]

if len(args) < 1:
    print('Usage:', sys.argv[0], 'infile', '[...]')
    sys.exit(1)


infilenames = args[0:]




outdir = 'devlib/fpga/xilinx'


filename_remove_suffixes = [
    'pkg',
]


comment_prefixes = [
    '//',
]


pinname_replace = {
    'RSVD' : 'NC',
    'N.C.' : 'NC',
}



unsplit_names = {
    'IO',
    'IP',
    'D',
}




part_numbers = {

    # Artix-7:
    'xc7a8',
    'xc7a15',
    'xc7a30t',
    'xc7a50t',
    'xc7a100t',
    'xc7a200t',
    'xc7a350t',

    # Kintex-7:
    'xc7k70t',
    'xc7k160t',
    'xc7k325t',
    'xc7k355t',
    'xc7k410t',
    'xc7k420t',
    'xc7k480t',

    # Virtex-7:
    'xc7v585t',
    'xc7v1500t',
    'xc7v2000t',
    'xc7vx330t',
    'xc7vx415t',
    'xc7vx485t',
    'xc7vx550t',
    'xc7vx690t',
    'xc7vx980t',
    'xc7vx1140t',

    # Spartan-3A:
    'xc3s50a',
    'xc3s200a',
    'xc3s400a',
    'xc3s700a',
    'xc3s1400a',

    # Spartan-3A DSP:
    'xc3sd1800a',
    'xc3sd3400a',

    # Spartan-6:
    'xc6slx4',
    'xc6slx9',
    'xc6slx16',
    'xc6slx25',
    'xc6slx45',
    'xc6slx75',
    'xc6slx100',
    'xc6slx150',
    'xc6slx25t',
    'xc6slx45t',
    'xc6slx75t',
    'xc6slx100t',
    'xc6slx150t',

    # Virtex-6:
    'xc6vlx75t',
    'xc6vlx130t',
    'xc6vlx195t',
    'xc6vlx240t',
    'xc6vlx365t',
    'xc6vlx550t',
    'xc6vlx760',
    'xc6vsx315t',
    'xc6vsx475t',
    'xc6vhx250t',
    'xc6vhx255t',
    'xc6vhx380t',
    'xc6vhx565t',

    # Virtex-5:
    'xc5vfx100t',
    'xc5vfx130t',
    'xc5vfx200t',
    'xc5vfx30t',
    'xc5vfx70t',
    'xc5vlx110',
    'xc5vlx110t',
    'xc5vlx155',
    'xc5vlx155t',
    'xc5vlx20t',
    'xc5vlx220',
    'xc5vlx220t',
    'xc5vlx30',
    'xc5vlx30t',
    'xc5vlx330',
    'xc5vlx330t',
    'xc5vlx50',
    'xc5vlx50t',
    'xc5vlx85',
    'xc5vlx85t',
    'xc5vsx240t',
    'xc5vsx35t',
    'xc5vsx50t',
    'xc5vsx95t',
    'xc5vtx150t',
    'xc5vtx240t',

}





# ReadAhead from
#   http://pythonconquerstheuniverse.wordpress.com/2011/07/22/read-ahead-and-python-generators/
def ReadAhead(sequence):
    for item in sequence:
        yield item
    yield None # return the "end of file mark" after the last item


def ReadAheadEnumerate(sequence, start = 0):
    for i, v in enumerate(sequence, start = start):
        yield i, v
    yield i, None   # return "end of file mark"


def ReadLine(Fin, state, linenumfield = 'linenum'):
    for L in Fin:
        state[linenumfield] += 1
        yield L.strip('\r\n')
    yield None  # return "end of file mark"






def parse_error(errmsg, line, state, specific = None):
    if specific:
        print(
            'ERR ',
            state['filename'],
            '[',
            state['linenum'],
            ']: ',
            errmsg,
            ' : ',
            repr(specific),
            sep='')

    else:
        print(
            'ERR ',
            state['filename'],
            '[',
            state['linenum'],
            ']: ',
            errmsg,
            sep='')

    sys.exit(1)





def remove_comments(line):
    for cp in comment_prefixes:
        commentindex = line.find(cp)
        if commentindex >= 0:
            line = line[:commentindex]

    return line




colnames_replace = {
    'bank' : 'bank',
    'bufio2' : 'bufio2',
    'col' : 'col',
    'column' : 'col',
    'diff_pair' : 'diff_pair',
    'i/o type' : 'io_type',
    'memory byte group' : 'mem_group',
    'no-connect' : 'noconn',
    'pin' : 'pin',
    'pin_#' : 'pin_num',
    'pin description' : 'name',
    'pin name' : 'name',
    'row' : 'row',
    'row_#' : 'row_num',
    'super logic region' : 'slr',
    'type' : 'type',
    'vccaux group' : 'vccaux_group',
}


def parse_colheaderline(line, state):
    # Remember the names of the columns

    colnames = line.split(state['delim'])

    if 'field2col' not in state:
        state['field2col'] = dict()


    for i, v in enumerate(colnames):
        v = v.lower()
        if v == state['partnum'].lower():
            v = colnames[i] = 'pin name'
            

        if v not in colnames_replace:
            parse_error('unknown column name:', line, state, (i, colnames[i]))


        v = colnames[i] = colnames_replace[v]

        state['field2col'][v] = i


    state['colnames'] = colnames
    state['numcols'] = len(colnames)



# Update the part number in the state dict
def update_partnum(line, state):

    partnum = state['partnum']

    partnum = partnum.strip().lower()

    # Hard-code 5- and 6-series devices to start with "xc"
    if partnum.startswith('5') or partnum.startswith('6'):
        partnum = 'xc' + partnum


    # Make sure this is a known part number, and determine package
    package = None
    for i in part_numbers:
        if partnum.startswith(i):
            package = partnum[len(i):]
            partnum = i
            break

    if not package:
        parse_error('unknown part number', line, state)


    state['partnum'] = partnum
    state['package'] = package



# Returns True if still in the header, False if now in pin section
def parse_headerline(line, state):
    line = remove_comments(line)

    line = line.strip()

    if 'delim' in state:
        line = line.lstrip(state['delim'])


    if not line:
        # Blank line.  Count as part of header.
        return True


    if not state.get('delim', None):
        # First non-blank line should include part number and delimiter

        commaindex = line.find(',')

        if commaindex >= 0:
            # COMMA delimiter
            state['delim'] = ','

        else:
            # TAB delimiter
            state['delim'] = '\t'


        if line.lower().startswith('device/package'):
            # Includes part number, creation date and time

            dummy, state['partnum'], state['partcreationtime'] = line.split(None, 2)


        elif line.lower().startswith('pin,bank,'):
            # Header includes column names.  Third column is part number
            dummy, dummy, state['partnum'], dummy = line.split(state['delim'], 3)

            parse_colheaderline(line, state)


        elif line.lower().startswith('pin,'):
            # Header includes column names.  Second column is part number
            dummy, state['partnum'], dummy = line.split(state['delim'], 2)

            parse_colheaderline(line, state)


        if 'partnum' not in state:
            parse_error('part number not found', line, state)


        update_partnum(line, state)

        return True



    if line.lower().startswith('pin'):
        # Column header

        parse_colheaderline(line, state)

        return True


    # Unknown!  Must be pin line.
    return False



def parse_bankdef(line, state):
    line = line.strip()

    if not line.lower().startswith('#bankstandards'):
        return None

    words = line.split()

    if len(words) != 3:
        parse_error('bank I/O standard line has incorrect number of columns',
                    line, state, (len(cols), 3))

    banks = words[1].split(',')
    standardnames = set(words[2].lower().split(','))

    for name in standardnames:
        try:
            devmanager.load_iostandard(name)
        except KeyError:
            parse_error('invalid I/O standard', line, state, name)

    return {k: standardnames for k in banks}



# Returns defined fields for pin line (or None if no data found)
def parse_pinline(line, state):
    line = remove_comments(line)

    line = line.strip()

    line = line.lstrip(state['delim'])


    if not line:
        return None


    if line.startswith('Total'):
        return None



    fields = dict()


    expectedlen = state['numcols']

    cols = [s.strip() for s in line.split(state['delim'], expectedlen - 1)]


    # If special line indicating NOPAD/UNCONNECTED
    for v in cols[1:]:
        if v.lower() in ('nopad/unconnected', 'float'):
            del cols[1:]

            for i in range(0, expectedlen - 1):
                cols.append('NC')

            break


    if len(cols) != expectedlen:
        parse_error('pin line has incorrect number of columns',
                    line, state, (len(cols), expectedlen))


    for i, v in enumerate(cols):
        colname = state['colnames'][i]

        fields[colname] = v.strip()


    ### Extract bank from name if bank is not defined
    if fields['bank'].lower() == 'na':
        words = fields['name'].split('_')
        v = words[-1]

        if v.isdigit():
            print('CORRECTING BANK from "{}" to {} for "{}"'.format(
                fields['bank'], v, fields['name'])
            )

            fields['bank'] = v



    
    return fields






pinname_suffixes = {
    'B',
    'N',
    'S',
    'L',
    'R',
    'LN',
    'LS',
    'RN',
    'RS',
}

pinname_suffix_re = [
    re.compile(r'^G\d+$'),
]


pinname_singlefunc_re = [
    re.compile(r'^VCCAUX_IO_G\d+$'),
]


def pinname_suffix_p(name, line, state):

    if name in pinname_suffixes:
        return True


    for r in pinname_suffix_re:
        if r.match(name):
            return True


    return False


def pinname_split(fields, line, state):

    name = fields['name']


    # Single function pin?
    for r in pinname_singlefunc_re:
        if r.match(name):
            yield name
            return


    # Handle extra delimiter for Spartan3A FPGAs
    name = name.replace('/', '_')

    words = name.split('_')

    # If a word is <= 2 chars in length, it probably is a suffix

    for i, v in enumerate(words):
        if not v:
            continue


        # Handle special case of X_*, where X in names to unsplit (can't be last word)
        if (v in unsplit_names) and (i != (len(words) - 1)):

            # Suppress bank identifier from names
            if (words[i+1] == fields['bank']) and words[i+1].isdigit():
                words[i+1] = ''



            # Only append next part if nonempty
            if words[i+1]:
                words[i] = '_'.join([
                    words[i],
                    words[i+1],
                ])

                words[i+1] = ''



        # Handle suffixes
        if pinname_suffix_p(v, line, state):

            if i == 0:
                continue

            words[i-1] = '_'.join([
                words[i-1],
                v,
            ])

            words[i] = ''


    for v in words:
        if v:
            # Suppress bank identifier from names
            if (v == fields['bank']) and v.isdigit():
                continue

            elif v.isdigit():
                print('ASSUMING BANK "{}" in "{}"'.format(v, name))
                continue

            if v in pinname_replace:
                v = pinname_replace[v]

            yield v







pinname2type_exact = {
    'ADV_B'         :   frozenset({ 'dig_out' }),
    'AVDD'          :   frozenset({ 'pow_in', 'REQ' }),
    'AVSS'          :   frozenset({ 'pow_in', 'REQ' }),
    'AWAKE'         :   frozenset({ 'dig_out' }),
    'BUSY'          :   frozenset({ 'dig_out' }),
    'CCLK'          :   frozenset({ 'dig_bidir', 'REQ' }),
    'CFGBVS'        :   frozenset({ 'pow_in', 'REQ' }),
    'CMPCLK'        :   frozenset({ 'dig_bidir' }),
    'CMPCS_B'       :   frozenset({ 'NC' }),
    'CMPMISO'       :   frozenset({ 'dig_bidir' }),
    'CMPMOSI'       :   frozenset({ 'dig_bidir' }),
    'CC'            :   frozenset({ 'dig_in' }),
    'CS_B'          :   frozenset({ 'dig_in' }),
    'CSI_B'         :   frozenset({ 'dig_in' }),
    'CSO_B'         :   frozenset({ 'dig_out' }),
    'DIN'           :   frozenset({ 'dig_in' }),
    'D_IN'          :   frozenset({ 'dig_in' }),
    'DONE'          :   frozenset({ 'dig_bidir' }),
    'DOUT'          :   frozenset({ 'dig_out' }),
    'D_OUT'         :   frozenset({ 'dig_out' }),
    'DQS'           :   frozenset({ 'dig_in' }),
    'DXN'           :   frozenset({ 'ana_out' }),
    'DXP'           :   frozenset({ 'ana_out' }),
    'EMCCLK'        :   frozenset({ 'dig_in' }),
    'FCS_B'         :   frozenset({ 'dig_out' }),
    'FOE_B'         :   frozenset({ 'dig_out' }),
    'FWE_B'         :   frozenset({ 'dig_out' }),
    'GC'            :   frozenset({ 'dig_in' }),
    'GND'           :   frozenset({ 'pow_in', 'REQ' }),
    'GNDADC'        :   frozenset({ 'pow_in', 'REQ' }),
    'HDC'           :   frozenset({ 'dig_out' }),
    'HSWAPEN'       :   frozenset({ 'dig_in', 'REQ' }),
    'INIT_B'        :   frozenset({ 'dig_bidir', 'REQ' }),
    'IO'            :   frozenset({ 'dig_bidir' }),
    'IP'            :   frozenset({ 'dig_in' }),
    'LDC'           :   frozenset({ 'dig_out' }),
    'MGTAVTTRCAL'   :   frozenset({ 'ana_in', 'REQ' }),
    'MGTRBIAS'      :   frozenset({ 'ana_in', 'REQ' }),
    'MGTRREF'       :   frozenset({ 'ana_in', 'REQ' }),
    'MGTVCCAUX'     :   frozenset({ 'pow_in', 'REQ' }),
    'MISO'          :   frozenset({ 'dig_in' }),
    'MOSI'          :   frozenset({ 'dig_out' }),
    'MRCC'          :   frozenset({ 'dig_in' }),
    'NC'            :   frozenset({ 'NC' }),
    'PROGRAM_B'     :   frozenset({ 'dig_in', 'REQ' }),
    'PROG_B'        :   frozenset({ 'dig_in', 'REQ' }),
    'PUDC_B'        :   frozenset({ 'dig_in', 'REQ' }),
    'RDWR_B'        :   frozenset({ 'dig_in' }),
    'RFUSE'         :   frozenset({ 'pow_in' }),
    'SRCC'          :   frozenset({ 'dig_in' }),
    'SUSPEND'       :   frozenset({ 'dig_in', 'REQ' }), # required pulldown during config
    'TCK'           :   frozenset({ 'dig_in', 'REQ' }),
    'TDI'           :   frozenset({ 'dig_in', 'REQ' }),
    'TDO'           :   frozenset({ 'dig_out', 'REQ' }),
    'TMS'           :   frozenset({ 'dig_in', 'REQ' }),
    'USERCCLK'      :   frozenset({ 'dig_in' }),
    'VBATT'         :   frozenset({ 'pow_in' }),
    'VFS'           :   frozenset({ 'pow_in' }),
    'VREF'          :   frozenset({ 'pow_in' }),
}

pinname2type_re = {
    re.compile(r'^A\d{1,2}$')           :   frozenset({ 'dig_out' }),
    re.compile(r'^(AD\d{1,2})([NP])()$') : frozenset({ 'ana_in', 'DIFF' }),
    re.compile(r'^D\d{1,2}$')           :   frozenset({ 'dig_bidir' }),
    re.compile(r'^FS[0-2]$')            :   frozenset({ 'dig_in' }),
    re.compile(r'^GCLK\d{1,2}$')        :   frozenset({ 'dig_in' }),
    re.compile(r'^IO_\d{1,2}$')         :   frozenset({ 'dig_bidir' }),
    re.compile(r'^(IO_L\d{1,2})([NP])()$') : frozenset({ 'dig_bidir', 'DIFF' }),
    re.compile(r'^(IP_L\d{1,2})([NP])()$') : frozenset({ 'dig_in', 'DIFF' }),
    re.compile(r'^[IT]RDY[1-2]$')       :   frozenset({ 'dig_out' }),
    re.compile(r'^LDC[0-2]$')           :   frozenset({ 'dig_out' }),
    re.compile(r'^[LR]HCLK\d{1,2}$')    :   frozenset({ 'dig_in' }),
    re.compile(r'^M[0-2]$')             :   frozenset({ 'dig_in', 'REQ' }),
    re.compile(r'^M\dA\d{1,2}$')        :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dBA[0-2]$')         :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dCASN$')            :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dCKE$')             :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dCLK$')             :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dCLKN$')            :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dDQ\d{1,2}$')       :   frozenset({ 'dig_bidir' }),
    re.compile(r'^M\d[LU]DM$')          :   frozenset({ 'dig_out' }),
    re.compile(r'^M\d[LU]DQSN?$')       :   frozenset({ 'dig_bidir' }),
    re.compile(r'^M\dODT$')             :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dRASN$')            :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dRESET$')           :   frozenset({ 'dig_out' }),
    re.compile(r'^M\dWE$')              :   frozenset({ 'dig_out' }),
    re.compile(r'^MGTAVCC.*$')          :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^MGTAVTT.*$')          :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^MGTHAGND.*$')         :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^MGTHAVCC.*$')         :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^MGTHAVTT.*$')         :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^(MGTREFCLK[01]?)([NP])*()$') : frozenset({ 'dig_in', 'DIFF' }),
    re.compile(r'^MGTVCCAUX_G\d{1,2}$') :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^(MGTX?RX)([NP])([0-3])$') : frozenset({ 'dig_in', 'DIFF' }),
    re.compile(r'^(MGTX?TX)([NP])([0-3])$') : frozenset({ 'dig_out', 'DIFF' }),
    re.compile(r'^MISO[0-3]$')          :   frozenset({ 'dig_bidir' }),
    re.compile(r'^RS[0-1]$')            :   frozenset({ 'dig_out' }),
    re.compile(r'^SCP[0-7]$')           :   frozenset({ 'dig_in' }),
    re.compile(r'^(SM\d{1,2})([NP])()$') : frozenset({ 'ana_in', 'DIFF' }),
    re.compile(r'^T[0-3]$')             :   frozenset({ 'dig_in' }),
    re.compile(r'^VCC.*$')              :   frozenset({ 'pow_in', 'REQ' }),
    re.compile(r'^V[NP]$')              :   frozenset({ 'ana_in' }),
    re.compile(r'^VREF[NP]$')           :   frozenset({ 'ana_in', 'REQ' }),
    re.compile(r'^VR[NP]$')             :   frozenset({ 'ana_in' }),
    re.compile(r'^VS[0-2]$')            :   frozenset({ 'dig_in' }),
}




def pinname2pintypes(name):
    
    if name in pinname2type_exact:
        return (set(pinname2type_exact[name]), None)

    for r, v in pinname2type_re.items():
        M = r.match(name)
        if M:
            return (set(v), M)


    raise ValueError('Unknown pin: ' + repr(name))



fpgas = device.DevCollection()


for infilename in infilenames:

    infileprefix, infileext = os.path.splitext(infilename)


    with open(infilename, 'r') as Fin:

        state = {
            'filename' : infilename,
            'linenum' : 0,
        }

        LineReader = ReadLine(Fin, state, linenumfield = 'linenum')

        # First parse header until first pin line

        InHeader = True
        
        while InHeader:
            L = next(LineReader)
            InHeader = parse_headerline(L, state)

            # If no longer part of header, save line for next section of parsing


        if 'colnames' not in state:
            parse_error('no header found', L, state)



        partnum = state['partnum']
        package = state['package']

        if partnum not in fpgas.devices:
            print('Creating device', partnum)

            fpgas.devices[partnum] = device.Device()

        D = fpgas.devices[partnum]

        if package not in D.packages:
            D.packages[package] = device.LogicalPackage()

        pack = D.packages[package]


        pin_longest_prefix = None


        # Now parse pin lines
        while L is not None:
            bankdef = parse_bankdef(L, state)

            if bankdef is None:
                fields = parse_pinline(L, state)
            else:
                fields = None

                for bank, standards in bankdef.items():
                    try:
                        bank = int(bank)
                    except ValueError:
                        pass

                    if bank in D.bank2iostd:
                        D.bank2iostd[bank] |= standards
                    else:
                        # Make separate copy of standards set
                        D.bank2iostd[bank] = set(standards)


            if fields:

                pin = device.LogicalPin()

                pack.pins[fields['pin']] = pin


                if pin_longest_prefix is None:
                    pin_longest_prefix = fields['pin']

                elif fields['pin'] != pin_longest_prefix:
                    prefixlen = 0
                    while fields['pin'][:prefixlen] == pin_longest_prefix[:prefixlen]:
                        prefixlen += 1
                    prefixlen -= 1

                    pin_longest_prefix = pin_longest_prefix[:prefixlen]


                bank = fields['bank']

                if bank.isdigit():
                    bank = int(bank)


                noconn = False

                for pinname in pinname_split(fields, L, state):
                    pintypes, match = pinname2pintypes(pinname)

                    params = {
                        'bank': bank,
                    }

                    if 'NC' in pintypes:
                        noconn = True
                        continue

                    req = False
                    if 'REQ' in pintypes:
                        req = True
                        pintypes.remove('REQ')

                    sigconfig = device.SINGLE
                    if 'DIFF' in pintypes:
                        pintypes.remove('DIFF')

                        name_prefix, p_n, name_suffix = match.groups()
                        other_p_n = {
                            'P': 'N',
                            'N': 'P',
                            'p': 'n',
                            'n': 'p',
                        }[p_n]

                        other_name = ''.join((
                            name_prefix, other_p_n, name_suffix
                        ))

                        params['diff_pair'] = other_name


                        sigconfig = {
                            'p': device.DIFF_P,
                            'n': device.DIFF_N,
                        }[p_n.lower()]


                    if len(pintypes) != 1:
                        parse_error('unexpected length of pintypes',
                                    L, state, (pinname, repr(pintypes)))

                    t = pintypes.pop()

                    try:
                        sigtype, sigdir = t.split('_')
                    except ValueError:
                        parse_error('unexpected pintype', L, state, repr(t))


                    pf = device.PinFunc(sigtype, sigdir, sigconfig,
                                        required=req, params=params)

                    pin.pinfuncs[pinname] = pf


                if pin.pinfuncs and noconn:
                    parse_error('noconn but have pinfuncs', L, state,
                                repr(pin.pinfuncs))


            L = next(LineReader)

        if pin_longest_prefix == 'P':
            # Convert to integer pin numbers

            prefixlen = len(pin_longest_prefix)

            for k in tuple(pack.pins):
                newkey = int(k[prefixlen:])

                pack.pins[newkey] = pack.pins[k]
                del pack.pins[k]


# Generate and test output files

sys.path.append(outdir)

for partnum, D in fpgas.devices.items():
    outfilename = os.path.join(outdir, partnum + '.py')

    with open(outfilename, 'w') as Fout:
        print('#!/usr/bin/env python3', file=Fout)
        print(file=Fout)
        print('"""FPGA device', partnum, file=Fout)
        print(file=Fout)
        print('Automatically generated by XilinxPinout.py', file=Fout)
        print(file=Fout)
        print('Latest tool revision:', VER, file=Fout)
        print('"""', file=Fout)
        print(file=Fout)
        print('import wallpy.cleanobject', file=Fout)
        print('import device', file=Fout)
        print(file=Fout)
        print(file=Fout)

        print('dev = ', end='', file=Fout)
        D.prettyprint(file = Fout)


    def errcallback(*args):
        print('ERR', *args)

    print('Generated ', outfilename, ' (valid: ', D.valid_p(errcallback), ')', sep='')

    tmp_module = __import__(partnum, globals = dict(), locals = dict())
    tmp_dev = tmp_module.dev

    print('\tChecking for equivalence...', 'good!' if tmp_dev == D else 'mismatch!')

