#!/usr/bin/env python3

"""Go through devices specified on commandline and evaluate BGA pin keys.

Ensure row letters are consistent (i.e. smaller set is subset of larger
set).
Ensure col numbers are consistent (i.e. range 1 .. max(colnum))

"""

import sys
import os.path

import devmanager
import schem


def rowsort(rowname):
    return (len(rowname), rowname.lower())


def consistent_range(smaller, larger, key=None):
    """Make sure the sorted larger begins with the sorted smaller.

    Return True if sorted(smaller) == sorted(larger)[:len(smaller)],
        or False otherwise.

    Arguments:
    smaller, larger -- iterables, len(smaller) <= len(larger)
    key -- sort key function

    """

    if len(smaller) > len(larger):
        return ValueError('smaller has more elements than longer')

    s_small = sorted(smaller, key=key)
    s_large = sorted(larger, key=key)

    return (s_large[:len(smaller)] == s_small)


def consistent_range_noorder(L, R, key=None):
    """Make sure the sorted larger of (L, R) begins with the sorted
    smaller of (L, R).

    Return True if sorted(smaller) == sorted(larger)[:len(smaller)],
        or False otherwise.

    Arguments:
    L, R -- iterables (no particular order required)
    key -- sort key function

    """

    if len(L) <= len(R):
        return consistent_range(L, R, key=key)
    else:
        return consistent_range(R, L, key=key)


rowset = set()  # Letters in pin key

args = sys.argv[1:]

for filename in args:
    rel = os.path.relpath(filename, start='devlib')
    devname, ext = os.path.splitext(rel)
    assert(ext == '.py')

    print('Evaluating', devname)
    dev = devmanager.load_device(devname)

    for packname, pack in dev.packages.items():
        rows = set()
        cols = set()

        for pinname in pack.pins:
            L, M, R = schem.name2ana(pinname)

            if not L:
                # Purely numeric pin name
                continue

            rows.add(L)
            cols.add(M)

        # Ensure row letters are consistent
        if not consistent_range_noorder(rows, rowset, key=rowsort):
            print('ROW ERR', devname, packname, sorted(rows, key=rowsort))
            sys.exit(1)

        if len(rows) > len(rowset):
            rowset = rows

        # Ensure column numbers are consistent
        if cols != set(range(1, len(cols)+1)):
            print('COL ERR', devname, packname, sorted(cols))
            sys.exit(1)

    devmanager.clear_caches()

print('Rows:', ' '.join(sorted(rowset, key=rowsort)))
