#!/usr/bin/env python3

import sys
import argparse
import random
import os.path
import time
import ast
import itertools
import code
import wallpy.units
import wallpy.version

import schem
import draw
import textimport


class MoveDrawHelper(object):
    """Helper class for drawing moves."""

    def __init__(self, filenamestr, boardsize, imagesize, D2dim, hightemp):
        self.filenamestr = filenamestr
        self.boardsize = boardsize
        self.imagesize = imagesize
        self.D2dim = D2dim
        self.hightemp = float(hightemp)

        self.count = itertools.count()


    def draw(self, P, extratext='', colors=None):
        fn = self.filenamestr.format(next(self.count))

        draw.draw_simple_placement(
            fn,
            P.D2coords,
            self.D2dim,
            self.boardsize,
            self.imagesize,
            extratext=extratext,
            colors=colors
        )


def print_title(title):
    bar = '#' * len(title)
    print()
    print(bar)
    print(title)
    print(bar)
    print()


def filename_add_info(filename, seed, version):
    fileprefix, filesuffix = os.path.splitext(filename)
    return '{}-{}-{}{}'.format(fileprefix, seed, version, filesuffix)


def dict_item(s):
    """Convert a string of format key=val to (arg, val).

    Raise ArgumentTypeError if string is of incorrect format.

    """

    words = s.split('=')

    if len(words) != 2:
        raise argparse.ArgumentTypeError(
            'Format is not key=val in {}'.format(s)
        )

    key, val = words

    try:
        val = ast.literal_eval(val)
    except ValueError:
        pass

    return (key, val)


def lists2dict(L):
    """Convert multiple lists of (k, v) pairs to a dict."""

    ret = dict()

    for kviter in L:
        ret.update(kviter)

    return ret


def first(iterable):
    """Return first element in iterable."""

    return next(iter(iterable))


if __name__ == '__main__':
    VERSION = wallpy.version.get_version()

    VERSION_HASHONLY = wallpy.version.get_version(
        require_prefix='',
        match_pattern='^DUMMY$' # No tag should match this
    )

    # Postprocess args using function (e.g. run dict on nargs=+)
    postparse_type = dict() # key: function

    # Commandline argument parser
    parser = argparse.ArgumentParser(
        description='Read and optionally solve a design file.',
        fromfile_prefix_chars='@'
    )

    parser.add_argument(
        '--version',
        action='version',
        version=VERSION,
    )

    parser.add_argument(
        '-1', '--solve1',
        action='store_true',
        default=False,
        help='Solve obvious Bundles'
    )

    ilp_group = parser.add_mutually_exclusive_group()
    ilp_group.add_argument(
        '-2', '--solveilp',
        action='store_true',
        default=False,
        help='Solve pin selection using ilp'
    )

    ilp_group.add_argument(
        '-3', '--solvebanks',
        action='store_true',
        default=False,
        help='Solve bank selection using ilp'
    )

    ilp_group.add_argument(
        '-4', '--solvebankpins',
        action='store_true',
        default=False,
        help='Solve pin selection given bank-to-bank assignments'
    )

    parser.add_argument(
        '--round-digits',
        nargs=1,
        type=int,
        default=[schem.DEFAULT_ROUND_DIGITS],
        help='Digits after decimal point to round ILP cost',
        metavar='int'
    )
    postparse_type['round_digits'] = first


    place_group = parser.add_mutually_exclusive_group()
    place_group.add_argument(
        '-p', '--place',
        nargs=2,
        help='Automatically place components',
        metavar=('board_width', 'board_height')
    )

    place_group.add_argument(
        '-b', '--place-and-banks',
        nargs=2,
        help='Automatically place components and make bank assignments',
        metavar=('board_width', 'board_height')
    )

    place_group.add_argument(
        '--import-placements',
        nargs=2,
        help='Import previously placed components',
        metavar=('filename', 'ref_unit')
    )


    parser.add_argument(
        '--place-params',
        action='append',
        nargs='+',
        type=dict_item,
        default=list(),
        help='Override placement parameter defaults',
        metavar='key=val'
    )
    postparse_type['place_params'] = lists2dict

    parser.add_argument(
        '-s', '--svg',
        nargs=1,
        default=None,
        help='Export final Schematic SVG to svgfile',
        metavar=('svgfile',)
    )

    parser.add_argument(
        '--svg-title-tag',
        nargs=1,
        default=None,
        help='Append tag to SVG title',
        metavar='tag'
    )

    parser.add_argument(
        '--svg-prefs',
        action='append',
        nargs='+',
        type=dict_item,
        default=list(),
        help='Override SVG preference defaults',
        metavar='key=val'
    )
    postparse_type['svg_prefs'] = lists2dict

    parser.add_argument(
        '--svg-styles',
        action='append',
        nargs='+',
        type=dict_item,
        default=list(),
        help='Override SVG style defaults',
        metavar='type.name.key=val'
    )
    postparse_type['svg_styles'] = lists2dict

    parser.add_argument(
        '-t', '--text',
        nargs=1,
        default=None,
        help='Export final Schematic as text to textfile',
        metavar='textfile'
    )

    parser.add_argument(
        '--valid',
        action='store_true',
        default=False,
        help='Show final Schematic validity'
    )

    parser.add_argument(
        '--seeds',
        nargs='+',
        type=int,
        default=[None],
        help='Seed random number generator for repeatable output',
        metavar='int'
    )

    parser.add_argument(
        '--best-seeds',
        type=int,
        default=3,
        help='Keep N best placements',
        metavar='N'
    )

    parser.add_argument(
        '--debug-bundles',
        action='store_true',
        default=False,
        help='Show all Bundles'
    )

    parser.add_argument(
        '--debug-ilp',
        nargs=1,
        default=[None],
        help='Save ILP variables and constraints',
        metavar='csvfile'
    )
    postparse_type['debug_ilp'] = first

    parser.add_argument(
        '--debug-banks',
        action='store_true',
        default=False,
        help='Show all DevInst banks within Bundles'
    )

    parser.add_argument(
        '-i', '--debug-interact',
        action='store_true',
        default=False,
        help='Drop into interactive shell after processing'
    )

    parser.add_argument(
        '--debug-draw-moves',
        nargs=4,
        default=[],
        help='Draw all moves to FILENAMEs of size WxH starting @T'
    )

    parser.add_argument(
        'designfile',
        help='Design intent input file'
    )


    prefs = parser.parse_args()

    for k, func in postparse_type.items():
        v = getattr(prefs, k)
        v = func(v)
        setattr(prefs, k, v)

    print('Tool version:', VERSION)

    print('Preferences:')
    for k, v in sorted(vars(prefs).items()):
        print('\t', k, ': ', repr(v), sep='')
    print()


    try:
        S = schem.from_file(prefs.designfile)
    except IOError:
        print('Could not open file', repr(prefs.designfile))
        sys.exit(1)

    if prefs.valid:
        def errcallback(*args, **kwargs):
            print('ERR', args, repr(kwargs))
        print('Valid?', S.valid_p(errcallback))


    if prefs.place:
        print_title('Automatic Initial Placement')

        board_dim = tuple(
            float(
                wallpy.units.Length(v)
                / wallpy.units.Length(1, 'mm')
            )
            for v in prefs.place
        )

        time_place = time.time()

        placed_results = S.find_best_placements(
            prefs.seeds,
            board_dim,
            params=prefs.place_params
        )

        time_place = time.time() - time_place

        print()
        print('Initial placement took', time_place, 'seconds')


        print_title('Placed results')

        for cost, P, seed in placed_results:
            print('Seed', seed, 'has placement cost', cost)


        ### Take the N best seeds
        N = prefs.best_seeds
        seed2placement = dict()
        seed2cost = dict()

        for cost, P, seed in placed_results[:prefs.best_seeds]:
            seed2placement[seed] = P
            seed2cost[seed] = cost


    elif prefs.place_and_banks:
        print_title('Automatic Initial Placement and Bank Assignment')

        unit_length = wallpy.units.Length(1, 'mm')

        board_dim = tuple(
            float(
                wallpy.units.Length(v)
                / unit_length
            )
            for v in prefs.place_and_banks
        )

        D2dim = {
            DIN: tuple(
                float(wallpy.units.Length(v) / unit_length)
                for v in DI.phypack.dimensions[:2]
            )
            for DIN, DI in S.devinst_n2o.items()
        }

        if prefs.debug_draw_moves:
            (
                filenamestr,
                image_width,
                image_height,
                start_temp
            ) = prefs.debug_draw_moves

            imagesize = tuple(map(int, (image_width, image_height)))
            post_move_arg = MoveDrawHelper(
                filenamestr,
                board_dim,
                imagesize,
                D2dim,
                start_temp
            )
            post_move_arg.textcount = itertools.count()

            def post_move_callback(P, accepted, cur_state, arg):
                if cur_state['T'] > arg.hightemp:
                    return
                if not P.movetype.startswith('DI_'):
                    return
                text = '#{:09d} @{:04g}. Cost {:04g} {} ({}) {}'.format(
                    next(arg.textcount),
                    cur_state['T'],
                    P.total_cost,
                    P.movetype,
                    P.debug_move,
                    'ACCEPTED' if accepted else 'rejected'
                )
                colors = dict()
                if accepted:
                    colors['bg'] = (0.9, 1, 0.9)
                else:
                    colors['bg'] = (1, 0.9, 0.9)
                arg.draw(P, extratext=text, colors=colors)

            prefs.place_params['post_move_callback'] = post_move_callback
            prefs.place_params['post_move_arg'] = post_move_arg

        time_place = time.time()

        placed_results = S.find_best_place_and_banks(
            prefs.seeds,
            board_dim,
            params=prefs.place_params
        )

        time_place = time.time() - time_place

        print()
        print('Initial placement took', time_place, 'seconds')


        print_title('Placed results')

        for cost, P, seed in placed_results:
            print('Seed', seed, 'has placement cost', cost)


        ### Take the N best seeds
        N = prefs.best_seeds
        seed2placement = dict()
        seed2cost = dict()

        for cost, P, seed in placed_results[:prefs.best_seeds]:
            seed2placement[seed] = P
            seed2cost[seed] = cost


    elif prefs.import_placements:
        filename, ref_unit = prefs.import_placements

        N = 0
        seed2placement = dict()
        seed2cost = dict()

        with open(filename, 'r') as F:
            for placement in textimport.import_Placements(F):

                placement.D2coords = {
                    DIN: schem.coords_angle_div_pi(coords)
                    for DIN, coords in placement.D2coords.items()
                }

                S.set_placement_2D(
                    placement.D2coords,
                    length_in_units=ref_unit,
                    angle_in_units='pi_rad'
                )

                placement.D2coords = {
                    DIN: DI.coords
                    for DIN, DI in S.devinst_n2o.items()
                }

                seed = 'i{}'.format(N)
                seed2placement[seed] = placement
                seed2cost[seed] = 0
                N += 1


    else:
        ### No automatic placement. Use original.
        seed2placement = {None: schem.Placement(D2coords=S.DIN2coords)}
        seed2cost = {None: -1}


    orig_nets = S.nets.copy()
    orig_logpin2net = S.logpin2net.copy()

    for seedindex, seed in enumerate(seed2placement, 1):
        print_title('Iteration {} (Seed: {!r})'.format(seedindex, seed))


        ### Clear other placements' potential pins.
        S.potential_pins.clear()
        S.potential_pins2.clear()


        ### Recall original nets
        S.nets = orig_nets.copy()
        S.logpin2net = orig_logpin2net.copy()


        ### Recall this placement
        for DIN, coords in seed2placement[seed].D2coords.items():
            S.devinst_n2o[DIN].coords = coords


        if prefs.solve1:
            S.solve_phase1_obvious()

        if prefs.solveilp:
            if prefs.solvebanks:
                print('--solveilp is mutually exclusive with --solvebanks')
                sys.exit(1)

            round_digits = prefs.round_digits

            debugfilename = prefs.debug_ilp
            if debugfilename:
                debugfilename = filename_add_info(
                    debugfilename, seed, VERSION_HASHONLY
                )

            S.solve_ilp(
                round_digits=round_digits,
                debugfilename=debugfilename
            )

        if prefs.solvebanks:
            for val, var in S.assign_banks():
                print(val, var)

        if prefs.solvebankpins:
            if not hasattr(seed2placement[seed], 'bundles'):
                print('Placement has no bundle information!')
                sys.exit(1)

            round_digits = prefs.round_digits

            debugfilename = prefs.debug_ilp
            if debugfilename:
                debugfilename = filename_add_info(
                    debugfilename, seed, VERSION_HASHONLY
                )

            S.solve_b2b_ilp(
                seed2placement[seed],
                round_digits=round_digits,
                debugfilename=debugfilename
            )

        if prefs.text:
            filename = prefs.text[0]

            filename = filename_add_info(filename, seed, VERSION_HASHONLY)

            with open(filename, 'w') as F:
                S.export(F)

        if prefs.svg:
            filename = prefs.svg[0]

            filename = filename_add_info(filename, seed, VERSION_HASHONLY)

            titletag = prefs.svg_title_tag
            if titletag:
                titletag = ' ' + titletag[0]

            comments = list()
            comments.append('Preferences:')
            for k, v in sorted(vars(prefs).items()):
                comments.append('\t{}: {}'.format(k, v))

            comments.append('Costs:')
            for line in str(seed2placement[seed].costs).split('\n'):
                comments.append('\t{}'.format(line))

            svg_prefs = {
                'title': '{} (cost={}){} ({})'.format(
                    os.path.basename(filename),
                    round(seed2cost[seed], 2),
                    titletag,
                    VERSION
                ),
                'pinlabels': False,
                'pins': True,
                'comments': comments,
            }

            svg_prefs.update(prefs.svg_prefs)

            svg_styles = dict(prefs.svg_styles)

            with open(filename, 'w') as F:
                S.export_svg(F, prefs=svg_prefs, styles=svg_styles)

        if prefs.valid:
            def errcallback(*args, **kwargs):
                print('ERR', args, repr(kwargs))
            print('Valid?', S.valid_p(errcallback))


    if prefs.debug_bundles:
        S.debug_bundles(include_results=True)

    if prefs.debug_banks:
        for bundlename, B in S.bundles.items():
            print(bundlename, 'BANKS BEGIN')
            for k, v in B.DIN2banks().items():
                print('\t{:}: {}'.format(k, v))
            print(bundlename, 'BANKS END')
            print()

    if prefs.debug_interact:
        code.interact(local=locals())


    print()
    print('Done.')
