#!/usr/bin/env python3

import sys
import math
from xml.etree import ElementTree as et
import wallpy.units

sys.path.append('.')

import devmanager



main_unit = 'mm'


extend_size = (     # in main units
    80.0,
    80.0,
)




styles = {
    'title'     :   {
        'font-family'       :   'Sans',
        'font-size'         :   '6px',
        'text-anchor'       :   'start',
    },


    'pad'       :   {
        'font-family'       :   'Sans',
        'font-size'         :   '2px',
        'text-anchor'       :   'middle',
        'dominant-baseline' :   'central',
    },

}


for k in styles:
    attr = styles[k]
    styles[k] = ';'.join((':'.join(kv) for kv in attr.items()))



def length2float(L):
    return float(L / wallpy.units.Length(1, main_unit))



def pinindex2angle(pinindex, numpins, fixedpin, fixedangle):
    i = pinindex - fixedpin
    return ((360 * i / numpins) + fixedangle) % 360



def trans_zoom(parent, x=0, y=0, zoom=1):
    g_trans = et.SubElement(
        parent, 'g',
        transform='translate({:0.06F}, {:0.06F})'.format(x, y))

    return et.SubElement(g_trans, 'g', transform='scale({:0.08E})'.format(zoom))



def phypack_svg(phypack_fn, zoom=100, file=sys.stdout,
               features=frozenset()):

    """Print the given PhysicalPackage as an SVG file."""

    ### Adapted from
    ### http://nick.onetwenty.org/index.php/2010/04/07/creating-svg-files-with-python/


    PP = devmanager.load_phypack(phypack_fn)
    dim_float = tuple(length2float(PP.dimensions[i]) for i in range(2))

    width, height = (((zoom * dim_float[i]) + extend_size[i]) for i in range(2))

    attrib = {
        'width'     :   '{:0.06F}'.format(width) + main_unit,

        'height'    :   '{:0.06F}'.format(height) + main_unit,

        'viewBox'   :   ' '.join('{:0.06F}'.format(v) for v in (0, 0, width, height)),
    }

    doc = et.Element('svg', version='1.1', xmlns='http://www.w3.org/2000/svg',
                     attrib=attrib)


    ### Print title
    title = et.SubElement(doc, 'title')
    title.text = 'PhysicalPackage: ' + phypack_fn


    ### Background for debug
    et.SubElement(doc, 'rect', x='0', y='0', width='{:0.06F}'.format(width),
                  height='{:0.06F}'.format(height), fill='lightcyan')


    title_label = et.SubElement(doc, 'text', x='1', y='5', style=styles['title'])
    title_label.text = phypack_fn



    ### Coordinate system of package
    dx, dy = (width/2, (zoom * dim_float[1])/2 + (extend_size[1]/2))

    pack_area = trans_zoom(doc, x=dx, y=dy, zoom=zoom)



    ### Show outer bounds of package
    pack_dim = tuple(length2float(PP.dimensions[i]) for i in range(2))
    x, y = ('{:0.06F}'.format(-pack_dim[i] / 2) for i in range(2))
    width, height = ('{:0.06F}'.format(pack_dim[i]) for i in range(2))

    et.SubElement(pack_area, 'rect', x=x, y=y, width=width, height=height, fill='lightgray')

    if (
        ('radial_line' in features)
        or ('map2circ' in features)
    ):

        # Determine distance from center to outer corner
        pack_radius = math.sqrt(sum((c * c) for c in pack_dim)) / 2



    if 'map2circ' in features:
        numpins = len(PP.pins)


        ### Draw circle of virtual pins
        et.SubElement(pack_area, 'circle', cx='0', cy='0', fill='none',
                      r='{:0.06F}'.format(pack_radius), attrib={
                          'stroke'          :   'blue',
                          'stroke-width'    :   '0.01',
                      })
                      


        ### Find a pin that aligns perfectly to the circle
        #       (x == 0) or (y == 0) or (|x| == |y|)


        zero_length = wallpy.units.Length(0, 'mm')


        found_pinkey = None
        found_angle = None

        for phypinkey, phypin in PP.pins.items():
            x, y = phypin.coords[:2]

            if x == zero_length:
                if y > zero_length:
                    found_angle = 90
                elif y < zero_length:
                    found_angle = -90
                else:
                    continue


            if y == zero_length:
                if x > zero_length:
                    found_angle = 0
                elif x < zero_length:
                    found_angle = 180
                else:
                    continue


            if abs(x) == abs(y):
                x_positive = x > zero_length
                y_positive = y > zero_length

                angles = {
                    (True, True)    :   45,
                    (False, True)   :   135,
                    (False, False)  :   225,
                    (True, False)   :   315,
                }

                found_angle = angles[(x_positive, y_positive)]

            if found_angle is not None:
                found_pinkey = phypinkey
                break



        if found_pinkey is None:
            raise ValueError('Cannot find suitable physical pin!')

        found_pin = PP.pins[found_pinkey]
        coords = tuple(length2float(found_pin.coords[i]) for i in range(2))


    ### Now show each pin
    for phypinkey, phypin in PP.pins.items():
        coords = tuple(length2float(phypin.coords[i]) for i in range(2))
        dimensions = tuple(length2float(phypin.dimensions[i]) for i in range(2))

        cx, cy = ('{:0.06F}'.format(coords[i]) for i in range(2))
        x, y = ('{:0.06F}'.format(coords[i] - (dimensions[i] / 2)) for i in range(2))
        width, height = ('{:0.06F}'.format(dimensions[i]) for i in range(2))




        if (
            ('radial_line' in features)
            or ('map2circ' in features)
        ):

            line_attrib = {
                'fill'          :   'none',
                'stroke'        :   'black',
                'stroke-width'  :   '0.01',
            }

            g_lines = et.SubElement(pack_area, 'g', attrib=line_attrib)






        if 'radial_line' in features:

            if False:
                dist = math.sqrt(sum((c * c) for c in coords))

                if dist >= 0.001:
                    scale = pack_radius / dist

            else:
                scale = 1


            endcoords = tuple((scale * coords[i]) for i in range(2))
            x2, y2 = ('{:0.06F}'.format(endcoords[i]) for i in range(2))
            et.SubElement(g_lines, 'line', x1='0', y1='0', x2=x2, y2=y2)






        if 'map2circ' in features:
            angle = pinindex2angle(phypinkey, numpins, found_pinkey, found_angle)

            endcoords = tuple(pack_radius * func(angle * math.pi / 180)
                              for func in (math.cos, lambda a : -math.sin(a)))

            x2, y2 = ('{:0.06F}'.format(endcoords[i]) for i in range(2))
            et.SubElement(g_lines, 'line', x1=cx, y1=cy, x2=x2, y2=y2)




    for phypinkey, phypin in PP.pins.items():
        coords = tuple(length2float(phypin.coords[i]) for i in range(2))
        dimensions = tuple(length2float(phypin.dimensions[i]) for i in range(2))

        x, y = ('{:0.06F}'.format(coords[i] - (dimensions[i] / 2)) for i in range(2))
        width, height = ('{:0.06F}'.format(dimensions[i]) for i in range(2))


        et.SubElement(pack_area, 'rect', x=x, y=y, width=width, height=height,
                      fill='white')

        pinlabel_area = trans_zoom(pack_area, x=coords[0], y=coords[1], zoom=(1.0 / zoom))
        pinlabel = et.SubElement(pinlabel_area, 'text', x='0', y='0', style=styles['pad'])
        pinlabel.text = str(phypinkey)





    ### ElementTree 1.2 doesn't write the SVG file header errata, so do that manually
    print('<?xml version="1.0" standalone="no"?>', file=file)
    print('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"', file=file)
    print('\t"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">', file=file)
    for L in et.tostringlist(doc):
        print(str(L, 'utf-8'), file=file)




if __name__ == '__main__':
    def usage():
        print('Usage:', sys.argv[0], 'phypack_fn zoom out_fn')
        sys.exit(1)

    args = sys.argv[1:]

    if len(args) < 3:
        usage()

    phypack_fn, zoom, out_fn = args[:3]
    features = set(args[3:])

    zoom = int(zoom, 0)

    with open(out_fn, 'w') as F:
        phypack_svg(phypack_fn, zoom=zoom, file=F, features=features)



