#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys
import io

import wallpy.random
import wallpy.misc

import wallpy.cleanobject





class Test_CleanObject(unittest.TestCase):

    class CleanObjectChildA(wallpy.cleanobject.CleanObject):
        posarg_names = [ 'p1', 'p2', 'p3' ]
        kwarg_names = [ 'kw1', 'kw2', 'kw3' ]

        def __init__(self, p1, p2, p3, kw1 = None, kw2 = None, kw3 = None):
            args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
            super().__init__(*args, **kwargs)

    class CleanObjectChildB(wallpy.cleanobject.CleanObject):
        posarg_names = [ 'p1', 'p2', 'p3' ]
        kwarg_names = [ 'kw1', 'kw2', 'kw3' ]

        def __init__(self, p1, p2, p3, kw1 = None, kw2 = None, kw3 = None):
            args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
            super().__init__(*args, **kwargs)

    class CleanObjectChildC(wallpy.cleanobject.CleanObject):
        posarg_names = [ 'p1', 'p2', 'p3' ]
        kwarg_names = [ 'kw1', 'kw2', 'kw3' ]

        def __init__(self, p1, p2, p3, kw1 = None, kw2 = None, kw3 = None):
            args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
            super().__init__(*args, **kwargs)


    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(0, 1000)

        posarg_names = type(self).CleanObjectChildA.posarg_names
        kwarg_names = type(self).CleanObjectChildA.kwarg_names

        # Positional argument values
        self.A_posarg_vals = list()
        for k in posarg_names:
            self.A_posarg_vals.append(next(self.RD))

        # Keyword argument values
        self.A_kwarg_vals = dict()
        for k in kwarg_names:
            self.A_kwarg_vals[k] = next(self.RD)


        self.A = type(self).CleanObjectChildA(*self.A_posarg_vals,
                                              **self.A_kwarg_vals)


    def test_init(self):
        for i, k in enumerate(type(self.A).posarg_names):
            self.assertEqual(getattr(self.A, k), self.A_posarg_vals[i],
                             'posattr mismatch')

        for k, v in self.A_kwarg_vals.items():
            self.assertEqual(getattr(self.A, k), v, 'kwattr mismatch')


    def test_eq_sameclass(self):

        A2 = type(self).CleanObjectChildA(*self.A_posarg_vals,
                                          **self.A_kwarg_vals)

        self.assertEqual(self.A, A2, 'equality error')

        posarg_names = type(self).CleanObjectChildA.posarg_names
        kwarg_names = type(self).CleanObjectChildA.kwarg_names

        # Positional argument values
        A4_posarg_vals = list()
        for k in posarg_names:
            A4_posarg_vals.append(next(self.RD))

        # Keyword argument values
        A4_kwarg_vals = dict()
        for k in kwarg_names:
            A4_kwarg_vals[k] = next(self.RD)


        A4 = type(self).CleanObjectChildA(*A4_posarg_vals,
                                               **A4_kwarg_vals)

        A5 = type(self).CleanObjectChildA(*A4_posarg_vals,
                                               **A4_kwarg_vals)

        self.assertEqual(A4, A5, 'equality error')


        self.assertNotEqual(self.A, A4, 'inequality error')
        self.assertNotEqual(self.A, A5, 'inequality error')
        self.assertNotEqual(A2, A5, 'inequality error')
        self.assertNotEqual(A2, A4, 'inequality error')



    def test_eq_otherclass(self):
        B = type(self).CleanObjectChildB(*self.A_posarg_vals,
                                         **self.A_kwarg_vals)

        self.assertEqual(self.A, B)

        B.p1 += 1

        self.assertNotEqual(self.A, B)



        C_posarg_vals = list(self.A_posarg_vals)
        C_kwarg_vals = dict(self.A_kwarg_vals)
        C_posarg_vals[0] += 1

        C = type(self).CleanObjectChildC(*C_posarg_vals, **C_kwarg_vals)

        self.assertNotEqual(self.A, C)
        self.assertEqual(B, C)


        C.p1 -= 1

        self.assertEqual(self.A, C)






class Test_CleanObject_Randomattr(unittest.TestCase):

    # Holds positional and keyword argument names
    class SimpleArgnameHolder(wallpy.cleanobject.CleanObject):
        posarg_names = []
        kwarg_names = []


    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(0, 1000)
        num_posargs = 4
        num_kwargs = 4

        posarg_names = [ 'pos_{:03}'.format(next(self.RD)) for i in range(num_posargs) ]
        kwarg_names = [ 'kw_{:03}'.format(next(self.RD)) for i in range(num_kwargs) ]

        type(self).SimpleArgnameHolder.posarg_names = posarg_names
        type(self).SimpleArgnameHolder.kwarg_names = kwarg_names


        ### Create values

        # Positional argument values
        posarg_vals = list()
        for k in posarg_names:
            posarg_vals.append(next(self.RD))

        # Keyword argument values
        kwarg_vals = dict()
        for k in kwarg_names:
            kwarg_vals[k] = next(self.RD)



        ### Create object for use in tests
        self.A = type(self).SimpleArgnameHolder(*posarg_vals, **kwarg_vals)



    def test_fields_k(self):

        expect = list()
        expect.extend(type(self.A).posarg_names)
        expect.extend(type(self.A).kwarg_names)


        result = list(self.A.fields_k())

        self.assertListEqual(expect, result, 'fields_k mismatch')




    def test_fields_v(self):

        expect = list()

        for k in type(self.A).posarg_names:
            expect.append(getattr(self.A, k))

        for k in type(self.A).kwarg_names:
            expect.append(getattr(self.A, k))


        result = list(self.A.fields_v())

        self.assertListEqual(expect, result, 'fields_v mismatch')



    def test_fields_kv(self):

        expect = list()

        for k in type(self.A).posarg_names:
            expect.append((k, getattr(self.A, k)))

        for k in type(self.A).kwarg_names:
            expect.append((k, getattr(self.A, k)))


        result = list(self.A.fields_kv())

        self.assertListEqual(expect, result, 'fields_kv mismatch')



    def test_fields_kv_filter(self):


        # Filter by key: starts with pos_
        expect = list()

        for k in type(self.A).posarg_names:
            expect.append((k, getattr(self.A, k)))


        def filt_pos(k, v):
            return k.startswith('pos_')


        result = list(self.A.fields_kv(filt = filt_pos))

        self.assertListEqual(expect, result, 'filtered (pos) fields_kv mismatch')



        # Filter by key: starts with kw_
        expect = list()

        for k in type(self.A).kwarg_names:
            expect.append((k, getattr(self.A, k)))


        def filt_kw(k, v):
            return k.startswith('kw_')


        result = list(self.A.fields_kv(filt = filt_kw))

        self.assertListEqual(expect, result, 'filtered (kw) fields_kv mismatch')



        # Filter by value: even values
        expect = list()

        for k in type(self.A).posarg_names:
            v = getattr(self.A, k)

            if (v & 1) == 0:
                expect.append((k, v))

        for k in type(self.A).kwarg_names:
            v = getattr(self.A, k)

            if (v & 1) == 0:
                expect.append((k, v))


        def filt_even(k, v):
            return (v & 1) == 0


        result = list(self.A.fields_kv(filt = filt_even))

        self.assertListEqual(expect, result, 'filtered (even) fields_kv mismatch')



    def test_prettyprint(self):
        # Make SimpleArgnameHolder available without prefix
        exec_globals = {
            'SimpleArgnameHolder' : type(self.A),
        }

        exec_locals = dict()

        with io.StringIO() as buf:
            self.A.prettyprint(file = buf)

            resultstring = buf.getvalue()

            exec('result = ' + resultstring, exec_globals, exec_locals)

        result = exec_locals['result']

        self.assertEqual(self.A, result)



    def test_repr(self):
        # Make SimpleArgnameHolder available without prefix
        exec_globals = {
            'SimpleArgnameHolder' : type(self.A),
        }

        exec_locals = dict()


        resultstring = repr(self.A)

        exec('result = ' + resultstring, exec_globals, exec_locals)

        result = exec_locals['result']

        self.assertEqual(self.A, result)




    def test_valid_p(self):

        class JustValid(object):
            def __init__(self, is_valid = True):
                self.is_valid = is_valid

            def valid_p(self, errcallback = None):
                if self.is_valid:
                    return True

                if errcallback:
                    errcallback('Set to be INVALID')

                return False


        k = type(self.A).posarg_names[0]


        self.assertEqual(self.A.valid_p(), True)


        setattr(self.A, k, JustValid(True))

        self.assertEqual(self.A.valid_p(), True)


        setattr(self.A, k, JustValid(False))

        self.assertEqual(self.A.valid_p(), False)



class Test_CleanObject_DFS(unittest.TestCase):
    
    class Node(wallpy.cleanobject.CleanObject):
        posarg_names = [ 'val' ]
        kwarg_names = [ 'children' ]

        def __init__(self, val, children = None):

            if children:
                children = list(children)
            else:
                children = list()

            args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
            super().__init__(*args, **kwargs)


    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)

        num_nodes = 10

        self.nodes = [type(self).Node(i) for i in range(num_nodes)]

        self.edges = [
            (0, 1),
            (0, 2),
            (0, 3),
            (1, 4),
            (2, 5),
            (2, 6),
            (3, 6),
            (3, 7),
            (4, 8),
            (5, 9),
            (8, 2),
            (8, 9),
        ]

        self.expect_order = (9, 8, 4, 1, 5, 6, 2, 7, 3, 0)


        # Construct graph
        for b, e in self.edges:
            self.nodes[b].children.append(self.nodes[e])


    def test_traverse(self):

        def childfunc(node):
            return node.children



        def callback(node, resultlist):
            resultlist.append(self.nodes.index(node))


        expect = list(self.expect_order)

        result = list()

        self.nodes[0].dfs(
            childfunc = childfunc,
            callback = callback,
            callback_posargs = [result]
        )


        self.assertListEqual(expect, result, 'incorrect DFS order')



    def test_search(self):
        def childfunc(node):
            return node.children


        def callback(node):
            if node.val == 5:
                raise StopIteration


        expect = self.nodes[5]

        result = self.nodes[0].dfs(
            childfunc = childfunc,
            callback = callback
        )


        self.assertEqual(expect, result, 'incorrect DFS returned node')



class Test_CleanObject_extrafields(unittest.TestCase):

    class ExtraField(wallpy.cleanobject.CleanObject):
        posarg_names = [ 'val1' ]
        kwarg_names = [ 'val2' ]

        extra_names = [ 'extra1', 'extra2' ]

        moduleprefix = 'type(self).'

        def __init__(self, val1, val2 = None):
            args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
            super().__init__(*args, **kwargs)


    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_init(self):
        val1, val2, extra = (next(self.RD) for i in range(3))


        with self.assertRaises(TypeError):
            badUUT = type(self).ExtraField(val1, val2 = val2, extra1 = extra)


        with self.assertRaises(TypeError):
            badUUT = type(self).ExtraField(val1, val2 = val2, extra2 = extra)



        UUT = type(self).ExtraField(val1, val2 = val2)

        with io.StringIO() as buf:
            UUT.prettyprint(file = buf)

            bufstr = buf.getvalue()


        newUUT = eval(bufstr)

        self.assertEqual(UUT, newUUT, 'evaluated prettyprint mismatch')


        self.assertEqual(hasattr(UUT, 'extra1'), True, 'missing attr extra1')
        self.assertEqual(hasattr(UUT, 'extra2'), True, 'missing attr extra2')



        UUT = type(self).ExtraField(val1, val2 = val2)
        UUT.extra1 = extra

        with io.StringIO() as buf:
            UUT.prettyprint(file = buf)

            bufstr = buf.getvalue()


        newUUT = eval(bufstr)

        self.assertNotEqual(UUT, newUUT, 'evaluated prettyprint should not match')


        newUUT.extra1 = extra

        self.assertEqual(UUT, newUUT, 'UUT mismatch')




        UUT = type(self).ExtraField(val1, val2 = val2)
        UUT.extra2 = extra

        with io.StringIO() as buf:
            UUT.prettyprint(file = buf)

            bufstr = buf.getvalue()


        newUUT = eval(bufstr)

        self.assertNotEqual(UUT, newUUT, 'evaluated prettyprint should not match')


        newUUT.extra2 = extra

        self.assertEqual(UUT, newUUT, 'UUT mismatch')




    def test_valid(self):

        class JustValid(object):
            def __init__(self, is_valid = True):
                self.is_valid = is_valid

            def valid_p(self, errcallback = None):
                if self.is_valid:
                    return True

                if errcallback:
                    errcallback('Set to be INVALID')

                return False



        UUT = type(self).ExtraField(JustValid(False))
        self.assertEqual(UUT.valid_p(), False, 'invalid viewed as valid')


        UUT.val1.is_valid = True
        self.assertEqual(UUT.valid_p(), True, 'valid viewed as invalid')


        UUT.extra1 = JustValid(False)
        self.assertEqual(UUT.valid_p(), False, 'invalid viewed as valid')


        UUT.extra1.is_valid = True
        self.assertEqual(UUT.valid_p(), True, 'valid viewed as invalid')


        UUT.extra2 = JustValid(False)
        self.assertEqual(UUT.valid_p(), False, 'invalid viewed as valid')


        UUT.extra2.is_valid = True
        self.assertEqual(UUT.valid_p(), True, 'valid viewed as invalid')




    def test_dfs(self):

        class JustVals(wallpy.cleanobject.CleanObject):
            posarg_names = [ 'val1' ]
            kwarg_names = [ 'val2' ]

            moduleprefix = 'type(self).'

            def __init__(self, val1, val2 = None):
                args, kwargs = wallpy.misc.args2pos_kw(type(self), locals())
                super().__init__(*args, **kwargs)



        boringval = next(self.RD)
        wantval = next(self.RD)


        obj_match = type(self).ExtraField(boringval)

        searchtree_leaf = JustVals(
            type(self).ExtraField(
                JustVals(obj_match)
            )
        )


        searchtree_branch = JustVals(
            type(self).ExtraField(
                type(self).ExtraField(boringval)
            )
        )

        searchtree_branch.val1.val1.extra1 = obj_match


        def callback(node, want):
            possible_keys = [ 'val1', 'val2', 'extra1', 'extra2' ]

            for k in possible_keys:
                if hasattr(node, k):
                    if getattr(node, k) == want:
                        raise StopIteration




        # Test matching node as a leaf along posargs
        result = searchtree_leaf.dfs(callback = callback, callback_posargs = [wantval])
        self.assertEqual(result, None, 'search was supposed to return nothing')

        for k in ('val1', 'val2', 'extra1', 'extra2'):
            setattr(obj_match, k, wantval)
            result = searchtree_leaf.dfs(callback = callback, callback_posargs = [wantval])
            self.assertEqual(result, obj_match, 'search was supposed to return obj_match')


            setattr(obj_match, k, None)
            result = searchtree_leaf.dfs(callback = callback, callback_posargs = [wantval])
            self.assertEqual(result, None, 'search was supposed to return nothing')





        # Test matching node as a leaf along extra fields
        result = searchtree_branch.dfs(callback = callback, callback_posargs = [wantval])
        self.assertEqual(result, None, 'search was supposed to return nothing')

        for k in ('val1', 'val2', 'extra1', 'extra2'):
            setattr(obj_match, k, wantval)
            result = searchtree_branch.dfs(callback = callback, callback_posargs = [wantval])
            self.assertEqual(result, obj_match, 'search was supposed to return obj_match')


            setattr(obj_match, k, None)
            result = searchtree_branch.dfs(callback = callback, callback_posargs = [wantval])
            self.assertEqual(result, None, 'search was supposed to return nothing')





class Test_CleanList(unittest.TestCase):

    def setUp(self):
        self.L = wallpy.cleanobject.CleanList(range(4))

        self.L.append(wallpy.cleanobject.CleanObject())

        self.L.extend(range(88, 94))



    def test_elements(self):
        expect = [
            0, 1, 2, 3,
            wallpy.cleanobject.CleanObject(),
            88, 89, 90, 91, 92, 93,
        ]


        self.assertListEqual(expect, list(self.L), 'elements differ')
        self.assertEqual(expect, list(self.L), 'lists differ')


        not_expect = list(expect)
        not_expect.append('fish')

        self.assertNotEqual(not_expect, list(self.L),
                            'lists do not differ')

        self.assertEqual(self.L, wallpy.cleanobject.CleanList(list(self.L)),
                         'equivalent CleanList does not match')


    def test_prettyprint(self):
        expect = """wallpy.cleanobject.CleanList([
  0,
  1,
  2,
  3,
  CleanObject(
  )
  ,
  88,
  89,
  90,
  91,
  92,
  93,
])
"""

        with io.StringIO() as buf:
            self.L.prettyprint(file = buf)

            actual = buf.getvalue()


        self.assertEqual(expect, actual, 'prettyprint differs from expected')



    def test_valid_p(self):
        
        class JustValid(object):
            def __init__(self, is_valid = True):
                self.is_valid = is_valid

            def valid_p(self, errcallback = None):
                if self.is_valid:
                    return True

                if errcallback:
                    errcallback('Set to be INVALID')

                return False


        L_valid = wallpy.cleanobject.CleanList([
            0,
            2,
            5,
            JustValid(True),
            False,
            JustValid(True),
            JustValid(True),
            -1,
        ])

        L_valid2 = wallpy.cleanobject.CleanList([
            wallpy.cleanobject.CleanList(),
            L_valid,
            99,
            -1,
            False,
            JustValid(True),
            L_valid,
        ])

        L_invalid = wallpy.cleanobject.CleanList([
            0,
            2,
            5,
            JustValid(True),
            JustValid(False),
            False,
            JustValid(True),
            JustValid(True),
            -1,
        ])

        L_invalid2 = wallpy.cleanobject.CleanList([
            wallpy.cleanobject.CleanList(),
            L_valid,
            99,
            -1,
            False,
            L_invalid,
            JustValid(True),
            L_valid,
        ])


        self.assertEqual(L_valid.valid_p(), True)
        self.assertEqual(L_valid2.valid_p(), True)

        self.assertEqual(L_invalid.valid_p(), False)
        self.assertEqual(L_invalid2.valid_p(), False)




class Test_CleanDict(unittest.TestCase):

    def setUp(self):
        D = wallpy.cleanobject.CleanDict({
            1 : 2,
            3 : 4,
            5 : 6,
            7 : False,
            None : 99,
            'abc' : 'def',
        })

        D2 = wallpy.cleanobject.CleanDict(D)
        D2[1] = -1
        D2[3] = -3
        D2[5] = -5
        D2[9] = None
        D2['abc'] += '_ghi'

        D[1234] = D2

        self.D = D


    def test_valid_p(self):

        class JustValidHashIsID(object):
            def __init__(self, is_valid = True):
                self.is_valid = is_valid

            def valid_p(self, errcallback = None):
                if self.is_valid:
                    return True

                if errcallback:
                    errcallback('Set to be INVALID')

                return False

            def __hash__(self):
                return id(self)

            def __repr__(self):
                return ''.join([
                    type(self).__name__,
                    '(',
                    repr(self.is_valid),
                    ')',
                ])


        D_valid = wallpy.cleanobject.CleanDict({
            1 : 2,
            3 : 4,
            5 : JustValidHashIsID(True),
            JustValidHashIsID(True) : 1,
            JustValidHashIsID(True) : 2,
            JustValidHashIsID(True) : 3,
            JustValidHashIsID(True) : JustValidHashIsID(True),
        })

        D_valid2 = wallpy.cleanobject.CleanDict({
            1 : 2,
            3 : 4,
            5 : JustValidHashIsID(True),
            JustValidHashIsID(True) : 1,
            JustValidHashIsID(True) : D_valid,
            JustValidHashIsID(True) : 3,
            JustValidHashIsID(True) : JustValidHashIsID(True),
        })

        D_invalid = wallpy.cleanobject.CleanDict({
            1 : 2,
            3 : 4,
            5 : JustValidHashIsID(True),
            JustValidHashIsID(True) : 1,
            JustValidHashIsID(False) : 2,
            JustValidHashIsID(True) : 3,
            JustValidHashIsID(True) : JustValidHashIsID(True),
        })

        D_invalid2 = wallpy.cleanobject.CleanDict({
            1 : 2,
            3 : 4,
            5 : JustValidHashIsID(True),
            JustValidHashIsID(True) : 1,
            JustValidHashIsID(True) : 2,
            JustValidHashIsID(True) : D_invalid,
            JustValidHashIsID(True) : JustValidHashIsID(True),
        })


        self.assertEqual(D_valid.valid_p(), True)
        self.assertEqual(D_valid2.valid_p(), True)

        self.assertEqual(D_invalid.valid_p(), False)
        self.assertEqual(D_invalid2.valid_p(), False)




    def test_prettyprint(self):
        expect = """wallpy.cleanobject.CleanDict({
  None : 99,
  1 : 2,
  1234 : 
    wallpy.cleanobject.CleanDict({
      None : 99,
      1 : -1,
      3 : -3,
      5 : -5,
      7 : False,
      9 : None,
      'abc' : 'def_ghi',
    })
  ,
  3 : 4,
  5 : 6,
  7 : False,
  'abc' : 'def',
})
"""

        with io.StringIO() as buf:
            self.D.prettyprint(file = buf)

            actual = buf.getvalue()


        self.assertEqual(sorted(expect), sorted(actual))



    def test_elements(self):
        expect = {
            1 : 2,
            3 : 4,
            5 : 6,
            7 : False,
            None : 99,
            'abc' : 'def',
            1234 : wallpy.cleanobject.CleanDict({
                1 : -1,
                3 : -3,
                5 : -5,
                7 : False,
                9 : None,
                None : 99,
                'abc' : 'def_ghi',
            }),
        }


        actual = dict(self.D)

        self.assertDictEqual(expect, actual, 'elements differ')

        self.assertEqual(self.D, wallpy.cleanobject.CleanDict(dict(self.D)),
                         'equivalent CleanDict does not match')



    def test_keep_only_matches(self):
        RD = wallpy.random.RandomDistinct(1000, 10000)

        def evenkeys(k, v):
            return (k % 2) == 0

        tD = wallpy.cleanobject.CleanDict({i : next(RD) for i in range(20)})

        expect = {k : tD[k] for k in range(0, 20, 2)}

        tD.keep_only_matches(evenkeys)

        actual = dict(tD)

        self.assertDictEqual(expect, actual, 'keep only evens: mismatch')



    def test_remove_matches(self):
        RD = wallpy.random.RandomDistinct(1000, 10000)

        def evenkeys(k, v):
            return (k % 2) == 0

        tD = wallpy.cleanobject.CleanDict({i : next(RD) for i in range(20)})

        expect = {k : tD[k] for k in range(1, 20, 2)}

        tD.remove_matches(evenkeys)

        actual = dict(tD)

        self.assertDictEqual(expect, actual, 'remove evens: mismatch')





__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_CleanObject),
    __TL.loadTestsFromTestCase(Test_CleanObject_Randomattr),
    __TL.loadTestsFromTestCase(Test_CleanObject_DFS),
    __TL.loadTestsFromTestCase(Test_CleanObject_extrafields),
    __TL.loadTestsFromTestCase(Test_CleanList),
    __TL.loadTestsFromTestCase(Test_CleanDict),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
