#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys
import io

import wallpy.random

import wallpy.misc





class Test_HelperFunctions(unittest.TestCase):
    debug = False

    def setUp(self):
        pass



    def test_args2pos_kw(self):
        RD = wallpy.random.RandomDistinct(0, 1000)
        num_posargs = 4
        num_kwargs = 4
        num_dummyvals = 8

        posarg_names = [ 'pos_{:03}'.format(next(RD)) for i in range(num_posargs) ]
        kwarg_names = [ 'kw_{:03}'.format(next(RD)) for i in range(num_kwargs) ]

        # Holds positional and keyword argument names
        class SimpleArgnameHolder(object):
            posarg_names = []
            kwarg_names = []


        C = SimpleArgnameHolder()
        type(C).posarg_names = posarg_names
        type(C).kwarg_names = kwarg_names


        ### Create values
        localvars = dict()

        # Positional argument values
        posarg_vals = list()
        for k in posarg_names:
            posarg_vals.append(next(RD))

        # Keyword argument values
        kwarg_vals = dict()
        for k in kwarg_names:
            kwarg_vals[k] = next(RD)

        # Other, unused values
        dummy_vals = dict()
        for i in range(num_dummyvals):
            k = 'dum_{:03}'.format(next(RD))
            dummy_vals[k] = next(RD)


        # Compose localvars
        localvars = dict()
        for k, v in zip(posarg_names, posarg_vals):
            localvars[k] = v

        for k in kwarg_names:
            localvars[k] = kwarg_vals[k]

        for k, v in dummy_vals.items():
            localvars[k] = v


        ### Now perform the test!
        result_posargs, result_kwargs = wallpy.misc.args2pos_kw(type(C), localvars)


        if self.debug:
            print()
            print()
            print('PA', ', '.join(
                '{0[0]}: {0[1]}'.format(x) for x in zip(posarg_names, posarg_vals)
            ))
            print('KW', ', '.join(
                '{0[0]}: {0[1]}'.format(x) for x in sorted(kwarg_vals.items())
            ))

            print('LV:')
            for k, v in sorted(localvars.items()):
                print('', repr(k), repr(v), sep='\t')


            print()
            print('Results:')
            print('PA', ', '.join(
                '{0[0]}: {0[1]}'.format(x) for x in zip(posarg_names, result_posargs)
            ))
            print('KW', ', '.join(
                '{0[0]}: {0[1]}'.format(x) for x in sorted(result_kwargs.items())
            ))

        self.assertListEqual(result_posargs, posarg_vals, 'incorrect positional args')
        self.assertDictEqual(result_kwargs, kwarg_vals, 'incorrect keyword args')




class Test_CachedMap(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_basic(self):
        CM = wallpy.misc.CachedMap(lambda x : x)

        keys = [next(self.RD) for i in range(100)]
        expect = list(keys)

        actual = [CM[k] for k in keys]

        self.assertListEqual(expect, actual, 'basic values differ')



        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'basic values differ on repeat #' + str(i))


    def test_posargs(self):
        argval = next(self.RD)

        CM = wallpy.misc.CachedMap(lambda x, arg0 : arg0, [argval])

        keys = [next(self.RD) for i in range(100)]
        expect = [argval for k in keys]

        actual = [CM[k] for k in keys]

        self.assertListEqual(expect, actual, 'posarg values differ')



        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'posarg values differ on repeat #' + str(i))




    def test_kwargs(self):
        argval = next(self.RD)

        def ret_kwarg(key, arg = None):
            return arg

        CM = wallpy.misc.CachedMap(ret_kwarg, func_kwargs = {'arg' : argval})

        keys = [next(self.RD) for i in range(100)]
        expect = [argval for k in keys]

        actual = [CM[k] for k in keys]

        self.assertListEqual(expect, actual, 'kwarg values differ')



        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'kwarg values differ on repeat #' + str(i))



    def test_reset_func(self):
        CM = wallpy.misc.CachedMap(lambda x : x)

        keys = [next(self.RD) for i in range(100)]



        expect = list(keys)

        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'orig func: values differ on repeat #' + str(i))


        CM.func = lambda x : x * 2

        expect = [k * 2 for k in keys]

        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'func 2: values differ on repeat #' + str(i))



        CM.func = lambda x : (x * 4) + 1

        expect = [((k * 4) + 1) for k in keys]

        for i in range(10):
            actual = [CM[k] for k in keys]

            self.assertListEqual(expect, actual,
                                 'func 3: values differ on repeat #' + str(i))



class Test_ArcList(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)

        self.vals = tuple(next(self.RD) for i in range(16))

        self.UUT = wallpy.misc.ArcList(self.vals, 0, len(self.vals))


    def test_init(self):

        actual = tuple(self.UUT)
        self.assertTupleEqual(self.vals, actual)


        for o in range(len(self.vals)):
            L = wallpy.misc.ArcList(self.vals, o, len(self.vals)-o-1)

            expect = self.vals[o:-1]
            actual = tuple(L)
            self.assertTupleEqual(expect, actual)



    def test_modify_subset(self):

        numvals = len(self.vals)

        for i in range(numvals):

            self.UUT.length = i
            self.assertEqual(i, self.UUT.length)

            expect = self.vals[:i]
            actual = tuple(self.UUT)
            self.assertTupleEqual(expect, actual)

            self.assertTupleEqual(self.vals, self.UUT.elements)



        self.UUT.length = 1

        for i in range(numvals):

            self.UUT.offset = i
            self.assertEqual(i, self.UUT.offset)

            expect = self.vals[i:i+1]
            actual = tuple(self.UUT)
            self.assertTupleEqual(expect, actual)

            self.assertTupleEqual(self.vals, self.UUT.elements)



    def test_new_elementlist(self):

        self.assertTupleEqual(self.vals, tuple(self.UUT))
        self.assertTupleEqual(self.vals, self.UUT.elements)

        for i in range(10):
            self.UUT.offset = 1
            self.UUT.length = 11

            newvals = tuple(next(self.RD) for i in range(10))
            self.UUT.elements = newvals

            self.assertEqual(0, self.UUT.offset)
            self.assertEqual(len(newvals), self.UUT.length)

            self.assertTupleEqual(newvals, tuple(self.UUT))
            self.assertTupleEqual(newvals, self.UUT.elements)



    def test_rotate(self):

        kwargs = {
            'can_rotate' :  True,
        }

        # ((iterable, start, len), expectval)
        args_expect = (
            (('abcdef', 0, 6), 'abcdef'),
            (('abcdef', 1, 6), 'bcdefa'),
            (('abcdef', 2, 6), 'cdefab'),
            (('abcdef', 3, 6), 'defabc'),
            (('abcdef', 4, 6), 'efabcd'),
            (('abcdef', 5, 6), 'fabcde'),
            (('abcdef', 6, 6), 'abcdef'),
            (('abcdef', 7, 6), 'bcdefa'),

            (('abcdef', 0, 3), 'abc'),
            (('abcdef', 1, 3), 'bcd'),
            (('abcdef', 2, 3), 'cde'),
            (('abcdef', 3, 3), 'def'),
            (('abcdef', 4, 3), 'efa'),
            (('abcdef', 5, 3), 'fab'),
            (('abcdef', 6, 3), 'abc'),
            (('abcdef', 7, 3), 'bcd'),

            (('abcdef', -1, 3), 'fab'),
            (('abcdef', -2, 3), 'efa'),
            (('abcdef', -3, 3), 'def'),
            (('abcdef', -4, 3), 'cde'),
            (('abcdef', -5, 3), 'bcd'),
            (('abcdef', -6, 3), 'abc'),
            (('abcdef', -7, 3), 'fab'),

            (('abcdef', 0, 99), 'abcdef'),
            (('abcdef', 1, 99), 'bcdefa'),
            (('abcdef', 2, 99), 'cdefab'),
            (('abcdef', 3, 99), 'defabc'),
            (('abcdef', 4, 99), 'efabcd'),
            (('abcdef', 5, 99), 'fabcde'),
            (('abcdef', 6, 99), 'abcdef'),
            (('abcdef', 7, 99), 'bcdefa'),
        )

        for args, expect in args_expect:
            UUT = wallpy.misc.ArcList(*args, **kwargs)

            self.assertSequenceEqual(tuple(UUT), expect)





__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_HelperFunctions),
    __TL.loadTestsFromTestCase(Test_CachedMap),
    __TL.loadTestsFromTestCase(Test_ArcList),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
