#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys

import wallpy.random




class Test_RandomDistinct(unittest.TestCase):
    debug = False

    def setUp(self):
        self.start = 0
        self.stop = 160
        self.step = 2
        self.maxtries = (abs(self.start - self.stop) // self.step) * 256

        self.RD = wallpy.random.RandomDistinct(
            self.start,
            self.stop,
            self.step,
            self.maxtries
        )


    def test_minargs(self):
        RD = wallpy.random.RandomDistinct(self.start, self.stop)


    def test_valid(self):
        valid_set = set(range(self.start, self.stop, self.step))

        for x in self.RD:
            self.assertIn(x, valid_set, 'invalid random number')


    def test_unique(self):
        seen_set = set()

        for x in self.RD:
            self.assertNotIn(x, seen_set, 'duplicate number found')

            seen_set.add(x)


    def test_len(self):
        actual_len = len(self.RD)
        expected_len = len(range(self.start, self.stop, self.step))

        self.assertEqual(actual_len, expected_len, 'maximum length mismatch')


    def test_range(self):
        # Make sure every number in the available range is returned eventually.

        remaining_set = set(range(self.start, self.stop, self.step))

        for x in self.RD:
            remaining_set.remove(x)

        self.assertSetEqual(remaining_set, set(), 'not every possible number was generated')



    def test_shortened_len(self):
        # Evaluate length for a more realistic setup

        range_size = 1000
        num_iter = 128
        maxtries = 16


        if self.debug:
            print()
            print('Testing with range size:', range_size)
            print('Number of iterations:', num_iter)
            print('Max tries for unique value:', maxtries)

        histogram = dict()

        for i in range(1, num_iter+1):
            R = wallpy.random.RandomDistinct(0, range_size, 1, maxtries = maxtries)
            
            L = len(R)

            if L in histogram:
                histogram[L] += 1
            else:
                histogram[L] = 1



        if self.debug:
            print('Results:')

            for k, v in sorted(histogram.items()):
                print(k, v)


        for k in histogram:
            self.assertGreaterEqual(k, range_size // 4, 'length shorter than expected')



    def test_rewind(self):
        start = 0
        stop = 8
        
        R = wallpy.random.RandomDistinct(start, stop)

        first_val = next(R)

        # No repeats before rewind
        for i in R:
            self.assertNotEqual(i, first_val, 'unexpected repeat')


        R.rewind()

        found_first_val = False
        for i in R:
            if i == first_val:
                found_first_val = True
                break


        self.assertTrue(found_first_val, 'did not repeat')



__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_RandomDistinct),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
