#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys
import random
import io
from xml.etree import ElementTree as et

import wallpy.random
import wallpy.svg



class Test_Canvas(unittest.TestCase):

    def setUp(self):
        self.UUT = wallpy.svg.Canvas((320, 240),
                                     title='Test Doc',
                                     main_unit='mm')


    def test_draw(self):
        doc = self.UUT.doc

        Rback = wallpy.svg.rect(doc, (0, 0), self.UUT.size, fill='azure')

        line_attrib = {
            'fill'          :   'none',
            'stroke'        :   'black',
            'stroke-width'  :   '1.0',
        }
        Grp = et.SubElement(doc, 'g', **line_attrib)

        R1 = wallpy.svg.rect(Grp, (160, 120), (50, 40), fill='lightblue')
        R2 = wallpy.svg.rect(Grp, (10, 10), (9, 9), fill='lightgreen')

        textstyle = {
            'font-family'       :   'Sans',
            'font-size'         :   '6px',
            'text-anchor'       :   'start',
            'dominant-baseline' :   'central',
        }
        T1 = wallpy.svg.text(doc, (163, 123+12), style=textstyle, text='Five')

        with open('/tmp/test_inherit.svg', 'w') as F:
            self.UUT.export(F)



    def test_export(self):
        doc = self.UUT.doc

        with io.StringIO(newline='\n') as F:
            expect = """
                <?xml version="1.0" standalone="no"?>
                <!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
                "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
                <svg
                height="240.000000mm"
                version="1.1"
                viewBox="0.000000 0.000000 320.000000 240.000000"
                width="320.000000mm"
                xmlns="http://www.w3.org/2000/svg"
                >
                <title
                >
                Test Doc
                </title>
                </svg>
                """

            expect_lines = tuple(filter(
                bool,
                (L.strip() for L in expect.split('\n'))
            ))

            self.UUT.export(F)
            actual = F.getvalue()

            actual_lines = tuple(filter(
                bool,
                (L.strip() for L in actual.split('\n'))
            ))

            self.assertEqual(expect_lines, actual_lines)



    def test_SubElement_export(self):
        doc = self.UUT.doc

        SE_outer = wallpy.svg.SubElement(doc, 'g')
        SE_inner = wallpy.svg.SubElement(SE_outer, 'g_empty')

        with io.StringIO(newline='\n') as F:
            expect = """
                <?xml version="1.0" standalone="no"?>
                <!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
                "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
                <svg
                height="240.000000mm"
                version="1.1"
                viewBox="0.000000 0.000000 320.000000 240.000000"
                width="320.000000mm"
                xmlns="http://www.w3.org/2000/svg"
                >
                <title
                >
                Test Doc
                </title>
                <g
                >
                <g_empty
                />
                </g>
                </svg>
                """

            expect_lines = tuple(filter(
                bool,
                (L.strip() for L in expect.split('\n'))
            ))

            self.UUT.export(F)
            actual = F.getvalue()

            actual_lines = tuple(filter(
                bool,
                (L.strip() for L in actual.split('\n'))
            ))

            self.assertEqual(expect_lines, actual_lines)



__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_Canvas),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)

