#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys
import os.path
import fractions

import wallpy.random
import wallpy.units





def check_init_for_type(testcase, T,
                       valid_vals, invalid_vals,
                       valid_units, invalid_units):


        ### Valid combinations of values and units

        for val in valid_vals:
            for unit in valid_units:
                UUT = T(val, unit)




        ### Invalid combinations of values and units

        for val in valid_vals:
            for unit in invalid_units:
                with testcase.assertRaises((ValueError, TypeError)):
                    UUT = T(val, unit)

        for val in invalid_vals:
            for unit in valid_units:
                with testcase.assertRaises((ValueError, TypeError)):
                    UUT = T(val, unit)

        for val in invalid_vals:
            for unit in invalid_units:
                with testcase.assertRaises((ValueError, TypeError)):
                    UUT = T(val, unit)




def check_arith_for_type(testcase, T, unit):


    # Make sure the original Measure's unit no longer is acceptable
    with testcase.assertRaises(ValueError):
        dummy = T(0, '')



    for i in range(8):
        tmpRD = wallpy.random.RandomDistinct(1, 10000)

        main_int = next(tmpRD)
        main_numer = next(tmpRD)
        main_denom = next(tmpRD)
        other_int = next(tmpRD)
        other_numer = next(tmpRD)
        other_denom = next(tmpRD)


        for main in (
            main_int,
            -main_int,
            fractions.Fraction(main_numer, main_denom),
            fractions.Fraction(-main_numer, main_denom),
        ):

            UUT_main = T(main, unit)



            ### Unary operators

            expect = T(-main, unit)
            actual = -UUT_main
            testcase.assertEqual(expect, actual, 'neg mismatch')

            expect = T(+main, unit)
            actual = +UUT_main
            testcase.assertEqual(expect, actual, 'pos mismatch')

            expect = T(abs(main), unit)
            actual = abs(UUT_main)
            testcase.assertEqual(expect, actual, 'abs mismatch')



            ### Binary operators with same units
            for other in (
                other_int,
                -other_int,
                fractions.Fraction(other_numer, other_denom),
                fractions.Fraction(-other_numer, other_denom),
            ):

                UUT_other = T(other, unit)


                expect = T(main + other, unit)
                actual = UUT_main + UUT_other
                testcase.assertEqual(expect, actual, 'add mismatch')
                actual = UUT_other + UUT_main
                testcase.assertEqual(expect, actual, 'add (swap) mismatch')

                expect = T(main - other, unit)
                actual = UUT_main - UUT_other
                testcase.assertEqual(expect, actual, 'add mismatch')

                expect = T(other - main, unit)
                actual = UUT_other - UUT_main
                testcase.assertEqual(expect, actual, 'sub (swap) mismatch')


                expect = main // other
                actual = UUT_main // UUT_other
                testcase.assertEqual(expect, actual, 'floordiv mismatch')

                expect = T(main % other, unit)
                actual = UUT_main % UUT_other
                testcase.assertEqual(expect, actual, 'mod mismatch')

                expect = (
                    main // other,
                    T(main % other, unit)
                )
                actual = divmod(UUT_main, UUT_other)
                testcase.assertEqual(expect, actual, 'divmod mismatch')




                # Make sure reversed divmod gets original value of main
                d, m = divmod(UUT_main, UUT_other)

                expect = UUT_main
                actual = (d * UUT_other) + m
                testcase.assertEqual(expect, actual, 'reversed divmod mismatch')





                UUT_bad = wallpy.units.Measure(other, '')

                with testcase.assertRaises(TypeError):
                    dummy = UUT_other + UUT_bad

                with testcase.assertRaises(TypeError):
                    dummy = UUT_bad + UUT_other

                with testcase.assertRaises(TypeError):
                    dummy = UUT_other - UUT_bad

                with testcase.assertRaises(TypeError):
                    dummy = UUT_bad - UUT_other

                with testcase.assertRaises(TypeError):
                    dummy = UUT_other // UUT_bad

                with testcase.assertRaises(TypeError):
                    dummy = UUT_bad // UUT_other

                with testcase.assertRaises(TypeError):
                    dummy = UUT_other % UUT_bad

                with testcase.assertRaises(TypeError):
                    dummy = UUT_bad % UUT_other

                with testcase.assertRaises(TypeError):
                    dummy = divmod(UUT_other, UUT_bad)

                with testcase.assertRaises(TypeError):
                    dummy = divmod(UUT_bad, UUT_other)




            ### Binary operators with unitless values
            for other in (
                other_int,
                -other_int,
                fractions.Fraction(other_numer, other_denom),
                fractions.Fraction(-other_numer, other_denom),
            ):

                for UUT_other in (
                    fractions.Fraction(other),
                    wallpy.units.Measure(other, ''),
                ):

                    expect = T(main * other, unit)
                    actual = UUT_main * UUT_other
                    testcase.assertEqual(expect, actual, 'mul mismatch')
                    actual = UUT_other * UUT_main
                    testcase.assertEqual(expect, actual, 'rmul mismatch')

                    expect = T(fractions.Fraction(main) / other, unit)
                    actual = UUT_main / UUT_other
                    testcase.assertEqual(expect, actual, 'truediv mismatch')



                    with testcase.assertRaises(TypeError):
                        dummy = UUT_other / UUT_main



                UUT_bad = T(other, unit)

                with testcase.assertRaises(TypeError):
                    dummy = UUT_main * UUT_bad

                with testcase.assertRaises(TypeError):
                    dummy = UUT_bad * UUT_main




def check_comparison_for_type(testcase, T, unit):

    # Test values
    small, large = sorted(next(testcase.RD) for i in range(2))

    while large < (small + 2):
        small, large = sorted(next(testcase.RD) for i in range(2))


    testcase.assertLess(small + 1, large, 'test setup error')


    UUT_small = T(small, unit)
    UUT_small_p1 = T(small+1, unit)
    UUT_large = T(large, unit)
    UUT_large_p1 = T(large+1, unit)


    # Equivalence
    testcase.assertEqual(UUT_small, UUT_small)
    testcase.assertIs(UUT_small, UUT_small)
    testcase.assertEqual(UUT_small, T(small, unit))
    testcase.assertIsNot(UUT_small, T(small, unit))
    testcase.assertNotEqual(UUT_small, UUT_small_p1)
    testcase.assertNotEqual(UUT_small, UUT_small_p1)

    testcase.assertEqual(UUT_large, UUT_large)
    testcase.assertIs(UUT_large, UUT_large)
    testcase.assertEqual(UUT_large, T(large, unit))
    testcase.assertIsNot(UUT_large, T(large, unit))
    testcase.assertNotEqual(UUT_large, UUT_large_p1)
    testcase.assertNotEqual(UUT_large, UUT_large_p1)

    testcase.assertNotEqual(UUT_small, UUT_large)



    # Inequalities with self
    testcase.assertLessEqual(UUT_small, UUT_small)
    testcase.assertGreaterEqual(UUT_small, UUT_small)

    testcase.assertLessEqual(UUT_large, UUT_large)
    testcase.assertGreaterEqual(UUT_large, UUT_large)


    # Inequalities with others
    testcase.assertLess(UUT_small, UUT_small_p1)
    testcase.assertLessEqual(UUT_small, UUT_small_p1)

    testcase.assertGreater(UUT_small_p1, UUT_small)
    testcase.assertGreaterEqual(UUT_small_p1, UUT_small)


    testcase.assertLess(UUT_large, UUT_large_p1)
    testcase.assertLessEqual(UUT_large, UUT_large_p1)

    testcase.assertGreater(UUT_large_p1, UUT_large)
    testcase.assertGreaterEqual(UUT_large_p1, UUT_large)


    testcase.assertLess(UUT_small, UUT_large)
    testcase.assertLessEqual(UUT_small, UUT_large)

    testcase.assertGreater(UUT_large, UUT_small)
    testcase.assertGreaterEqual(UUT_large, UUT_small)




def check_strings(testcase, T, valid_units, invalid_units):

    testvals = (
        ('1'        ,   fractions.Fraction(1, 1)),
        ('1.1'      ,   fractions.Fraction(11, 10)),
        ('0'        ,   fractions.Fraction(0, 1)),
        ('-1'       ,   fractions.Fraction(-1, 1)),
        ('-2.559'   ,   fractions.Fraction(-2559, 1000)),
    )

    for strval, fracval in testvals:

        for unit in valid_units:
            for sep in (' ', '  ', '\t', ' \t\t '):
                s = sep.join((strval, unit))

                expect = T(fracval, unit)
                actual = T(s)

                testcase.assertEqual(expect, actual)


            s = ' '.join((strval, unit))
            expect = T(fracval, unit)
            actual = T(str(expect))
            testcase.assertEqual(expect, actual)



        for unit in invalid_units:
            if unit.isspace():
                continue

            for sep in (' ', '  ', '\t', ' \t\t '):
                s = sep.join((strval, unit))

                with testcase.assertRaises((ValueError, TypeError)):
                    UUT = T(s)




class Test_Measure(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_init(self):

        valid_vals = (
            next(self.RD),
            fractions.Fraction(next(self.RD), 1),
            fractions.Fraction(1, next(self.RD)),
            fractions.Fraction(next(self.RD), next(self.RD)),
        )

        invalid_vals = (
            1+2j,
            3j,
            'a',
            '',
            object(),
        )


        valid_units = (
            '',
        )

        invalid_units = (
            'a',
            'm',
            'mm',
            ' ',
            '1',
            '0',
        )


        check_init_for_type(self, wallpy.units.Measure,
                           valid_vals, invalid_vals,
                           valid_units, invalid_units)


        check_strings(self, wallpy.units.Measure, valid_units, invalid_units)



    def test_arith(self):

        class Measure_OneUnit(wallpy.units.Measure):
            moduleprefix = ''
            native_unit = 'a'
            unit2native_mult = {
                'a' :   1,
            }


        check_arith_for_type(self, Measure_OneUnit, 'a')



    def test_comparison(self):
        check_comparison_for_type(self, wallpy.units.Measure, '')





class Test_Length(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_init(self):

        valid_vals = (
            next(self.RD),
            fractions.Fraction(next(self.RD), 1),
            fractions.Fraction(1, next(self.RD)),
            fractions.Fraction(next(self.RD), next(self.RD)),
        )

        invalid_vals = (
            1+2j,
            3j,
            'a',
            '',
            object(),
        )


        valid_units = (
            'nm',
            'micron',
            'mm',
            'cm',
            'm',

            'thou',
            'mil',
            'in',
            'ft',
            'yd',
        )

        invalid_units = (
            '',
            ' ',
            '1',
            '0',
            'l',
            'L',
            'a',
        )


        check_init_for_type(self, wallpy.units.Length,
                           valid_vals, invalid_vals,
                           valid_units, invalid_units)


        check_strings(self, wallpy.units.Length, valid_units, invalid_units)




    def test_arith(self):
        check_arith_for_type(self, wallpy.units.Length, 'mm')



    def test_comparison(self):
        check_comparison_for_type(self, wallpy.units.Length, 'mm')



        order_l2g = (
            wallpy.units.Length(1, 'nm'),
            wallpy.units.Length(1, 'micron'),
            wallpy.units.Length(1, 'mil'),
            wallpy.units.Length(1, 'mm'),
            wallpy.units.Length(1, 'cm'),
            wallpy.units.Length(fractions.Fraction(253, 100), 'cm'),
            wallpy.units.Length(1, 'in'),
            wallpy.units.Length(fractions.Fraction(255, 100), 'cm'),
            wallpy.units.Length(1, 'ft'),
            wallpy.units.Length(fractions.Fraction(298, 100), 'ft'),
            wallpy.units.Length(91, 'cm'), #   91 cm ~= 2.98557 feet
            wallpy.units.Length(1, 'yd'),  #   1 yd = 91.44 cm
            wallpy.units.Length(92, 'cm'),
            wallpy.units.Length(1, 'm'),
        )

        order_rand = list(order_l2g)
        
        expect = order_l2g
        actual = tuple(sorted(order_rand))
        self.assertTupleEqual(expect, actual)



        self.assertLess(
            wallpy.units.Length(fractions.Fraction(2539, 1000), 'cm'),
            wallpy.units.Length(1, 'in'))

        self.assertLessEqual(
            wallpy.units.Length(fractions.Fraction(2539, 1000), 'cm'),
                             wallpy.units.Length(1, 'in'))


        self.assertEqual(
            wallpy.units.Length(fractions.Fraction(254, 100), 'cm'),
            wallpy.units.Length(1, 'in'))

        self.assertLessEqual(
            wallpy.units.Length(fractions.Fraction(254, 100), 'cm'),
            wallpy.units.Length(1, 'in'))

        self.assertGreaterEqual(
            wallpy.units.Length(fractions.Fraction(254, 100), 'cm'),
            wallpy.units.Length(1, 'in'))





        # Check equivalence between units
        self.assertEqual(
            wallpy.units.Length(1000, 'nm'), wallpy.units.Length(1, 'micron'))

        self.assertEqual(
            wallpy.units.Length(1000, 'micron'), wallpy.units.Length(1, 'mm'))

        self.assertEqual(
            wallpy.units.Length(10, 'mm'), wallpy.units.Length(1, 'cm'))

        self.assertEqual(
            wallpy.units.Length(100, 'cm'), wallpy.units.Length(1, 'm'))

        self.assertEqual(
            wallpy.units.Length(fractions.Fraction(254, 10), 'mm'),
            wallpy.units.Length(1, 'in'))

        self.assertEqual(
            wallpy.units.Length(1000, 'mil'), wallpy.units.Length(1, 'in'))

        self.assertEqual(
            wallpy.units.Length(1000, 'thou'), wallpy.units.Length(1, 'in'))

        self.assertEqual(
            wallpy.units.Length(1, 'mil'), wallpy.units.Length(1, 'thou'))

        self.assertEqual(
            wallpy.units.Length(12, 'in'), wallpy.units.Length(1, 'ft'))

        self.assertEqual(
            wallpy.units.Length(12, 'in'), wallpy.units.Length(1, 'ft'))

        self.assertEqual(
            wallpy.units.Length(3, 'ft'), wallpy.units.Length(1, 'yd'))





class Test_Angle(unittest.TestCase):

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)


    def test_init(self):

        valid_vals = (
            next(self.RD),
            fractions.Fraction(next(self.RD), 1),
            fractions.Fraction(1, next(self.RD)),
            fractions.Fraction(next(self.RD), next(self.RD)),
        )

        invalid_vals = (
            1+2j,
            3j,
            'a',
            '',
            object(),
        )


        valid_units = (
            'deg',
            'pi_rad',
        )

        invalid_units = (
            '',
            ' ',
            '1',
            '0',
            'l',
            'L',
            'a',
            'mm',
            'ft',
            'rad',
        )


        check_init_for_type(self, wallpy.units.Angle,
                           valid_vals, invalid_vals,
                           valid_units, invalid_units)


        check_strings(self, wallpy.units.Angle, valid_units, invalid_units)




    def test_arith(self):
        check_arith_for_type(self, wallpy.units.Angle, 'deg')
        check_arith_for_type(self, wallpy.units.Angle, 'pi_rad')



    def test_comparison(self):
        check_comparison_for_type(self, wallpy.units.Angle, 'deg')
        check_comparison_for_type(self, wallpy.units.Angle, 'pi_rad')



        order_l2g = (
            wallpy.units.Angle(1, 'deg'),
            wallpy.units.Angle(179, 'deg'),
            wallpy.units.Angle(1, 'pi_rad'),
            wallpy.units.Angle(181, 'deg'),
            wallpy.units.Angle(359, 'deg'),
            wallpy.units.Angle(2, 'pi_rad'),
            wallpy.units.Angle(361, 'deg'),
        )

        order_rand = list(order_l2g)
        
        expect = order_l2g
        actual = tuple(sorted(order_rand))
        self.assertTupleEqual(expect, actual)




        # Check equivalence between units
        self.assertEqual(
            wallpy.units.Angle(180, 'deg'), wallpy.units.Angle(1, 'pi_rad'))




__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_Measure),
    __TL.loadTestsFromTestCase(Test_Length),
    __TL.loadTestsFromTestCase(Test_Angle),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)

