#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

import unittest
import sys
import random
import io

import wallpy.random
import wallpy.viz





class Test_viz_grid(unittest.TestCase):
    num_dimensions = 2

    def setUp(self):
        self.RD = wallpy.random.RandomDistinct(1000, 10000)

        self.indices = [
            'ABCDEFGHJKLMNPTUWXYZ',
            'one two three four five six seven eight nine ten'.split(),
        ]

        self.G = wallpy.viz.Grid(self.num_dimensions, indices = self.indices)


    def test_defaults(self):
        tG = wallpy.viz.Grid(self.num_dimensions)


        expect = (tG.default_dimlen,) * self.num_dimensions
        self.assertTupleEqual(expect, tG.dimlens, 'dimlens differ')


        expect = (tuple(range(tG.default_dimlen)),) * self.num_dimensions
        actual = tuple(tuple(i) for i in tG.indices)
        self.assertTupleEqual(expect, actual, 'indices differ')


    def test_indices(self):

        expect = (20, 10)
        self.assertTupleEqual(expect, self.G.dimlens, 'dimlens differ')


        expect = tuple(tuple(i) for i in self.indices)
        actual = self.G.indices
        self.assertTupleEqual(expect, actual, 'indices differ')


    def test_conversions(self):
        test_n2i = {
            (0, 0) : ('A', 'one'),
            (1, 2) : ('B', 'three'),
            (9, 9) : ('K', 'ten'),
            (2, 1) : ('C', 'two'),
            (19, 9) : ('Z', 'ten'),
        }

        
        for k, expect in test_n2i.items():
            actual = self.G.native2indexed(k)

            self.assertTupleEqual(expect, actual, 'tuple mismatch: ' + str(k))


        for expect, k in test_n2i.items():
            actual = self.G.indexed2native(k)

            self.assertTupleEqual(expect, actual, 'tuple mismatch: ' + str(k))


    def test_items(self):
        test_data = {
            ('A', 'one') : next(self.RD),
            ('Z', 'five') : next(self.RD),
            ('D', 'nine') : next(self.RD),
            ('D', 'ten') : next(self.RD),
            ('X', 'four') : next(self.RD),
        }

        # Insert data
        for k, v in test_data.items():
            self.G[k] = v

        expect = dict(test_data)
        actual = { k : v for k, v in self.G.indexed_items()}
        self.assertDictEqual(expect, actual, 'inserted data mismatch')


        # Check data
        for k, v in test_data.items():
            actual = self.G[k]
            self.assertEqual(v, actual, 'getitem mismatch')


        # Insert bad data
        bad_data = {
            ('A', 'banana') : next(self.RD),
            ('beaker', 'one') : next(self.RD),
            ('fish', 'food') : next(self.RD),
        }

        for k, v in bad_data.items():
            with self.assertRaises(ValueError):
                self.G[k] = v


        # Overwrite and insert data
        overwrite = {
            ('A', 'one') : next(self.RD),
            ('C', 'three') : next(self.RD),
        }

        for k, v in overwrite.items():
            self.G[k] = v
            expect[k] = v

        actual = { k : v for k, v in self.G.indexed_items()}
        self.assertDictEqual(expect, actual, 'overwritten data mismatch')


        # Delete data
        delete_keys = (
            ('X', 'four'),
            ('D', 'nine'),
        )

        for k in delete_keys:
            del self.G[k]
            del expect[k]

        actual = { k : v for k, v in self.G.indexed_items()}
        self.assertDictEqual(expect, actual, 'deleted data mismatch')


        # Delete nonexistent data
        nonexist_keys = (
            ('M', 'one'),
            ('K', 'two'),
        )

        for k in nonexist_keys:
            with self.assertRaises(KeyError):
                del self.G[k]


    def test_indexed_keys(self):
        data = dict()

        for i in range(20):
            key = tuple(random.choice(self.indices[n])
                        for n in range(self.num_dimensions))

            val = next(self.RD)
            data[key] = val
            self.G[key] = val


        expect = frozenset(data.keys())
        actual = frozenset(self.G.indexed_keys())
        self.assertSetEqual(expect, actual, 'indexed_keys mismatch')



    def test_import_tabtext(self):

        # Test 2D Grid
        text = """
1	2	3
4	5	6
7	8	9
10	11	12
""".strip()

        data = {
            (0, 0) : 1,
            (1, 0) : 2,
            (2, 0) : 3,
            (0, 1) : 4,
            (1, 1) : 5,
            (2, 1) : 6,
            (0, 2) : 7,
            (1, 2) : 8,
            (2, 2) : 9,
            (0, 3) : 10,
            (1, 3) : 11,
            (2, 3) : 12,
        }


        tG = wallpy.viz.Grid(2)

        tG.import_tabtext(text)

        actual = { k : v for k, v in tG.indexed_items()}
        self.assertDictEqual(data, actual, 'imported 2D tabtext data mismatch')



        # Test 1D Grid
        data = {(i,) : next(self.RD) for i in range(20)}

        text = '\t'.join(str(v) for k, v in sorted(data.items()))

        tG = wallpy.viz.Grid(1)

        tG.import_tabtext(text)

        actual = { k : v for k, v in tG.indexed_items()}
        self.assertDictEqual(data, actual, 'imported 1D tabtext data mismatch')



        # Test invalid dimensions
        for dim in (3, 4, 5, 99):
            tG = wallpy.viz.Grid(dim)

            with self.assertRaises(ValueError):
                tG.import_tabtext('None')



    def test_export_tabtext(self):
        size_x = 8
        size_y = 24
        prob_val = 10   # In percent


        ### Test 2D Grid
        tG = wallpy.viz.Grid(2)

        # Construct tabtext for grid
        def get_val(x, y):
            if (x == (size_x - 1)) and (y == (size_y - 1)):
                # Force grid to be given size
                return next(self.RD)
            elif random.randrange(100) < prob_val:
                return next(self.RD)
            return None

        with io.StringIO() as buf_in:
            for y in range(size_y):
                vals = tuple(repr(get_val(x, y)) for x in range(size_x))
                print(*vals, sep='\t', file=buf_in)

            text = buf_in.getvalue()

        tG.import_tabtext(text)

        with io.StringIO() as buf_out:
            tG.export_tabtext(buf_out)

            actual = buf_out.getvalue()

        self.assertEqual(text, actual, 'exported 2D tabtext differs')



        ### Test 1D Grid
        tG = wallpy.viz.Grid(1)

        size_x = 24

        # Construct tabtext for grid
        def get_val(x):
            if x == (size_x - 1):
                # Force grid to be given size
                return next(self.RD)
            elif random.randrange(100) < prob_val:
                return next(self.RD)
            return None

        with io.StringIO() as buf_in:
            vals = tuple(repr(get_val(x)) for x in range(size_x))
            print(*vals, sep='\t', file=buf_in)

            text = buf_in.getvalue()

        tG.import_tabtext(text)

        with io.StringIO() as buf_out:
            tG.export_tabtext(buf_out)

            actual = buf_out.getvalue()

        self.assertEqual(text, actual, 'exported 1D tabtext differs')



    def test_contains(self):
        data = {
            (1, 2) : next(self.RD),
            (4, 3) : next(self.RD),
            (0, 1) : next(self.RD),
            (3, 3) : next(self.RD),
        }

        tG = wallpy.viz.Grid(2, nativedict = data)


        for key in data:
            self.assertEqual((key in tG), True, 'key should be present')

        nonexistent_keys = (
            (1, 3),
            (4, 4),
            (0, 0),
            (1, 1),
        )

        for key in nonexistent_keys:
            self.assertEqual((key in tG), False, 'key should not be present')



    def test_values(self):
        data = {
            (next(self.RD), next(self.RD)) : next(self.RD) for i in range(20)
        }

        tG = wallpy.viz.Grid(2, nativedict = data)

        expect = frozenset(data.values())
        actual = frozenset(tG.values())

        self.assertSetEqual(expect, actual, 'values differ')



    def test_nativedict(self):
        data = dict()
        expect = dict()

        for i in range(20):
            nativekey = tuple(random.randrange(len(self.indices[n]))
                              for n in range(self.num_dimensions))
            indexedkey = tuple(self.indices[n][nativekey[n]]
                               for n in range(self.num_dimensions))

            val = next(self.RD)
            data[indexedkey] = val
            self.G[indexedkey] = val
            expect[nativekey] = val


        actual = self.G.nativedict()

        self.assertDictEqual(expect, actual, 'nativedict mismatch')


        # Create a new Grid using nativedict

        tG = wallpy.viz.Grid(self.num_dimensions, indices = self.indices,
                             nativedict = self.G.nativedict())

        self.assertEqual(self.G, tG, 'new Grid mismatch')



    def test_copy(self):
        data = dict()

        for i in range(20):
            key = tuple(random.choice(self.indices[n])
                        for n in range(self.num_dimensions))

            val = next(self.RD)
            data[key] = val
            self.G[key] = val


        actual = self.G.copy()
        self.assertEqual(self.G, actual, 'copy mismatch')

        self.assertIsNot(self.G, actual, 'copy returned same object')




class Test_viz_list2perimeter(unittest.TestCase):

    def setUp(self):
        pass


    def test_range4(self):
        # y=0     3
        # y=1   0   2
        # y=2     1

        expect = {
            (0, 1) : 0,
            (1, 2) : 1,
            (2, 1) : 2,
            (1, 0) : 3,
        }

        actual = {
            coords : v for coords, v in wallpy.viz.list2perimeter(range(4))
        }

        self.assertDictEqual(expect, actual, 'range(4) coords differ')


    def test_range8(self):
        # y=0     7 6
        # y=1   0     5
        # y=2   1     4
        # y=3     2 3

        expect = {
            (0, 1) : 0,
            (0, 2) : 1,
            (1, 3) : 2,
            (2, 3) : 3,
            (3, 2) : 4,
            (3, 1) : 5,
            (2, 0) : 6,
            (1, 0) : 7,
        }

        actual = {
            coords : v for coords, v in wallpy.viz.list2perimeter(range(8))
        }

        self.assertDictEqual(expect, actual, 'range(8) coords differ')




__TL = unittest.TestLoader()

alltests = unittest.TestSuite((
    __TL.loadTestsFromTestCase(Test_viz_grid),
    __TL.loadTestsFromTestCase(Test_viz_list2perimeter),
))


if __name__ == '__main__':
    unittest.TextTestRunner(verbosity=2).run(alltests)
