#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Classes supporting prettyprinting.

Exported Classes:

CleanObject -- Superclass implementing repr, prettyprint, valid_p, etc.
CleanList -- list CleanObject
CleanDict -- dict CleanObject
CleanSet -- set CleanObject

"""


import sys
import io
import random

from . import prettyprint




moduleprefix = __name__ + '.'




class CleanObject(object):
    """Superclass implementing repr, prettyprint, valid_p, etc.

    Public functions:

    __repr__ -- Returns a string with the Python representation of the class

    prettyprint -- Pretty-prints the object to the specified file.

    valid_p -- Returns True iff all the fields are valid.


    Restrictions:

    Does not support cyclical graphs (e.g. A.x = B, B.y = A).
    """

    posarg_names = []
    kwarg_names = []

    # Used in fields_*, valid_p, and dfs but NOT in __init__ or prettyprint
    #   Extra fields start out with value = None
    extra_names = []


    def __init__(self, *args, **kwargs):
        """Set the fields given by args and kwargs.
        
        args must have exactly as many elements as type(self).posarg_names
        """

        posarg_names = type(self).posarg_names
        posarg_nameset = frozenset(posarg_names)

        kwarg_names = type(self).kwarg_names
        kwarg_nameset = frozenset(kwarg_names)

        if len(args) != len(posarg_nameset):
            raise TypeError(' '.join([
                '__init__',
                'takes exactly',
                str(len(posarg_nameset)),
                'positional',
                'argument' if len(posarg_nameset) == 1 else 'arguments',
                'but',
                str(len(args)),
                'given',
            ]))

        if set(kwargs) != kwarg_nameset:
            raise TypeError(' '.join([
                'kwargs do not match required keyword args:',
                'given',
                repr(set(kwargs)),
                'but require',
                repr(kwarg_nameset),
            ]))

        for i, k in enumerate(posarg_names):
            setattr(self, k, args[i])

        for k, v in kwargs.items():
            setattr(self, k, v)


        self.searchflag = None


        # Set extra fields' values to None
        for k in type(self).extra_names:
            setattr(self, k, None)



    def __eq__(self, other):
        if not hasattr(other, '__dict__'):
            return False

        return self.__dict__ == other.__dict__


    def prettyprint(self, indent=0, file=sys.stdout):
        PP = prettyprint.PrettyPrinter(indent, file)

        moduleprefix = ''
        initialsep = prettyprint.InitialSeparator('', ', ')

        if hasattr(type(self), 'moduleprefix'):
            moduleprefix = type(self).moduleprefix

        PP.print(
            moduleprefix,
            type(self).__name__,
            '(',
            extraspaces = 0)


        # Positional arguments
        posarg_names = type(self).posarg_names

        for k in posarg_names:
            v = getattr(self, k)

            sep = next(initialsep)

            PP.print(sep, '# ', k, ' = ', extraspaces = 2)


            if hasattr(v, 'prettyprint'):
                v.prettyprint(indent = indent + 4, file = file)

            else:
                PP.print(
                    repr(v),
                    extraspaces = 4)


        # Keyword arguments
        kwarg_names = type(self).kwarg_names

        for k in kwarg_names:
            v = getattr(self, k)

            sep = next(initialsep)

            PP.print(sep, k, ' = ', extraspaces = 2)


            if hasattr(v, 'prettyprint'):
                v.prettyprint(indent = indent + 4, file = file)

            else:
                PP.print(
                    repr(v),
                    extraspaces = 4)


        PP.print(
            ')',
            extraspaces = 0)




    def __repr__(self):

        with io.StringIO() as outbuf:
            self.prettyprint(file = outbuf)

            def filterlines(buf):
                for L in buf.getvalue().split('\n'):

                    L = L.strip()

                    commentindex = L.find('#')
                    if commentindex >= 0:
                        L = L[:commentindex]

                    if L:
                        yield L


            return ' '.join(filterlines(outbuf))





    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args):
                pass


        # Positional arguments
        posarg_names = type(self).posarg_names

        for k in posarg_names:
            v = getattr(self, k)

            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid', k)
                    return False


        # Keyword arguments
        kwarg_names = type(self).kwarg_names

        for k in kwarg_names:
            v = getattr(self, k)

            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid', k)
                    return False


        # Extra fields
        extra_names = type(self).extra_names

        for k in extra_names:
            v = getattr(self, k)

            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid', k)
                    return False


        return True




    def fields_k(self, filt = None, filt_posargs = None, filt_kwargs = None):

        if not filt:
            def filt(k, v, *args, **kwargs):
                return True

        if filt_posargs:
            filt_posargs = tuple(filt_posargs)
        else:
            filt_posargs = tuple()

        if not filt_kwargs:
            filt_kwargs = dict()


        for k in type(self).posarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield k


        for k in type(self).kwarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield k


        for k in type(self).extra_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield k




    def fields_v(self, filt = None, filt_posargs = None, filt_kwargs = None):

        if not filt:
            def filt(k, v, *args, **kwargs):
                return True

        if filt_posargs:
            filt_posargs = tuple(filt_posargs)
        else:
            filt_posargs = tuple()

        if not filt_kwargs:
            filt_kwargs = dict()


        for k in type(self).posarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield v


        for k in type(self).kwarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield v


        for k in type(self).extra_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield v




    def fields_kv(self, filt = None, filt_posargs = None, filt_kwargs = None):

        if not filt:
            def filt(k, v, *args, **kwargs):
                return True

        if filt_posargs:
            filt_posargs = tuple(filt_posargs)
        else:
            filt_posargs = tuple()

        if not filt_kwargs:
            filt_kwargs = dict()


        for k in type(self).posarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield (k, v)


        for k in type(self).kwarg_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield (k, v)


        for k in type(self).extra_names:
            v = getattr(self, k)

            if filt(k, v, *filt_posargs, **filt_kwargs):
                yield (k, v)




    def dfs(self, childfunc = None, childfunc_posargs = None,
            childfunc_kwargs = None, callback = None,
            callback_posargs = None, callback_kwargs = None,
            searchflag = None):


        if childfunc:
            if childfunc_posargs:
                childfunc_posargs = tuple(childfunc_posargs)
            else:
                childfunc_posargs = tuple()

            if not childfunc_kwargs:
                childfunc_kwargs = dict()


        else:

            def childfunc(node):
                return node.fields_v()

            childfunc_posargs = tuple()
            childfunc_kwargs = dict()



        if not callback:
            def callback(*args, **kwargs):
                pass

        if callback_posargs:
            callback_posargs = tuple(callback_posargs)
        else:
            callback_posargs = tuple()

        if not callback_kwargs:
            callback_kwargs = dict()

        if searchflag is None:
            # Pick a new, distinct searchflag

            searchflag = self.searchflag

            while searchflag == self.searchflag:
                searchflag = random.randint(1, 1 << 32)

        else:
            if searchflag == self.searchflag:
                raise ValueError(' '.join([
                    'Cannot use searchflag equivalent to',
                    'previous searchflag value',
                ]))



        def child_filter(k, v, searchflag):
            if not hasattr(v, 'searchflag'):
                return False

            if not hasattr(v, 'dfs'):
                return False

            return v.searchflag != searchflag



        # Perform DFS
        to_view = [ self ]
        self.searchflag = searchflag

        while to_view:
            cur = to_view[-1]


            children = []

            for k, v in enumerate(
                childfunc(
                    cur,
                    *childfunc_posargs,
                    **childfunc_kwargs)
            ):
                if child_filter(k, v, searchflag):
                    v.searchflag = searchflag
                    children.insert(0, v)


            to_view.extend(children)


            if children:
                # Need to handle children first!  Wait for now.
                continue


            del to_view[-1]

            try:
                callback(cur, *callback_posargs, **callback_kwargs)
            except StopIteration:
                # End search
                return cur


        return None




class CleanList(list):
    """list with prettyprint function."""

    moduleprefix = moduleprefix


    def prettyprint(self, indent=0, file=sys.stdout):
        PP = prettyprint.PrettyPrinter(indent, file)

        moduleprefix = ''
        initialsep = prettyprint.InitialSeparator('', ', ')

        if hasattr(type(self), 'moduleprefix'):
            moduleprefix = type(self).moduleprefix

        PP.print(
            moduleprefix,
            type(self).__name__,
            '([',
            extraspaces = 0)


        # Elements
        for v in self:
            if hasattr(v, 'prettyprint'):
                v.prettyprint(indent = indent + 2, file = file)
                PP.print(',', extraspaces = 2)

            else:
                PP.print(
                    repr(v),
                    ',',
                    extraspaces = 2)


        PP.print(
            '])',
            extraspaces = 0)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args):
                pass


        for i, v in enumerate(self):
            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid at index', i)
                    return False


        return True



class CleanDict(dict):
    """dict with prettyprint function."""

    moduleprefix = moduleprefix


    def prettyprint(self, indent=0, file=sys.stdout):
        PP = prettyprint.PrettyPrinter(indent, file)

        moduleprefix = ''
        initialsep = prettyprint.InitialSeparator('', ', ')

        if hasattr(type(self), 'moduleprefix'):
            moduleprefix = type(self).moduleprefix

        PP.print(
            moduleprefix,
            type(self).__name__,
            '({',
            extraspaces = 0)


        # Elements
        for k, v in self.items():
            if hasattr(v, 'prettyprint'):
                PP.print(repr(k),  ' : ', extraspaces = 2)
                v.prettyprint(indent = indent + 4, file = file)
                PP.print(',', extraspaces = 2)

            else:
                PP.print(
                    repr(k),
                    ' : ',
                    repr(v),
                    ',',
                    extraspaces = 2)


        PP.print(
            '})',
            extraspaces = 0)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args):
                pass


        for k, v in self.items():
            if hasattr(k, 'valid_p'):
                if not k.valid_p(errcallback):
                    errcallback('Invalid key', repr(k))
                    return False

            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid value at key', repr(k))
                    return False


        return True


    def keep_only_matches(self, keepfunc, *args, **kwargs):
        """Keep only the elements s.t. bool(keepfunc(k, v, *args, **kwargs)) == True."""

        todel = list()

        for k, v in self.items():
            if keepfunc(k, v, *args, **kwargs):
                continue
            todel.append(k)

        for k in todel:
            del self[k]


    def remove_matches(self, removefunc, *args, **kwargs):
        """Remove the elements s.t. bool(removefunc(k, v, *args, **kwargs)) == True."""

        todel = list()

        for k, v in self.items():
            if removefunc(k, v, *args, **kwargs):
                todel.append(k)

        for k in todel:
            del self[k]




class CleanSet(set):
    """set with prettyprint function."""

    moduleprefix = moduleprefix


    def prettyprint(self, indent=0, file=sys.stdout):
        PP = prettyprint.PrettyPrinter(indent, file)

        moduleprefix = ''
        initialsep = prettyprint.InitialSeparator('', ', ')

        if hasattr(type(self), 'moduleprefix'):
            moduleprefix = type(self).moduleprefix

        PP.print(
            moduleprefix,
            type(self).__name__,
            '({',
            extraspaces = 0)


        # Elements
        for v in self:
            if hasattr(v, 'prettyprint'):
                v.prettyprint(indent = indent + 2, file = file)
                PP.print(',', extraspaces = 2)

            else:
                PP.print(
                    repr(v),
                    ',',
                    extraspaces = 2)


        PP.print(
            '})',
            extraspaces = 0)


    def valid_p(self, errcallback = None):

        if not errcallback:
            def errcallback(*args):
                pass


        for v in self:
            if hasattr(v, 'valid_p'):
                if not v.valid_p(errcallback):
                    errcallback('Invalid element')
                    return False


        return True

