#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Node supporting depth-first-search.

Exported Classes:

Node -- standard graph node supporting DFS

"""

import unittest
import random

from . import cleanobject
from . import misc



class Node(cleanobject.CleanObject):
    posarg_names = [ 'val' ]
    kwarg_names = [ 'children' ]

    def __init__(self, val, children = None):
        self.flag = None

        if children:
            children = list(children)
        else:
            children = list()

        args, kwargs = misc.args2pos_kw(type(self), locals())
        super().__init__(*args, **kwargs)



    def dfs_recursive(self, callback = None, callback_posargs = None,
                      callback_kwargs = None, flag = None):

        if not callback:
            def callback(*args, **kwargs):
                pass

        if callback_posargs:
            callback_posargs = tuple(callback_posargs)
        else:
            callback_posargs = tuple()

        if callback_kwargs:
            callback_kwargs = callback_kwargs
        else:
            callback_kwargs = dict()


        if not flag:
            # Pick a new, distinct flag

            flag = self.flag

            while flag == self.flag:
                flag = random.randrange(0, 1 << 32)


        self.flag = flag


        # Run DFS on children first
        for c in self.children:
            if c.flag == flag:
                continue

            c.dfs_recursive(callback = callback,
                            callback_posargs = callback_posargs,
                            callback_kwargs = callback_kwargs,
                            flag = flag)


        # Now run callback, since all children have finished
        callback(self, *callback_posargs, **callback_kwargs)



    def dfs_iterative(self, callback = None, callback_posargs = None,
                      callback_kwargs = None, flag = None):

        if not callback:
            def callback(*args, **kwargs):
                pass

        if callback_posargs:
            callback_posargs = tuple(callback_posargs)
        else:
            callback_posargs = tuple()

        if callback_kwargs:
            callback_kwargs = callback_kwargs
        else:
            callback_kwargs = dict()


        if not flag:
            # Pick a new, distinct flag

            flag = self.flag

            while flag == self.flag:
                flag = random.randrange(0, 1 << 32)



        to_view = [self]

        while to_view:
            cur = to_view[-1]
            cur.flag = flag

            appended_child = False

            # Make order consistent with recursive
            children = list(cur.children)
            children.reverse()

            for c in children:
                if c.flag != flag:
                    to_view.append(c)
                    appended_child = True

            if appended_child:
                # Need to handle children first!  Wait for now.
                continue
            

            del to_view[-1]

            callback(cur, *callback_posargs, **callback_kwargs)


