#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Miscellaneous supporting functions and classes.

Exported Functions:

arg2pos_kw -- use classes's keywords and localvars dict to form args, kwargs

Exported Classes:

CachedMap -- caching for expensive function calls
ArcList -- list with offset and length, optional rotation

"""


def args2pos_kw(C, localvars):
    """Compose positional arg list and keyword arg dict from variable dict localvars.

    Arguments:
    C -- class holding positional and keyword argument names (posarg_names and kwarg_names)
    localvars -- dict containing variable names and values

    Returns posargs, kwargs:
    posargs -- list containing the values of localvars specified by posarg_names
    kwargs -- dict containing the names and values of localvars specified by kwarg_names
    """


    posarg_names = C.posarg_names
    kwarg_names = C.kwarg_names


    posarg_vals = list()
    kwarg_vals = dict()

    for k in posarg_names:
        posarg_vals.append(localvars[k])

    for k in kwarg_names:
        kwarg_vals[k] = localvars[k]


    return posarg_vals, kwarg_vals



class CachedMap(object):
    """Calls the given function for each key and caches the result."""

    def __init__(self, func, func_posargs = None, func_kwargs = None):
        self.__func = func

        if func_posargs:
            self.__func_posargs = tuple(func_posargs)
        else:
            self.__func_posargs = tuple()

        if func_kwargs:
            self.__func_kwargs = dict(func_kwargs)
        else:
            self.__func_kwargs = dict()


        self.__cache = dict()



    @property
    def func(self):
        return self.__func

    @func.setter
    def func(self, val):
        self.__func = val
        self.__cache.clear()



    @property
    def func_posargs(self):
        return self.__func_posargs

    @func_posargs.setter
    def func_posargs(self, val):
        self.__func_posargs = val
        self.__cache.clear()



    @property
    def func_kwargs(self):
        return self.__func_kwargs

    @func_kwargs.setter
    def func_kwargs(self, val):
        self.__func_kwargs = val
        self.__cache.clear()



    def __getitem__(self, key):

        if key in self.__cache:
            return self.__cache[key]

        return self.__func(key, *self.__func_posargs, **self.__func_kwargs)



    def __call__(self, key):
        return self[key]



    def __eq__(self, other):
        if (
            hasattr(other, 'func')
            and hasattr(other, 'func_posargs')
            and hasattr(other, 'func_kwargs')
        ):

            return (
                (self.func.__code__ == other.func.__code__)
                and (self.func_posargs == other.func_posargs)
                and (self.func_kwargs == other.func_kwargs)
            )

        return False


    def __ne__(self, other):
        return not self.__eq__(other)




class ArcList(object):
    """Ordered list, with stored offset and length.

    Can modify offset, but will be capped such that the selected sublist
        remains within the stored list.

    Arguments:
    iterable -- Ordered elements for internal list
    offset -- Initial offset of subset
    length -- Length of subset
    can_rotate -- Whether or not the sublist can be a rotation of the internal list
    """


    def __init__(self, iterable, offset, length, can_rotate = False):

        self.__L = tuple(iterable)
        self.__len = len(self.__L)
        self.__can_rotate = bool(can_rotate)

        if length < 0:
            length = 0
        elif length > self.__len:
            length = self.__len

        self.__sl = length


        self.__so = self.bound_offset(offset)



    def bound_offset(self, desired_offset):
        if self.__can_rotate:
            return desired_offset

        if desired_offset < 0:
            desired_offset = 0
        elif (desired_offset + self.__sl) > self.__len:
            desired_offset = self.__len - self.__sl

        return desired_offset




    def __iter__(self):
        if self.__can_rotate:
            extra = self.__so + self.__sl - self.__len

            while self.__so < 0:
                extra += self.__len
                self.__so += self.__len

            while self.__so >= self.__len:
                extra -= self.__len
                self.__so -= self.__len

            if extra > 0:
                return iter(self.__L[self.__so:] + self.__L[:extra])

        return iter(self.__L[self.__so : self.__so + self.__sl])



    @property
    def length(self):
        return self.__sl

    @length.setter
    def length(self, value):

        # Correct out-of-bounds value
        if value < 0:
            value = 0
        elif value > self.__len:
            value = self.__len

        self.__sl = value



    @property
    def offset(self):
        return self.__so

    @offset.setter
    def offset(self, value):
        self.__so = self.bound_offset(value)



    @property
    def elements(self):
        return self.__L

    @elements.setter
    def elements(self, value):
        self.__L = tuple(value)
        self.__len = len(self.__L)

        # Correct out-of-bounds length and offset
        if self.__sl > self.__len:
            self.__sl = self.__len

        if not self.__can_rotate and (self.__so + self.__sl) > self.__len:
            self.__so = self.__len - self.__sl


    @property
    def can_rotate(self):
        return self.__can_rotate

    @can_rotate.setter
    def can_rotate(self, value):
        self.__can_rotate = bool(value)
        self.__so = self.bound_offset(value)




    def update_offset(self, new_offset):
        """Try to set offset to new_offset.

        Return True if offset changed, False if offset remained the same."""

        old_offset = self.offset
        self.offset = new_offset

        return (self.offset != old_offset)

