#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Extra random-number functions and classes.

Exported Classes:

RandomDistinct -- Return numbers unique for the life of the instance.
"""


import random




class RandomDistinct(object):
    """Return unique random numbers from range(start, stop, step).
    
    If no available number is generated within maxtries attempts,
        raise StopIteration.
    """

    fast_max_size = 1024

    def __init__(self, start, stop, step = 1, maxtries = 16):
        self.__start = int(start)
        self.__stop = int(stop)
        self.__step = int(step)
        self.maxtries = int(maxtries)

        self.__pastvals = set()


        self.update_optimization()


    def __iter__(self):
        return self


    def __next__(self):

        if self.__nextvals is not None:
            ### Use the "fast" method

            if self.__nextvals:
                return self.__nextvals.pop(0)
            else:
                raise StopIteration('No more random numbers are available')

        ### Use the "slow" method (for larger ranges)

        for i in range(self.maxtries):
            r = random.randrange(self.__start, self.__stop, self.__step)

            if r in self.__pastvals:
                r = None
            else:
                break

        if r is None:
            raise StopIteration('No more random numbers are available')

        self.__pastvals.add(r)
        return r


    def __len__(self):
        for i, r in enumerate(self, start = 1):
            pass

        return i


    def rewind(self):
        self.__pastvals.clear()
        self.update_optimization()


    def update_optimization(self):

        R = range(self.__start, self.__stop, self.__step)

        if len(R) < self.fast_max_size:
            nextvals = list(R)
            random.shuffle(nextvals)
            self.__nextvals = [x for x in nextvals if x not in self.__pastvals]
        else:
            self.__nextvals = None


    def set_start(self, val):
        self.__start = int(val)
        self.update_optimization()

    def set_stop(self, val):
        self.__stop = int(val)
        self.update_optimization()

    def set_step(self, val):
        self.__step = int(val)
        self.update_optimization()


    start = property(lambda self : self.__start, set_start)
    stop = property(lambda self : self.__stop, set_stop)
    step = property(lambda self : self.__step, set_step)

