#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""SVG output support.

Exported Functions:

length2float -- convert a units.Length to a float
trans_zoom -- return a group with translate, zoom
str_format -- convert to string a scalar or iterable of str or numeric
rect -- return a rectangle element
circle -- return a circle element
text -- return a text element
line -- return a line element
SubElement -- return a named subelement

Exported Classes:

Canvas -- convenience wrapper for SVG drawing area

"""

import sys
import math
from xml.etree import ElementTree as et

from . import units

### Some code adapted from
### http://nick.onetwenty.org/index.php/2010/04/07/creating-svg-files-with-python/



def length2float(L, main_unit):
    return float(L / units.Length(1, main_unit))


def trans_zoom(parent, x=0, y=0, zoom=1):
    transforms = (
        'translate({:0.06F},{:0.06F})'.format(x, y),
        'scale({:0.08E})'.format(zoom),
    )

    g_trans = SubElement(
        parent, 'g',
        transform=' '.join(transforms)
    )

    return g_trans


def str_format(a):
    if str(a) is a:
        return a

    try:
        return map(str_format, a)
    except TypeError:
        return '{:0.06F}'.format(a)



def rect(parent, coords, size, **kwargs):
    x, y = map(str_format, coords)
    width, height = map(str_format, size)
    return SubElement(
        parent, 'rect',
        x = x, y = y,
        width = width, height = height,
        **kwargs
    )


def circle(parent, ccoords, r, **kwargs):
    cx, cy = map(str_format, ccoords)
    r = str_format(r)
    return SubElement(
        parent, 'circle',
        cx = cx, cy = cy, r = r,
        **kwargs
    )


def text(parent, coords, text='', **kwargs):
    x, y = map(str_format, coords)
    ret = SubElement(
        parent, 'text',
        x = x, y = y,
        **kwargs
    )
    ret.text = text
    return ret


def line(parent, startcoords, endcoords, **kwargs):
    x1, y1 = map(str_format, startcoords)
    x2, y2 = map(str_format, endcoords)
    return SubElement(
        parent, 'line',
        x1 = x1, y1 = y1,
        x2 = x2, y2 = y2,
        **kwargs
    )


def SubElement(parent, name, **kwargs):
    return et.SubElement(
        parent, name,
        **kwargs
    )




class Canvas(object):
    """SVG convenience wrapper"""

    def __init__(self, size, title='', main_unit='mm', **kwargs):

        # Ensure valid main_unit
        dummy = units.Length(0, main_unit)
        self.main_unit = main_unit

        attrib = dict(kwargs)

        width, height = size
        self.size = (width, height)

        attrib['width'] = '{:0.06F}{:}'.format(width, self.main_unit)
        attrib['height'] = '{:0.06F}{:}'.format(height, self.main_unit)
        attrib['viewBox'] = ' '.join(
            '{:0.06F}'.format(v) for v in (0, 0, width, height)
        )

        self.doc = et.Element(
            'svg', version='1.1',
            xmlns='http://www.w3.org/2000/svg',
            attrib=attrib
        )

        if title:
            t_el = SubElement(self.doc, 'title')
            t_el.text = title



    def export(self, file):

        ### ElementTree 1.2 doesn't write the SVG file header errata,
        ###     so do that manually
        print('<?xml version="1.0" standalone="no"?>', file=file)
        print('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"', file=file)
        print('\t"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">',
              file=file)
        for L in et.tostringlist(self.doc):
            print(str(L, 'utf-8'), file=file)


