#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Units for internal representations of dimensions.

Exported Functions:

unitless_frac -- return a Measurement in unitless form (native unit)

Exported Classes:

Measure -- generic measurement superclass, itself with no units
Length -- subclass of Measure, supporting length units
Angle -- subclass of Measure, supporting angle units

"""

import sys
import fractions


moduleprefix = __name__ + '.'



def unitless_frac(val):
    if hasattr(val, 'native_val') and hasattr(val, 'unit'):
        if val.unit:
            raise TypeError('Cannot convert to unitless value from unit: '
                            + repr(val.unit))
        return val.native_val

    return fractions.Fraction(val)





class Measure(object):

    __slots__ = (
        '__native_val',
        '__unit',
    )

    moduleprefix = moduleprefix

    native_unit = ''
    unit2native_mult = {
        ''  :   1,
    }


    def __init__(self, val, unit = None):

        if unit is None:
            # Must be a string value including units,
            #   OR another compatible Measure

            if hasattr(val, 'unit'):
                unit = val.unit

            else:

                if not hasattr(val, 'split'):
                    raise TypeError('For no units specified, must be string value')

                words = val.split()

                if not (1 <= len(words) <= 2):
                    raise ValueError('String format must be "<val> <unit>"')

                val = fractions.Fraction(words[0])

                if len(words) < 2:
                    unit = ''
                else:
                    unit = words[1]


        if unit not in type(self).unit2native_mult:
            raise ValueError('Unrecognized unit: ' + repr(unit))

        self.__unit = unit


        if hasattr(val, 'native_val') and hasattr(val, 'unit'):
            if val.unit not in type(self).unit2native_mult:
                raise TypeError('Cannot convert value with unit: ' + repr(val.unit))

            self.__native_val = val.native_val

        else:
            self.__native_val = fractions.Fraction(val) * type(self).unit2native_mult[unit]



    def __repr__(self):
        valstr = ''.join((
            'fractions.',
            repr(self.val_in_units),
        ))

        return ''.join((
            moduleprefix,
            type(self).__name__,
            '(',
            valstr,
            ', ',
            repr(self.__unit),
            ')',
        ))


    def __str__(self):
        return ' '.join((str(self.val_in_units), self.__unit))



    @property
    def native_val(self):
        return self.__native_val


    @property
    def val_in_units(self):
        return fractions.Fraction(
            self.__native_val,
            type(self).unit2native_mult[self.__unit]
        )


    @property
    def unit(self):
        return self.__unit



    def same_measure_type(self, other):
        if hasattr(other, 'native_val') and hasattr(other, 'unit'):
            return other.unit in type(self).unit2native_mult
        return False



    def measure_eq(self, other):
        if not self.same_measure_type(other):
            raise TypeError('Not the same measurement type')
        return self.native_val == other.native_val





    ### Rich comparison functions

    def __eq__(self, other):
        if not self.same_measure_type(other):
            return False
        return (self.native_val == other.native_val)


    def __ne__(self, other):
        if not self.same_measure_type(other):
            return True
        return (self.native_val != other.native_val)


    def __lt__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.native_val < other.native_val


    def __le__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.native_val <= other.native_val


    def __gt__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.native_val > other.native_val


    def __ge__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.native_val >= other.native_val






    def __bool__(self):
        return bool(self.__native_val)



    def __new_nativeval(self, nativeval):
        """Return a copy of self but with the given native value."""

        nv_measure = type(self)(nativeval, type(self).native_unit)

        return nv_measure.convert(self.__unit)


    def convert(self, newunit):
        """Return a copy of self but with the given new unit."""

        return type(self)(self, newunit)



    ### Arithmetic operations

    def __add__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.__new_nativeval(self.native_val + other.native_val)


    def __sub__(self, other):
        if not self.same_measure_type(other):
            return NotImplemented
        return self.__new_nativeval(self.native_val - other.native_val)


    def __mul__(self, other):
        # Ensure other is valid (unitless and exact)
        try:
            other_val = unitless_frac(other)
        except TypeError:
            return NotImplemented
        return self.__new_nativeval(self.native_val * other_val)


    def __truediv__(self, other):
        # Same type: return unitless fraction
        if self.same_measure_type(other):
            return fractions.Fraction(self.native_val, other.native_val)

        # Otherwise, ensure other is valid (unitless and exact)
        try:
            other_val = unitless_frac(other)
        except TypeError:
            return NotImplemented
        return self.__new_nativeval(self.native_val / other_val)


    def __floordiv__(self, other):
        # Accept only Measures of same type and return int
        if not self.same_measure_type(other):
            return NotImplemented
        return self.native_val // other.native_val


    def __mod__(self, other):
        # Accept only Measures of same type and return Measure
        if not self.same_measure_type(other):
            return NotImplemented
        return self.__new_nativeval(self.native_val % other.native_val)


    def __divmod__(self, other):
        # Accept only Measures of same type and return tuple(int, Measure)
        if not self.same_measure_type(other):
            return NotImplemented
        return (self // other, self % other)


    def __rmul__(self, other):
        return self * other


    def __neg__(self):
        return self.__new_nativeval(- self.native_val)


    def __pos__(self):
        return self.__new_nativeval(+ self.native_val)


    def __abs__(self):
        return self.__new_nativeval(abs(self.native_val))





class Length(Measure):
    native_unit = 'nm'
    unit2native_mult = {
        'm'         :   1000000000,
        'cm'        :   10000000,
        'mm'        :   1000000,
        'micron'    :   1000,
        'nm'        :   1,

        'yd'        :   36 * 25400000,
        'ft'        :   12 * 25400000,
        'in'        :   25400000,
        'mil'       :   25400,
        'thou'      :   25400,
    }



class Angle(Measure):
    native_unit = 'deg'
    unit2native_mult = {
        'deg'       :   1,
        'pi_rad'    :   180,
    }


