#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Determine version number from source control.

Exported Functions:

git_query_version -- query git source control for tag name + commits
get_version -- retrieve and check version string

"""

import sys
import subprocess
import os.path
import re

# Adapted from:
# https://github.com/warner/python-ecdsa/blob/9e21c3388cc98ba90877a1e4dbc2aaf66c67d365/setup.py


def git_query_version(fallback_version='',
                      match_pattern='v[0-9]*'):
    """Retrieve a version string from a git repository if possible.
    Otherwise return fallback_version.

    Keyword arguments:
    fallback_version -- return this if no version string available
    search_prefix -- try to find a tag with this prefix

    """
    try:
        p = subprocess.Popen(
            [
                'git', 'describe',
                '--tags',
                '--dirty=-mod',
                '--always',
                '--match',
                match_pattern,
            ],
            stdout=subprocess.PIPE
        )

    except EnvironmentError:
        return fallback_version

    stdout = str(p.communicate()[0].strip(), 'utf-8')
    if p.returncode != 0:
        return fallback_version

    return stdout



def get_version(fallback_version='',
                require_prefix='v',
                match_pattern='v[0-9]*'):
    """Retrieve a version string if possible.
    Otherwise, return fallback_version.

    Keyword arguments:
    fallback_version -- return this if no version string available
    require_prefix -- if set, then retrieved string must begin with this
                        (if it doesn't, then return fallback_version)
    match_pattern -- if set, then retrieve a version string matching
                        this pattern

    """

    if not match_pattern:
        if require_prefix:
            match_pattern = require_prefix + '*'
        else:
            match_pattern = '*'

    ver = git_query_version(fallback_version=fallback_version,
                            match_pattern=match_pattern)

    if require_prefix:
        if not ver.startswith(require_prefix):
            return fallback_version

    return ver
