#!/usr/bin/env python3

######################## BEGIN LICENSE BLOCK ########################
# Copyright (c) 2012 Wall Software, LLC
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
######################### END LICENSE BLOCK #########################

"""Graph / grid visualizations.

Exported Functions:

list2perimeter -- space values equally on four sides of a square

Exported Classes:

Grid -- multi-dimensional sparse matrix with specified index sets

"""

import sys

from . import prettyprint




class Grid(object):
    """Multi-dimensional sparse matrix with specified index sets."""

    default_dimlen = 64

    def __init__(self, num_dimensions, nativedict = None, indices = None):
        """Arguments:
        num_dimensions -- number of dimensions for coordinate keys

        Keyword arguments:
        nativedict -- copy this to internal dict
        indices -- ordered index values for each dimension

        """

        num_dimensions = int(num_dimensions)

        if num_dimensions < 1:
            raise ValueError('Must have positive number of dimensions')

        self.__num_dimensions = num_dimensions

        if nativedict:
            self.__D = dict(nativedict)
        else:
            self.__D = dict()

        if indices:
            if len(indices) != self.__num_dimensions:
                raise ValueError('indices do not match number of dimensions')

            dimlens = [None] * self.__num_dimensions
            tmp_indices = [None] * self.__num_dimensions

            for i, L in enumerate(indices):
                dl = len(L)

                if dl < 1:
                    raise ValueError(' '.join([
                        'dimension',
                        str(i),
                        'must have at least one index value',
                    ]))

                dimlens[i] = dl
                tmp_indices[i] = tuple(L)

            self.__dimlens = tuple(dimlens)
            self.__indices = tuple(tmp_indices)


        else:
            index_tuple = tuple(range(self.default_dimlen))
            self.__dimlens = (self.default_dimlen, ) * self.__num_dimensions
            self.__indices = (index_tuple, ) * self.__num_dimensions



    num_dimensions = property(lambda self : self.__num_dimensions)
    dimlens = property(lambda self : self.__dimlens)
    indices = property(lambda self : self.__indices)


    def native2indexed(self, nativekey):
        if len(nativekey) != self.__num_dimensions:
            raise ValueError('nativekey does not match number of dimensions')

        for i in range(self.__num_dimensions):
            if not (0 <= nativekey[i] < self.__dimlens[i]):
                raise ValueError(' '.join([
                    'coordinate for dimension',
                    str(i),
                    'out of range',
                ]))

        return tuple(self.__indices[i][x] for i, x in enumerate(nativekey))


    def indexed2native(self, indexedkey):
        if len(indexedkey) != self.__num_dimensions:
            raise ValueError('nativekey does not match number of dimensions')

        ret = list()

        for i in range(self.__num_dimensions):
            nativeval = self.__indices[i].index(indexedkey[i])

            ret.append(nativeval)

        return tuple(ret)


    def __eq__(self, other):
        return self.__dict__ == other.__dict__


    def __setitem__(self, indexedkey, val):
        nativekey = self.indexed2native(indexedkey)
        self.__D[nativekey] = val


    def __getitem__(self, indexedkey):
        nativekey = self.indexed2native(indexedkey)
        return self.__D[nativekey]


    def __delitem__(self, indexedkey):
        nativekey = self.indexed2native(indexedkey)
        del self.__D[nativekey]


    def __contains__(self, indexedkey):
        nativekey = self.indexed2native(indexedkey)
        return nativekey in self.__D


    def indexed_items(self):
        for nativekey, val in self.__D.items():
            yield (self.native2indexed(nativekey), val)


    def indexed_keys(self):
        for nativekey in self.__D:
            yield self.native2indexed(nativekey)


    def values(self):
        for val in self.__D.values():
            yield val


    def nativedict(self):
        return self.__D.copy()


    def import_tabtext(self, text):

        if self.__num_dimensions == 1:
            for x, vstr in enumerate(text.split('\t')):
                vstr = vstr.rstrip()

                if not vstr:
                    continue

                val = eval(vstr)

                self[(x,)] = val

        elif self.__num_dimensions == 2:
            for y, L in enumerate(text.split('\n')):
                for x, vstr in enumerate(L.split('\t')):
                    vstr = vstr.rstrip()

                    if not vstr:
                        continue

                    val = eval(vstr)

                    self[(x, y)] = val

        else:
            raise ValueError('Cannot import tabtext for > 2D grid')



    def export_tabtext(self, file, default=None):

        if self.__num_dimensions == 1:
            # Find maximum values for X
            max_x = 0

            for k in self.__D:
                x = k[0]
                max_x = max(max_x, x)

            # Now print grid
            initialsep = prettyprint.InitialSeparator('', '\t')

            for x in range(max_x + 1):
                val = repr(self.__D.get((x,), default))

                print(next(initialsep), val, sep='', end='', file=file)

            print(file=file)


        elif self.__num_dimensions == 2:
            # Find maximum values for X and Y
            max_x = 0
            max_y = 0

            for k in self.__D:
                x, y = k
                max_x = max(max_x, x)
                max_y = max(max_y, y)

            # Now print grid
            for y in range(max_y + 1):
                initialsep = prettyprint.InitialSeparator('', '\t')

                for x in range(max_x + 1):
                    val = repr(self.__D.get((x, y), default))

                    print(next(initialsep), val, sep='', end='', file=file)

                print(file=file)

        else:
            raise ValueError('Cannot export tabtext for > 2D grid')


    def copy(self):
        return type(self)(
            self.__num_dimensions,
            indices = self.__indices,
            nativedict = self.__D)




def list2perimeter(iterable):
    """Return a dict of (x, y) : val, with values moving counterclockwise
        around a rectangular perimeter."""

    tup = tuple(iterable)
    iterlen = len(tup)
    sidelen = iterlen // 4

    if (iterlen % 4) != 0:
        raise ValueError('Can support iterables only with lengths evenly divisible by 4')

    def i2xy(i):

        # West side?
        if i < sidelen:
            return (0, i + 1)

        # South side?
        i -= sidelen
        if i < sidelen:
            return (i + 1, sidelen + 1)

        # East side?
        i -= sidelen
        if i < sidelen:
            return (sidelen + 1, sidelen - i)

        # North side?
        i -= sidelen
        if i < sidelen:
            return (sidelen - i, 0)

        raise AssertionError('i should never be >= 4*sidelen')


    for i, v in enumerate(tup):
        coords = i2xy(i)
        yield (coords, v)

